/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include<stdio.h>
#include<string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <pimglib.h>
#include <picowidget.h>
#include <display.h>
#include <eventloop.h>

#define NORMAL_FONT "-picopeta-simputer-medium-r-normal--11-*-*-*-p-*-fontspecific-0"
#define BOLD_FONT 	"-picopeta-simputer-bold-r-normal--11-*-*-*-p-*-fontspecific-0"

px_display *disp;
px_eventloop *eloop;
px_widget *toplvl;
px_widget *butt[5];
px_widget *msglbl;
px_color col, acol, gcol;
border_types btype;

// Resets the font of the previous clicked button to normal font.
// btype remembers the previously pressed button.
void reset_previously_pressed_button( void )
{
	int widx;

	switch (btype) {
		case RECTANGLE_BORDER:
			widx = 0;
			break;
		case BEVELLED_BORDER:
			widx = 1;
			break;
		case ELLIPTIC_BORDER:
			widx = 2;
			break;
		case NO_BORDER:
			widx = 3;
			break;
		default:
			return;
	}
	butt[widx]->specific.button.set_font(butt[widx],NORMAL_FONT);
	butt[widx]->redraw(butt[widx]);
}

// BUTTON_PRESS callback for the rectangle button
Bool butt1_callback(px_widget* widget, px_event* event)
{
	reset_previously_pressed_button();

	btype = RECTANGLE_BORDER;
	butt[0]->specific.button.set_font(butt[0], BOLD_FONT);
	butt[0]->redraw(butt[0]);
	msglbl->specific.label.set_text(msglbl,"Rectangular button pressed");
	msglbl->redraw(msglbl);
	return False;
}

// BUTTON_PRESS callback for the bevelled button
Bool butt2_callback(px_widget* widget, px_event* event)
{
	reset_previously_pressed_button();
	btype = BEVELLED_BORDER;

	butt[1]->specific.button.set_font(butt[1], BOLD_FONT);
	butt[1]->redraw(butt[1]);
	msglbl->specific.label.set_text(msglbl,"Bevelled button pressed");
	msglbl->redraw(msglbl);
	return False;
}

// BUTTON_PRESS callback for the elliptic button
Bool butt3_callback(px_widget* widget, px_event* event)
{
	reset_previously_pressed_button();
	btype = ELLIPTIC_BORDER;

	butt[2]->specific.button.set_font(butt[2], BOLD_FONT);
	butt[2]->redraw(butt[2]);
	msglbl->specific.label.set_text(msglbl,"Elliptic button pressed");
	msglbl->redraw(msglbl);
	return False;
}

// BUTTON_PRESS callback for the no border button
Bool butt4_callback(px_widget* widget, px_event* event)
{
	reset_previously_pressed_button();
	btype = NO_BORDER;

	butt[3]->specific.button.set_font(butt[3], BOLD_FONT);
	butt[3]->redraw(butt[3]);
	msglbl->specific.label.set_text(msglbl,"No-Border button pressed");
	msglbl->redraw(msglbl);
	return False;
}

// BUTTON_PRESS callback for the image background button
Bool butt5_callback(px_widget* widget, px_event* event)
{
	reset_previously_pressed_button();
	btype = -1;

	msglbl->specific.label.set_text(msglbl,"Pixmap button pressed");
	msglbl->redraw(msglbl);
	return False;
}

void create_buttons(px_widget *parent)
{
	butt[0] = new_widget(BUTTON, parent, 50, 20, 140, 30, "Hello");
	butt[0]->specific.button.set_font(butt[0], NORMAL_FONT);
	butt[0]->specific.button.set_text(butt[0], "Me - Rectangle");
	butt[0]->set_border_type(butt[0], RECTANGLE_BORDER);
	butt[0]->set_background_color(butt[0], &col);
	butt[0]->specific.button.set_active_background_color(butt[0], &acol);
	butt[0]->specific.button.set_active_font_color(butt[0], "white");
	butt[0]->callbacks[BUTTON_PRESSED] = butt1_callback;
	butt[0]->redraw(butt[0]);
	butt[0]->show(butt[0]);

	butt[1] = new_widget(BUTTON, parent, 50, 60, 140, 30, "Hello");
	butt[1]->specific.button.set_font(butt[1], NORMAL_FONT);
	butt[1]->specific.button.set_text(butt[1], "Me - Bevelled");
	butt[1]->set_border_type(butt[1], BEVELLED_BORDER);
	butt[1]->set_background_color(butt[1], &col);
	butt[1]->specific.button.set_active_background_color(butt[1], &acol);
	butt[1]->specific.button.set_active_font_color(butt[1], "white");
	butt[1]->callbacks[BUTTON_PRESSED] = butt2_callback;
	butt[1]->redraw(butt[1]);
	butt[1]->show(butt[1]);

	butt[2] = new_widget(BUTTON, parent, 50, 100, 140, 30, "Hello");
	butt[2]->specific.button.set_font(butt[2], NORMAL_FONT);
	butt[2]->specific.button.set_text(butt[2], "Me - Ellipse");
	butt[2]->set_border_type(butt[2], ELLIPTIC_BORDER);
	butt[2]->set_background_color(butt[2], &col);
	butt[2]->specific.button.set_active_background_color(butt[2], &acol);
	butt[2]->specific.button.set_active_font_color(butt[2], "white");
	butt[2]->callbacks[BUTTON_PRESSED] = butt3_callback;
	butt[2]->redraw(butt[2]);
	butt[2]->show(butt[2]);

	butt[3] = new_widget(BUTTON, parent, 50, 140, 140, 30, "Hello");
	butt[3]->specific.button.set_font(butt[3], NORMAL_FONT);
	butt[3]->specific.button.set_text(butt[3], "Me - No Border");
	butt[3]->set_border_type(butt[3], NO_BORDER);
	butt[3]->set_background_color(butt[3], &col);
	butt[3]->specific.button.set_active_background_color(butt[3], &acol);
	butt[3]->specific.button.set_active_font_color(butt[3], "white");
	butt[3]->callbacks[BUTTON_PRESSED] = butt4_callback;
	butt[3]->redraw(butt[3]);
	butt[3]->show(butt[3]);

	butt[4] = new_widget(BUTTON, parent, 10, 20, 25, 150, "Hello");
	butt[4]->specific.button.set_font(butt[4], NORMAL_FONT);
	butt[4]->specific.button.set_text(butt[4], "");
	butt[4]->set_border_type(butt[4], RECTANGLE_BORDER);
	button_set_normal_pixmap_from_file(butt[4], "bg1.png");
	button_set_active_pixmap_from_file(butt[4], "bg2.png");
	butt[4]->callbacks[BUTTON_PRESSED] = butt5_callback;
	butt[4]->redraw(butt[4]);
	butt[4]->show(butt[4]);
}

int main(int av, char **ac)
{
	disp = display_get_display();
	
	if (disp->open_display(disp, (char *)getenv("DISPLAY"))) {
		col.type = TRANSPARENT;
		acol.type = SOLID;
		widget_xcolor_from_str(&acol.first_color, "black");

		// Create the toplevel window
		toplvl = new_widget(TOPLEVEL, NULL, 0, 0, 240, 270, NULL);
		toplvl->specific.toplevel.set_title(toplvl, "Example");
		toplvl->redraw(toplvl);
		toplvl->show(toplvl);

		// Add buttons
		create_buttons(toplvl);

		// Create the message label
		msglbl = new_widget(LABEL, toplvl, 10, 180, 220, 60, NULL);
		label_set_normal_color(msglbl, "#336699");
		msglbl->specific.label.set_font(msglbl, BOLD_FONT);
		msglbl->specific.label.set_text(msglbl, "Message Area.");
		gcol.type = SOLID;
		widget_xcolor_from_str(&gcol.first_color, "grey");
		widget_set_background_color(msglbl, &gcol);
		msglbl->redraw(msglbl);
		msglbl->show(msglbl);

		// Get into eventloop
		eloop = eventloop_get_eventloop();
		eloop->loop(eloop);

		display_destroy();
	} else {
		fprintf(stderr, "Cannot open display.\n");
		exit(1);
	}
	exit(0);
}
