/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

/*
 * Checkbutton Example:
 *   2 questions will multiple correct answers are posed. The answers for
 *   each question appears as 4 checkbuttons. There are buttons to reset
 *   the checkbuttons and to show the correct answers.
 *
 *   Checkbuttons have the standard routines for setting font, setting
 *   text, justification, fg/bg color, etc. The function to select and
 *   deselect a checkbutton needs to be noted.
 */
#include <stdio.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <pimglib.h>
#include <picowidget.h>
#include <display.h>
#include <eventloop.h>

#define TINY_FONT "-picopeta-simputer-medium-r-normal--7-*-*-*-p-*-fontspecific-0"
#define NORMAL_FONT "-picopeta-simputer-medium-r-normal--11-*-*-*-p-*-fontspecific-0"
#define BOLD_FONT "-picopeta-simputer-bold-r-normal--11-*-*-*-p-*-fontspecific-0"

px_display *disp;
px_eventloop *eloop;
px_widget *toplvl;
px_widget *one_cb[4], *two_cb[4];
px_widget *rstbutt, *ansbutt;
px_widget *quesn1, *quesn2;
px_color col, acol;

// Create the checkbuttons for the first question.
void create_one_cbs(px_widget *parent)
{
	one_cb[0] = new_widget(CHECKBUTTON, parent, 10, 55, 110, 20, "Hello");
	one_cb[0]->specific.checkbutton.set_font(one_cb[0], TINY_FONT);
	one_cb[0]->specific.checkbutton.set_text(one_cb[0], "Jawaharlal Nehru");
	one_cb[0]->specific.checkbutton.set_justification(one_cb[0], JUSTIFY_LEFT);
	one_cb[0]->redraw(one_cb[0]);
	one_cb[0]->show(one_cb[0]);

	one_cb[1] = new_widget(CHECKBUTTON, parent, 130, 55, 110, 20, "Hello");
	one_cb[1]->specific.checkbutton.set_font(one_cb[1], TINY_FONT);
	one_cb[1]->specific.checkbutton.set_text(one_cb[1], "Mahatma Gandhi");
	one_cb[1]->specific.checkbutton.set_justification(one_cb[1], JUSTIFY_LEFT);
	one_cb[1]->redraw(one_cb[1]);
	one_cb[1]->show(one_cb[1]);

	one_cb[2] = new_widget(CHECKBUTTON, parent, 10, 75, 110, 20, "Hello");
	one_cb[2]->specific.checkbutton.set_font(one_cb[2], TINY_FONT);
	one_cb[2]->specific.checkbutton.set_text(one_cb[2], "Netaji SC Bose");
	one_cb[2]->specific.checkbutton.set_justification(one_cb[2], JUSTIFY_LEFT);
	one_cb[2]->redraw(one_cb[2]);
	one_cb[2]->show(one_cb[2]);

	one_cb[3] = new_widget(CHECKBUTTON, parent, 130, 75, 110, 20, "Hello");
	one_cb[3]->specific.checkbutton.set_font(one_cb[3], TINY_FONT);
	one_cb[3]->specific.checkbutton.set_text(one_cb[3], "Lal Bahadur Sastri");
	one_cb[3]->specific.checkbutton.set_justification(one_cb[3], JUSTIFY_LEFT);
	one_cb[3]->redraw(one_cb[3]);
	one_cb[3]->show(one_cb[3]);
}

// Checkbuttons for the second question.
void create_two_cbs(px_widget *parent)
{
	two_cb[0] = new_widget(CHECKBUTTON, parent, 10, 160, 110, 20, "Hello");
	two_cb[0]->specific.checkbutton.set_font(two_cb[0], TINY_FONT);
	two_cb[0]->specific.checkbutton.set_text(two_cb[0], "Brahmaputra");
	two_cb[0]->specific.checkbutton.set_justification(two_cb[0], JUSTIFY_LEFT);
	two_cb[0]->redraw(two_cb[0]);
	two_cb[0]->show(two_cb[0]);

	two_cb[1] = new_widget(CHECKBUTTON, parent, 130, 160, 110, 20, "Hello");
	two_cb[1]->specific.checkbutton.set_font(two_cb[1], TINY_FONT);
	two_cb[1]->specific.checkbutton.set_text(two_cb[1], "Ganges");
	two_cb[1]->specific.checkbutton.set_justification(two_cb[1], JUSTIFY_LEFT);
	two_cb[1]->redraw(two_cb[1]);
	two_cb[1]->show(two_cb[1]);

	two_cb[2] = new_widget(CHECKBUTTON, parent, 10, 180, 110, 20, "Hello");
	two_cb[2]->specific.checkbutton.set_font(two_cb[2], TINY_FONT);
	two_cb[2]->specific.checkbutton.set_text(two_cb[2], "Narmada");
	two_cb[2]->specific.checkbutton.set_justification(two_cb[2], JUSTIFY_LEFT);
	two_cb[2]->redraw(two_cb[2]);
	two_cb[2]->show(two_cb[2]);

	two_cb[3] = new_widget(CHECKBUTTON, parent, 130, 180, 110, 20, "Hello");
	two_cb[3]->specific.checkbutton.set_font(two_cb[3], TINY_FONT);
	two_cb[3]->specific.checkbutton.set_text(two_cb[3], "Indus");
	two_cb[3]->specific.checkbutton.set_justification(two_cb[3], JUSTIFY_LEFT);
	two_cb[3]->redraw(two_cb[3]);
	two_cb[3]->show(two_cb[3]);
}

// The two questions appear as label widgets
void create_labels(px_widget *parent)
{
	quesn1 = new_widget(LABEL, parent, 10, 5, 220, 45, NULL);
	quesn1->specific.label.set_font(quesn1, NORMAL_FONT);
	quesn1->specific.label.set_normal_color(quesn1, "#336699");
	quesn1->specific.label.set_text(quesn1, "1. Who has NOT held the post of the Prime Minister of India?");
	quesn1->specific.label.set_justification(quesn1, JUSTIFY_LEFT);
	widget_set_wrap_type(quesn1, WRAP_WORD);
	quesn1->redraw(quesn1);
	quesn1->show(quesn1);

	quesn2 = new_widget(LABEL, parent, 10, 110, 220, 45, NULL);
	quesn2->specific.label.set_font(quesn2, NORMAL_FONT);
	quesn2->specific.label.set_normal_color(quesn2, "#cc9966");
	quesn2->specific.label.set_text(quesn2, "2. Which river(s) cut across international boundaries?");
	quesn2->specific.label.set_justification(quesn2, JUSTIFY_LEFT);
	widget_set_wrap_type(quesn2, WRAP_WORD);
	quesn2->redraw(quesn2);
	quesn2->show(quesn2);
}

// Resets the checkbuttons (needs an explicit redraw())
Bool reset_cbs(px_widget *wdgt, px_event *ev)
{
	checkbutton_deselect(one_cb[0]);
	one_cb[0]->redraw(one_cb[0]);
	checkbutton_deselect(one_cb[1]);
	one_cb[1]->redraw(one_cb[1]);
	checkbutton_deselect(one_cb[2]);
	one_cb[2]->redraw(one_cb[2]);
	checkbutton_deselect(one_cb[3]);
	one_cb[3]->redraw(one_cb[3]);

	checkbutton_deselect(two_cb[0]);
	two_cb[0]->redraw(two_cb[0]);
	checkbutton_deselect(two_cb[1]);
	two_cb[1]->redraw(two_cb[1]);
	checkbutton_deselect(two_cb[2]);
	two_cb[2]->redraw(two_cb[2]);
	checkbutton_deselect(two_cb[3]);
	two_cb[3]->redraw(two_cb[3]);
	return(False);
}

// Displays the correct answers.
Bool show_answers(px_widget *wdgt, px_event *ev)
{
	checkbutton_deselect(one_cb[0]);
	one_cb[0]->redraw(one_cb[0]);
	checkbutton_select(one_cb[1]);
	one_cb[1]->redraw(one_cb[1]);
	checkbutton_select(one_cb[2]);
	one_cb[2]->redraw(one_cb[2]);
	checkbutton_deselect(one_cb[3]);
	one_cb[3]->redraw(one_cb[3]);

	checkbutton_select(two_cb[0]);
	two_cb[0]->redraw(two_cb[0]);
	checkbutton_select(two_cb[1]);
	two_cb[1]->redraw(two_cb[1]);
	checkbutton_deselect(two_cb[2]);
	two_cb[2]->redraw(two_cb[2]);
	checkbutton_select(two_cb[3]);
	two_cb[3]->redraw(two_cb[3]);
	return(False);
}

// Create the reset and show answes buttons. Associate appropriate callbacks.
void create_buttons(px_widget *parent)
{
	rstbutt = new_widget(BUTTON,parent,40,230,70,30,"Reset");
	rstbutt->specific.button.set_font(rstbutt, BOLD_FONT);
	rstbutt->specific.button.set_text(rstbutt, "Reset");
	rstbutt->set_border_type(rstbutt, BEVELLED_BORDER);
	rstbutt->set_background_color(rstbutt, &col);
	rstbutt->specific.button.set_active_background_color(rstbutt, &acol);
	rstbutt->specific.button.set_active_font_color(rstbutt, "white");
	rstbutt->callbacks[BUTTON_PRESSED] = reset_cbs;
	rstbutt->redraw(rstbutt);
	rstbutt->show(rstbutt);

	ansbutt = new_widget(BUTTON,parent,130,230,70,30,"Answers");
	ansbutt->specific.button.set_font(ansbutt, BOLD_FONT);
	ansbutt->specific.button.set_text(ansbutt, "Answers");
	ansbutt->set_border_type(ansbutt, BEVELLED_BORDER);
	ansbutt->set_background_color(ansbutt, &col);
	ansbutt->specific.button.set_active_background_color(ansbutt, &acol);
	ansbutt->specific.button.set_active_font_color(ansbutt, "white");
	ansbutt->callbacks[BUTTON_PRESSED] = show_answers;
	ansbutt->redraw(ansbutt);
	ansbutt->show(ansbutt);
}

int main(int av, char **ac)
{
	disp = display_get_display();
	
	if (disp->open_display(disp, (char *)getenv("DISPLAY"))) {
		col.type = TRANSPARENT;
		acol.type = SOLID;
		widget_xcolor_from_str(&acol.first_color, "black");

		// Create the toplevel window
		toplvl = new_widget(TOPLEVEL, NULL, 0, 0, 240, 270, NULL);
		toplvl->specific.toplevel.set_title(toplvl, "Example");
		toplvl->redraw(toplvl);
		toplvl->show(toplvl);

		// Create the widgets
		create_one_cbs(toplvl);
		create_two_cbs(toplvl);
		create_labels(toplvl);
		create_buttons(toplvl);

		// Get into eventloop
		eloop = eventloop_get_eventloop();
		eloop->loop(eloop);

		display_destroy();
	} else {
		fprintf(stderr, "Cannot open display.\n");
		exit(1);
	}
	exit(0);
}
