/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include<stdio.h>
#include<string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <pimglib.h>
#include <picowidget.h>
#include <display.h>
#include <eventloop.h>

#define NORMAL_FONT "-picopeta-simputer-medium-r-normal--11-*-*-*-p-*-fontspecific-0"
#define BOLD_FONT "-picopeta-simputer-bold-r-normal--11-*-*-*-p-*-fontspecific-0"

px_display *disp;
px_eventloop *eloop;
px_widget *toplvl;
px_widget *lbls[3], *entries[4];
px_widget *resetbutt, *quitbutt;
px_color col, acol;

// Create entry widgets to accept
//    1. name  (default value is "Enter name")
//    2. address1 (default value is "Enter address 1")
//    3. address2 (Default value is "Enter address 2")
//    4. phone (Default value is "Enter phone");
void create_entries(px_widget *parent)
{
	entries[0] = new_widget(ENTRY, parent, 85, 5, 130, 20, "Hello");
	entries[0]->specific.entry.set_font(entries[0], NORMAL_FONT);
	entries[0]->specific.entry.set_text(entries[0], "Enter name");
	entries[0]->redraw(entries[0]);
	entries[0]->show(entries[0]);

	entries[1] = new_widget(ENTRY, parent, 85, 50, 130, 20, "Hello");
	entries[1]->specific.entry.set_font(entries[1], NORMAL_FONT);
	entries[1]->specific.entry.set_text(entries[1], "Enter address 1");
	entries[1]->redraw(entries[1]);
	entries[1]->show(entries[1]);

	entries[2] = new_widget(ENTRY, parent, 85, 75, 130, 20, "Hello");
	entries[2]->specific.entry.set_font(entries[2], NORMAL_FONT);
	entries[2]->specific.entry.set_text(entries[2], "Enter address 2");
	entries[2]->redraw(entries[2]);
	entries[2]->show(entries[2]);

	entries[3] = new_widget(ENTRY, parent, 85, 120, 130, 20, "Hello");
	entries[3]->specific.entry.set_font(entries[3], NORMAL_FONT);
	entries[3]->specific.entry.set_text(entries[3], "Enter phone");
	entries[3]->redraw(entries[3]);
	entries[3]->show(entries[3]);
}

// Create the labels for the entry widgets.
void create_labels(px_widget *parent)
{
	lbls[0] = new_widget(LABEL, parent, 10, 10, 70, 20, NULL);
	lbls[0]->specific.label.set_font(lbls[0], BOLD_FONT);
	lbls[0]->specific.label.set_normal_color(lbls[0], "#336699");
	lbls[0]->specific.label.set_text(lbls[0], "Name: ");
	lbls[0]->specific.label.set_justification(lbls[0], JUSTIFY_LEFT);
	lbls[0]->redraw(lbls[0]);
	lbls[0]->show(lbls[0]);

	lbls[1] = new_widget(LABEL, parent, 10, 55, 70, 20, NULL);
	lbls[1]->specific.label.set_font(lbls[1], BOLD_FONT);
	lbls[1]->specific.label.set_normal_color(lbls[1], "#336699");
	lbls[1]->specific.label.set_text(lbls[1], "Address: ");
	lbls[1]->specific.label.set_justification(lbls[1], JUSTIFY_LEFT);
	lbls[1]->redraw(lbls[1]);
	lbls[1]->show(lbls[1]);

	lbls[2] = new_widget(LABEL, parent, 10, 125, 70, 20, NULL);
	lbls[2]->specific.label.set_font(lbls[2], BOLD_FONT);
	lbls[2]->specific.label.set_normal_color(lbls[2], "#336699");
	lbls[2]->specific.label.set_text(lbls[2], "Phone: ");
	lbls[2]->specific.label.set_justification(lbls[2], JUSTIFY_LEFT);
	lbls[2]->redraw(lbls[2]);
	lbls[2]->show(lbls[2]);
}

// Callback for the "Reset" button where the entry widget values are reset.
// A redraw() needs to be called for the changes to be visible.
Bool reset_entries(px_widget *wdgt, px_event *ev)
{
	entries[0]->specific.entry.set_text(entries[0], "");
	entries[0]->redraw(entries[0]);
	entries[1]->specific.entry.set_text(entries[1], "");
	entries[1]->redraw(entries[1]);
	entries[2]->specific.entry.set_text(entries[2], "");
	entries[2]->redraw(entries[2]);
	entries[3]->specific.entry.set_text(entries[3], "");
	entries[3]->redraw(entries[3]);

	return(False);
}

// Callback on the Quit button. Destroys the entry widgets and then the
// toplevel.
Bool quit_app(px_widget *wdgt, px_event *ev)
{
	entry_destroy(entries[0]);
	entry_destroy(entries[1]);
	entry_destroy(entries[2]);
	entry_destroy(entries[3]);

	toplvl->destroy(toplvl);
	return(False);
}

void create_buttons(px_widget *parent)
{
	resetbutt = new_widget(BUTTON,parent,40,230,70,30,"Reset");
	resetbutt->specific.button.set_font(resetbutt, BOLD_FONT);
	resetbutt->specific.button.set_text(resetbutt, "Reset");
	resetbutt->set_border_type(resetbutt, BEVELLED_BORDER);
	resetbutt->set_background_color(resetbutt, &col);
	resetbutt->specific.button.set_active_background_color(resetbutt,&acol);
	resetbutt->specific.button.set_active_font_color(resetbutt, "white");
	resetbutt->callbacks[BUTTON_PRESSED] = reset_entries;
	resetbutt->redraw(resetbutt);
	resetbutt->show(resetbutt);

	quitbutt = new_widget(BUTTON,parent,130,230,70,30,"Quit");
	quitbutt->specific.button.set_font(quitbutt, BOLD_FONT);
	quitbutt->specific.button.set_text(quitbutt, "Quit");
	quitbutt->set_border_type(quitbutt, BEVELLED_BORDER);
	quitbutt->set_background_color(quitbutt, &col);
	quitbutt->specific.button.set_active_background_color(quitbutt, &acol);
	quitbutt->specific.button.set_active_font_color(quitbutt, "white");
	quitbutt->callbacks[BUTTON_PRESSED] = quit_app;
	quitbutt->redraw(quitbutt);
	quitbutt->show(quitbutt);
}

int main(int av, char **ac)
{
	disp = display_get_display();
	
	if (disp->open_display(disp, (char *)getenv("DISPLAY"))) {
		col.type = TRANSPARENT;
		acol.type = SOLID;
		widget_xcolor_from_str(&acol.first_color, "black");

		// Create the toplevel window
		toplvl = new_widget(TOPLEVEL, NULL, 0, 0, 240, 270, NULL);
		toplvl->specific.toplevel.set_title(toplvl, "Example");
		toplvl->redraw(toplvl);
		toplvl->show(toplvl);

		// Create the widgets
		create_entries(toplvl);
		create_labels(toplvl);
		create_buttons(toplvl);

		// Get into eventloop
		eloop = eventloop_get_eventloop();
		eloop->loop(eloop);

		display_destroy();
	} else {
		fprintf(stderr, "Cannot open display.\n");
		exit(1);
	}
	exit(0);
}
