/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include<stdio.h>
#include<string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <pimglib.h>
#include <picowidget.h>
#include <display.h>
#include <eventloop.h>

#define NORMAL_FONT "-picopeta-simputer-medium-r-normal--11-*-*-*-p-*-fontspecific-0"
#define BOLD_FONT 	"-picopeta-simputer-bold-r-normal--11-*-*-*-p-*-fontspecific-0"

px_display *disp;
px_eventloop *eloop;
px_widget *toplvl;
px_widget *lbls[3];
px_widget *msglbl;
px_color col, acol, gcol, cols[3];

// Single callback for a click on the label widget. The y-coordinate
// will determine the label on which the click happened.
Bool lbl_clicked(px_widget *wdgt, px_event *ev)
{
	px_label *lbl=&(msglbl->specific.label);

	if (wdgt->y == 20) {
		lbl->set_text(msglbl, "You clicked on Left aligned label.");
		lbl->set_justification(msglbl, JUSTIFY_LEFT);
	} else if (wdgt->y == 60) {
		lbl->set_text(msglbl, "You clicked on Right aligned label.");
		lbl->set_justification(msglbl, JUSTIFY_RIGHT);
	} else if (wdgt->y == 100) {
		lbl->set_text(msglbl, "You clicked on Centered label.");
		lbl->set_justification(msglbl, JUSTIFY_CENTER);
	}
	msglbl->redraw(msglbl);
	return(False);
}

// Create 3 labels - left-aligned, right-aligned and centered labels.
// The same callback is associated for a click.
void create_labels(px_widget *parent)
{
	lbls[0] = new_widget(LABEL, parent, 20, 20, 200, 30, "Hello");
	lbls[0]->specific.label.set_font(lbls[0], NORMAL_FONT);
	lbls[0]->specific.label.set_text(lbls[0], "Left Aligned label");
	lbls[0]->specific.label.set_justification(lbls[0], JUSTIFY_LEFT);
	lbls[0]->set_background_color(lbls[0], &cols[0]);
	lbls[0]->callbacks[BUTTON_PRESSED] = lbl_clicked;
	lbls[0]->redraw(lbls[0]);
	lbls[0]->show(lbls[0]);

	lbls[1] = new_widget(LABEL, parent, 20, 60, 200, 30, "Hello");
	lbls[1]->specific.label.set_font(lbls[1], NORMAL_FONT);
	lbls[1]->specific.label.set_text(lbls[1], "Right Aligned");
	lbls[1]->specific.label.set_justification(lbls[1], JUSTIFY_RIGHT);
	lbls[1]->set_background_color(lbls[1], &cols[1]);
	lbls[1]->callbacks[BUTTON_PRESSED] = lbl_clicked;
	lbls[1]->redraw(lbls[1]);
	lbls[1]->show(lbls[1]);

	lbls[2] = new_widget(LABEL, parent, 20, 100, 200, 30, "Hello");
	lbls[2]->specific.label.set_font(lbls[2], NORMAL_FONT);
	lbls[2]->specific.label.set_text(lbls[2], "Centered label");
	lbls[2]->specific.label.set_justification(lbls[2], JUSTIFY_CENTER);
	lbls[2]->set_background_color(lbls[2], &cols[2]);
	lbls[2]->callbacks[BUTTON_PRESSED] = lbl_clicked;
	lbls[2]->redraw(lbls[2]);
	lbls[2]->show(lbls[2]);
}

// Colors are loaded for use in the label widgets.
void load_colors( void )
{
	cols[0].type = SOLID; cols[1].type = SOLID;
	cols[2].type = SOLID;
	widget_xcolor_from_str(&cols[0].first_color, "#6699bb");
	widget_xcolor_from_str(&cols[1].first_color, "#99bbee");
	widget_xcolor_from_str(&cols[2].first_color, "#bbee22");

	col.type = TRANSPARENT;
	acol.type = SOLID;
	gcol.type = SOLID;
	widget_xcolor_from_str(&acol.first_color, "black");
	widget_xcolor_from_str(&gcol.first_color, "grey");
}

int main(int av, char **ac)
{
	disp = display_get_display();
	
	if (disp->open_display(disp, (char *)getenv("DISPLAY"))) {
		load_colors();

		// Create the toplevel window
		toplvl = new_widget(TOPLEVEL, NULL, 0, 0, 240, 270, NULL);
		toplvl->specific.toplevel.set_title(toplvl, "Label Example");
		toplvl->redraw(toplvl);
		toplvl->show(toplvl);

		// Add labels
		create_labels(toplvl);

		// Create the message label
		msglbl = new_widget(LABEL, toplvl, 10, 180, 220, 60, NULL);
		label_set_normal_color(msglbl, "#336699");
		msglbl->specific.label.set_font(msglbl, BOLD_FONT);
		msglbl->specific.label.set_text(msglbl, "Message Area.");
		msglbl->specific.label.set_wrap_type(msglbl, WRAP_WORD);
		widget_set_background_color(msglbl, &gcol);
		msglbl->redraw(msglbl);
		msglbl->show(msglbl);

		// Get into eventloop
		eloop = eventloop_get_eventloop();
		eloop->loop(eloop);

		display_destroy();
	} else {
		fprintf(stderr, "Cannot open display.\n");
		exit(1);
	}
	exit(0);
}
