/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include<stdio.h>
#include<string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <pimglib.h>
#include <picowidget.h>
#include <display.h>
#include <eventloop.h>
#include <acctrl.h>

#define NORMAL_FONT "-picopeta-simputer-medium-r-normal--11-*-*-*-p-*-fontspecific-0"
#define BOLD_FONT 	"-picopeta-simputer-bold-r-normal--11-*-*-*-p-*-fontspecific-0"

enum _screen {
	MAIN_SCREEN,
	LABEL_SCREEN,
	RADIO_SCREEN,
	CHECK_SCREEN
};
px_display *display;
px_eventloop *eloop;
px_widget *toplevel;
int screen;

///////////[Label Page]
px_widget *labelpage_frame, *label[2];

void create_labelpage_screen()
{
	labelpage_frame = new_widget(FRAME, toplevel, 0, 0, 240, 270, "");
	labelpage_frame->redraw(labelpage_frame);
	labelpage_frame->show(labelpage_frame);

	// Add labels
	label[0] = new_widget(LABEL, labelpage_frame, 10, 40, 220, 40, "");
	label[0]->specific.label.set_font(label[0], BOLD_FONT);
	label[0]->specific.label.set_justification(label[0], JUSTIFY_CENTER);
	label[0]->specific.label.set_text(label[0], "Label 1");
	label[0]->redraw(label[0]);
	label[0]->show(label[0]);

	label[1] = new_widget(LABEL, labelpage_frame, 10, 120, 220, 40, "");
	label[1]->specific.label.set_font(label[1], BOLD_FONT);
	label[1]->specific.label.set_justification(label[1], JUSTIFY_CENTER);
	label[1]->specific.label.set_text(label[1], "Label 2");
	label[1]->redraw(label[1]);
	label[1]->show(label[1]);
}

void show_labelpage_screen()
{
	if(!labelpage_frame)
		create_labelpage_screen();

	toplevel->specific.toplevel.set_title(toplevel, "LabelPage");
	screen = LABEL_SCREEN;

	labelpage_frame->show(labelpage_frame);

	ac_control_win(AC_CMD_SHOW_UP , "", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_SET_ANNOTATE_KEY , "label", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_SHOW_PAGE_NAV , "", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_DISABLE_PAGE_PREV , "", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_ENABLE_PAGE_NEXT , "", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_SET_PAGE_NUM , "1", display_get_Xdisplay(), toplevel->win);
}

void hide_labelpage_screen()
{
	labelpage_frame->hide(labelpage_frame);
}

///////////[RadioButton Page]
px_widget *radiopage_frame, *radio[2];

void create_radiopage_screen()
{
	radiopage_frame = new_widget(FRAME, toplevel, 0, 0, 240, 270, "");
	radiopage_frame->redraw(radiopage_frame);
	radiopage_frame->show(radiopage_frame);

	// Add labels
	radio[0] = new_widget(RADIOBUTTON, radiopage_frame, 10, 40, 220, 40, "");
	radio[0]->specific.radiobutton.set_font(radio[0], BOLD_FONT);
	radio[0]->specific.radiobutton.set_justification(radio[0], JUSTIFY_LEFT);
	radio[0]->specific.radiobutton.set_text(radio[0], "RadioButton 1");
	radio[0]->specific.radiobutton.select(radio[0]);
	radio[0]->redraw(radio[0]);
	radio[0]->show(radio[0]);

	radio[1] = new_widget(RADIOBUTTON, radiopage_frame, 10, 120, 220, 40, "");
	radio[1]->specific.radiobutton.set_font(radio[1], BOLD_FONT);
	radio[1]->specific.radiobutton.set_justification(radio[1], JUSTIFY_LEFT);
	radio[1]->specific.radiobutton.set_text(radio[1], "RadioButton 2");
	radio[1]->specific.radiobutton.group_together(radio[1], radio[0]);
	radio[1]->redraw(radio[1]);
	radio[1]->show(radio[1]);
}

void show_radiopage_screen()
{
	if(!radiopage_frame)
		create_radiopage_screen();

	toplevel->specific.toplevel.set_title(toplevel, "RadioButtonPage");
	screen = RADIO_SCREEN;

	radiopage_frame->show(radiopage_frame);

	ac_control_win(AC_CMD_SHOW_UP , "", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_SET_ANNOTATE_KEY , "radio", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_SHOW_PAGE_NAV , "", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_ENABLE_PAGE_PREV , "", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_ENABLE_PAGE_NEXT , "", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_SET_PAGE_NUM , "2", display_get_Xdisplay(), toplevel->win);
}

void hide_radiopage_screen()
{
	radiopage_frame->hide(radiopage_frame);
}

///////////[CheckButton Page]
px_widget *checkpage_frame, *check[2];

void create_checkpage_screen()
{
	checkpage_frame = new_widget(FRAME, toplevel, 0, 0, 240, 270, "");
	checkpage_frame->redraw(checkpage_frame);
	checkpage_frame->show(checkpage_frame);

	// Add labels
	check[0] = new_widget(CHECKBUTTON, checkpage_frame, 10, 40, 220, 40, "");
	check[0]->specific.checkbutton.set_font(check[0], BOLD_FONT);
	check[0]->specific.checkbutton.set_justification(check[0], JUSTIFY_LEFT);
	check[0]->specific.checkbutton.set_text(check[0], "CheckButton 1");
	check[0]->redraw(check[0]);
	check[0]->show(check[0]);

	check[1] = new_widget(CHECKBUTTON, checkpage_frame, 10, 120, 220, 40, "");
	check[1]->specific.checkbutton.set_font(check[1], BOLD_FONT);
	check[1]->specific.checkbutton.set_justification(check[1], JUSTIFY_LEFT);
	check[1]->specific.checkbutton.set_text(check[1], "CheckButton 2");
	check[1]->redraw(check[1]);
	check[1]->show(check[1]);
}

void show_checkpage_screen()
{
	if(!checkpage_frame)
		create_checkpage_screen();

	toplevel->specific.toplevel.set_title(toplevel, "CheckButtonPage");
	screen = CHECK_SCREEN;

	checkpage_frame->show(checkpage_frame);

	ac_control_win(AC_CMD_SHOW_TICK , "", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_SET_ANNOTATE_KEY , "check", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_SHOW_PAGE_NAV , "", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_ENABLE_PAGE_PREV , "", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_DISABLE_PAGE_NEXT , "", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_SET_PAGE_NUM , "3", display_get_Xdisplay(), toplevel->win);
}

void hide_checkpage_screen()
{
	checkpage_frame->hide(checkpage_frame);
}

///////////[Main Page]
px_widget *mainpage_frame, *button[3];
Bool mainpage_button_callback(px_widget *widget, px_event *event);

void create_mainpage_screen()
{
	int i;
	char num[5], *bname[4] = {"Label Page", "RadioButton Page", "CheckButton Page"};
	px_color col, acol;

	mainpage_frame = new_widget(FRAME, toplevel, 0, 0, 240, 270, "");
	mainpage_frame->redraw(mainpage_frame);
	mainpage_frame->show(mainpage_frame);

	col.type = TRANSPARENT;
	acol.type = SOLID;
	widget_xcolor_from_str(&acol.first_color, "black");

	// Add buttons
	for(i=0; i<3; i++) {
		sprintf(num, "%d", i);
		button[i] = new_widget(BUTTON, toplevel, 20, 30 + i*60 +i*20, 200, 40, num);
		button[i]->specific.button.set_font(button[i], BOLD_FONT);
		button[i]->specific.button.set_text(button[i], bname[i]);
		button[i]->set_border_type(button[i], BEVELLED_BORDER);
		button[i]->set_background_color(button[i], &col);
		button[i]->specific.button.set_active_background_color(button[i], &acol);
		button[i]->specific.button.set_active_font_color(button[i], "white");
		button[i]->callbacks[BUTTON_RELEASED] = mainpage_button_callback;
		button[i]->redraw(button[i]);
		button[i]->show(button[i]);
	}
}

void show_mainpage_screen()
{
	if(!mainpage_frame)
		create_mainpage_screen();

	toplevel->specific.toplevel.set_title(toplevel, "PageNavDemo");
	screen = MAIN_SCREEN;

	mainpage_frame->show(mainpage_frame);

	ac_control_win(AC_CMD_SHOW_CLOSE , "", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_SET_ANNOTATE_KEY , "main", display_get_Xdisplay(), toplevel->win);
	ac_control_win(AC_CMD_HIDE_PAGE_NAV , "", display_get_Xdisplay(), toplevel->win);
}

void hide_mainpage_screen()
{
	mainpage_frame->hide(mainpage_frame);
}

Bool mainpage_button_callback(px_widget *widget, px_event *event)
{

	hide_mainpage_screen();
	
	switch(atoi(widget->tag)) {
		case 0:
			show_labelpage_screen();
			break;
		case 1:
			show_radiopage_screen();
			break;
		case 2:
			show_checkpage_screen();
			break;
	}
	
	return False;
}

////////////////[Main Function]
Bool toplevel_client_message(px_widget *widget, px_event *event)
{
	long atom = event->xev->xclient.data.l[0];
     
     if(atom==display_get_atom(display, "WM_DELETE_WINDOW")) {
		printf("Cleanup any stuff, if required.\n");
		return False;
	}

     if(atom==display_get_atom(display, "_AC_BUTTON_UP_REL")) {
		if(screen==LABEL_SCREEN)
			hide_labelpage_screen();
		if(screen==RADIO_SCREEN)
			hide_radiopage_screen();
	
		show_mainpage_screen();
		
		return False;
	}

     if(atom==display_get_atom(display, "_AC_BUTTON_TICK_REL")) {
		if(screen==CHECK_SCREEN)
			hide_checkpage_screen();
	
		show_mainpage_screen();
		
		return False;
	}
	
     if(atom==display_get_atom(display, "_AC_BUTTON_PREVPAGE_REL")) {
		if(screen==RADIO_SCREEN) {
			show_labelpage_screen();
			hide_radiopage_screen();
		} else if(screen==CHECK_SCREEN) {
			show_radiopage_screen();
			hide_checkpage_screen();
		}

		return False;
	}

     if(atom==display_get_atom(display, "_AC_BUTTON_NEXTPAGE_REL")) {
		if(screen==LABEL_SCREEN) {
			show_radiopage_screen();
			hide_labelpage_screen();
		} else if(screen==RADIO_SCREEN) {
			show_checkpage_screen();
			hide_radiopage_screen();
		}

		return False;
	}

	return False;
}

int main(int av, char **ac)
{
	int i;
	char mesg[50];
	px_widget *frame, *label;
	display = display_get_display();
	
	if(display->open_display(display, (char *)getenv("DISPLAY"))) {
		// Create the toplevel window
		toplevel = new_widget(TOPLEVEL, NULL, 0, 0, 240, 270, NULL);
		ac_init_app(display_get_Xdisplay(), toplevel->win);
		toplevel->specific.toplevel.set_title(toplevel, "PageNavDemo");
		toplevel->callbacks[CLIENT_MESSAGE] = toplevel_client_message;
		toplevel->redraw(toplevel);
		toplevel->show(toplevel);

		// Show the main page, containing buttons.
		show_mainpage_screen();

		// Register window protocols.
	   	display_register_wm_protocol(display, "WM_DELETE_WINDOW");
     	display_register_wm_protocol(display, "_AC_BUTTON_UP_REL");
	     display_register_wm_protocol(display, "_AC_BUTTON_TICK_REL");
          display_register_wm_protocol(display, "_AC_BUTTON_NEXTPAGE_REL");
          display_register_wm_protocol(display, "_AC_BUTTON_PREVPAGE_REL");

		// Get into eventloop
		eloop = eventloop_get_eventloop();
		eloop->loop(eloop);

		display_destroy();
	} else {
		fprintf(stderr, "Cannot open display.\n");
		exit(1);
	}
	exit(0);
}
