/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

/*
 *  Sketch example:
 *    Creates a name and signature sketch widget on which the user can
 *    scribble her/his name and signature. Two buttons are attached to
 *    the sketch widgets to clear their respective contents. A save button
 *    is provided to temporarily save the name and signature.
 *
 *    Note, the routines for loading and saving sketches. Also, each sketch
 *    widget can be associated with a different pen color and width.
 */
#include <stdio.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <pimglib.h>
#include <picowidget.h>
#include <display.h>
#include <eventloop.h>

#define NORMAL_FONT "-picopeta-simputer-medium-r-normal--11-*-*-*-p-*-fontspecific-0"
#define BOLD_FONT "-picopeta-simputer-bold-r-normal--11-*-*-*-p-*-fontspecific-0"

px_display *disp;
px_eventloop *eloop;
px_widget *toplvl;
px_widget *sigsktch, *nmsktch;
px_widget *sigclr, *nmclr, *okbutt;
px_widget *lbl1, *lbl2;
px_color col, acol, gcol;
border_types btype;

// Creates the sketch widgets for name and signature.
void create_sketch(px_widget *parent)
{
	nmsktch = new_widget(SKETCH, parent, 20, 20, 200, 40, "Hello");
	nmsktch->specific.sketch.load_from_file(nmsktch, "/tmp/name");
	nmsktch->specific.sketch.set_bg_color(nmsktch, 0xe8e8e8);
	nmsktch->specific.sketch.set_pen_width(nmsktch, 1);
	nmsktch->redraw(nmsktch);
	nmsktch->show(nmsktch);

	sigsktch = new_widget(SKETCH, parent, 20, 100, 200, 40, "Hello");
	sigsktch->specific.sketch.load_from_file(sigsktch, "/tmp/signature");
	sigsktch->specific.sketch.set_pen_width(sigsktch, 2);
	sigsktch->specific.sketch.set_pen_color(sigsktch, 0x0000ff);
	sigsktch->redraw(sigsktch);
	sigsktch->show(sigsktch);
}

// Callback for the clear button attached to the signature
Bool clr1_callback(px_widget *wdgt, px_event *ev)
{
	sigsktch->specific.sketch.clear_sketches(sigsktch);
	sigsktch->redraw(sigsktch);
}

// Callback for the clear button attached to the name
Bool clr2_callback(px_widget *wdgt, px_event *ev)
{
	nmsktch->specific.sketch.clear_sketches(nmsktch);
	nmsktch->redraw(nmsktch);
}

// Saves the signature temporarily to files in /tmp
Bool save_sketches(px_widget *wdgt, px_event *ev)
{
	nmsktch->specific.sketch.save_to_file(nmsktch, "/tmp/name");
	sigsktch->specific.sketch.save_to_file(sigsktch, "/tmp/signature");
}

// Creates the label widgets for Name and Signature
void create_labels(px_widget *parent)
{
	lbl1 = new_widget(LABEL, parent, 10, 5, 75, 15, NULL);
	lbl1->specific.label.set_font(lbl1, NORMAL_FONT);
	lbl1->specific.label.set_text(lbl1, "Name:");
	lbl1->specific.label.set_justification(lbl1, JUSTIFY_LEFT);
	lbl1->redraw(lbl1);
	lbl1->show(lbl1);

	lbl2 = new_widget(LABEL, parent, 10, 85, 75, 15, NULL);
	lbl2->specific.label.set_font(lbl2, NORMAL_FONT);
	lbl2->specific.label.set_text(lbl2, "Signature:");
	lbl2->specific.label.set_justification(lbl2, JUSTIFY_LEFT);
	lbl2->redraw(lbl2);
	lbl2->show(lbl2);
}

// Creates the clear buttons for name and signature. Associates callbacks.
void create_buttons(px_widget *parent)
{
	nmclr = new_widget(BUTTON,parent,180,65,40,20,"Clear");
	nmclr->specific.button.set_font(nmclr, NORMAL_FONT);
	nmclr->specific.button.set_text(nmclr, "Clear");
	nmclr->set_border_type(nmclr, BEVELLED_BORDER);
	nmclr->set_background_color(nmclr, &col);
	nmclr->specific.button.set_active_background_color(nmclr, &acol);
	nmclr->specific.button.set_active_font_color(nmclr, "white");
	nmclr->callbacks[BUTTON_PRESSED] = clr2_callback;
	nmclr->redraw(nmclr);
	nmclr->show(nmclr);

	sigclr = new_widget(BUTTON,parent,180,145,40,20,"Clear");
	sigclr->specific.button.set_font(sigclr, NORMAL_FONT);
	sigclr->specific.button.set_text(sigclr, "Clear");
	sigclr->set_border_type(sigclr, BEVELLED_BORDER);
	sigclr->set_background_color(sigclr, &col);
	sigclr->specific.button.set_active_background_color(sigclr, &acol);
	sigclr->specific.button.set_active_font_color(sigclr, "white");
	sigclr->callbacks[BUTTON_PRESSED] = clr1_callback;
	sigclr->redraw(sigclr);
	sigclr->show(sigclr);

	okbutt = new_widget(BUTTON,parent,80,180,80,30,"Save");
	okbutt->specific.button.set_font(okbutt, BOLD_FONT);
	okbutt->specific.button.set_text(okbutt, "Save");
	okbutt->set_border_type(okbutt, BEVELLED_BORDER);
	okbutt->set_background_color(okbutt, &col);
	okbutt->specific.button.set_active_background_color(okbutt, &acol);
	okbutt->specific.button.set_active_font_color(okbutt, "white");
	okbutt->callbacks[BUTTON_PRESSED] = save_sketches;
	okbutt->redraw(okbutt);
	okbutt->show(okbutt);
}

int main(int av, char **ac)
{
	disp = display_get_display();
	
	if (disp->open_display(disp, (char *)getenv("DISPLAY"))) {
		col.type = TRANSPARENT;
		acol.type = SOLID;
		widget_xcolor_from_str(&acol.first_color, "black");

		// Create the toplevel window
		toplvl = new_widget(TOPLEVEL, NULL, 0, 0, 240, 270, NULL);
		toplvl->specific.toplevel.set_title(toplvl, "Example");
		toplvl->redraw(toplvl);
		toplvl->show(toplvl);

		// Add buttons
		create_sketch(toplvl);
		create_labels(toplvl);
		create_buttons(toplvl);

		// Get into eventloop
		eloop = eventloop_get_eventloop();
		eloop->loop(eloop);

		display_destroy();
	} else {
		fprintf(stderr, "Cannot open display.\n");
		exit(1);
	}
	exit(0);
}
