/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

/*
 * Tab example:
 *   Creates a container tab with 3 tabs. The first two tabs have a button
 *   each. The button in Tab 1 and the tab both have callbacks associated.
 *   The button in Tab 2 has a callback.
 *
 *   Note that, one can start with a particular tab raised. Also, when a
 *   tab is selected, a callback can be associated to initialize the
 *   contents of the tab.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/extensions/XTest.h>
#include <X11/keysym.h>
#include <X11/extensions/shape.h>
#include <X11/xpm.h>

#include <picowidget.h>
#include <display.h>
#include <eventloop.h>

#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <acctrl.h>

#define BOLD_FONT "-picopeta-simputer-bold-r-normal--11-*-*-*-p-*-fontspecific-0"

px_eventloop* eventloop;
px_widget *toplvl;
px_widget *cframe;	// Container frame
px_widget *ctab;	// Container tab
px_widget *tabs[3];	// Child tabs
px_widget *butt[3];	// Buttons in tabs
px_color black_col, white_col;

// Tab and button-in-tab callbacks.
Bool tab1_cb(px_widget *wdgt, px_event *ev)
{
	printf("Tab 1 brought on top\n");
	return(False);
}

Bool tab2_cb(px_widget *wdgt, px_event *ev)
{
	printf("Tab 2 brought on top\n");
	return(False);
}

Bool tab3_cb(px_widget *wdgt, px_event *ev)
{
	printf("Tab 3 brought on top\n");
	return(False);
}

Bool tab1_butt_cb(px_widget *wdgt, px_event *ev)
{
	printf("Button in Tab 1 clicked\n");
	return False;
}

Bool tab1_click_cb(px_widget *wdgt, px_event *ev)
{
	printf("Clicked in Tab 1\n");
	return(False);
}

Bool tab2_butt_cb(px_widget *wdgt, px_event *ev)
{
	printf("Clicked on button in Tab 2\n");
	return(False);
}

int main(int argc, char **argv)
{
	char *display_name = (char *)getenv("DISPLAY");
	px_display* display = get_display();

	if (display->open_display(display, display_name)) {
		white_col.type = SOLID;
		widget_xcolor_from_str(&(white_col.first_color), "white");
			
		black_col.type = SOLID;
		widget_xcolor_from_str(&(black_col.first_color), "black");
     
		toplvl = new_widget(TOPLEVEL, NULL,  0, 0, 240, 270, NULL);
		toplvl->specific.toplevel.set_title(toplvl, "Testing");

		// Create the container frame
		cframe = new_widget(FRAME, toplvl, 0, 0, 240, 270, "child");
		cframe->redraw(cframe);
		cframe->show(cframe);

		// Create the container 'parent' tab
		ctab = new_widget(TAB, cframe, 0, 0, 240, 270, "Tabs");
		ctab->show(ctab);
		ctab->redraw(ctab);

		tabs[0] = ctab->specific.tab.add_tab(ctab,"Tab 1",tab1_cb);
		tabs[1] = ctab->specific.tab.add_tab(ctab,"Tab 2",tab2_cb);
		tabs[2] = ctab->specific.tab.add_tab(ctab,"Tab 3",tab3_cb);

// NOTE: no need to do a redraw and show for the individual tabs

		// Associate a callback for a click on the Tab 1 main area
		tabs[0]->callbacks[BUTTON_RELEASED] = tab1_click_cb;

		butt[0] = new_widget(BUTTON, tabs[0], 30, 40, 80, 30, NULL);
		butt[0]->specific.button.set_font(butt[0], BOLD_FONT);
		butt[0]->callbacks[BUTTON_RELEASED] = tab1_butt_cb;
		butt[0]->set_border_type(butt[0], BEVELLED_BORDER);
		butt[0]->specific.button.set_text(butt[0], "Click Me!");
		butt[0]->set_background_color(butt[0], &white_col);
		butt[0]->specific.button.set_active_background_color(butt[0], &black_col);
		butt[0]->specific.button.set_active_font_color(butt[0],"white");
		butt[0]->redraw(butt[0]);
		butt[0]->show(butt[0]);

		butt[1] = new_widget(BUTTON, tabs[1], 50, 70, 120, 30, NULL);
		butt[1]->specific.button.set_font(butt[1], BOLD_FONT);
		butt[1]->callbacks[BUTTON_RELEASED] = tab2_butt_cb;
		butt[1]->set_border_type(butt[1], BEVELLED_BORDER);
		butt[1]->specific.button.set_text(butt[1], "Another Button");
		butt[1]->specific.button.set_active_background_color(butt[1], &black_col);
		butt[1]->specific.button.set_active_font_color(butt[1],"white");
		butt[1]->redraw(butt[1]);
		butt[1]->show(butt[1]);

		ctab->specific.tab.raise_tab(ctab, 1);
// NOTE: no need to do a redraw after a raise_tab()

		toplvl->redraw(toplvl);
		toplvl->show(toplvl);
        
		eventloop = eventloop_get_eventloop();
		eventloop->loop(eventloop);
	}
}
