/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

/*
 *  Text listbox example:
 *    Creates a simple text listbox with 3 columns. To start with the listbox
 *    has one column with the width of the listbox. When a column is added,
 *    this width is subtracted from the first column's width. Every subsequent
 *    column uses the first column's width to make itself space.
 *
 *    Note, the routines for getting the currently selected row and element,
 *    adding elements to the columns, removing rows/elements, etc.
 *    The text listbox allows the user to drag on it to change the selection.
 *    Auto-scroll happens when the user drags out of the listbox boundaries.
 *
 *    Multiple selection listbox allows multiple rows to be simultaneously
 *    selected. The type of a text listbox can be changed on the fly. The
 *    list of selected rows can be retrieved via txtlb_get_all_selecn().
 *    Note that, the integer array of selected row numbers needs to be freed
 *    after usage.
 */
#include <stdio.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <pimglib.h>
#include <picowidget.h>
#include <display.h>
#include <eventloop.h>

#define NORMAL_FONT "-picopeta-simputer-medium-r-normal--11-*-*-*-p-*-fontspecific-0"

px_display *disp;
px_eventloop *eloop;
px_widget *toplvl, *lbox;
px_widget *typeButt, *getButt;
px_color col, acol, gcol;
Bool multi_flg=False;

// Callback when a selection in the listbox has changed. Note that, 
// txtlb_get_element allocates a char string; which needs to be freed.
void lb_selecn_change(px_widget *wdgt, int idx)
{
	char *str;

	txtlb_get_element(wdgt, &str, idx, 0);
	printf("selecn changed to %s in column one\n", str);
	free(str);
}

// Button callback to illustrate the use of removing rows, repopulating
// the listbox entries and toggling the type of the listbox.
Bool change_listbox_type(px_widget *wdgt, px_event *ev)
{
	int i;
	char str[32];

	i = txtlb_get_no_of_rows(lbox);
	for (; i > 0; i--)
		lbox->specific.txtlb.remove_lb_row(lbox, 0, False);
	for (i=0; i < 100; i++) {
		sprintf(str, "%d-th row", i);
		lbox->specific.txtlb.add_lb_elem(lbox, str, 0);
		sprintf(str, "[%d,1]", i);
		lbox->specific.txtlb.add_lb_elem(lbox, str, 1);
		sprintf(str, "[%d,2]", i);
		lbox->specific.txtlb.add_lb_elem(lbox, str, 2);
	}
	if (multi_flg == False) {
		txtlb_set_type(lbox, TXTLB_MULTI_SELECN);
		typeButt->specific.button.set_text(typeButt, "Single Selection");
		typeButt->redraw(typeButt);
		multi_flg = True;
	} else {
		txtlb_set_type(lbox, TXTLB_SINGLE_SELECN);
		typeButt->specific.button.set_text(typeButt, "Multi Selection");
		typeButt->redraw(typeButt);
		multi_flg = False;
	}
	lbox->redraw(lbox);
	return(False);
}

// Callback to get the selected listbox elements.
Bool get_selection(px_widget *wdgt, px_event *ev)
{
	if (multi_flg == True) {
		int *arr;
		int sel_count, i;

		sel_count = txtlb_get_all_selecn(lbox, &arr);
		printf("Selected elements:- ");
		for (i=0; i < sel_count; i++)
			printf(" %d", arr[i]);
		printf("\n");
		free(arr);
	} else {
		int selecn;

		selecn = txtlb_get_selecn(lbox);
		printf("%d-th row is selected\n", selecn);
	}
	return(False);
}

// Create the type and get selection buttons.
void create_button(px_widget *par)
{
	col.type = TRANSPARENT;
	acol.type = SOLID;
	widget_xcolor_from_str(&acol.first_color, "black");

	typeButt = new_widget(BUTTON, par, 10, 235, 110, 25, "Hello");
	typeButt->specific.button.set_font(typeButt, NORMAL_FONT);
	typeButt->specific.button.set_text(typeButt, "Multi Selection");
	typeButt->set_border_type(typeButt, BEVELLED_BORDER);
	typeButt->set_background_color(typeButt, &col);
	typeButt->specific.button.set_active_background_color(typeButt, &acol);
	typeButt->specific.button.set_active_font_color(typeButt, "white");
	typeButt->callbacks[BUTTON_RELEASED] = change_listbox_type;
	typeButt->redraw(typeButt);
	typeButt->show(typeButt);

	getButt = new_widget(BUTTON, par, 130, 235, 100, 25, "Hello");
	getButt->specific.button.set_font(getButt, NORMAL_FONT);
	getButt->specific.button.set_text(getButt, "Get Selection");
	getButt->set_border_type(getButt, BEVELLED_BORDER);
	getButt->set_background_color(getButt, &col);
	getButt->specific.button.set_active_background_color(getButt, &acol);
	getButt->specific.button.set_active_font_color(getButt, "white");
	getButt->callbacks[BUTTON_RELEASED] = get_selection;
	getButt->redraw(getButt);
	getButt->show(getButt);
}

int main(int av, char **ac)
{
	int i;
	char str[32];

	disp = display_get_display();
	
	if(disp->open_display(disp, (char *)getenv("DISPLAY"))) {
		// Create the toplevel window
		toplvl = new_widget(TOPLEVEL,NULL,0,0,240,270,"TOPLEVEL_TAG");
		toplvl->specific.toplevel.set_title(toplvl, "Example");
		toplvl->redraw(toplvl);
		toplvl->show(toplvl);

		// Create the text listbox and add two columns
		lbox = new_widget(TXTLB, toplvl, 10, 10, 220, 220, NULL);

		lbox->specific.txtlb.add_lb_column(lbox, 60);
		lbox->specific.txtlb.add_lb_column(lbox, 60);
		
		// Initially populate the listbox
		for (i=0; i < 100; i++) {
			sprintf(str, "%d-th row", i);
//			lbox->specific.txtlb.add_lb_elem(lbox, str, 0);
			lbox->specific.txtlb.insert_lb_elem(lbox, str, 0, i);
			sprintf(str, "[%d,1]", i);
//			lbox->specific.txtlb.add_lb_elem(lbox, str, 1);
			lbox->specific.txtlb.insert_lb_elem(lbox, str, 1, i);
			sprintf(str, "[%d,2]", i);
//			lbox->specific.txtlb.add_lb_elem(lbox, str, 2);
			lbox->specific.txtlb.insert_lb_elem(lbox, str, 2, i);
		}
//	Makes a plain listbox
//		lbox->specific.txtlb.change_lb_color(lbox,TXTLB_BACKGROUND,"#ffffff");
//		lbox->specific.txtlb.change_lb_color(lbox,TXTLB_ROW_BACKGROUND,"#ffffff");
//		lbox->specific.txtlb.change_lb_color(lbox,TXTLB_TROUGH,"#ffffff");
//		lbox->specific.txtlb.change_lb_color(lbox,TXTLB_SELECN_HIGHLIGHT,"#000000");
		lbox->redraw(lbox);
		lbox->show(lbox);
		txtlb_set_intimate_change(lbox, lb_selecn_change);

		create_button(toplvl);

		// Get into eventloop
		eloop = eventloop_get_eventloop();
		eloop->loop(eloop);

		display_destroy();
	} else {
		fprintf(stderr, "Cannot open display.\n");
		exit(1);
	}
	exit(0);
}
