/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include <X11/extensions/shape.h>
#include <X11/xpm.h>
#include <stdlib.h>
#include <string.h>
#include "button.h"
#include "image.h"
#include "font.h"
#include "eventloop.h"
#include "widget_factory.h"
#include "display.h"
#include "theme.h"

px_widget* button_create(px_widget* parent, int x, int y, int width, int height, const char* tag)
{
	px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
	Display* display = display_get_Xdisplay();
	
	widget->win = XCreateWindow(display, parent->win, x, y, width, height, 0, CopyFromParent, CopyFromParent, CopyFromParent, 0, NULL);

#ifdef __debug__
	printf("%s(): %p - Created Window with parent window = %lu, x = %d, y = %d, width = %d, height = %d, new window = %lu\n\n", 
			__func__, widget, parent->win, x, y, width, height, widget->win);

	XSync(display, False);
#endif
	// Initialize common widget variables
	widget->x = x;
	widget->y = y;
	widget->height = height;
	widget->width = width;

	if (tag != NULL)
	{
		widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	}
	else
		widget->tag = NULL;
	widget->wtype = BUTTON;
	widget->child_of = NULL;

	widget->use_pixmap = False;
	widget->background.background_color = (px_color*)malloc(sizeof(px_color));
	widget->background.background_color->type = SOLID;
	widget_xcolor_from_str(&(widget->background.background_color->first_color), "white");

	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);

	return widget;
}

void button_set_parameters(px_widget* widget)
{
	px_button* button = &(widget->specific.button);
	
	widget->default_callbacks[BUTTON_PRESSED] = button_default_button_press;
	widget->default_callbacks[BUTTON_RELEASED] = button_default_button_released;
	widget->default_callbacks[MOUSE_ENTER] = button_default_mouse_enter;
	widget->default_callbacks[MOUSE_LEAVE] = button_default_mouse_leave;

	widget->set_size = button_set_size;
	widget->redraw = button_redraw;
	widget->destroy = button_destroy;
	
	// Initialize widget specific variables

	button->layout.wrap = WRAP_CHAR;
	button->layout.just = JUSTIFY_CENTER;
	button->layout.text = NULL;
	button->layout.x = 0;
	button->layout.y = 0;
	button->layout.width = widget->width;
	button->layout.height = widget->height;
	button->layout.font = NULL;
	button->layout.font_color = NULL;
	button->layout.chunks = NULL;
	button->layout.no_of_chunks = 0;

	button->fontname = NULL;

	button->auto_repeat = False;
	button->delay = 170;
	button->speed = 80;
	button->auto_repeat_timer = NULL;

	button->normal_font = NULL;
	button->active_font = NULL;
	button->disabled_font = NULL;
	
	button->active_color = NULL;
	button->disabled_color = NULL;
	
	button->normal = NULL;
	button->active = NULL;
	button->disabled = NULL;

	button->active_border_color = NULL;
	button->disabled_border_color = NULL;

	button->change_size = False;

	button->set_normal_pixmap_from_file = button_set_normal_pixmap_from_file;
	button->set_active_pixmap_from_file = button_set_active_pixmap_from_file;
	button->set_disabled_pixmap_from_file = button_set_disabled_pixmap_from_file;
	
	button->set_normal_pixmap_from_pixmap = button_set_normal_pixmap_from_pixmap;
	button->set_active_pixmap_from_pixmap = button_set_active_pixmap_from_pixmap;
	button->set_disabled_pixmap_from_pixmap = button_set_disabled_pixmap_from_pixmap;

	button->set_text = button_set_text;
	button->set_justification = button_set_justification;
	button->set_font = button_set_font;

	button->get_justification = button_get_justification;
	button->get_text = button_get_text;

	button->set_auto_repeat = button_set_auto_repeat;
	button->set_auto_repeat_delay = button_set_auto_repeat_delay;
	button->set_auto_repeat_speed = button_set_auto_repeat_speed;

	button->set_normal_font_color = button_set_normal_font_color;
	button->set_active_font_color = button_set_active_font_color;
	button->set_disabled_font_color = button_set_disabled_font_color;
	
	button->set_active_background_color = button_set_active_background_color;
	button->set_disabled_background_color = button_set_disabled_background_color;

	button->set_active_border_color = button_set_active_border_color;
	button->set_disabled_border_color = button_set_disabled_border_color;
	button->set_wrap_type = widget_set_wrap_type;
	button->size_depends_on_contents = widget_size_depends_on_contents;

#if 0
	// Santhosh
	button_set_font(widget, "-picopeta-simputer-medium-r-normal--11-*-*-*-p-*-fontspecific-0");
	{
		px_theme* theme = theme_get_theme();
		int ret;
		char *imgpath;
		struct Image *bimg;
		
		imgpath = (char*)malloc(strlen(theme->theme_path)+25);
		
		sprintf(imgpath, "%s/button_normal", theme->theme_path);
		ret = pimg_load(imgpath, display_get_Xdisplay(), &bimg, 0);
		if(ret>0) {
			struct Image *zimg;
			px_pixmap pxmp;

			pimg_zoom(display_get_Xdisplay(), &bimg, &zimg, 
						(float)widget->width/(float)bimg->width, (float)widget->height/(float)bimg->height,
						ZOOM_LOW_QUALITY);

			image_initialize_pixmap(&pxmp);
			image_create_pixmap(widget->win, &pxmp, zimg->width, zimg->height);
			image_copy_image(widget->win, &pxmp, zimg);

			button_set_normal_pixmap_from_pixmap(widget, &pxmp);

			pimg_free(&zimg);
			pimg_free(&bimg);
			image_destroy(&pxmp);
		}
		
		sprintf(imgpath, "%s/button_active", theme->theme_path);
		ret = pimg_load(imgpath, display_get_Xdisplay(), &bimg, 0);
		if(ret>0) {
			struct Image *zimg;
			px_pixmap pxmp;

			pimg_zoom(display_get_Xdisplay(), &bimg, &zimg, 
						(float)widget->width/(float)bimg->width, (float)widget->height/(float)bimg->height,
						ZOOM_LOW_QUALITY);

			image_initialize_pixmap(&pxmp);
			image_create_pixmap(widget->win, &pxmp, zimg->width, zimg->height);
			image_copy_image(widget->win, &pxmp, zimg);

			button_set_active_pixmap_from_pixmap(widget, &pxmp);

			pimg_free(&zimg);
			pimg_free(&bimg);
			image_destroy(&pxmp);
		}
		
		free(imgpath);
	}
#endif
}

void button_set_size(px_widget* widget, unsigned long width, unsigned long height)
{
	if (!widget->specific.button.change_size)
	{
		widget_set_size(widget, width, height);
		widget->specific.button.layout.width = widget->width;
		widget->specific.button.layout.height = widget->height;

		widget_compute_textlayout(&(widget->specific.button.layout), widget);
	}
}

Bool button_default_button_press(px_widget* widget, px_event* event)
{
	widget->state =  ACTIVE;
	widget->redraw(widget);
	widget->show(widget);

	if (widget->specific.button.auto_repeat && widget->callbacks[BUTTON_PRESSED] != NULL)
	{
		px_eventloop *eventloop = eventloop_get_eventloop();

		if (widget->specific.button.auto_repeat_timer != NULL) {
			eventloop->delete_timer(eventloop, widget->specific.button.auto_repeat_timer);
			widget->specific.button.auto_repeat_timer = NULL;
		}
		widget->specific.button.auto_repeat_timer =
			eventloop->add_timer(eventloop, 0, 
				widget->specific.button.delay * 1000, 
				button_handle_press_delay,widget,event,False);
	}

	return True;
}

Bool button_handle_press_delay(px_widget* widget, px_event* event)
{
	px_eventloop* eventloop = eventloop_get_eventloop();
	
	widget->specific.button.auto_repeat_timer = NULL;
	
	widget->specific.button.auto_repeat_timer = 
		eventloop->add_timer(eventloop, 0,
			widget->specific.button.speed * 1000, 
			widget->callbacks[BUTTON_PRESSED],widget,event,True);

	return True;
}

Bool button_default_button_released(px_widget* widget, px_event* event)
{
	widget->state =  NORMAL;
	widget->redraw(widget);
	widget->show(widget);

	if (widget->specific.button.auto_repeat_timer != NULL)
	{
		px_eventloop *eventloop=eventloop_get_eventloop();

		eventloop->delete_timer(eventloop, widget->specific.button.auto_repeat_timer);
		widget->specific.button.auto_repeat_timer = NULL;
	}

	return True;
}


Bool button_default_mouse_enter(px_widget* widget, px_event* event)
{
	widget->state =  ACTIVE;
	widget->redraw(widget);
	widget->show(widget);

	if (widget->specific.button.auto_repeat && widget->callbacks[BUTTON_PRESSED] != NULL)
	{
		px_eventloop* eventloop = eventloop_get_eventloop();

		if (widget->specific.button.auto_repeat_timer != NULL) {
			eventloop->delete_timer(eventloop,
				widget->specific.button.auto_repeat_timer);
			widget->specific.button.auto_repeat_timer = NULL;
		}
		widget->specific.button.auto_repeat_timer =
			eventloop->add_timer(eventloop, 0,
				widget->specific.button.delay * 1000, 
				button_handle_press_delay,widget,event,False);
	}
	return True;
}

Bool button_default_mouse_leave(px_widget* widget, px_event* event)
{
	widget->state = NORMAL;
	(*widget->redraw)(widget);
	widget->show(widget);

	if (widget->specific.button.auto_repeat_timer != NULL)
	{
		px_eventloop* eventloop = eventloop_get_eventloop();

		eventloop->delete_timer(eventloop,
				widget->specific.button.auto_repeat_timer);
		widget->specific.button.auto_repeat_timer = NULL;
	}
	return True;
}

void button_redraw(px_widget* widget)
{
	px_pixmap* image_to_show = NULL;
	px_color* bgd_color = NULL;
	XColor* border = NULL, *backup;
	Display* display = display_get_Xdisplay();
	px_button *button = &(widget->specific.button);
	
#ifdef __debug__
	printf("%s() : %p : Begin\n", __func__, widget);
#endif

	if (widget->state == DISABLED)
	{
		image_to_show = button->disabled;
		button->layout.font_color = button->disabled_font;
		bgd_color = button->disabled_color;
		border = button->disabled_border_color;
	}
	else if (widget->state == ACTIVE)
	{
		image_to_show = button->active;
		button->layout.font_color = button->active_font;
		bgd_color = button->active_color;
		border = button->active_border_color;
	}
	else
	{
		image_to_show = button->normal;
		button->layout.font_color = button->normal_font;
		bgd_color = widget->background.background_color;
	}

	if (widget->state != NORMAL && (image_to_show == NULL && bgd_color == NULL))
	{
			
		image_to_show = button->normal;
		bgd_color = widget->background.background_color;
	}
	
	if (border == NULL)
		border = widget->border_color;
	
	if (image_to_show == NULL || image_to_show->w < widget->width || image_to_show->h < widget->height)
	{
		if (widget->use_pixmap == True)
		{
			image_copy_pixmap(widget->win, &(widget->backing_store), widget->background.background_pixmap);
		}
		else if (bgd_color != NULL)
		{
			widget_draw_px_color(&(widget->backing_store), bgd_color);
		}
	}

	if (image_to_show != NULL)
	{
		int destx, desty;
		GC gc;

		gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
		XSetGraphicsExposures(display, gc, False);

		destx = (widget->width - image_to_show->w) / 2;
		desty = (widget->height - image_to_show->h) / 2;

		if (image_to_show->pxmp != None)
		XCopyArea(display, image_to_show->pxmp, widget->backing_store.pxmp, gc,
				0, 0, 
				image_to_show->w, image_to_show->h,
				destx, desty);
		if (image_to_show->mask != None)
		{
			GC mask_gc;
			int created = 0;
			
			if (widget->backing_store.mask == None)
			{
				widget->backing_store.mask = XCreatePixmap(display, widget->win, 
						widget->backing_store.w, widget->backing_store.h, 1);
				created = 1;
			}
				
			mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
			XSetGraphicsExposures(display, mask_gc, False);

			XSetForeground(display, mask_gc, 1);
			
			if (created)
			{
				XFillRectangle(display, widget->backing_store.mask, mask_gc, 0, 0, 
						widget->backing_store.w, widget->backing_store.h);
			}
			
			XCopyArea(display, image_to_show->mask, widget->backing_store.mask, mask_gc,
					0, 0, 
					image_to_show->w, image_to_show->h,
					destx, desty);
			XFreeGC(display, mask_gc);
		}
		else if (widget->backing_store.mask != None)
		{
			XFreePixmap(display, widget->backing_store.mask);
			widget->backing_store.mask = None;
		}
		XFreeGC(display, gc);
	}

	widget_draw_text(&(button->layout), &(widget->backing_store));
	
	backup = widget->border_color;
	widget->border_color = border;
	widget_draw_border(widget);
	widget->border_color = backup;
	
	XClearWindow(display, widget->win);
	widget_set_event_mask(widget, widget->emask);
	
#ifdef __debug__
	XSync(display, False);
	printf("%s() : %p : End\n", __func__, widget);
#endif
}


void button_destroy(px_widget* widget)
{
	px_button* button = &(widget->specific.button);
	px_textchunk* chunk, *next;
	
	if (button->layout.text != NULL)
		free(button->layout.text);

	if (button->normal_font != NULL)
	{
		free(widget->specific.button.normal);
	}
	
	if (button->active_font != NULL)
	{
		
		free(button->active_font);
	}

	if (button->disabled_font != NULL)
	{
		free(button->disabled_font);
	}

	if (button->active_color != NULL)
	{
		free(button->active_color);
	}

	if (button->disabled_color != NULL)
	{
		free(button->disabled_color);
	}
	
	if (button->fontname != NULL)
	{
		font_free_font(button->fontname);
		free(button->fontname);
	}

	if (button->normal != NULL)
	{
		image_destroy(button->normal);
		free(button->normal);
	}
	
	if (button->active != NULL)
	{
		image_destroy(button->active);
		free(button->active);
	}

	if (button->disabled != NULL)
	{
		image_destroy(button->disabled);
		free(button->disabled);
	}

	chunk = button->layout.chunks;
	while (chunk != NULL)
	{
		next = chunk->next;
		free(chunk);
		chunk = next;
	}
	
	widget_destroy(widget);
}

void button_set_normal_pixmap_from_file(px_widget* widget, char*path)
{
	px_button* specific = &(widget->specific.button);
	
	if (specific->normal == NULL)
	{
		specific->normal = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->normal);
	}

	if (!image_load_file(widget->win, specific->normal, path))
	{
		image_destroy(specific->normal);
		free(specific->normal);
		specific->normal = NULL;
	}
}


void button_set_active_pixmap_from_file(px_widget* widget, char*path)
{
	px_button* specific = &(widget->specific.button);
	
	if (specific->active == NULL)
	{
		specific->active = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->active);
	}

	if (!image_load_file(widget->win, specific->active, path))
	{
		image_destroy(specific->active);
		free(specific->active);
		specific->active = NULL;
	}
}

void button_set_disabled_pixmap_from_file(px_widget* widget, char*path)
{
	px_button* specific = &(widget->specific.button);
	
	if (specific->disabled == NULL)
	{
		specific->disabled = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->disabled);
	}

	image_load_file(widget->win, specific->disabled, path);
	if (!image_load_file(widget->win, specific->disabled, path))
	{
		image_destroy(specific->disabled);
		free(specific->disabled);
		specific->disabled = NULL;
	}
}

void button_set_normal_pixmap_from_pixmap(px_widget* widget, const px_pixmap* pixmap)
{
	px_button* specific = &(widget->specific.button);
	
	if (specific->normal == NULL)
	{
		specific->normal = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->normal);
	}

	if (!image_copy_pixmap(widget->win, specific->normal, pixmap))
	{
		image_destroy(specific->normal);
		free(specific->normal);
		specific->normal = NULL;
	}
}

void button_set_active_pixmap_from_pixmap(px_widget* widget, const px_pixmap* pixmap)
{
	px_button* specific = &(widget->specific.button);
	
	if (specific->active == NULL)
	{
		specific->active = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->active);
	}

	if (!image_copy_pixmap(widget->win, specific->active, pixmap))
	{
		image_destroy(specific->active);
		free(specific->active);
		specific->active = NULL;
	}
}

void button_set_disabled_pixmap_from_pixmap(px_widget* widget, const px_pixmap* pixmap)
{
	px_button* specific = &(widget->specific.button);
	
	if (specific->disabled == NULL)
	{
		specific->disabled = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->disabled);
	}

	if (!image_copy_pixmap(widget->win, specific->disabled, pixmap))
	{
		image_destroy(specific->disabled);
		free(specific->disabled);
		specific->disabled = NULL;
	}
}

void button_set_normal_font_color(px_widget* widget, const char* colour)
{
	if (widget->specific.button.normal_font == NULL)
	{
		widget->specific.button.normal_font  = (XColor*)malloc(sizeof(XColor));
	}
	widget_xcolor_from_str(widget->specific.button.normal_font, colour);
}


void button_set_active_font_color(px_widget* widget, const char* colour)
{
	if (widget->specific.button.active_font == NULL)
	{
		widget->specific.button.active_font = (XColor*)malloc(sizeof(XColor));
	}
	widget_xcolor_from_str(widget->specific.button.active_font, colour);
}

void button_set_disabled_font_color(px_widget* widget, const char* colour)
{
	if (widget->specific.button.disabled_font == NULL)
	{
		widget->specific.button.disabled_font = (XColor*)malloc(sizeof(XColor));
	}
	widget_xcolor_from_str(widget->specific.button.disabled_font, colour);
}

void button_set_active_background_color(px_widget* widget, const px_color* col)
{
	if (widget->specific.button.active_color == NULL)
	{
		widget->specific.button.active_color = (px_color*)malloc(sizeof(px_color));
	}
	
	(*widget->specific.button.active_color) = (*col);
}

void button_set_disabled_background_color(px_widget* widget, const px_color* col)
{
	if (widget->specific.button.disabled_color == NULL)
	{
		widget->specific.button.disabled_color = (px_color*)malloc(sizeof(px_color));
	}
	
	(*widget->specific.button.disabled_color) = (*col);
}

void button_set_active_border_color(px_widget* widget, const char* colour)
{
	if (widget->specific.button.active_border_color == NULL)
	{
		widget->specific.button.active_border_color  = (XColor*)malloc(sizeof(XColor));
	}
	widget_xcolor_from_str(widget->specific.button.active_border_color, colour);
}

void button_set_disabled_border_color(px_widget* widget, const char* colour)
{
	if (widget->specific.button.disabled_border_color == NULL)
	{
		widget->specific.button.disabled_border_color  = (XColor*)malloc(sizeof(XColor));
	}
	widget_xcolor_from_str(widget->specific.button.disabled_border_color, colour);
}

void button_set_text(px_widget* widget, const char* t)
{
	if (widget->specific.button.layout.text != NULL)
	{
		free(widget->specific.button.layout.text);
		widget->specific.button.layout.text = NULL;
	};

	if (t != NULL)
	{
		widget->specific.button.layout.text = strdup(t);
		widget_compute_textlayout(&(widget->specific.button.layout), widget);
		if (widget->specific.button.change_size)
		{
			widget_set_size(widget, widget->specific.button.layout.width < 1 ? 1 : widget->specific.button.layout.width,
					widget->specific.button.layout.height < 1 ? 1 : widget->specific.button.layout.height);
		}

	}
}

void button_set_font(px_widget* widget, const char* fontname)
{
	if (widget->specific.button.fontname != NULL)
	{
		font_free_font(widget->specific.button.fontname);
		free(widget->specific.button.fontname);
		widget->specific.button.fontname = NULL;
	}

	widget->specific.button.fontname = strdup(fontname);
	widget->specific.button.layout.font = font_load_font(widget->specific.button.fontname);
	widget_compute_textlayout(&(widget->specific.button.layout), widget);
	if (widget->specific.button.change_size)
	{
		widget_set_size(widget, widget->specific.button.layout.width < 1 ? 1 : widget->specific.button.layout.width,
				widget->specific.button.layout.height < 1 ? 1 : widget->specific.button.layout.height);
	}
}

void button_set_justification(px_widget* widget, justification j)
{
	widget->specific.button.layout.just = j;
}

justification button_get_justification(const px_widget* widget)
{
	return widget->specific.button.layout.just;
}

const char* button_get_text(const px_widget* widget)
{
	return widget->specific.button.layout.text;
}

void button_set_auto_repeat(px_widget* widget, Bool repeat)
{
	widget->specific.button.auto_repeat = repeat;
}

void button_set_auto_repeat_delay(px_widget* widget, unsigned int delay)
{
	widget->specific.button.delay = delay;
}

void button_set_auto_repeat_speed(px_widget* widget, unsigned int speed)
{
	widget->specific.button.speed = speed;
}

