/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include <X11/extensions/shape.h>
#include <X11/xpm.h>
#include <stdlib.h>
#include <string.h>
#include "checkbutton.h"
#include "image.h"
#include "font.h"
#include "theme.h"
#include "widget_factory.h"
#include "display.h"

px_widget* checkbutton_create(px_widget* parent, int x, int y, int width, int height, const char* tag)
{
	px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
	Display* display = display_get_Xdisplay();
	widget->win = XCreateWindow(display, parent->win, x, y, width, height, 0, CopyFromParent, CopyFromParent, CopyFromParent, 0, NULL);

#ifdef __debug__
	printf("%s(): %p - Created Window with parent window = %lu, x = %d, y = %d, width = %d, height = %d, new window = %lu\n\n", 
			__func__, widget, parent->win, x, y, width, height, widget->win);

	XSync(display, False);
#endif
	// Initialize common widget variables
	widget->x = x;
	widget->y = y;
	widget->height = height;
	widget->width = width;

	if (tag != NULL)
	{
		widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	}
	else
		widget->tag = NULL;
	widget->wtype = CHECKBUTTON;
	widget->child_of = NULL;

	widget->use_pixmap = False;
	widget->background.background_color = (px_color*)malloc(sizeof(px_color));
	widget->background.background_color->type = SOLID;
	widget_xcolor_from_str(&(widget->background.background_color->first_color), "white");

	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);

	return widget;
}

void checkbutton_set_parameters(px_widget* widget)
{
	px_checkbutton* checkbutton = &(widget->specific.checkbutton);
	
	widget->default_callbacks[BUTTON_PRESSED] = checkbutton_default_checkbutton_pressed;

	widget->set_size = checkbutton_set_size;
	widget->redraw = checkbutton_redraw;
	widget->destroy = checkbutton_destroy;

	
	// Initialize widget specific variables
	checkbutton->layout.just = JUSTIFY_CENTER;
	checkbutton->layout.wrap = WRAP_CHAR;
	checkbutton->layout.text = NULL;
	checkbutton->layout.x = 0;
	checkbutton->layout.y = 0;
	checkbutton->layout.width = widget->width;
	checkbutton->layout.height = widget->height;
	checkbutton->layout.font = NULL;
	checkbutton->layout.font_color = NULL;
	checkbutton->layout.chunks = NULL;
	checkbutton->layout.no_of_chunks = 0;

	checkbutton->fontname = NULL;
	checkbutton->indicator_on = True;
	checkbutton->value = 0;
	
	checkbutton->normal_font = NULL;
	checkbutton->disabled_font = NULL;
	
	checkbutton->normal = NULL;
	checkbutton->selected = NULL;
	checkbutton->disabled = NULL;

	checkbutton->change_size = False;
	
	checkbutton->set_normal_pixmap_from_file = checkbutton_set_normal_pixmap_from_file;
	checkbutton->set_selected_pixmap_from_file = checkbutton_set_selected_pixmap_from_file;
	checkbutton->set_disabled_pixmap_from_file = checkbutton_set_disabled_pixmap_from_file;
	
	checkbutton->set_normal_pixmap_from_pixmap = checkbutton_set_normal_pixmap_from_pixmap;
	checkbutton->set_selected_pixmap_from_pixmap = checkbutton_set_selected_pixmap_from_pixmap;
	checkbutton->set_disabled_pixmap_from_pixmap = checkbutton_set_disabled_pixmap_from_pixmap;

	checkbutton->set_text = checkbutton_set_text;
	checkbutton->set_justification = checkbutton_set_justification;
	checkbutton->set_font = checkbutton_set_font;

	checkbutton->get_justification = checkbutton_get_justification;
	checkbutton->get_text = checkbutton_get_text;

	checkbutton->set_normal_font_color = checkbutton_set_normal_font_color;
	checkbutton->set_disabled_font_color = checkbutton_set_disabled_font_color;
	
	checkbutton->show_indicator = checkbutton_show_indicator;
	checkbutton->hide_indicator = checkbutton_hide_indicator;

	checkbutton->select = checkbutton_select;
	checkbutton->deselect = checkbutton_deselect;

	checkbutton->set_wrap_type = widget_set_wrap_type;
	checkbutton->size_depends_on_contents = widget_size_depends_on_contents;
}

void checkbutton_set_size(px_widget* widget, unsigned long width, unsigned long height)
{
	widget_set_size(widget, width, height);
	widget->specific.checkbutton.layout.width = widget->width;
	widget->specific.checkbutton.layout.height = widget->height;

	widget_compute_textlayout(&(widget->specific.checkbutton.layout), widget);
}


Bool checkbutton_default_checkbutton_pressed(px_widget* widget, px_event* event)
{
	if (widget->specific.checkbutton.value)
	{
		widget->specific.checkbutton.deselect(widget);
	}
	else
	{
		widget->specific.checkbutton.select(widget);
	}
	widget->redraw(widget);
	widget->show(widget);

	return True;
}

void checkbutton_redraw(px_widget* widget)
{
	GC gc;
	px_pixmap* image_to_show = NULL;
	px_pixmap* indicator = NULL;
	unsigned int drawareax = 0 ;
	px_checkbutton* checkbutton = &(widget->specific.checkbutton);
	Display* display = display_get_Xdisplay();
	
#ifdef __debug__
	printf("%s() : %p : Begin\n", __func__, widget);
#endif
	gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
	XSetGraphicsExposures(display, gc, False);

	if (widget->state == DISABLED)
	{
		image_to_show = checkbutton->disabled;
		checkbutton->layout.font_color = checkbutton->disabled_font;
	}
	else
	{
		if (checkbutton->value)
			image_to_show = checkbutton->selected;

		if (image_to_show == NULL)
			image_to_show = checkbutton->normal;

		checkbutton->layout.font_color = checkbutton->normal_font;
	}

	if (widget->use_pixmap == True)
	{
		image_copy_pixmap(widget->win, &(widget->backing_store), widget->background.background_pixmap);
	}
	else
	{
		widget_draw_px_color(&(widget->backing_store), widget->background.background_color);
	}

	if (checkbutton->indicator_on)
	{
		px_theme* theme = theme_get_theme();
		
		if (checkbutton->value)
			indicator = theme->get_pixmap(theme, CHECKBUTTON_ON);
		else
			indicator = theme->get_pixmap(theme, CHECKBUTTON_OFF);

		if (indicator != NULL && indicator->pxmp != None)
		{
			int desty;

			desty = (widget->height - indicator->h) / 2;

			drawareax = indicator->w + 2;
			
			if (indicator->mask != None)
			{
				GC gcxor, gcand;
				gcxor = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
				XSetGraphicsExposures(display, gcxor, False);

			    	XSetForeground(display, gcxor,BlackPixel(display, DefaultScreen(display)));
				XSetBackground(display, gcxor,WhitePixel(display, DefaultScreen(display)));
				XSetFunction(display, gcxor, GXxor);
				XSetFillStyle(display, gcxor, FillSolid);
				XSetGraphicsExposures(display, gcxor, False);
	
				gcand = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
				XSetGraphicsExposures(display, gcand, False);

				XSetForeground(display, gcand, 0);
				XSetBackground(display, gcand, ~0);
				XSetFunction(display, gcand, GXand);
				XSetFillStyle(display, gcand, FillSolid);
				XSetGraphicsExposures(display, gcand, False);
	
				XCopyArea(display, indicator->pxmp, widget->backing_store.pxmp, gcxor, 
					0, 0, 
					indicator->w, indicator->h,
					1, desty); 

				XCopyPlane(display, indicator->mask, widget->backing_store.pxmp, gcand, 
					0, 0, 
					indicator->w, indicator->h,
					1, desty, 1L); 

				XCopyArea(display, indicator->pxmp, widget->backing_store.pxmp, gcxor, 
					0, 0,
					indicator->w, indicator->h,
					1, desty); 

				XFreeGC(display, gcxor);
				XFreeGC(display, gcand);
			}
			else
			{
				XCopyArea(display, indicator->pxmp, widget->backing_store.pxmp, gc,
					0, 0, 
					indicator->w, indicator->h,
					1, desty);
			}
		}
	}

	
	if (image_to_show != NULL)
	{
		int destx, desty;

		destx = ((widget->width - drawareax) - image_to_show->w) / 2 + drawareax;
		desty = (widget->height - image_to_show->h) / 2;

		if (image_to_show->pxmp != None)
			XCopyArea(display, image_to_show->pxmp, widget->backing_store.pxmp, gc,
				0, 0, 
				image_to_show->w, image_to_show->h,
				destx, desty);
		if (image_to_show->mask != None)
		{
			GC mask_gc;
			int created = 0;
			
			if (widget->backing_store.mask == None)
			{
				widget->backing_store.mask = XCreatePixmap(display, widget->win, 
						widget->backing_store.w, widget->backing_store.h, 1);
				created = 1;
			}
				
			mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
			XSetGraphicsExposures(display, mask_gc, False);

			XSetForeground(display, mask_gc, 1);
			
			if (created)
			{
				XFillRectangle(display, widget->backing_store.mask, mask_gc, 0, 0, 
						widget->backing_store.w, widget->backing_store.h);
			}
			
			XCopyArea(display, image_to_show->mask, widget->backing_store.mask, mask_gc,
					0, 0, 
					image_to_show->w, image_to_show->h,
					destx, desty);
			XFreeGC(display, mask_gc);
		}
		else if (widget->backing_store.mask != None)
		{
			XFreePixmap(display, widget->backing_store.mask);
			widget->backing_store.mask = None;
		}
	}

	if (checkbutton->indicator_on && widget->backing_store.mask != None && 
			indicator != NULL && indicator->mask != None)
	{
		int desty;
		GC mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
		XSetGraphicsExposures(display, mask_gc, False);

		desty = (widget->height - indicator->h) / 2;
		drawareax = indicator->w + 2;
		XCopyArea(display, indicator->mask, widget->backing_store.mask, mask_gc,
			0, 0, 
			indicator->w, indicator->h,
			1, desty);
		XFreeGC(display, mask_gc);
	}

	checkbutton->layout.x = drawareax;
	widget_draw_text(&(checkbutton->layout), &(widget->backing_store));

	XFreeGC(display, gc);
	widget_draw_border(widget);
	XClearWindow(display, widget->win);
	widget_set_event_mask(widget, widget->emask);

#ifdef __debug__
	XSync(display, False);
	printf("%s() : %p : End\n", __func__, widget);
#endif
}

void checkbutton_destroy(px_widget* widget)
{
	px_textchunk* chunk, *next;
	px_checkbutton* checkbutton = &(widget->specific.checkbutton);

	if (widget->specific.checkbutton.layout.text != NULL)
		free(widget->specific.checkbutton.layout.text);

	if (widget->specific.checkbutton.normal_font != NULL)
	{
		free(widget->specific.checkbutton.normal);
	}
	
	if (widget->specific.checkbutton.disabled_font != NULL)
	{
		free(widget->specific.checkbutton.disabled);
	}

	if (widget->specific.checkbutton.fontname != NULL)
	{
		font_free_font(widget->specific.checkbutton.fontname);
		free(widget->specific.checkbutton.fontname);
	}

	if (widget->specific.checkbutton.normal != NULL)
	{
		image_destroy(widget->specific.checkbutton.normal);
		free(widget->specific.checkbutton.normal);
	}
	
	if (widget->specific.checkbutton.selected != NULL)
	{
		image_destroy(widget->specific.checkbutton.selected);
		free(widget->specific.checkbutton.selected);
	}
	
	if (widget->specific.checkbutton.disabled != NULL)
	{
		image_destroy(widget->specific.checkbutton.disabled);
		free(widget->specific.checkbutton.disabled);
	}

	chunk = checkbutton->layout.chunks;
	while (chunk != NULL)
	{
		next = chunk->next;
		free(chunk);
		chunk = next;
	}

	widget_destroy(widget);
}

void checkbutton_set_normal_pixmap_from_file(px_widget* widget, char*path)
{
	px_checkbutton* specific = &(widget->specific.checkbutton);
	
	if (specific->normal == NULL)
	{
		specific->normal = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->normal);
	}

	if (!image_load_file(widget->win, specific->normal, path))
	{
		image_destroy(specific->normal);
		free(specific->normal);
		specific->normal = NULL;
	}
}

void checkbutton_set_selected_pixmap_from_file(px_widget* widget, char*path)
{
	px_checkbutton* specific = &(widget->specific.checkbutton);
	
	if (specific->selected == NULL)
	{
		specific->selected = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->selected);
	}

	if (!image_load_file(widget->win, specific->selected, path))
	{
		image_destroy(specific->selected);
		free(specific->selected);
		specific->selected = NULL;
	}
}

void checkbutton_set_disabled_pixmap_from_file(px_widget* widget, char*path)
{
	px_checkbutton* specific = &(widget->specific.checkbutton);
	
	if (specific->disabled == NULL)
	{
		specific->disabled = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->disabled);
	}

	image_load_file(widget->win, specific->disabled, path);
	if (!image_load_file(widget->win, specific->disabled, path))
	{
		image_destroy(specific->disabled);
		free(specific->disabled);
		specific->disabled = NULL;
	}
}

void checkbutton_set_normal_pixmap_from_pixmap(px_widget* widget, const px_pixmap* pixmap)
{
	px_checkbutton* specific = &(widget->specific.checkbutton);
	
	if (specific->normal == NULL)
	{
		specific->normal = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->normal);
	}

	if (!image_copy_pixmap(widget->win, specific->normal, pixmap))
	{
		image_destroy(specific->normal);
		free(specific->normal);
		specific->normal = NULL;
	}
}

void checkbutton_set_selected_pixmap_from_pixmap(px_widget* widget, const px_pixmap* pixmap)
{
	px_checkbutton* specific = &(widget->specific.checkbutton);
	
	if (specific->selected == NULL)
	{
		specific->selected = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->selected);
	}

	if (!image_copy_pixmap(widget->win, specific->selected, pixmap))
	{
		image_destroy(specific->selected);
		free(specific->selected);
		specific->selected = NULL;
	}
}

void checkbutton_set_disabled_pixmap_from_pixmap(px_widget* widget, const px_pixmap* pixmap)
{
	px_checkbutton* specific = &(widget->specific.checkbutton);
	
	if (specific->disabled == NULL)
	{
		specific->disabled = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->disabled);
	}

	if (!image_copy_pixmap(widget->win, specific->disabled, pixmap))
	{
		image_destroy(specific->disabled);
		free(specific->disabled);
		specific->disabled = NULL;
	}
}

void checkbutton_set_normal_font_color(px_widget* widget, const char* colour)
{
	if (widget->specific.checkbutton.normal_font == NULL)
	{
		widget->specific.checkbutton.normal_font  = (XColor*)malloc(sizeof(XColor));
	}
	widget_xcolor_from_str(widget->specific.checkbutton.normal_font, colour);
}


void checkbutton_set_disabled_font_color(px_widget* widget, const char* colour)
{
	if (widget->specific.checkbutton.disabled_font == NULL)
	{
		widget->specific.checkbutton.disabled_font = (XColor*)malloc(sizeof(XColor));
	}
	widget_xcolor_from_str(widget->specific.checkbutton.disabled_font, colour);
}

void checkbutton_set_text(px_widget* widget, const char* t)
{
	if (widget->specific.checkbutton.layout.text != NULL)
	{
		free(widget->specific.checkbutton.layout.text);
		widget->specific.checkbutton.layout.text = NULL;
	};

	if (t != NULL)
	{
		widget->specific.checkbutton.layout.text = strdup(t);
		widget_compute_textlayout(&(widget->specific.checkbutton.layout), widget);
	}
}

void checkbutton_set_font(px_widget* widget, const char* fontname)
{
	if (widget->specific.checkbutton.fontname != NULL)
	{
		font_free_font(widget->specific.checkbutton.fontname);
		free(widget->specific.checkbutton.fontname);
		widget->specific.checkbutton.fontname = NULL;
	}

	widget->specific.checkbutton.fontname = strdup(fontname);
	widget->specific.checkbutton.layout.font = font_load_font(widget->specific.checkbutton.fontname);
	widget_compute_textlayout(&(widget->specific.checkbutton.layout), widget);
}

void checkbutton_set_justification(px_widget* widget, justification j)
{
	widget->specific.checkbutton.layout.just = j;
}

justification checkbutton_get_justification(const px_widget* widget)
{
	return widget->specific.checkbutton.layout.just;
}

const char* checkbutton_get_text(const px_widget* widget)
{
	return widget->specific.checkbutton.layout.text;
}

void checkbutton_show_indicator(px_widget* widget)
{
	widget->specific.checkbutton.indicator_on = True;
}

void checkbutton_hide_indicator(px_widget* widget)
{
	widget->specific.checkbutton.indicator_on = False;
}

void checkbutton_select(px_widget* widget)
{
	widget->specific.checkbutton.value = True;
}

void checkbutton_deselect(px_widget* widget)
{
	widget->specific.checkbutton.value = False;
}
	
