/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

// PicoWidget : ComboBox
// Copyrights (c) PicoPeta Simputers Pvt. Ltd.
// Programmer: P R Subramanya

#include "combobox.h"
#include "display.h"
#include "image.h"
#include "theme.h"
#include "font.h"
#include "widget_factory.h"
#include "txtlb.h"
#include "label.h"

#include <string.h>
#include <stdlib.h>

#define DFLT_COMBO_FNT "-picopeta-simputer-medium-r-normal--9-*-*-*-p-*-fontspecific-0"
#define DFLT_VISIBLE_OPTIONS_COUNT 4

struct _combobox {
	px_widget *cframe, *clbl, *clist, *cbutt;
	char *fontname;
	int visible_options;
	int selecn;
	Bool drop_flg;
};
	
static void intern_combobox_set_size(px_widget* wdgt, unsigned long width,
		unsigned long height)
{
	struct _combobox *cbox=wdgt->specific.combobox.internal;
	px_xynode* ynodes = NULL;
	MajorAxis axis;

	if (wdgt->wtype != COMBOBOX)
		return;
	if ((wdgt->width == width) && (wdgt->height == height))
		return;

	ynodes = wdgt->child_of->specific.toplevel.ynodes;
	axis = wdgt->child_of->specific.toplevel.axis;

	ynodes = widget_remove_child(ynodes, wdgt, axis == YAXIS ? YDIRECTION : XDIRECTION, axis);

	XResizeWindow(display_get_Xdisplay(), wdgt->win, width, height);
	wdgt->width = width; wdgt->height = height;

	if (cbox->cframe->width != width) {
		cbox->cframe->set_size(cbox->cframe, width, cbox->cframe->height);
		cbox->clbl->set_size(cbox->clbl, width-20, cbox->clbl->height);
		cbox->clbl->redraw(cbox->clbl);
		cbox->cbutt->set_position(cbox->cbutt, width-20, 0);
		cbox->cbutt->redraw(cbox->cbutt);
		cbox->clist->set_size(cbox->clist,width,cbox->clist->height);
		cbox->clist->redraw(cbox->clist);
		cbox->cframe->redraw(cbox->cframe);
	}
		
	ynodes = widget_add_child(ynodes, wdgt, axis == YAXIS ? YDIRECTION : XDIRECTION, axis);
	wdgt->child_of->specific.toplevel.ynodes = ynodes;
}

void combobox_set_size(px_widget* wdgt, unsigned long width,
		unsigned long height)
{
	struct _combobox *cbox;

	if ((wdgt == NULL) || (wdgt->wtype != COMBOBOX))
		return;
	cbox = wdgt->specific.combobox.internal;
	if (cbox->drop_flg == True) {
		intern_combobox_set_size(wdgt, width, wdgt->height);
	} else {
		intern_combobox_set_size(wdgt, width, 20);
	}
}

Bool combobox_button_release_cb(px_widget *wdgt, px_event *ev)
{
	if (ev->private_data != NULL) {
		px_widget *cbox=ev->private_data;
		struct _combobox *intern=cbox->specific.combobox.internal;

		if (intern->drop_flg == False) {
			intern_combobox_set_size(cbox, cbox->width,intern->clist->height+20);
			intern->drop_flg = True;
		} else {
			intern_combobox_set_size(cbox,cbox->width,20);
			intern->drop_flg = False;
		}
		intern->cframe->redraw(intern->cframe);
	}
	return(False);
}

void combobox_selecn_change_cb(px_widget *wdgt, int selecn)
{
	px_widget *cbox=wdgt->callback_parameters[BUTTON_RELEASED];
	struct _combobox *intern=cbox->specific.combobox.internal;
	char *selitem=NULL;

	txtlb_get_element(wdgt, &selitem, selecn, 0);
	intern->clbl->specific.label.set_text(intern->clbl, selitem);
	intern->clbl->redraw(intern->clbl);
	free(selitem);
	if (intern->drop_flg == True) {
		intern_combobox_set_size(cbox, cbox->width, 20);
		intern->drop_flg = False;
	}
	intern->cframe->redraw(intern->cframe);
	if (cbox->specific.combobox.intimate_selecn_change != NULL) {
		cbox->specific.combobox.intimate_selecn_change(cbox, selecn);
	}
}

px_widget* combobox_create(px_widget* parent, int x, int y, int width, int height, const char* tag)
{
	px_widget* wdgt = (px_widget*)malloc(sizeof(px_widget));
	Display* disp=display_get_Xdisplay();

	// Initialize common widget variables
	wdgt->win = XCreateWindow(disp, parent->win, x, y, width, 20, 
			0,CopyFromParent,CopyFromParent,CopyFromParent,0,NULL);
	XSelectInput(disp, wdgt->win, PropertyChangeMask);
	
	wdgt->x = x; wdgt->y = y;
	wdgt->width = width; wdgt->height = 20;

	if (tag != NULL) {
		wdgt->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(wdgt->tag, tag);
	} else
		wdgt->tag = NULL;

	wdgt->wtype = COMBOBOX;
	wdgt->child_of = NULL;
	wdgt->use_pixmap = False;
	wdgt->background.background_color = (px_color*)malloc(sizeof(px_color));
	wdgt->background.background_color->type = SOLID;
	widget_xcolor_from_str(&(wdgt->background.background_color->first_color), "#ffffff");

	// Initializing backing store
	image_initialize_pixmap(&(wdgt->backing_store));
	image_create_pixmap(wdgt->win, &(wdgt->backing_store), width, 20);
	widget_draw_px_color(&(wdgt->backing_store), wdgt->background.background_color);
	return(wdgt);
}

Bool combobox_handle_ev(px_widget *wdgt, px_event *ev)
{
	struct _combobox *cbox=wdgt->specific.combobox.internal;

	if ((ev->type != NR_CALLBACKS) &&
	((ev->type == EXPOSE) || (wdgt->get_state(wdgt) != DISABLED))) {
		return(widget_handle_event(cbox->cframe,ev));
	}
	return(True);
}

void combobox_redraw(px_widget *wdgt)
{
	Display *disp=display_get_Xdisplay();
	struct _combobox *cbox=(struct _combobox *)wdgt->specific.combobox.internal;

	if (wdgt->hidden == False) {
		XClearWindow(disp, wdgt->win);
	}
	cbox->clist->redraw(cbox->clist);
	cbox->clbl->redraw(cbox->clbl);
	cbox->cframe->redraw(cbox->cframe);
}

void combobox_show(px_widget *wdgt)
{
	Display *disp=display_get_Xdisplay();

	XSetWindowBackgroundPixmap(disp, wdgt->win, wdgt->backing_store.pxmp);
	XMapWindow(disp, wdgt->win);
	wdgt->hidden = False;
}

void combobox_destroy(px_widget *wdgt)
{
	struct _combobox *cbox=wdgt->specific.combobox.internal;

	cbox->cframe->destroy(cbox->cframe);
	widget_destroy(wdgt);
}

int combobox_add_option(px_widget *wdgt, char * optstr, Bool redraw_flg)
{
	struct _combobox *cbox=wdgt->specific.combobox.internal;

	if (cbox != NULL) {
		px_widget *lb=cbox->clist;

		lb->specific.txtlb.add_lb_elem(lb, optstr,0);
		if (redraw_flg == True)
			lb->redraw(lb);
		return(txtlb_get_no_of_rows(lb));
	}
	return(-1);
}

int combobox_remove_option(px_widget *wdgt, int idx)
{
	struct _combobox *cbox=wdgt->specific.combobox.internal;

	if (cbox != NULL) {
		px_widget *lb=cbox->clist;

		if (idx >= 0) {
			lb->specific.txtlb.remove_lb_row(lb, idx, False);
			if (idx <= cbox->selecn) {
				char *selitem=NULL;

				txtlb_get_element(lb,&selitem,cbox->selecn,0);
				if (selitem != NULL) {
					cbox->clbl->specific.label.set_text(
						cbox->clbl, selitem);
					free(selitem);
					cbox->selecn = 0;
				} else {
					cbox->clbl->specific.label.set_text(
						cbox->clbl, " ");
					cbox->selecn = -1;
				}
				cbox->clbl->redraw(cbox->clbl);
			}
		}
		return(txtlb_get_no_of_rows(lb));
	}
	return(-1);
}

void combobox_set_font(px_widget* wdgt, const char *fntname)
{
	struct _combobox *cbox;

	if ((wdgt == NULL) || (wdgt->wtype != COMBOBOX))
		return;
	cbox = wdgt->specific.combobox.internal;

	free(cbox->fontname);
	cbox->fontname = strdup(fntname);
	label_set_font(cbox->clbl, fntname);
	cbox->clbl->redraw(cbox->clbl);
}

int combobox_clear_options(px_widget *wdgt)
{
	struct _combobox *cbox=wdgt->specific.combobox.internal;

	if (cbox != NULL) {
		px_widget *lb=cbox->clist;

		txtlb_clear_all_rows(lb);
		cbox->selecn = -1;
		cbox->clbl->specific.label.set_text(cbox->clbl," ");
		cbox->clbl->redraw(cbox->clbl);
		return(1);
	}
	return(-1);
}

int combobox_get_no_of_options(px_widget *wdgt)
{
	struct _combobox *cbox;

	if ((wdgt == NULL) || (wdgt->wtype != COMBOBOX))
		return(-1);
	cbox = wdgt->specific.combobox.internal;
	return(txtlb_get_no_of_rows(cbox->clist));
}

void combobox_get_option(px_widget *wdgt, char ** option, int idx)
{
	struct _combobox *cbox;

	if ((wdgt == NULL) || (wdgt->wtype != COMBOBOX)) {
		(*option) = NULL;
		return;
	}
	cbox = wdgt->specific.combobox.internal;
	txtlb_get_element(cbox->clist, option, idx, 0);
}

void combobox_get_selecn(px_widget *wdgt, char ** option, int * idx)
{
	struct _combobox *cbox=wdgt->specific.combobox.internal;

	if (cbox != NULL) {
		px_widget *lb=cbox->clist;

		txtlb_get_element(lb,option,cbox->selecn,0);
		(*idx) = cbox->selecn;
	}
}

void combobox_set_selecn(px_widget *wdgt, int selecn)
{
	struct _combobox *cbox=wdgt->specific.combobox.internal;

	if ((cbox != NULL) && (selecn >= 0) && (cbox->selecn != selecn)) {
		px_widget *lb=cbox->clist;

		if (txtlb_get_no_of_rows(lb) > selecn) {
			char *elem=NULL;

			txtlb_set_selection(lb, selecn);
			txtlb_get_element(lb, &elem, selecn, 0);
			if (elem != NULL)
				label_set_text(cbox->clbl, elem);
			else
				label_set_text(cbox->clbl, " ");
			cbox->selecn = selecn;
		}
	}
}

void combobox_change_color(px_widget *wdgt, combo_colors col, char *colstr)
{
	struct _combobox *cbox;

	if ((wdgt == NULL) || (wdgt->wtype != COMBOBOX))
		return;
	cbox = wdgt->specific.combobox.internal;
	switch (col) {
		case COMBO_LABEL_BACKGROUND: {
			px_color bgcol;

			bgcol.type = SOLID;
			widget_xcolor_from_str(&bgcol.first_color, colstr);
	                cbox->clbl->set_background_color(cbox->clbl,&bgcol);
			cbox->clbl->redraw(cbox->clbl);
			}
			break;
		case COMBO_LABEL_FOREGROUND:
			cbox->clbl->specific.label.set_normal_color(cbox->clbl,
					colstr);
			cbox->clbl->redraw(cbox->clbl);
			break;
		case COMBO_LIST_BACKGROUND:
			cbox->clist->specific.txtlb.change_lb_color(cbox->clist,
					TXTLB_BACKGROUND, colstr);
			cbox->clist->redraw(cbox->clist);
			break;
		case COMBO_LIST_FOREGROUND:
			cbox->clist->specific.txtlb.change_lb_color(cbox->clist,
					TXTLB_FOREGROUND, colstr);
			cbox->clist->redraw(cbox->clist);
			break;
		case COMBO_LIST_ROW_BACKGROUND:
			cbox->clist->specific.txtlb.change_lb_color(cbox->clist,
					TXTLB_ROW_BACKGROUND, colstr);
			cbox->clist->redraw(cbox->clist);
			break;
		case COMBO_LIST_PEAK:
			cbox->clist->specific.txtlb.change_lb_color(cbox->clist,
					TXTLB_PEAK, colstr);
			cbox->clist->redraw(cbox->clist);
			break;
		case COMBO_LIST_TROUGH:
			cbox->clist->specific.txtlb.change_lb_color(cbox->clist,
					TXTLB_TROUGH, colstr);
			cbox->clist->redraw(cbox->clist);
			break;
		case COMBO_LIST_SELECN_HIGHLIGHT:
			cbox->clist->specific.txtlb.change_lb_color(cbox->clist,
					TXTLB_SELECN_HIGHLIGHT, colstr);
			cbox->clist->redraw(cbox->clist);
			break;
		case COMBO_LIST_SELECN_FOREGROUND:
			cbox->clist->specific.txtlb.change_lb_color(cbox->clist,
					TXTLB_SELECN_FOREGROUND, colstr);
			cbox->clist->redraw(cbox->clist);
			break;
	}
}

void combobox_set_intimate_selecn_change(px_widget *cbox, 
		void (*change_fn)(px_widget *, int))
{
	if ((cbox == NULL) || (cbox->wtype != COMBOBOX))
		return;
	cbox->specific.combobox.intimate_selecn_change = change_fn;
}

void combobox_set_parameters(px_widget* wdgt)
{
	struct _combobox *intern=NULL;
	px_color bgcol;
	px_theme* theme=theme_get_theme();
		
	wdgt->handle_event = combobox_handle_ev;
	wdgt->redraw = combobox_redraw;
	wdgt->show = combobox_show;
	wdgt->destroy = combobox_destroy;
	wdgt->set_size = combobox_set_size;

	intern = (struct _combobox *)malloc(sizeof(struct _combobox));
	
	intern->fontname = strdup(DFLT_COMBO_FNT);

	intern->cframe = new_widget(FRAME,wdgt,0,0,wdgt->width,120,"");
	intern->cframe->redraw(intern->cframe);
	intern->cframe->show(intern->cframe);

	bgcol.type = SOLID;
	widget_xcolor_from_str(&bgcol.first_color, "#d3d3dd");

	intern->clbl = new_widget(LABEL,intern->cframe,0,0,
				wdgt->width-20,20,"");
	intern->clbl->specific.label.set_font(intern->clbl, DFLT_COMBO_FNT);
	intern->clbl->set_background_color(intern->clbl, &bgcol);
	intern->clbl->specific.label.set_text(intern->clbl, "-Select-");
	intern->clbl->redraw(intern->clbl);
	intern->clbl->show(intern->clbl);

	intern->cbutt = new_widget(BUTTON,intern->cframe,
				wdgt->width-20,0,20,20,"");
	intern->cbutt->specific.button.set_font(intern->cbutt, DFLT_COMBO_FNT);
	intern->cbutt->specific.button.set_normal_pixmap_from_pixmap(
		intern->cbutt, theme_get_pixmap(theme, COMBOBOX_DOWN));
	intern->cbutt->specific.button.set_active_pixmap_from_pixmap(
		intern->cbutt, theme_get_pixmap(theme, COMBOBOX_DOWN_ACTIVE));
	intern->cbutt->callbacks[BUTTON_RELEASED] = combobox_button_release_cb;
	intern->cbutt->callback_parameters[BUTTON_RELEASED] = (void *)wdgt;
	intern->cbutt->redraw(intern->cbutt);
	intern->cbutt->show(intern->cbutt);

	intern->clist = new_widget(TXTLB,intern->cframe,0,20,wdgt->width,100,"");
	intern->clist->specific.txtlb.add_lb_elem(intern->clist,"-Select-",0);
	txtlb_set_selection(intern->clist, 0);
	intern->clist->callback_parameters[BUTTON_RELEASED] = (void *)wdgt;
	txtlb_set_intimate_change(intern->clist, combobox_selecn_change_cb);
	intern->clist->redraw(intern->clist);
	intern->clist->show(intern->clist);
	
	wdgt->specific.combobox.add_option = combobox_add_option;
	wdgt->specific.combobox.remove_option = combobox_remove_option;
	wdgt->specific.combobox.clear_options = combobox_clear_options;
	wdgt->specific.combobox.get_selecn = combobox_get_selecn;
	wdgt->specific.combobox.set_selecn = combobox_set_selecn;
	wdgt->specific.combobox.change_color = combobox_change_color;
	wdgt->specific.combobox.intimate_selecn_change = NULL;

	wdgt->specific.combobox.internal = intern;
}

void combobox_set_visible_num_of_options(px_widget *wdgt, int visible_num)
{
	struct _combobox *cbox;

	if ((wdgt == NULL) || (wdgt->wtype != COMBOBOX))
		return;
	cbox = wdgt->specific.combobox.internal;
	if ((visible_num <= 0) || (visible_num == cbox->visible_options))
		return;
	txtlb_set_visible_no_of_rows(cbox->clist, visible_num);
	cbox->visible_options = visible_num;
	cbox->cframe->set_size(cbox->cframe, cbox->cframe->width,
			cbox->clist->height+20);
	if (cbox->drop_flg == True) {
		intern_combobox_set_size(wdgt, cbox->cframe->width,
					cbox->cframe->height);
	} else {
		intern_combobox_set_size(wdgt, cbox->cframe->width, 20);
	}
	cbox->clist->redraw(cbox->clist);
	cbox->cframe->redraw(cbox->cframe);
}

void combobox_disable_events(px_widget *wdgt)
{
	struct _combobox *cbox;

	if ((wdgt == NULL) || (wdgt->wtype != COMBOBOX))
		return;
	cbox = wdgt->specific.combobox.internal;

	cbox->cbutt->default_callbacks[BUTTON_PRESSED] = NULL;
	cbox->cbutt->default_callbacks[BUTTON_RELEASED] = NULL;
	cbox->cbutt->callbacks[BUTTON_PRESSED] = NULL;
	cbox->cbutt->callbacks[BUTTON_RELEASED] = NULL;
	cbox->clist->default_callbacks[BUTTON_PRESSED] = NULL;
	cbox->clist->default_callbacks[MOTION] = NULL;
	cbox->clist->default_callbacks[BUTTON_RELEASED] = NULL;
	cbox->clist->callbacks[BUTTON_PRESSED] = NULL;
	cbox->clist->callbacks[MOTION] = NULL;
	cbox->clist->callbacks[BUTTON_RELEASED] = NULL;
}

void combobox_set_callback(px_widget *wdgt, widget_callbacks eventtype,
		Bool (*ev_handler)(px_widget *wdgt, px_event *ev))
{
	struct _combobox *cbox;

	if ((wdgt == NULL) || (wdgt->wtype != COMBOBOX))
		return;

	cbox = wdgt->specific.combobox.internal;
	cbox->cbutt->callbacks[eventtype] = ev_handler;
	cbox->clbl->callbacks[eventtype] = ev_handler;
	cbox->clist->callbacks[eventtype] = ev_handler;
}

XColor * combobox_get_fg_color(px_widget *wdgt)
{
	struct _combobox *cbox;

	if ((wdgt == NULL) || (wdgt->wtype != COMBOBOX))
		return(NULL);
	cbox = (struct _combobox *)wdgt->specific.combobox.internal;
	return(cbox->clbl->specific.label.normal);
}

XColor * combobox_get_bg_color(px_widget *wdgt)
{
	struct _combobox *cbox;

	if ((wdgt == NULL) || (wdgt->wtype != COMBOBOX))
		return(0L);
	cbox = (struct _combobox *)wdgt->specific.combobox.internal;
	return(&(cbox->clbl->background.background_color->first_color));
}

char *combobox_get_fontname(px_widget *wdgt)
{
	struct _combobox *cbox;

	if ((wdgt == NULL) || (wdgt->wtype != COMBOBOX))
		return(NULL);
	cbox = (struct _combobox *)wdgt->specific.combobox.internal;
	return(cbox->fontname);
}
