/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include "display.h"
#include "clipboard.h"
#include <stdlib.h>
#include <string.h>
#include <X11/Xatom.h>

px_display* picowidget_display_variable = NULL;
px_display* display_get_display()
{

	if (picowidget_display_variable == NULL)
	{
		picowidget_display_variable = (px_display*)malloc(sizeof(px_display));
		picowidget_display_variable->display = NULL;

		picowidget_display_variable->wm_protocols = (Atom*) malloc(sizeof(Atom) * 5);
		picowidget_display_variable->protocol_names = (char**) malloc(sizeof(char*) * 5);
		picowidget_display_variable->no_of_protocols = 0;
		picowidget_display_variable->protocols_allocated = 5;

		picowidget_display_variable->xa_primary = NULL;
		picowidget_display_variable->xa_secondary = NULL;

		picowidget_display_variable->register_wm_protocol = display_register_wm_protocol;
		picowidget_display_variable->get_atom = display_get_atom;

		picowidget_display_variable->open_display = display_open_display;
		picowidget_display_variable->close_display = display_close_display;
		picowidget_display_variable->set_display = display_set_display;
		
		picowidget_display_variable->get_clipboard = display_get_clipboard;
		
		picowidget_display_variable->flush = display_flush;
		picowidget_display_variable->sync = display_sync;
	}

	return picowidget_display_variable;
}

Display* display_get_Xdisplay()
{
	px_display* pxdisplay = display_get_display();
	
	if (pxdisplay->display == NULL)
	{
		fprintf(stderr, "Display has not been set. Please set the display by calling either \"open_display\" or \"set_display\" function in px_display structure before calling any function in the picowidget library\n");
		exit(1);
	}

	return pxdisplay->display;
}

Bool display_open_display(px_display* pxdisplay, char* disp)
{
	pxdisplay->display = XOpenDisplay(disp);

	if (pxdisplay->display == NULL)
		return False;
	
	pxdisplay->wm_protocols[pxdisplay->no_of_protocols] = XInternAtom(pxdisplay->display, "WM_DELETE_WINDOW",False);
	pxdisplay->protocol_names[pxdisplay->no_of_protocols] = strdup("WM_DELETE_WINDOW");
	pxdisplay->no_of_protocols++;
	pxdisplay->wm_protocols[pxdisplay->no_of_protocols] = XInternAtom(pxdisplay->display, "WM_PROTOCOLS",False);
	pxdisplay->protocol_names[pxdisplay->no_of_protocols] = strdup("WM_PROTOCOLS");
	pxdisplay->no_of_protocols++;
	pxdisplay->wm_protocols[pxdisplay->no_of_protocols] = XInternAtom(pxdisplay->display, "WM_NORMAL_HINTS",False);
	pxdisplay->protocol_names[pxdisplay->no_of_protocols] = strdup("WM_NORMAL_HINTS");
	pxdisplay->no_of_protocols++;

	return True;
}

void display_close_display(px_display* pxdisplay)
{
	unsigned int count;
	XCloseDisplay(pxdisplay->display);
	pxdisplay->display = NULL;

	for (count = 0; count < pxdisplay->no_of_protocols;count++)
	{
		free(pxdisplay->protocol_names[count]);
		pxdisplay->protocol_names[count] = NULL;
	}

	pxdisplay->no_of_protocols = 0;
}

void display_set_display(px_display* pxdisplay, Display* d)
{
	pxdisplay->display = d;

	pxdisplay->wm_protocols[pxdisplay->no_of_protocols] = XInternAtom(pxdisplay->display, "WM_DELETE_WINDOW",False);
	pxdisplay->protocol_names[pxdisplay->no_of_protocols] = strdup("WM_DELETE_WINDOW");
	pxdisplay->no_of_protocols++;
	pxdisplay->wm_protocols[pxdisplay->no_of_protocols] = XInternAtom(pxdisplay->display, "WM_PROTOCOLS",False);
	pxdisplay->protocol_names[pxdisplay->no_of_protocols] = strdup("WM_PROTOCOLS");
	pxdisplay->no_of_protocols++;
	pxdisplay->wm_protocols[pxdisplay->no_of_protocols] = XInternAtom(pxdisplay->display, "WM_NORMAL_HINTS",False);
	pxdisplay->protocol_names[pxdisplay->no_of_protocols] = strdup("WM_NORMAL_HINTS");
	pxdisplay->no_of_protocols++;
}

void display_register_wm_protocol(px_display* pxdisplay, const char* protocol)
{
	if (pxdisplay->protocols_allocated == pxdisplay->no_of_protocols)
	{
		unsigned int count;
				
		Atom* new_list = (Atom*)malloc(sizeof(Atom) * (pxdisplay->protocols_allocated + 5));
		char** new_strlist = (char**)malloc(sizeof(char*) * (pxdisplay->protocols_allocated + 5));
			
		for (count = 0; count < pxdisplay->no_of_protocols; count++)
		{
			new_list[count] = pxdisplay->wm_protocols[count];
			new_strlist[count] = pxdisplay->protocol_names[count];
		}

		free(pxdisplay->wm_protocols);
		free(pxdisplay->protocol_names);
		pxdisplay->wm_protocols = new_list;
		pxdisplay->protocol_names = new_strlist;
		pxdisplay->protocols_allocated += 5;
	}
	pxdisplay->wm_protocols[pxdisplay->no_of_protocols] = XInternAtom(pxdisplay->display, protocol, False);
	pxdisplay->protocol_names[pxdisplay->no_of_protocols] = strdup(protocol);
	pxdisplay->no_of_protocols++;
}

Atom display_get_atom(px_display* pxdisplay, const char* protocol)
{
	int count = 0;

	for (count = 0; count < pxdisplay->no_of_protocols; count++)
		if (strcmp(pxdisplay->protocol_names[count], protocol) == 0)
			return pxdisplay->wm_protocols[count];

	return pxdisplay->wm_protocols[0];
}

px_clipboard* display_get_clipboard(px_display* display, const char* atom)
{
	if (atom != NULL && atom[0] == 's')
	{
		// Get XA_secondary

		if (display->xa_secondary == NULL)
		{
			display->xa_secondary = clipboard_create();
			display->xa_secondary->selection = XA_SECONDARY;
		}

		return display->xa_secondary;
	}
	
	// Get XA_PRIMARY
		
	if (display->xa_primary == NULL)
	{
		display->xa_primary = clipboard_create();
		display->xa_primary->selection = XA_PRIMARY;
	}

	return display->xa_primary;
}

Window display_get_root_window()
{
	px_display* pxdisplay = display_get_display();
	Window win;
	
	if (pxdisplay->display == NULL)
	{
		fprintf(stderr, "Display has not been set. Please set the display by calling either \"open_display\" or \"set_display\" function in px_display structure before calling any function in the picowidget library\n");
		exit(1);
	}

	win = RootWindow(pxdisplay->display, DefaultScreen(pxdisplay->display));
	return win;
}

void display_flush(px_display* display)
{
	XFlush(display->display);
}

void display_sync(px_display* display)
{
	XSync(display->display, False);
}

void display_destroy()
{
	picowidget_display_variable->close_display(picowidget_display_variable);

	free(picowidget_display_variable);

	picowidget_display_variable = NULL;
}
