/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include "eventloop.h"
#include "display.h"

#include "font.h"
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <X11/Xatom.h>

struct _redraw_queue {
	px_widget *widget;
	int show;
	struct _redraw_queue *next;
};
static void redraw_queue_add(px_eventloop *eloop, px_widget *widget, int show);
static void redraw_queue_process(px_eventloop *eloop);
static px_widget *recent_redraw_widget;

void eventloop_handle_key_events(px_eventloop* eventloop, px_event* event);
px_eventloop* eventloop_get_eventloop()
{
	static px_eventloop* eventloop = NULL;

	if (eventloop == NULL)
	{
		eventloop = eventloop_create();
	}

	return eventloop;
}

px_eventloop* eventloop_create()
{
	px_eventloop* eventloop = (px_eventloop*)malloc(sizeof(px_eventloop));

	// Initialize all structure variables
	eventloop->run = 1;
	eventloop->grabbed_widget = NULL;
	eventloop->focussed_widget = NULL;

	eventloop->toplevels = (px_widget**)malloc(sizeof(px_widget*) * 5);
	eventloop->no_of_toplevels = 0;
	eventloop->toplevels_allocated = 5;

	eventloop->condemned = (px_widget**)malloc(sizeof(px_widget*) * 5);
	eventloop->no_of_condemned = 0;
	eventloop->condemned_allocated = 5;
	
	eventloop->timers = NULL;
	eventloop->secondary_timers = NULL;
	eventloop->timer_to_delete = NULL;
	eventloop->preserve_timers = False;

	eventloop->delay.tv_sec = 0;
	eventloop->delay.tv_nsec = 10000L;

	// Initialize function pointers
	eventloop->grab = eventloop_grab;
	eventloop->focus = eventloop_focus;
	eventloop->register_toplevel = eventloop_register_toplevel;
	eventloop->remove_toplevel = eventloop_remove_toplevel;
	eventloop->set_delay = eventloop_set_delay;
	eventloop->get_toplevel = eventloop_get_toplevel;
	
	eventloop->add_timer = eventloop_add_timer;
	eventloop->add_timer_struct = eventloop_add_timer_struct;
	eventloop->delete_timer = eventloop_delete_timer;
	eventloop->fire_timers = eventloop_fire_timers;

	eventloop->condemn = eventloop_condemn;

	eventloop->destroy = eventloop_destroy;
	eventloop->exit = eventloop_exit;

	eventloop->loop = eventloop_loop;
	eventloop->do_one_event = eventloop_do_one_event;
	eventloop->handle_event = eventloop_handle_event;
	
	eventloop->ignore = eventloop_ignore;

	eventloop->redraw_queue = NULL;
	eventloop->redraw_queue_process = redraw_queue_process;
	eventloop->redraw_queue_add = redraw_queue_add;

	return eventloop;
}

void eventloop_grab(px_eventloop* eventloop, px_widget* widget)
{
	eventloop->grabbed_widget = widget;
}

void eventloop_focus(px_eventloop* eventloop, px_widget* widget)
{
	if (eventloop->focussed_widget != NULL)
	{
		eventloop->focussed_widget->set_focus(eventloop->focussed_widget, False);
	}
	eventloop->focussed_widget = widget;

	if (widget != NULL)
	{
		px_widget* parent = widget;
		Window focussed_win = None;
		Display* display = display_get_Xdisplay();
		while (parent->child_of != NULL)
			parent = parent->child_of;

		if (!parent->hidden)
		{
			int dummy;
			XGetInputFocus(display, &focussed_win, &dummy);
			
			if (focussed_win != parent->win)
			{
				XGrabServer(display);
				XSetInputFocus(display, parent->win, RevertToPointerRoot, CurrentTime);
				XUngrabServer(display);
			}

			widget->set_focus(widget, True);
		}
	}
}

void eventloop_register_toplevel(px_eventloop* eventloop, px_widget* widget)
{
	if (eventloop->toplevels_allocated == eventloop->no_of_toplevels)
	{
		unsigned int count;
				
		px_widget** new_list = (px_widget**)malloc(sizeof(px_widget*) * 
						(eventloop->toplevels_allocated + 5));
			
		for (count = 0; count < eventloop->no_of_toplevels; count++)
			new_list[count] = eventloop->toplevels[count];

		free(eventloop->toplevels);
		eventloop->toplevels = new_list;
		eventloop->toplevels_allocated += 5;
	}
	eventloop->toplevels[eventloop->no_of_toplevels] = widget;
	eventloop->no_of_toplevels++;
}

void eventloop_remove_toplevel(px_eventloop* eventloop, px_widget* widget)
{
	unsigned int count = 0, index = eventloop->no_of_toplevels + 1;
	

	for (count = 0; count < eventloop->no_of_toplevels && index > eventloop->no_of_toplevels; count++)
	{
		if (eventloop->toplevels[count] == widget)
			index = count;
	}

	if (index < eventloop->no_of_toplevels)
	{
		for (count = index + 1; count < eventloop->no_of_toplevels; count++)
			eventloop->toplevels[count-1] = eventloop->toplevels[count];
		eventloop->no_of_toplevels--;
	}
}
		
void eventloop_set_delay(px_eventloop* eventloop, const struct timespec* delay)
{
	eventloop->delay.tv_sec = delay->tv_sec;
	eventloop->delay.tv_nsec = delay->tv_nsec;
}

px_widget* eventloop_get_toplevel(px_eventloop* eventloop, Window window)
{
	int count;

	for (count = 0; count < eventloop->no_of_toplevels; count++)
		if (window == eventloop->toplevels[count]->win)
			return eventloop->toplevels[count];

	return NULL;
}

px_timer* eventloop_add_timer(px_eventloop* eventloop, long delay_secs, long delay_usecs, 
		Bool (*fn_ptr)(void*, void*), void* param1, void* param2, Bool repeat)
{
	px_timer* timer = (px_timer*)malloc(sizeof(px_timer));
	struct timeval start;
	const long MILLION = 1000000;
	
	gettimeofday(&start, NULL);
	timer->delay.tv_sec = delay_secs;
	timer->delay.tv_usec = delay_usecs;
	
	timer->fire_at.tv_usec=start.tv_usec + delay_usecs;
	timer->fire_at.tv_sec=start.tv_sec + delay_secs;
	if (timer->fire_at.tv_usec > MILLION)
	{
		timer->fire_at.tv_usec = timer->fire_at.tv_usec - MILLION;
		timer->fire_at.tv_sec++;
	}

	timer->function = fn_ptr;
	timer->parameter_1 = param1;
	timer->parameter_2 = param2;
	timer->repeat = repeat;
	timer->next = NULL;

	if (eventloop->preserve_timers)
	{
		eventloop->secondary_timers = eventloop_add_timer_struct(eventloop, eventloop->secondary_timers, timer);
	}
	else
	{
		eventloop->timers = eventloop_add_timer_struct(eventloop, eventloop->timers, timer);
	}

	return timer;
}

px_timer* eventloop_add_timer_struct(px_eventloop* eventloop, px_timer* list_start, px_timer* timer)
{
	if (list_start == NULL) {
		list_start = timer;
	} else {
		px_timer *node=list_start, *prev=NULL;

		while ((node != NULL) && (node != timer)) {
			node = node->next;
		}
		if (node == NULL) {
			node = list_start;
			while ((node != NULL) && 
				(node->fire_at.tv_sec <= 
				 	timer->fire_at.tv_sec) && 
				(node->fire_at.tv_usec <= 
				 	timer->fire_at.tv_usec)) {
				prev = node;
				node = node->next;
			}
			if (prev == NULL) {
				timer->next = list_start;
				list_start = timer;
			} else {
				timer->next = prev->next;
				prev->next = timer;
			}
		}
	}
	return list_start;
}

void eventloop_delete_timer(px_eventloop* eventloop, px_timer* timer)
{
	if (eventloop->timers != NULL)
	{
		if (eventloop->preserve_timers)
		{
			timer->function = NULL;
			timer->fire_at.tv_sec = 0;
			timer->fire_at.tv_usec = 0;
			timer->repeat = False;
		}
		else
		{
			px_timer *node=eventloop->timers, *prev=NULL;

			while (node != NULL && node != timer)
			{
				prev = node;
				node = node->next;
			}

			if (node == timer)
			{
				if (prev == NULL)
				{
					eventloop->timers = timer->next;
				}
				else
				{
					prev->next = node->next;
				}
				free(timer);
			}
		}
	}
#if 0
	{
		px_timer* node = eventloop->timers;

		printf("%s() : The timer list contains : ", __func__);
		while (node != NULL)
		{
			printf("%p (%d): ", node, node->repeat);
			node = node->next;
		}
		printf("\n");
	}
#endif
}

void eventloop_fire_timers(px_eventloop* eventloop)
{
	if (eventloop->timers != NULL)
	{
		struct timeval now;
		px_timer *node=eventloop->timers, *nodecopy=eventloop->timers;
		
		eventloop->preserve_timers = True;
		gettimeofday(&now, NULL);

/*		if (node != NULL)
		{
			printf ("NOW = %d : %d & node = %d : %d  diff = %d : %d\n", now.tv_sec, now.tv_usec, node->fire_at.tv_sec,  node->fire_at.tv_usec, now.tv_sec - node->fire_at.tv_sec, now.tv_usec - node->fire_at.tv_usec);
		}
*/
		while ((node != NULL) && 
			(node->fire_at.tv_sec < now.tv_sec ? 1 : 
			(node->fire_at.tv_sec == now.tv_sec &&
			node->fire_at.tv_usec <= now.tv_usec))) {
#ifdef __debug__
			//printf("Will fire timer %p\n", node);
#endif
			if (node->function != NULL)
				node->function(node->parameter_1, node->parameter_2);

			if (node->repeat) {
				const long MILLION = 1000000;
				px_timer *current=node->next, *prev=NULL;
				
				node->fire_at.tv_usec = now.tv_usec + node->delay.tv_usec;
				node->fire_at.tv_sec = now.tv_sec + node->delay.tv_sec;

				while (node->fire_at.tv_usec >= MILLION) {
					node->fire_at.tv_usec -= MILLION;
					node->fire_at.tv_sec++;
				}
				
				while ((current != NULL) &&  
					(current->fire_at.tv_sec <= 
					 	node->fire_at.tv_sec) &&
					(current->fire_at.tv_usec <= 
					 	node->fire_at.tv_usec)) {
					prev = current;
					current = current->next;
				}

				if (prev != NULL) {
					if (nodecopy == eventloop->timers) {
						eventloop->timers = node->next;
						nodecopy = eventloop->timers;
					}
					node->next = prev->next;
					prev->next = node;
				}
			} else {
				if (nodecopy == eventloop->timers) {
					eventloop->timers = node->next;
					nodecopy = eventloop->timers;
				} else {
					px_timer *prev=eventloop->timers;

					while (prev->next != node) {
						prev = prev->next;
					}
					if (prev != NULL) {
						prev->next = node->next;
						nodecopy = prev;
					}
				}
				free(node); node = NULL;
			}

			if (nodecopy != NULL)
				node = nodecopy->next;
			else
				node = NULL;
			nodecopy = node;
		}
		eventloop->preserve_timers = False;

		if (eventloop->secondary_timers != NULL)
		{
			px_timer *sec=eventloop->secondary_timers, *next=NULL;

			while (sec != NULL)
			{
				next = sec->next;
				sec->next = NULL;
				eventloop->timers = eventloop->add_timer_struct(eventloop, eventloop->timers, sec);
				sec = next;
			}
			eventloop->secondary_timers = NULL;
		}
	}

#if 0
	{
		px_timer* node = eventloop->timers;

		printf("%s() : The timer list contains : ", __func__);
		while (node != NULL)
		{
			printf("%p : ", node);
			node = node->next;
		}
		printf("\n");
	}
#endif
}

void eventloop_destroy(px_eventloop* eventloop)
{
	px_timer* node = eventloop->timers, *next = NULL;
	unsigned int count;
	
	while (node != NULL)
	{
		next = node->next;
		free(node);
		node = next;
	
	}

	for (count = 0; count < eventloop->no_of_toplevels; count++)
	{
		eventloop->toplevels[count]->destroy(eventloop->toplevels[count]);
		free(eventloop->toplevels[count]);
	}
}

void eventloop_condemn(px_eventloop* e, px_widget* w)
{
	if (e->no_of_condemned == e->condemned_allocated)
	{
		
		px_widget** list = (px_widget**)malloc(sizeof(px_widget*) * (e->condemned_allocated + 5));
		int count;

		for (count = 0; count < e->no_of_condemned;count++)
			list[count] = e->condemned[count];

		free(e->condemned);
		e->condemned_allocated += 5;
		e->condemned = list;
	}

	e->condemned[e->no_of_condemned++] = w;
}
		
void eventloop_do_one_event(px_eventloop* eventloop)
{
	Display* display = display_get_Xdisplay();
	if ( XPending(display)) 
 	{
		px_widget* toplevel;
		XEvent an_event;
		px_event event;
		XNextEvent(display, &an_event);
#ifdef __debug__
//		printf("%s() : Got event %d on %d\n", __func__, an_event.type, an_event.xany.window);
#endif

		event.xev = &an_event;
		event.private_data = NULL;
		event.type = NR_CALLBACKS;
		event.x = 0;
		event.y = 0;
		toplevel = eventloop->get_toplevel(eventloop, event.xev->xany.window);

		eventloop_handle_event(eventloop, &event, toplevel);
	}
}

void eventloop_handle_event(px_eventloop* eventloop, px_event* event, px_widget* toplevel)
{
	int x, y;	
	px_display* pxdisplay = display_get_display();
	Display* display = pxdisplay->display;

	switch (event->xev->type) 
	{
		case ClientMessage:
			event->type = CLIENT_MESSAGE;
  			break;
	       		
		case ConfigureNotify:
			event->type = CONFIGURE;
#ifdef __debug__
			printf("%s() : Configure event for %lu , width = %d, height = %d\n", __func__, 
					event->xev->xany.window, event->xev->xconfigure.width,
					event->xev->xconfigure.height);
#endif
			
  			break;
	       
		case Expose:
			event->type = EXPOSE;
	 		break;

		case MapNotify:
			event->type = MAP;
			break;

		case UnmapNotify:
			event->type = UNMAP;
			break;

		case VisibilityNotify:
			event->type = VISIBILITY;
			break;
			
		case ButtonPress:
			event->type = BUTTON_PRESSED;
			x = event->xev->xbutton.x;
			y = event->xev->xbutton.y;
			break;

		case ButtonRelease:
			event->type = BUTTON_RELEASED;
			x = event->xev->xbutton.x;
			y = event->xev->xbutton.y;
			break;

		case MotionNotify:
			event->type = MOTION;
			x = event->xev->xmotion.x;
			y = event->xev->xmotion.y;
			break;

		case LeaveNotify:
			event->type = MOUSE_LEAVE;
			break;

		case EnterNotify:
			event->type = MOUSE_ENTER;
			break;
					
		case KeyPress:
			event->type = KEY_PRESSED;
			break;

		case KeyRelease:
			event->type = KEY_RELEASED;
			break;

		case MappingNotify:
			XRefreshKeyboardMapping(&(event->xev->xmapping));
			break;

		case FocusIn:
			event->type = FOCUSIN;
			if (eventloop->focussed_widget != NULL)
			{
				eventloop->focussed_widget->handle_event(eventloop->focussed_widget, event);
			}
			else if (toplevel != NULL)
			{
				eventloop->focus(eventloop, toplevel);
			}
			event->type = NR_CALLBACKS;
			break;
		case FocusOut:
			event->type = FOCUSOUT;
			if (eventloop->focussed_widget != NULL)
			{
				eventloop->focussed_widget->handle_event(eventloop->focussed_widget, event);
			}
			event->type = NR_CALLBACKS;
			break;
		case SelectionClear:
			{
				px_clipboard* clip = NULL;
				if (event->xev->xselectionclear.selection == XA_SECONDARY)
					clip = pxdisplay->get_clipboard(pxdisplay, "secondary");
				else
					clip = pxdisplay->get_clipboard(pxdisplay, "primary");

				if (clip != NULL)
				{
					clip->selection_clear(clip, event);
				}
			}

			event->type = NR_CALLBACKS;
			break;
				
		case SelectionNotify:
			{
				px_clipboard* clip = NULL;
				if (event->xev->xselection.selection == XA_SECONDARY)
					clip = pxdisplay->get_clipboard(pxdisplay, "secondary");
				else
					clip = pxdisplay->get_clipboard(pxdisplay, "primary");

				if (clip != NULL)
				{
					clip->selection_notify(clip, event);
				}
			}
			
			event->type = NR_CALLBACKS;
			break;

		case SelectionRequest:
			{
				px_clipboard* clip = NULL;
				if (event->xev->xselectionrequest.selection == XA_SECONDARY)
					clip = pxdisplay->get_clipboard(pxdisplay, "secondary");
				else
					clip = pxdisplay->get_clipboard(pxdisplay, "primary");

				if (clip != NULL)
				{
					clip->send_selection(clip, event);
				}
			}
			event->type = NR_CALLBACKS;
			break;
			
		case PropertyNotify :
			{
				Atom inc = XInternAtom(display, "INCR", False);

				if (event->xev->xproperty.atom == inc)
				{
					px_clipboard* clip = NULL;
					if (event->xev->xproperty.atom == XA_SECONDARY)
						clip = pxdisplay->get_clipboard(pxdisplay, "secondary");
					else
						clip = pxdisplay->get_clipboard(pxdisplay, "primary");
	
					if (clip != NULL)
					{
						clip->handle_incr_mechanism(clip, event);
					}
				}
			}
			event->type = NR_CALLBACKS;
			break;
						
		default:
			break;
    	}

	if (event->type != NR_CALLBACKS) {
		if (eventloop->grabbed_widget != NULL) {
			if (event->type < 3) {
				event->x = x - eventloop->grabbed_widget->get_absolute_x
					(eventloop->grabbed_widget, event->xev->xany.window);
				event->y = y - eventloop->grabbed_widget->get_absolute_y
					(eventloop->grabbed_widget, event->xev->xany.window);
			}
			eventloop->grabbed_widget->handle_event(eventloop->grabbed_widget, event);
		} else if ((event->type == KEY_PRESSED || event->type == KEY_RELEASED) && 
				eventloop->focussed_widget != NULL)
		{
			eventloop_handle_key_events(eventloop, event);
		}
		else if (toplevel != NULL) {
			if (event->type < 3) {
				if (event->xev->xany.window != toplevel->win) {
					event->x = x - toplevel->x;
					event->y = y - toplevel->y;
				} else {
					event->x = x;
					event->y = y;
				}
			}

			toplevel->handle_event(toplevel, event);
		}
		if (eventloop->no_of_toplevels == 0)
			eventloop->exit(eventloop);

	}
}
	
void eventloop_loop(px_eventloop* eventloop)
{
	unsigned int count;

	if (eventloop->no_of_toplevels == 0)
	{
		printf("There are no active toplevels!!\n");
	}
	else
	{
		eventloop->redraw_queue_process(eventloop);
		while (eventloop->run)
		{
			eventloop->do_one_event(eventloop);
			eventloop->redraw_queue_process(eventloop);
			eventloop->fire_timers(eventloop);

			// Santhosh : Clear the condemned widgets.
			for (count = 0; count < eventloop->no_of_condemned; count++) {
				eventloop->condemned[count]->destroy(eventloop->condemned[count]);
				free(eventloop->condemned[count]);
			}
			eventloop->no_of_condemned = 0;
			
			nanosleep(&(eventloop->delay),NULL); // sleep for a 10th of a second 
		}
	}
}

void eventloop_exit(px_eventloop* eventloop)
{
	eventloop->run = 0;
}

void eventloop_handle_key_events(px_eventloop* eventloop, px_event* event)
{
	px_widget* parent = eventloop->focussed_widget;
	Bool ret = True;
	
	while(ret && parent != NULL)
	{
		ret = parent->handle_event(parent, event);
		parent = parent->child_of;
	}
}

void eventloop_ignore(px_eventloop* eventloop, widget_callbacks type, int no_of_events)
{
	Display* display = display_get_Xdisplay();
	int event_type;
	XEvent xev;

	int num = 0;
	
	switch (type) 
	{
		case CLIENT_MESSAGE:
			event_type = ClientMessage;
  			break;
	       		
		case CONFIGURE:
			event_type = ConfigureNotify;
  			break;
	       
		case EXPOSE:
			event_type = Expose;
	 		break;

		case MAP:
			event_type = MappingNotify;
			break;

		case UNMAP:
			event_type = UnmapNotify;
			break;

		case VISIBILITY:
			event_type = VisibilityNotify;
			break;
			
		case BUTTON_PRESSED:
			event_type = ButtonPress;
			break;

		case BUTTON_RELEASED:
			event_type = ButtonRelease;
			break;

		case MOTION:
			event_type = MOTION;
			break;

		case MOUSE_LEAVE:
			event_type = LeaveNotify;
			break;

		case MOUSE_ENTER:
			event_type = EnterNotify;
			break;
					
		case KEY_PRESSED:
			event_type = KeyPress;
			break;

		case KEY_RELEASED:
			event_type = KeyRelease;
			break;
		case FOCUSIN:
			event_type = FocusIn;
			break;
		case FOCUSOUT:
			event_type = FocusOut;
			break;
		default:
			return;
    	}
	while(XPending(display) && num < no_of_events) {
		XPeekEvent(display, &xev);
		
		if (xev.type == event_type)
		{
			XNextEvent(display_get_Xdisplay(), &xev);
			num++;
		}
		else
			break;
	}
}

static void redraw_queue_add(px_eventloop *eloop, px_widget *widget, int show)
{
	struct _redraw_queue *node, *p, *q;

	if(eloop->redraw_queue == NULL) {
		node = (struct _redraw_queue*)malloc(sizeof(struct _redraw_queue));
		node->widget = widget;
		node->show = show;
		node->next =  NULL;
		eloop->redraw_queue = node;
		recent_redraw_widget = widget;
	} else {
		p = eloop->redraw_queue;

		while(p) {
			if(recent_redraw_widget == widget || p->widget == widget) {
				recent_redraw_widget = widget;
				break;
			}
			q = p;
			p = p->next;
		}

		if(p == NULL) {
			node = (struct _redraw_queue*)malloc(sizeof(struct _redraw_queue));
			node->widget = widget;
			node->show = show;
			node->next = NULL;
			q->next = node;
			recent_redraw_widget = widget;
		}
	}
}

static void redraw_queue_process(px_eventloop *eloop)
{
	struct _redraw_queue *p;

	// Redraw all the widgets in the queue first.
	p = eloop->redraw_queue;
	while(p) {
		p->widget->redraw(p->widget);
		p = p->next;
	}
	
	// Show all the widgets in the queue.
	while(eloop->redraw_queue) {
		p = eloop->redraw_queue;
		if(p->show)
			p->widget->show(p->widget);

		eloop->redraw_queue = p->next;
		free(p);
	}
}

