/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include "font.h"
#include "display.h"
#include <string.h>
#include <stdlib.h>

FontCache* fontcache = NULL;

FontCache* font_search_font(const char*);
void font_cleanup_list();

px_font* font_load_font(const char* fontname)
{
	FontCache* node = font_search_font(fontname);
	XFontStruct* f = NULL;
	Display* display = display_get_Xdisplay();

#ifdef __debug__
	printf("%s() : Searching for %s\n", __func__, fontname);
#endif
	if (node == NULL)
	{
		int count;
		f = XLoadQueryFont(display, fontname);
		if (f == NULL) 
		{
			fprintf(stderr, "label: failed loading font '%s'\n", fontname);
			return NULL;
      		}
		node = (FontCache*)malloc(sizeof(FontCache));
		node->font_string = (char*)malloc(sizeof(char) * (strlen(fontname) + 1));
		strcpy(node->font_string, fontname);
		node->font = (px_font*)malloc(sizeof(px_font));
		node->font->font = f;
		node->references = 0;
		node->next = fontcache;
		fontcache = node;
#ifdef __debug__
		printf("Did not find it. created %p\n", fontcache);
#endif
	/*	printf("min_char = %u, max_char = %u, min_byte = %u, max_byte=%u, all chars = %d, per_char = %p\n", f->min_char_or_byte2, f->max_char_or_byte2, f->min_byte1, f->max_byte1, f->all_chars_exist, f->per_char);
		for (count = f->min_char_or_byte2; count < f->max_char_or_byte2; count++)
		{
			printf("%c - lbearing = %d, rbearing = %d, width = %d, acsent = %d, desenct = %d\n", count, (f->per_char+count-f->min_char_or_byte2)->lbearing, (f->per_char+count-f->min_char_or_byte2)->rbearing, (f->per_char + count-f->min_char_or_byte2)->width, (f->per_char + count-f->min_char_or_byte2)->ascent, (f->per_char + count-f->min_char_or_byte2)->descent);
		}*/
	}

	node->references++;
#ifdef __debug__
	printf("References for %s is %d\n", fontname, node->references);
#endif
	node->lifespan = 2;
	return node->font;
}

void font_free_font(const char* fontname)
{
	FontCache* node = font_search_font(fontname);
	if (node != NULL)
	{
		node->references--;
		font_cleanup_list();
	}
}

FontCache* font_search_font(const char* fontname)
{
	FontCache* node = fontcache;

	while (node != NULL)
	{
		if (strcmp(fontname, node->font_string) == 0)
			return node;
		node = node->next;
	}

	return NULL;
}

void font_cleanup_list()
{
	FontCache* node = fontcache, *prevnode = NULL;

	while (node != NULL)
	{
		if (node->references == 0)
		{
			if (node->lifespan == 0)
			{
				Display* display = display_get_Xdisplay();
				if (prevnode != NULL)
					prevnode->next = node->next;
				else
					fontcache = node->next;

				free(node->font_string);
				XFreeFont(display, node->font->font);
				free(node->font);
				free(node);

				if (prevnode != NULL)
					node = prevnode;
				else
					node = fontcache;
			}
			else
			{
				node->lifespan--;
			}
		}
		prevnode = node;
		node = node->next;
	}
}

void font_destroy_all()
{
	FontCache* node = fontcache;
	Display* display = display_get_Xdisplay();
	
	while (node != NULL)
	{
		FontCache* next = node->next;

		free(node->font_string);
		XFreeFont(display, node->font->font);
		free(node->font);
		free(node);

		node = next;
	}
}

int font_get_width_of_char(const px_font* font, const char c)
{
	if (font->font->min_byte1 == 0 && font->font->max_byte1 == 0)
	{
		if (c >= font->font->min_char_or_byte2 && c <= font->font->max_char_or_byte2)
		{
			return (font->font->per_char + c - font->font->min_char_or_byte2)->width;
		}

	}

	return 0;
}

int font_get_width(const px_font* font, const char* text, unsigned int length)
{
	int width;
	
	if (font->font != NULL)
		width = XTextWidth(font->font, text, length);
	else
		width = 0;

	return width;
}

int font_get_height(const px_font* font, const char* text)
{
	int height;
	
	if (font->font != NULL)
		height = font->font->ascent + font->font->descent;
	else
		height = 0;

	return height;
}


