/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include <X11/extensions/shape.h>
#include <stdlib.h>
#include <string.h>
#include "frame.h"
#include "image.h"
#include "widget_factory.h"
#include "display.h"

px_widget* frame_create(px_widget* parent, int x, int y, int width, int height, const char* tag)
{
	px_widget* widget = _frame_create(parent->win, x, y, width, height, tag);
	return widget;
}


px_widget* _frame_create(Window win, int x, int y, int width, int height, const char* tag)
{
	px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
	Display* display = display_get_Xdisplay();

	// Initialize common widget variables
	
	widget->win = XCreateWindow(display, win, x, y, width, height, 0,
			CopyFromParent, CopyFromParent, CopyFromParent, 0, NULL);


	widget->x = x;
	widget->y = y;
	widget->height = height;
	widget->width = width;
	widget->wtype = FRAME;
	widget->child_of = NULL;
	widget->use_pixmap = False;
	if (tag != NULL)
	{
		widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	}
	else
		widget->tag = NULL;

	widget->background.background_color = (px_color*)malloc(sizeof(px_color));
	widget->background.background_color->type = SOLID;
	widget_xcolor_from_str(&(widget->background.background_color->first_color), "white");


	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);

#ifdef __debug__
	printf("%s(): %p - Created Window with parent window = %lu, x = %d, y = %d, width = %d, height = %d, new window = %lu\n\n", 
			__func__, widget, win, x, y, width, height, widget->win);
#endif
	

	return widget;
}

void frame_set_parameters(px_widget* widget)
{
	int i;

	widget->default_callbacks[BUTTON_PRESSED] = frame_default_button_press;
	widget->default_callbacks[MOTION] = frame_default_motion;
	widget->default_callbacks[BUTTON_RELEASED] = frame_default_button_release;
	widget->default_callbacks[MOUSE_ENTER] = frame_default_button_press;
	widget->default_callbacks[MOUSE_LEAVE] = frame_default_mouse_leave;
	widget->default_callbacks[EXPOSE] = frame_default_expose;

	widget->handle_event = frame_handle_event;
	widget->redraw = frame_redraw;
	widget->show = frame_show;
	widget->destroy = frame_destroy;
	widget->is_transparent = toplevel_frame_is_transparent;
	widget->cleanup_internal_structures = frame_cleanup_data_structures;


	// Initialize widget specific variables 

	widget->specific.frame.widget_list = (px_widget**)malloc(sizeof(px_widget*) * 5);

	//added by Jyotirmoy
	for (i = 0; i < 5; i++)
		widget->specific.frame.widget_list[i] = NULL;
	
	
	widget->specific.frame.allocated = 5;
	widget->specific.frame.no_of_children = 0;
	widget->specific.frame.event_within = NULL;
	widget->specific.frame.ynodes = NULL;
	widget->specific.frame.axis = YAXIS;
	widget->specific.frame.count_within = -1;

	widget->specific.frame.add_child = frame_add_child;
	widget->specific.frame.remove_child = frame_remove_child;
}

void frame_add_child(px_widget* widget, px_widget* child)
{
	px_frame* frame = &(widget->specific.frame);

	if (frame->allocated == frame->no_of_children)
	{
		unsigned int count;
				
		px_widget** new_list = (px_widget**)malloc(sizeof(px_widget*) * 
				(frame->allocated + 10));
			
		for (count = 0; count < frame->no_of_children; count++)
			new_list[count] = frame->widget_list[count];

		free(frame->widget_list);
		frame->widget_list = new_list;
		frame->allocated += 10;
	}
	frame->widget_list[frame->no_of_children] = child;
	frame->no_of_children++;
	if (!widget->preserve_internal_structures)
	{
		frame->ynodes = widget_add_child(frame->ynodes, child, frame->axis == YAXIS ? YDIRECTION : XDIRECTION, frame->axis);
#ifdef __debug__
		printf("------- For frame %p -------------\n", widget);
		widget_print_frame_buffer(frame->ynodes, frame->axis == YAXIS ? YDIRECTION : XDIRECTION, frame->axis);
		printf("------- End For frame %p -------------\n", widget);
#endif
	}
	else
	{
		widget->dirty = True;
	}
}

void frame_remove_child(px_widget* widget, px_widget* child)
{
	px_frame* frame = &(widget->specific.frame);
	unsigned int count;

	for (count = 0; count < frame->no_of_children && frame->widget_list[count] != child ; count++);

	if (count < frame->no_of_children)
	{
		for (count++;count < frame->no_of_children;count++)
			frame->widget_list[count-1] = frame->widget_list[count];
		frame->no_of_children--;

		if (!widget->preserve_internal_structures)
		{
			frame->ynodes = widget_remove_child(frame->ynodes, child, 
					frame->axis == YAXIS ? YDIRECTION : XDIRECTION, frame->axis);

			
#ifdef __debug__
			printf("------- For frame %p -------------\n", widget);
			widget_print_frame_buffer(frame->ynodes, frame->axis == YAXIS ? YDIRECTION : XDIRECTION, frame->axis);
			printf("------- End For frame %p -------------\n", widget);
#endif
		}
		else
		{
			widget->dirty = True;
		}
	}
}

Bool frame_handle_event(px_widget* widget, px_event* event)
{
	unsigned int count = 0;
	Bool return_value = True;
#ifdef __debug__
	printf("%s() : %p Event type = %d at %d %d\n", __func__, widget, event->type, event->x, event->y);
#endif
	if (event->type != NR_CALLBACKS && (event->type == EXPOSE || widget->get_state(widget) != DISABLED))
	{

		return_value = widget_handle_event(widget, event);

		if (event->type > 6)
		{
			for (count = 0; return_value && count < widget->specific.frame.no_of_children; count++)
			{
				px_widget* child = widget->specific.frame.widget_list[count];
#ifdef __debug__
				printf("%s() : %p Found widget %p at %d : hidden %d\n", __func__,widget, child, 
						count, child->hidden);
#endif
				if (!child->hidden)
					return_value = child->handle_event(child, event);
			}
		}
	}
	return return_value;
}

Bool frame_default_button_press(px_widget* widget, px_event* event)
{
	px_frame* frame = &( widget->specific.frame);

	px_xynode* node = widget_get_xynode(frame->ynodes, event->x, event->y, frame->axis == YAXIS ? YDIRECTION : XDIRECTION,
			frame->axis);
	Bool return_value = True;
	
	widget->preserve_internal_structures = True;
	
	if (node != NULL)
	{
		int count;
		int event_given_to = -1;
		
		for (count = 0; count < node->no_of_widgets;count++)
		{
			if (!node->widget_list[count]->hidden && 
					node->widget_list[count]->get_state(node->widget_list[count]) != DISABLED)
			{
				int wx, wy;
				Bool transparency;
				
				wx = event->x - node->widget_list[count]->x;
				wy = event->y - node->widget_list[count]->y;
	
				transparency = node->widget_list[count]->is_transparent(node->widget_list[count], wx, wy);

				if (node->widget_list[count]->emask != EVENT_TRANSPARENT && 
						!(node->widget_list[count]->emask == EVENT_MASK && transparency))
				{
					event->x = wx;
					event->y = wy;
					if (return_value)
					{
						return_value = node->widget_list[count]->handle_event
							(node->widget_list[count], event);
					}
					event_given_to = count;
					event->x += node->widget_list[count]->x;
					event->y += node->widget_list[count]->y;
				}

				if (!transparency && !node->widget_list[count]->hidden)
				{
					break;
				}
			}
		}

		frame->count_within = event_given_to;
		frame->event_within = node;
		frame->last_x = event->x;
		frame->last_y = event->y;
	}

	widget->preserve_internal_structures = False;
	return return_value;
}

Bool frame_default_motion(px_widget* widget, px_event* event)
{
	px_frame* frame = &( widget->specific.frame);

	px_xynode* node = widget_get_xynode(frame->ynodes, event->x, event->y, frame->axis == YAXIS ? YDIRECTION : XDIRECTION,
			frame->axis);
	px_xynode* event_within = frame->event_within;
	int count_within = frame->count_within;
	Bool return_value = True;
	widget->preserve_internal_structures = True;
	
	if (!(node == NULL && event_within == NULL))
	{
		if (node == event_within)
		{
			// Give the motion event to all children
			int count;
			int event_given_to = -1;
			
			for (count = 0; count < node->no_of_widgets;count++)
			{
				if (!node->widget_list[count]->hidden && 
					node->widget_list[count]->get_state(node->widget_list[count]) != DISABLED)
				{
					int wx, wy;
					Bool transparency;
					wx = event->x - node->widget_list[count]->x;
					wy = event->y - node->widget_list[count]->y;
	
					transparency = node->widget_list[count]->is_transparent(node->widget_list[count], wx, wy);

					if (node->widget_list[count]->emask != EVENT_TRANSPARENT && 
							!(node->widget_list[count]->emask == EVENT_MASK && transparency))
					{
						event->x = wx;
						event->y = wy;

						if (return_value && count <= count_within)
						{
							return_value = node->widget_list[count]->handle_event
								(node->widget_list[count], event);
						}
						event_given_to = count;
						event->x += node->widget_list[count]->x;
						event->y += node->widget_list[count]->y;
					}

					if (!transparency && !node->widget_list[count]->hidden)
					{
						break;
					}
				}
			}

			if (event_given_to < count_within)
			{
				int ctr;
				Bool leave_return_value = True;
				event->type = MOUSE_LEAVE;
				for (ctr = event_given_to + 1; leave_return_value && ctr <= count_within;ctr++)
				{
					if (!node->widget_list[ctr]->hidden && 
							node->widget_list[ctr]->get_state(node->widget_list[ctr]) != DISABLED)
					{
						int wx, wy;
						Bool transparency;
						wx = frame->last_x - node->widget_list[ctr]->x;
						wy = frame->last_y - node->widget_list[ctr]->y;
		
						transparency = node->widget_list[ctr]->is_transparent
							(node->widget_list[ctr], wx, wy);
						if (node->widget_list[ctr]->emask != EVENT_TRANSPARENT && 
							!(node->widget_list[ctr]->emask == EVENT_MASK && transparency))
						{
#ifdef __debug__				
							printf("%s() 1. : %p Generating leave event on %p\n", __func__, widget, 
								node->widget_list[ctr]);
#endif
							leave_return_value = node->widget_list[ctr]->handle_event
								(node->widget_list[ctr], event);
						}
					}
				}
				event->type = MOTION;
			}
			else if (event_given_to > count_within)
			{
				int ctr;
				Bool enter_return_value = True;
				event->type = MOUSE_ENTER;
				for (ctr = count_within + 1; enter_return_value && ctr <= event_given_to;ctr++)
				{
					if (!node->widget_list[ctr]->hidden && 
							node->widget_list[ctr]->get_state(node->widget_list[ctr]) != DISABLED)
					{
						int wx, wy;
						Bool transparency;
						wx = event->x - node->widget_list[ctr]->x;
						wy = event->y - node->widget_list[ctr]->y;
	
						transparency = node->widget_list[ctr]->is_transparent
							(node->widget_list[ctr], wx, wy);
#ifdef __debug__				
						printf("%s() 1. : %p Generating Enter event on %p\n", __func__, widget, 
								node->widget_list[ctr]);
#endif
						if (node->widget_list[ctr]->emask != EVENT_TRANSPARENT && 
							!(node->widget_list[ctr]->emask == EVENT_MASK && transparency))
						{
							event->x = wx;
							event->y = wy;
							enter_return_value = node->widget_list[ctr]->handle_event
								(node->widget_list[ctr], event);
							event->x += node->widget_list[ctr]->x;
							event->y += node->widget_list[ctr]->y;
						}
							
					}
				}
				event->type = MOTION;
			}

			frame->count_within = event_given_to;
			frame->last_x = event->x;
			frame->last_y = event->y;
		}
		else if (node == NULL)
		{
			// Give leave events for all the widgets in event_within
			int count;
			Bool leave_return_value = True;

			for (count = 0; leave_return_value && count <= count_within;count++)
			{
				if (!event_within->widget_list[count]->hidden &&
						event_within->widget_list[count]->
						get_state(event_within->widget_list[count]) != DISABLED)
				{
					int wx, wy;
					Bool transparency;
					wx = frame->last_x - event_within->widget_list[count]->x;
					wy = frame->last_y - event_within->widget_list[count]->y;
		
					transparency = event_within->widget_list[count]->is_transparent
							(event_within->widget_list[count], wx, wy);
					event->type = MOUSE_LEAVE;
#ifdef __debug__				
					printf("%s() 2. : %p Generating leave event on %p\n", __func__, widget, 
							event_within->widget_list[count]);
#endif
							
					leave_return_value = event_within->widget_list[count]->handle_event
							(event_within->widget_list[count], event);
		
					event->type = MOTION;
				}
			}
			frame->event_within = NULL;
			frame->count_within = -1;
		}
		else if (event_within == NULL)
		{
			// Give Enter events for all the widgets in node
			int count;
			int event_given_to = -1;
			Bool enter_return_value = True;
			event->type = MOUSE_ENTER;
			for (count = 0; enter_return_value && count < node->no_of_widgets;count++)
			{
				if (!node->widget_list[count]->hidden && 
					node->widget_list[count]->get_state(node->widget_list[count]) != DISABLED)
				{
					int wx, wy;
					Bool transparency;
				
					wx = event->x - node->widget_list[count]->x;
					wy = event->y - node->widget_list[count]->y;
	
					transparency = node->widget_list[count]->is_transparent(node->widget_list[count], wx, wy);

					if (node->widget_list[count]->emask != EVENT_TRANSPARENT && 
							!(node->widget_list[count]->emask == EVENT_MASK && transparency))
					{
#ifdef __debug__				
						printf("%s() 2. : %p Generating Enter event on %p\n", __func__, widget, 
							node->widget_list[count]);
#endif
						event->x = wx;
						event->y = wy;
						enter_return_value = node->widget_list[count]->handle_event
							(node->widget_list[count], event);
						event_given_to = count;
						event->x += node->widget_list[count]->x;
						event->y += node->widget_list[count]->y;
					}

					if (!transparency && !node->widget_list[count]->hidden)
					{
						break;
					}
				}
			}
			event->type = MOTION;
			frame->count_within = event_given_to;
			frame->event_within = node;
			frame->last_x = event->x;
			frame->last_y = event->y;
		}
		else
		{
			// Find out the widgets that are there in node and not in event_within
			// and give them enter events
			
			int count;
			int new_count_within;
			int event_given_to = -1;
			Bool enter_return_value = True;	

			for (count = 0; enter_return_value && count < node->no_of_widgets;count++)
			{
				Bool transparency;
				int ctr;
				int wx, wy;

				wx = event->x - node->widget_list[count]->x;
				wy = event->y - node->widget_list[count]->y;
	
				transparency = node->widget_list[count]->is_transparent(node->widget_list[count], wx, wy);
			
				if (!node->widget_list[count]->hidden && 
					node->widget_list[count]->get_state(node->widget_list[count]) != DISABLED &&
					node->widget_list[count]->emask != EVENT_TRANSPARENT && 
					!(node->widget_list[count]->emask == EVENT_MASK && transparency))
				{
					for (ctr = 0; ctr <= count_within && enter_return_value && 
						event_within->widget_list[ctr]->depth  >=  node->widget_list[count]->depth &&
						node->widget_list[count] != event_within->widget_list[ctr];ctr++);

					if (ctr > count_within || 
							(node->widget_list[count] != event_within->widget_list[ctr]))
					{
#ifdef __debug__
						printf("%s() 3. : %p Generating ENter event on %p\n", __func__, widget, 
								node->widget_list[count]);
#endif
						event->x = wx;
						event->y = wy;
						event->type = MOUSE_ENTER;
						enter_return_value = node->widget_list[count]->handle_event
							(node->widget_list[count], event);
						event->x += node->widget_list[count]->x;
						event->y += node->widget_list[count]->y;
						event_given_to = count;
					}
					else
					{
						event->x = wx;
						event->y = wy;
						event->type = MOTION;
						enter_return_value = node->widget_list[count]->handle_event
							(node->widget_list[count], event);
						event->x += node->widget_list[count]->x;
						event->y += node->widget_list[count]->y;
						event_given_to = count;
					}
						
				}
				if (!transparency && !node->widget_list[count]->hidden)
					break;
			}
			event->type = MOTION;
			
			new_count_within = event_given_to;

			// Find out the widgets that are there in event_within and not in node
			// and give them leave events

			enter_return_value = True;
			event->type = MOUSE_LEAVE;
			for (count = 0; enter_return_value && count <= count_within;count++)
			{
				Bool transparency;
				int ctr;
				int wx, wy;

				wx = frame->last_x - event_within->widget_list[count]->x;
				wy = frame->last_y - event_within->widget_list[count]->y;
	
				transparency = event_within->widget_list[count]->is_transparent
					(event_within->widget_list[count], wx, wy);

				if (!event_within->widget_list[count]->hidden &&
					event_within->widget_list[count]->
					get_state(event_within->widget_list[count]) != DISABLED
					&& event_within->widget_list[count]->emask != EVENT_TRANSPARENT && 
					!(event_within->widget_list[count]->emask == EVENT_MASK && transparency))
				{
					for (ctr = 0; ctr <= new_count_within && enter_return_value && 
						node->widget_list[ctr]->depth >= event_within->widget_list[count]->depth &&
						node->widget_list[ctr] != event_within->widget_list[count];ctr++);

					if (ctr > new_count_within || 
							node->widget_list[ctr] != event_within->widget_list[count])
					{
						
#ifdef __debug__				
						printf("%s() 3. : %p Generating leave event on %p\n", __func__, widget, 
								event_within->widget_list[count]);
#endif
					
						event_within->widget_list[count]->handle_event
								(event_within->widget_list[count], event);
	
					}
				}
			}
			event->type = MOTION;

			frame->count_within = new_count_within;
			frame->event_within = node;
			frame->last_x = event->x;
			frame->last_y = event->y;
		}
	}
	widget->preserve_internal_structures = False;

	return return_value;
}

Bool frame_default_button_release(px_widget* widget, px_event* event)
{
	px_frame* frame = &( widget->specific.frame);

	int event_given_to = -1;
	
	px_xynode* node = widget_get_xynode(frame->ynodes, event->x, event->y, frame->axis == YAXIS ? YDIRECTION : XDIRECTION,
			frame->axis);
	
	Bool return_value = True;
	widget->preserve_internal_structures = True;
	
	if (node != NULL)
	{
		int count;

		for (count = 0; count < node->no_of_widgets;count++)
		{
			if (!node->widget_list[count]->hidden && 
				node->widget_list[count]->get_state(node->widget_list[count]) != DISABLED)
			{
				int wx, wy;
				Bool transparency;
				
				wx = event->x - node->widget_list[count]->x;
				wy = event->y - node->widget_list[count]->y;
	
				transparency = node->widget_list[count]->is_transparent(node->widget_list[count], wx, wy);

				if (node->widget_list[count]->emask != EVENT_TRANSPARENT && 
						!(node->widget_list[count]->emask == EVENT_MASK && transparency))
				{
					event->x = wx;
					event->y = wy;

					if (return_value)
					{
						return_value = node->widget_list[count]->handle_event
							(node->widget_list[count], event);
					}
					event_given_to = -1;
					event->x += node->widget_list[count]->x;
					event->y += node->widget_list[count]->y;
				}

				if (!transparency && !node->widget_list[count]->hidden)
				{
					count = node->no_of_widgets;
				}
			}
		}
	}

	if (frame->event_within != NULL && frame->event_within != node)
	{
		event->type = MOUSE_LEAVE;
		if (node != NULL)
		{
			// Find out the widgets that are there in event_within and not in node
			// and give them leave events

			int count;
			px_xynode* event_within = frame->event_within;
			Bool leave_return_value = True;
			
			for (count = 0; leave_return_value && count <= frame->count_within;count++)
			{
				int wx, wy;
				Bool transparency;
				int ctr;
				
				wx = frame->last_x - event_within->widget_list[count]->x;
				wy = frame->last_y - event_within->widget_list[count]->y;
	
				transparency = event_within->widget_list[count]->is_transparent
					(event_within->widget_list[count], wx, wy);


				if (!event_within->widget_list[count]->hidden &&
						event_within->widget_list[count]->
						get_state(event_within->widget_list[count]) != DISABLED &&
						event_within->widget_list[count]->emask != EVENT_TRANSPARENT && 
						!(event_within->widget_list[count]->emask == EVENT_MASK && transparency))
				{
					for (ctr = 0; ctr < event_given_to && leave_return_value && 
						node->widget_list[ctr]->depth >= event_within->widget_list[count]->depth &&
						node->widget_list[ctr] != event_within->widget_list[count];ctr++);

					if (ctr >= event_given_to || 
							node->widget_list[ctr] != event_within->widget_list[count])
					{
#ifdef __debug__				
						printf("%s() : %p Generating leave event on %p\n", __func__, widget, 
								event_within->widget_list[count]);
#endif
					
						leave_return_value = event_within->widget_list[count]->handle_event
								(frame->event_within->widget_list[count], event);
					}
				}
			}
		}
		else
		{
			// Have to give leave events for everything;
			int count;
			px_xynode* event_within = frame->event_within;
			Bool leave_return_value = True;
			
			for (count = 0; leave_return_value && count <= frame->count_within;count++)
			{
				if (!event_within->widget_list[count]->hidden &&
						event_within->widget_list[count]->
						get_state(event_within->widget_list[count]) != DISABLED)
				{
					int wx, wy;
					Bool transparency;
				
					wx = frame->last_x - event_within->widget_list[count]->x;
					wy = frame->last_y - event_within->widget_list[count]->y;
	
					transparency = event_within->widget_list[count]->is_transparent
						(event_within->widget_list[count], wx, wy);

					if (event_within->widget_list[count]->emask != EVENT_TRANSPARENT && 
							!(event_within->widget_list[count]->emask == EVENT_MASK && transparency))
					{
#ifdef __debug__				
						printf("%s() : %p Generating leave event on %p\n", __func__, widget, 
							event_within->widget_list[count]);
#endif
					
						leave_return_value = event_within->widget_list[count]->handle_event
							(event_within->widget_list[count], event);
					}
				}
			}
		}
		event->type = BUTTON_RELEASED;
	}
		
		
	frame->event_within = NULL;
	frame->count_within = -1;
	widget->preserve_internal_structures = False;

	return return_value;
}
	
	
Bool frame_default_mouse_leave(px_widget* widget, px_event* event)
{
	px_frame* frame = &( widget->specific.frame);
	Bool return_value = True;
	widget->preserve_internal_structures = True;
	
	if (frame->event_within != NULL)
	{
		// Have to give leave events for everything;
		int count;
		px_xynode* event_within = frame->event_within;

		for (count = 0; return_value && count <= frame->count_within;count++)
		{
			if (!event_within->widget_list[count]->hidden &&
					event_within->widget_list[count]->
					get_state(event_within->widget_list[count]) != DISABLED)
			{
				int wx, wy;
				Bool transparency;
				
				wx = frame->last_x - event_within->widget_list[count]->x;
				wy = frame->last_y - event_within->widget_list[count]->y;
	
				transparency = event_within->widget_list[count]->is_transparent
					(event_within->widget_list[count], wx, wy);

				if (event_within->widget_list[count]->emask != EVENT_TRANSPARENT && 
						!(event_within->widget_list[count]->emask == EVENT_MASK && transparency))
				{
#ifdef __debug__				
					printf("%s() : %p Generating leave event on %p\n", __func__, widget, 
						event_within->widget_list[count]);
#endif
					
					return_value = event_within->widget_list[count]->handle_event
						(event_within->widget_list[count], event);
				}
			}
		}
		frame->event_within = NULL;
		frame->count_within = -1;
	}
	widget->preserve_internal_structures = False;

	return return_value;
}

Bool frame_default_expose(px_widget* widget, px_event* event)
{
	widget->show(widget);
	return True;
}

void frame_redraw(px_widget* widget)
{
	unsigned int count;
	Display* display = display_get_Xdisplay();

#ifdef __debug__
	printf("%s(): %p\n", __func__, widget);
#endif
	
	if (widget->use_pixmap == True)
	{
		image_copy_pixmap(widget->win, &(widget->backing_store), widget->background.background_pixmap);
	}
	else
	{
		widget_draw_px_color(&(widget->backing_store), widget->background.background_color);
	}

	if (widget->use_pixmap == False && widget->background.background_color->type == TRANSPARENT)
	{
		GC  mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
		XSetGraphicsExposures(display, mask_gc, False);
		XSetForeground(display, mask_gc, 1);
		for (count = 0; count < widget->specific.frame.no_of_children; count++)
		{
			px_widget* child = widget->specific.frame.widget_list[count];
		
			XCopyArea(display, child->backing_store.mask,
					widget->backing_store.mask, mask_gc, 0, 0, 
					child->backing_store.w,
					child->backing_store.h,
					child->x,
					child->y);
		}
		XFreeGC(display, mask_gc);
	}
	widget_draw_border(widget);
	XClearWindow(display, widget->win);
	widget_set_event_mask(widget, widget->emask);

#ifdef __debug__
	printf("%s(): %p - done\n", __func__, widget);
#endif
}


void frame_show(px_widget* widget)
{
	unsigned int count;
	Display* display = display_get_Xdisplay();

#ifdef __debug__
	printf("%s(): %p Start\n", __func__, widget);
#endif
	XSetWindowBackgroundPixmap(display, widget->win, widget->backing_store.pxmp);

	if (widget->backing_store.mask != None)
		XShapeCombineMask (display, widget->win, ShapeBounding, 0, 0, 
			widget->backing_store.mask, ShapeSet);

	XMapWindow(display, widget->win);

	for (count = 0; count < widget->specific.frame.no_of_children; count++)
	{
		px_widget* child = widget->specific.frame.widget_list[count];

		if (!child->hidden)
			child->show(child);
	}
	widget->hidden = False;
#ifdef __debug__
	XSync(display, False);
	printf("%s() : %p : End\n", __func__, widget);
#endif
	
}

void frame_destroy(px_widget* widget)
{
	int count;
	px_frame* frame = &( widget->specific.frame);
	for (count = 0; count < frame->no_of_children; count++)
	{
		frame->ynodes = widget_remove_child(frame->ynodes, frame->widget_list[count], 
				frame->axis == YAXIS ? YDIRECTION : XDIRECTION, frame->axis);
		frame->widget_list[count]->destroy(frame->widget_list[count]);
		free(frame->widget_list[count]);
	}
	free(frame->widget_list);

	widget_destroy(widget);
}

void frame_cleanup_data_structures(px_widget* widget)
{
	px_frame* frame = &(widget->specific.frame);
	int count;
	
	px_xynode* ynode = frame->ynodes, *next = NULL;

	while (ynode != NULL)
	{
		px_xynode* xnode = ynode->xnodes, *xnext = NULL;
		next = ynode->next;

		free(ynode->widget_list);
		
		while (xnode != NULL)
		{
			free(xnode->widget_list);
			xnext = xnode->next;
			free(xnode);
			xnode = xnext;
		}
		free(ynode);
		ynode = next;
	}
	
	frame->ynodes= NULL;
	
	for (count = 0; count < frame->no_of_children; count++)
	{
		if (!frame->widget_list[count]->condemned)
			frame->ynodes = widget_add_child(frame->ynodes, frame->widget_list[count], 
				frame->axis == YAXIS ? YDIRECTION : XDIRECTION, frame->axis);
	}
}

