/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

// PicoWidget : Horizontal Scrollbar
// Copyrights (c) PicoPeta Simputers Pvt. Ltd.
// Programmer: Santhosh N

#include "hscrollbar.h"
#include "image.h"
#include "widget_factory.h"
#include <X11/extensions/shape.h>
#include <stdlib.h>
#include <string.h>
#include "theme.h"
#include "eventloop.h"
#include "display.h"

#define HSCROLL_DELAY 130000

static struct _hscrollbar {
	float slider_start, slider_length, increment;
};

typedef enum _hscrollbar_states 
{
	HSCROLL_NONE, HSCROLL_LEFT, HSCROLL_RIGHT, HSCROLL_SLIDER, HSCROLL_TROUGH
} hscrollbar_states;

static hscrollbar_states sbar_state;
static int prev_x;
static px_timer *timer=NULL;

px_widget* hscrollbar_create(px_widget* parent, int x, int y, int width, int height, const char *tag)
{
	px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
	Display* display = display_get_Xdisplay();
	px_theme* theme = theme_get_theme();

	widget->win = XCreateWindow(display, parent->win, x, y, width, height, 0,
							CopyFromParent, CopyFromParent, CopyFromParent, 0, NULL);

	// Initialize common widget variables
	widget->x = x;
	widget->y = y;
	widget->height = 16;
	widget->width = width;
	widget->wtype = HSCROLLBAR;
	widget->child_of = NULL;
	widget->use_pixmap = True;
	if(tag != NULL) {
		widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	}else
		widget->tag = NULL;

	widget->background.background_pixmap = theme_get_pixmap(theme, HSCROLLBAR_TROUGH);

	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);
	
	return widget;
}

void hscrollbar_set_parameters(px_widget* widget)
{
	px_hscrollbar *hscrollbar = &(widget->specific.hscrollbar);
	struct _hscrollbar *internal;
		
	// Initialize common widget default callbacks
	widget->default_callbacks[BUTTON_PRESSED] = hscrollbar_default_button_press;
	widget->default_callbacks[BUTTON_RELEASED] = hscrollbar_default_button_release;
	widget->default_callbacks[MOUSE_LEAVE] = hscrollbar_default_button_release;
	widget->default_callbacks[MOTION] = hscrollbar_default_button_motion;

	// Create internal data structure
	hscrollbar->internal = internal = (struct _hscrollbar *)malloc(sizeof(struct _hscrollbar));
	
	// Initialize internal widget structure variables
	internal->slider_start = 0.0;
	internal->slider_length = 1.0;
	internal->increment = 0.1;

	// Initalizing widget specific callbacks
	widget->redraw = hscrollbar_redraw;
	widget->destroy = hscrollbar_destroy;
	hscrollbar->set_slider_position = hscrollbar_set_slider_position;
	hscrollbar->set_slider_length = hscrollbar_set_slider_length;
	hscrollbar->set_increment = hscrollbar_set_increment;
	hscrollbar->intimate_change = NULL;
	hscrollbar->get_slider_position = hscrollbar_get_slider_position;
	hscrollbar->get_increment = hscrollbar_get_increment;
}

Bool hscrollbar_handle_up_arrow(px_widget* widget, px_event* event)
{
	px_hscrollbar *hscrollbar = &(widget->specific.hscrollbar);
     struct _hscrollbar *internal = (struct _hscrollbar *)hscrollbar->internal;
	float nstart;

	nstart = internal->slider_start - internal->increment;

	if(nstart < 0.0) {
		nstart = 0.0;
		if(timer != NULL) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			eventloop->delete_timer(eventloop, timer);
			timer = NULL;
		}
	}

	internal->slider_start = nstart;
	hscrollbar_redraw_slider(widget);
		
	if(hscrollbar->intimate_change != NULL)
		hscrollbar->intimate_change(widget->child_of, internal->slider_start);

	return True;
}

Bool hscrollbar_handle_down_arrow(px_widget* widget, px_event* event)
{
	px_hscrollbar *hscrollbar = &(widget->specific.hscrollbar);
     struct _hscrollbar *internal = (struct _hscrollbar *)hscrollbar->internal;
	float nstart;

	nstart = internal->slider_start + internal->increment;

	if(nstart > 1.0) {
		nstart = 1.0 ;
		if(timer != NULL) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			eventloop->delete_timer(eventloop, timer);
			timer = NULL;
		}
	}

	internal->slider_start = nstart;
	hscrollbar_redraw_slider(widget);
		
	if(hscrollbar->intimate_change != NULL)
		hscrollbar->intimate_change (widget->child_of, internal->slider_start);

	return True;
}
	
Bool hscrollbar_default_button_press(px_widget* widget, px_event* event)
{
	struct _hscrollbar *internal = (struct _hscrollbar *)widget->specific.hscrollbar.internal;
	int slider_start, slider_length;
	px_eventloop* eventloop = eventloop_get_eventloop();

	slider_length = (int)(internal->slider_length*(widget->width-32));
	if(slider_length<24)
		slider_length = 24;
	slider_start = (int)(internal->slider_start*(widget->width-32-slider_length));

	if(event->x >= 0 && event->x <= 16) {	// Left arrow
		sbar_state = HSCROLL_LEFT;
		if(internal->slider_start > 0.0) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			
			hscrollbar_handle_up_arrow(widget, event);
			timer = eventloop->add_timer(eventloop, 0, HSCROLL_DELAY, hscrollbar_handle_up_arrow, widget, event, True);
		}
		hscrollbar_redraw_left_button(widget);
	} else if(event->x >= (widget->width-16) && event->x <= (widget->width-1)) {	// Right arrow
		sbar_state = HSCROLL_RIGHT;
		if(widget->width-16 > slider_start+slider_length+16) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			
			hscrollbar_handle_down_arrow(widget, event);
			timer = eventloop->add_timer(eventloop, 0, HSCROLL_DELAY, hscrollbar_handle_down_arrow, widget, event, True);
		}
		hscrollbar_redraw_right_button(widget);
	} else if(event->x > 11 && event->x < slider_start+16) {	// Trough above slider
		sbar_state = HSCROLL_TROUGH;
		if(internal->slider_start > 0) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			
			hscrollbar_handle_up_arrow(widget, event);
			timer = eventloop->add_timer(eventloop, 0, HSCROLL_DELAY, hscrollbar_handle_up_arrow, widget, event, True);
		}
	} else if(event->x > slider_start+slider_length+16 && event->x < widget->width-16) { // Trough below slider
		sbar_state = HSCROLL_TROUGH;
		if(slider_start+slider_length+16 < widget->width-16) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			
			hscrollbar_handle_down_arrow(widget, event);
			timer = eventloop->add_timer(eventloop, 0, HSCROLL_DELAY, hscrollbar_handle_down_arrow, widget, event, True);
		}
	} else {	// Slider
		sbar_state = HSCROLL_SLIDER;
		prev_x = event->x;
	}
		
	eventloop->grab(eventloop, widget);

	return True;
}
		
Bool hscrollbar_default_button_release(px_widget* widget, px_event* event)
{
	px_eventloop* eventloop = eventloop_get_eventloop();

	eventloop->grab(eventloop, NULL);
	
	if(timer != NULL) {
		eventloop->delete_timer(eventloop, timer);
		timer = NULL;
	}

	sbar_state = HSCROLL_NONE;
	prev_x = 0;

	widget->redraw(widget);
	widget->show(widget);

	return True;
}

Bool hscrollbar_default_button_motion(px_widget* widget, px_event* event)
{
	px_hscrollbar* hscrollbar = &(widget->specific.hscrollbar);
	struct _hscrollbar *internal = (struct _hscrollbar *)hscrollbar->internal;
	int fg;
	XEvent xev, xev_copy;

	// Check if within Slider
	if(sbar_state == HSCROLL_SLIDER && event->x >= 16 && event->x <= (widget->width-16)) {
		float nstart;
		int slider_length;

		slider_length = (int)(internal->slider_length*(widget->width-32));
		if(slider_length<24)
			slider_length = 24;
		nstart = (internal->slider_start*(widget->width-32-slider_length)+(event->x-prev_x))/(widget->width-32-slider_length);

		// Make sure that slider is within the designated area
		if(nstart<0.0)
			nstart = 0.0;
		else if(nstart>1.0)
			nstart = 1.0;

		internal->slider_start = nstart;
		hscrollbar_redraw_slider(widget);
		
		if(hscrollbar->intimate_change != NULL)
			hscrollbar->intimate_change(widget->child_of, internal->slider_start);

		prev_x = event->x;
	}

	// Bypass extra motion events
	fg = 0;
	while(XPending(display_get_Xdisplay())) {
		XNextEvent(display_get_Xdisplay(), &xev);
		if(!fg && xev.type==MotionNotify) {
			fg = 1;
			xev_copy = xev;
		}
		if(xev.type==ButtonRelease) {
			fg = 1;
			xev_copy = xev;
			break;
		}
	}
	if(fg)
		XPutBackEvent(display_get_Xdisplay(), &xev_copy);

	return True;
}

void hscrollbar_redraw_left_button(px_widget* widget)
{
	Display* display = display_get_Xdisplay();
	GC gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
	px_pixmap* show_pixmap = NULL;
	px_theme* theme = theme_get_theme();

	if(sbar_state == HSCROLL_LEFT)
		show_pixmap = theme_get_pixmap(theme, HSCROLLBAR_LEFT_ACTIVE);
	else
		show_pixmap = theme_get_pixmap(theme, HSCROLLBAR_LEFT);
	
	XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc, 
				0, 0, 16, 16, 0, 0); 
	
	XClearArea(display, widget->win, 0, 0, 16, 16, False);

	XFreeGC(display, gc);
}

void hscrollbar_redraw_right_button(px_widget* widget)
{
	Display* display = display_get_Xdisplay();
	GC gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
	px_pixmap* show_pixmap = NULL;
	px_theme* theme = theme_get_theme();
	
	if(sbar_state == HSCROLL_RIGHT)
		show_pixmap = theme_get_pixmap(theme, HSCROLLBAR_RIGHT_ACTIVE);
	else
		show_pixmap = theme_get_pixmap(theme, HSCROLLBAR_RIGHT);
	
	XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc, 
				0, 0, 16, 16, widget->width-16, 0); 

	XClearArea(display, widget->win,  widget->width-16, 0, 16, 16, False);
	
	XFreeGC(display, gc);
}
	
void hscrollbar_redraw_slider(px_widget* widget)
{
	struct _hscrollbar *internal = (struct _hscrollbar *)widget->specific.hscrollbar.internal;
	Display* display = display_get_Xdisplay();
	px_theme* theme = theme_get_theme();
	px_pixmap *show_pixmap = NULL;
	GC gc, gcxor, gcand;
	int slider_start, slider_length;

	slider_length = (int)(internal->slider_length*(widget->width-32));
	if(slider_length<24) 
		slider_length = 24;
	slider_start = (int)(internal->slider_start*(widget->width-32-slider_length));

	if(widget->use_pixmap && (slider_start+slider_length)<widget->width) {
		// Create required GCs
		gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);

		gcxor = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
		XSetForeground(display, gcxor,WhitePixel(display, DefaultScreen(display)));
		XSetBackground(display, gcxor,WhitePixel(display, DefaultScreen(display)));
		XSetFunction(display, gcxor, GXxor);
		XSetFillStyle(display, gcxor, FillSolid);
		XSetGraphicsExposures(display, gcxor, False);

		gcand = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
		XSetForeground(display, gcand, 0);
		XSetBackground(display, gcand, ~0);
		XSetFunction(display, gcand, GXand);
		XSetFillStyle(display, gcand, FillSolid);
		XSetGraphicsExposures(display, gcand, False);

		// Draw the hscrollbar pixmaps
		show_pixmap = theme_get_pixmap(theme, HSCROLLBAR_TROUGH);
		XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc,
							0, 0, widget->width-32, 16, 16, 0);
		
		show_pixmap = theme_get_pixmap(theme, HSCROLLBAR_SLIDER);
		if(show_pixmap->mask!=None) {		// Transperant Slider
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 0, slider_length/2+1, 16, slider_start+16, 0);
			XCopyPlane(display, show_pixmap->mask, widget->backing_store.pxmp, gcand,
								0, 0, slider_length/2+1, 16, slider_start+16, 0, 1L);
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 0, slider_length/2+1, 16, slider_start+16, 0);

			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								320-slider_length/2, 0, slider_length/2, 16, slider_start+16+slider_length/2, 0);
			XCopyPlane(display, show_pixmap->mask, widget->backing_store.pxmp, gcand,
								320-slider_length/2, 0, slider_length/2, 16, slider_start+16+slider_length/2, 0, 1L);
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								320-slider_length/2, 0, slider_length/2, 16, slider_start+16+slider_length/2, 0);
		} else {
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc,
								0, 0, slider_length/2+1, 16, slider_start+16, 0);
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc,
								320-slider_length/2, 0, slider_length/2, 16, slider_start+16+slider_length/2, 0);
		}

		show_pixmap = theme_get_pixmap(theme, HSCROLLBAR_NOTCH);
		if(show_pixmap->mask!=None) {		// Transperant Slider Notch
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 0, 10, 10, (2*slider_start+slider_length)/2-5+16, 3);
			XCopyPlane(display, show_pixmap->mask, widget->backing_store.pxmp, gcand,
								0, 0, 10, 10, (2*slider_start+slider_length)/2-5+16, 3, 1L);
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 0, 10, 10, (2*slider_start+slider_length)/2-5+16, 3);
		} else {
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc,
								0, 0, 10, 10, (2*slider_start+slider_length)/2-5+16, 3);
		}
		
		XClearArea(display, widget->win, 16, 0, widget->width-32, widget->height, False);

		XFreeGC(display, gcand);
		XFreeGC(display, gcxor);
		XFreeGC(display, gc);
	}
}

void hscrollbar_redraw(px_widget* widget)
{
	// Draw the arrows
	hscrollbar_redraw_left_button(widget);
	hscrollbar_redraw_right_button(widget);

	// Draw the slider
	hscrollbar_redraw_slider(widget);

	widget_set_event_mask(widget, widget->emask);
}

void hscrollbar_destroy(px_widget* widget)
{
	widget_destroy(widget);
}

void hscrollbar_set_slider_position(px_widget* widget, float s)
{
	px_hscrollbar* hscrollbar = &(widget->specific.hscrollbar);
	struct _hscrollbar *internal = (struct _hscrollbar *)widget->specific.hscrollbar.internal;

	if(s<0.0)
		internal->slider_start = 0.0;
	else if(s>1.0)
		internal->slider_start = 1.0;
	else
		internal->slider_start = s;
	
	if(hscrollbar->intimate_change != NULL)
		hscrollbar->intimate_change(widget->child_of, internal->slider_start);
}

void hscrollbar_set_slider_length(px_widget* widget, float s)
{
	struct _hscrollbar *internal = (struct _hscrollbar *)widget->specific.hscrollbar.internal;

	if(s<0.0)
		internal->slider_length = 0.0;
	else if(s>1.0)
		internal->slider_length = 1.0;
	else
		internal->slider_length = s;
}

void hscrollbar_set_increment(px_widget* widget, float i)
{
	struct _hscrollbar *internal = (struct _hscrollbar *)widget->specific.hscrollbar.internal;

	if(i>=0.0 && i<=1.0)
		internal->increment = i;
}

float hscrollbar_get_slider_position(px_widget* widget)
{
	struct _hscrollbar *internal = (struct _hscrollbar *)widget->specific.hscrollbar.internal;
	
	return internal->slider_start;
}

float hscrollbar_get_increment(px_widget* widget)
{
	struct _hscrollbar *internal = (struct _hscrollbar *)widget->specific.hscrollbar.internal;

	return internal->increment;
}
