/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

// PicoWidget : Horizontal Slider
// Copyrights (c) PicoPeta Simputers Pvt. Ltd.
// Programmer: Vishweshwarayya. G. Mathad

#include "hslider.h"
#include "image.h"
#include "widget_factory.h"
#include <X11/extensions/shape.h>
#include <stdlib.h>
#include <string.h>
#include<math.h>
#include "theme.h"
#include "eventloop.h"
#include "display.h"
#include "font.h"

static struct _hslider {
	int hslider_start, hslider_length,min,max;
     Bool hslider_view,min_calib;
     hslider_calibration_side calib;
     char* fontname;
     px_font* font;
};

typedef struct _hslider pw_hslider;
static int hslider_x;
typedef enum _hslider_states 
{
	HSLIDER_SCROLL_NONE, 
	HSLIDER_SCROLL_SLIDER, 
	HSLIDER_SCROLL_TROUGH
} hslider_states;

hslider_states state;
static int trough_height,hslider_height,hslider_width;

px_widget* hslider_create(px_widget* parent, int x, int y, int width, int height, const char* tag)
{
	px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
	Display* display = display_get_Xdisplay();
	px_pixmap* show_pixmap = NULL;
	px_theme* theme = theme_get_theme();
	widget->win = XCreateWindow(display, parent->win, x, y, width, height, 0, CopyFromParent, CopyFromParent, CopyFromParent, 0, NULL);
	
	show_pixmap = theme_get_pixmap(theme, HBACK_SLIDER);
	trough_height=show_pixmap->h;
	show_pixmap = theme_get_pixmap(theme, HFORE_SLIDER);
	hslider_height=show_pixmap->h;
	hslider_width=show_pixmap->w;
#ifdef __debug__
	printf("%s(): %p - Created Window with parent window = %lu, x = %d, y = %d, width = %d, height = %d, new window = %lu\n\n", 
			__func__, widget, parent->win, x, y, width, height, widget->win);

	XSync(display, False);
#endif
	// Initialize common widget variables
	widget->x = x;
	widget->y = y;
	widget->height = height;
	widget->width = width;
	widget->child_of = parent;


	if (tag != NULL)
	{
		widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	}
	else
		widget->tag = NULL;
	widget->wtype = HSLIDER;
	widget->child_of = NULL;

	
	widget->use_pixmap = False;
	widget->background.background_color = (px_color*)malloc(sizeof(px_color));
	widget->background.background_color->type = SOLID;
	widget_xcolor_from_str(&(widget->background.background_color->first_color), "white");


	
	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);
		
	return widget;
}

void hslider_set_parameters(px_widget* widget)
{
	px_hslider *hslider=&(widget->specific.hslider);
	pw_hslider *internal;
	
	widget->default_callbacks[BUTTON_PRESSED] = hslider_default_button_press;
	widget->default_callbacks[BUTTON_RELEASED] = hslider_default_button_release;
	widget->default_callbacks[MOUSE_LEAVE] = hslider_default_button_release;
	widget->default_callbacks[MOTION] = hslider_default_button_motion;

	widget->redraw = hslider_redraw;
	widget->destroy = hslider_destroy;

	hslider->internal=(pw_hslider *)malloc(sizeof(pw_hslider));
	internal=hslider->internal;

	internal->hslider_start = 15;
	internal->hslider_view=0;
	internal->min_calib=False;
	internal->min = 0;
	internal->max = 100;
	internal->calib=HSLIDER_NONE;
	internal->hslider_length = hslider_height;
	hslider_x = 15;
	internal->fontname  = strdup("-picopeta-simputer-medium-r-normal--11-120-101-108-p-120-fontspecific-0");
	internal->font = font_load_font(internal->fontname);

	// Initalizing widget specific callbacks
	widget->specific.hslider.get_position = hslider_get_position;
	widget->specific.hslider.show_calibration=hslider_show_calibration;
	widget->specific.hslider.show_min_calibration=hslider_show_minor_calibration;
	widget->specific.hslider.set_position = hslider_set_value;
	widget->specific.hslider.show_value=hslider_show_value;
	widget->specific.hslider.set_range=hslider_set_range;
	widget->specific.hslider.set_font = hslider_set_font;
	widget->specific.hslider.intimate_slider_release = NULL;
	widget->specific.hslider.intimate_slider_motion = NULL;
	
}

Bool hslider_default_button_press(px_widget* widget, px_event* event)
{
	pw_hslider* hslider = widget->specific.hslider.internal;	
	if (!(event->x >= 15 && event->x <= (widget->width-15) &&
			event->y >= (widget->height/2-hslider_height/2) && event->y <= (widget->height/2 + hslider_height/2)))
	{
		state = HSLIDER_SCROLL_NONE;
		widget->show(widget);

	}
	else if ( (event->x >= 15 && event->x <= (widget->width-15) &&
			event->y >= (widget->height/2-hslider_height/2) && event->y <= (widget->height/2 + hslider_height/2)))
	{
		
		int new_start;	
		if(event->x>15 && event->x< (widget->width-15))
		{
			state =HSLIDER_SCROLL_TROUGH;

			if(event->x<= (15+hslider_width))
				new_start=15;
			else if (event->x>= (widget->width - 15 -hslider_width))
				new_start=(widget->width -15-hslider_width);
			else
				new_start=event->x-hslider_width/2;
			widget->redraw(widget);
			widget->show(widget);
			hslider->hslider_start = new_start;
				
						
			state =HSLIDER_SCROLL_SLIDER;	
			px_eventloop* eventloop = eventloop_get_eventloop();
			eventloop->grab(eventloop, widget);
			hslider_x = event->x;

			widget->redraw(widget);
			widget->show(widget);
		}			
			
	}

	return True;
}

Bool hslider_default_button_release(px_widget* widget, px_event* event)
{
	px_eventloop* eventloop = eventloop_get_eventloop();
	
	eventloop->grab(eventloop, NULL);

	if (widget->specific.hslider.intimate_slider_release != NULL)
				widget->specific.hslider.intimate_slider_release(hslider_get_position(widget));
	
	state = HSLIDER_SCROLL_NONE;
	hslider_x = 15;
	widget->redraw(widget);
	widget->show(widget);

	return True;
}

Bool hslider_default_button_motion(px_widget* widget, px_event* event)
{
	pw_hslider* hslider = widget->specific.hslider.internal;
	int fg=0;
	XEvent xev, xev_copy;
	
	if (state == HSLIDER_SCROLL_SLIDER && event->x >= 15 && event->x <= (widget->width-15))
	{
		{
			int diffy = (event->x - hslider_x);
			int endy = (hslider->hslider_start + hslider_width);
			
			if ( diffy< 0 && hslider->hslider_start >=15)
			{
				int new_start;
				if (hslider->hslider_start >= (-diffy) && (hslider->hslider_start + diffy) >= 15)
					new_start = hslider->hslider_start + diffy;
				else
					new_start = 15;
				
				widget->redraw(widget);
				widget->show(widget);
				hslider->hslider_start = new_start;
				
				if (widget->specific.hslider.intimate_slider_motion != NULL)
					widget->specific.hslider.intimate_slider_motion(hslider_get_position(widget));
			}
			else if (diffy >= 0 && endy<= (widget->width-15))
			{
				int diff = widget->width-15-endy, new_start;
		
				if (diff >=diffy && (hslider->hslider_start+diffy)<=(widget->width-15))
				{
					new_start = hslider->hslider_start + diffy;
				}
				else
				{
					new_start = (widget->width-15) - hslider_width;
				}
				
				widget->redraw(widget);
				widget->show(widget);
				hslider->hslider_start = new_start;
				
				if (widget->specific.hslider.intimate_slider_motion != NULL)
					widget->specific.hslider.intimate_slider_motion(hslider_get_position(widget));

			}
			hslider_x = event->x;
		}
	}

	// Bypass extra motion events
	while(XPending(display_get_Xdisplay())) {
		XNextEvent(display_get_Xdisplay(), &xev);
		if(xev.type==MotionNotify) {
			fg = 1;
			xev_copy = xev;
		}
		if(xev.type==ButtonRelease) {
			fg = 1;
			xev_copy = xev;
			break;
		}
	}
	if(fg) {
		XPutBackEvent(display_get_Xdisplay(), &xev_copy);
	}
	
	
	return True;
}

void hslider_redraw(px_widget* widget)
{
	unsigned int startx;
	char buffer[20];
	int i;
	Display* display = display_get_Xdisplay();
	GC gcxor, gcand;
	px_pixmap* show_pixmap = NULL;
	GC gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
	px_theme* theme = theme_get_theme();
	pw_hslider* internal = widget->specific.hslider.internal;


	if (widget->use_pixmap)
	{
		image_copy_pixmap(widget->win, &(widget->backing_store), widget->background.background_pixmap);
	}
	else
	{
		widget_draw_px_color(&(widget->backing_store), widget->background.background_color);
	}

	gcxor = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
    	XSetForeground(display, gcxor,WhitePixel(display, DefaultScreen(display)));
	XSetBackground(display, gcxor,WhitePixel(display, DefaultScreen(display)));
	XSetFunction(display, gcxor, GXxor);
	XSetFillStyle(display, gcxor, FillSolid);
	XSetGraphicsExposures(display, gcxor, False);

	gcand = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
	XSetForeground(display, gcand, 0);
	XSetBackground(display, gcand, ~0);
	XSetFunction(display, gcand, GXand);
	XSetFillStyle(display, gcand, FillSolid);
	XSetGraphicsExposures(display, gcand, False);
	
	
	XSetFont(display, gc, internal->font->font->fid);
	XSetForeground(display, gc, BlackPixel(display, DefaultScreen(display)));
	show_pixmap = theme_get_pixmap(theme, HBACK_SLIDER);
	if(show_pixmap->mask!=None)
	{
		XCopyArea(display,show_pixmap->pxmp, widget->backing_store.pxmp, gcxor, 0, 0, widget->width/2-15,trough_height, 15,widget->height/2-trough_height/2);
		XCopyPlane(display,show_pixmap->mask, widget->backing_store.pxmp, gcand, 0, 0,widget->width/2-15,trough_height, 15,widget->height/2-trough_height/2, 1L);
		XCopyArea(display,show_pixmap->pxmp, widget->backing_store.pxmp, gcxor, 0, 0, widget->width/2-15,trough_height, 15,widget->height/2-trough_height/2);
		XCopyArea(display,show_pixmap->pxmp, widget->backing_store.pxmp, gcxor, 320-widget->width/2+15,0, widget->width/2-15, trough_height, widget->width/2,widget->height/2-trough_height/2);
		XCopyPlane(display,show_pixmap->mask, widget->backing_store.pxmp, gcand, 320-widget->width/2+15,0, widget->width/2-15, trough_height, widget->width/2,widget->height/2-trough_height/2,1L);
		XCopyArea(display,show_pixmap->pxmp, widget->backing_store.pxmp, gcxor, 320-widget->width/2+15,0, widget->width/2-15, trough_height, widget->width/2,widget->height/2-trough_height/2);
	}
	else
	{
		XCopyArea(display,show_pixmap->pxmp, widget->backing_store.pxmp, gc, 0, 0, widget->width/2-15,trough_height, 15,widget->height/2-trough_height/2);
		XCopyArea(display,show_pixmap->pxmp, widget->backing_store.pxmp, gc, 320-widget->width/2+15,0, widget->width/2-15, trough_height,widget->width/2,widget->height/2-trough_height/2);
	}
		
	

	startx = (internal->hslider_start );
	show_pixmap = theme_get_pixmap(theme, HFORE_SLIDER);
		
	if(show_pixmap->mask!=None)
	{
		XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor, 0, 0, hslider_width, hslider_height, startx, widget->height/2- hslider_height/2);
		XCopyPlane(display, show_pixmap->mask, widget->backing_store.pxmp, gcand, 0, 0, hslider_width,  hslider_height, startx, widget->height/2- hslider_height/2,1L);
		XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor, 0, 0, hslider_width,  hslider_height, startx, widget->height/2- hslider_height/2);
	}
	else
		XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor, 0, 0, hslider_width,  hslider_height, startx, widget->height/2- hslider_height/2);

	if(internal->calib==SLIDER_BOTTOM)
	{
		XDrawLine(display, widget->backing_store.pxmp, gc, 15 ,widget->height/2+trough_height/2+hslider_height/2 , widget->width-16, widget->height/2+trough_height/2+hslider_height/2);
		XDrawLine(display, widget->backing_store.pxmp, gc, 15 ,widget->height/2+trough_height/2+hslider_height/2 , 15, widget->height/2+trough_height/2+hslider_height/2+10);
		if(internal->min_calib==True)
		{
			XDrawLine(display, widget->backing_store.pxmp, gc, widget->width/2 ,widget->height/2+trough_height/2+hslider_height/2 , widget->width/2, widget->height/2+trough_height/2+hslider_height/2+10);
			XDrawLine(display, widget->backing_store.pxmp, gc, widget->width/4 +7 ,widget->height/2+trough_height/2+hslider_height/2 , widget->width/4 +7, widget->height/2+trough_height/2+hslider_height/2+10);
			XDrawLine(display, widget->backing_store.pxmp, gc,widget->width-( widget->width/4 +7) ,widget->height/2+trough_height/2+hslider_height/2 , widget->width-( widget->width/4 +7), widget->height/2+trough_height/2+hslider_height/2+10);
		}
		XDrawLine(display, widget->backing_store.pxmp, gc,widget->width-16 , widget->height/2+trough_height/2+hslider_height/2 , widget->width-16 ,widget->height/2+trough_height/2+hslider_height/2+10 );
		sprintf(buffer,"%d",(int)internal->min);
		XDrawString(display,widget->backing_store.pxmp,gc,5,widget->height/2+trough_height/2+hslider_height/2+30,buffer,strlen(buffer));
		sprintf(buffer,"%d",(int)internal->max);
		XDrawString(display,widget->backing_store.pxmp,gc,widget->width-font_get_width(internal->font, buffer, strlen(buffer)),widget->height/2+trough_height/2+hslider_height/2+30,buffer,strlen(buffer));
	}
	if(internal->calib==SLIDER_TOP)
	{
		XDrawLine(display, widget->backing_store.pxmp, gc, 15 ,widget->height/2-trough_height/2-hslider_height/2 , widget->width-16, widget->height/2-trough_height/2-hslider_height/2);
		XDrawLine(display, widget->backing_store.pxmp, gc,15 ,widget->height/2-trough_height/2-10-hslider_height/2 , 15, widget->height/2-trough_height/2-hslider_height/2);
		if(internal->min_calib==True)
		{
			XDrawLine(display, widget->backing_store.pxmp, gc, widget->width/2 ,widget->height/2-trough_height/2-hslider_height/2 , widget->width/2, widget->height/2-trough_height/2-hslider_height/2-10);
			XDrawLine(display, widget->backing_store.pxmp, gc, widget->width/4 +7 ,widget->height/2-trough_height/2-hslider_height/2 , widget->width/4 +7, widget->height/2-trough_height/2-hslider_height/2-10);
			XDrawLine(display, widget->backing_store.pxmp, gc,widget->width-( widget->width/4 +7) ,widget->height/2-trough_height/2-hslider_height/2 , widget->width-( widget->width/4 +7), widget->height/2-trough_height/2-hslider_height/2-10);
		}
		XDrawLine(display, widget->backing_store.pxmp, gc,widget->width-16 , widget->height/2-trough_height/2-10-hslider_height/2 , widget->width-16, widget->height/2-trough_height/2-hslider_height/2 );
		sprintf(buffer,"%d",(int)internal->min);
		XDrawString(display,widget->backing_store.pxmp,gc,5,widget->height/2-trough_height/2-15-hslider_height/2,buffer,strlen(buffer));
		sprintf(buffer,"%d",(int)internal->max);
		XDrawString(display,widget->backing_store.pxmp,gc,widget->width-font_get_width(internal->font, buffer, strlen(buffer)),widget->height/2-trough_height/2-15-hslider_height/2,buffer,strlen(buffer));
	}

	i=font_get_width(internal->font, buffer, strlen(buffer))/2;
	if(internal->hslider_start+hslider_width/2-font_get_width(internal->font, buffer, strlen(buffer))/2 <= 0)
	i=14;
	if(internal->hslider_start+hslider_width/2+font_get_width(internal->font, buffer, strlen(buffer))/2 >= widget->width)
		i=font_get_width(internal->font, buffer, strlen(buffer))-15;
	sprintf(buffer,"%d",hslider_get_position(widget));
	if(internal->hslider_view && internal->calib==SLIDER_TOP)
	XDrawString(display,widget->backing_store.pxmp,gc,internal->hslider_start+hslider_width/2-i,widget->height/2+trough_height/2+25,buffer,strlen(buffer));
	else if(internal->hslider_view && internal->calib==SLIDER_BOTTOM)
	XDrawString(display,widget->backing_store.pxmp,gc,internal->hslider_start+hslider_width/2-i,widget->height/2-trough_height/2-10,buffer,strlen(buffer));
	
	XClearWindow(display, widget->win);
		
	widget_set_event_mask(widget, widget->emask);
	
	XFreeGC(display, gc);
	XSync(display, False);
#ifdef __debug__
	XSync(display, False);
	printf("%s() : %p\n", __func__, widget);
#endif
}

void hslider_destroy(px_widget* widget)
{
	widget_destroy(widget);
}

void hslider_set_range(px_widget* widget,int min, int max)
{
	pw_hslider *internal=widget->specific.hslider.internal;
	if(min<max)
	{
		internal->min=min;
		internal->max=max;
	}
}
void hslider_show_calibration(px_widget* widget,hslider_calibration_side side)
{
	pw_hslider *internal=widget->specific.hslider.internal;
	internal->calib=side;
}
void hslider_show_minor_calibration(px_widget* widget,Bool show)
{
	pw_hslider *internal=widget->specific.hslider.internal;
	internal->min_calib=show;
}
void hslider_show_value(px_widget* widget, Bool state)
{
	pw_hslider *internal=widget->specific.hslider.internal;
	internal->hslider_view=state;
}
void hslider_set_value(px_widget* widget, int value)
{
	pw_hslider *internal=widget->specific.hslider.internal;
	int  hslider_max, hslider_min;

	hslider_max=internal->max;
	hslider_min=internal->min;

	if(value>=hslider_min && value<=hslider_max)
	{
		internal->hslider_start=(value-hslider_min)*((float)((widget->width-30)-hslider_width)/(float)(hslider_max-hslider_min))+15;
	}
	else if (value<hslider_min)
		internal->hslider_start=15;
	else if (value>hslider_max)
		internal->hslider_start=(widget->width-15)-hslider_width;
	printf("%d\n",hslider_width);

}

int hslider_get_position(px_widget* widget)
{
	int  hslider_max, hslider_min;
	float pos;
	pw_hslider *internal=widget->specific.hslider.internal;
	
	hslider_max=internal->max;
	hslider_min=internal->min;

	pos=(internal->hslider_start-15)/((float)((widget->width-30)-hslider_width)/(float)(hslider_max-hslider_min));
	pos=ceilf(pos);
	
	pos=pos+hslider_min;
	return(pos);
}
void hslider_set_font(px_widget* widget, const char* fontname)
{
	pw_hslider *internal=widget->specific.hslider.internal;
	if(internal->fontname == NULL) {
		internal->fontname = (char*)malloc(sizeof(char) * (strlen(fontname) + 1));
	} else {
		font_free_font(internal->fontname);	
	}
	internal->font = font_load_font(fontname);
}
