/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include <stdio.h>
#include <X11/Xlib.h>
#include <time.h>
#include <sys/time.h>

// Theme specific pixmaps
typedef enum _theme_pixmaps {
	SCROLLBAR_UP,
	SCROLLBAR_UP_ACTIVE,
	SCROLLBAR_DOWN,
	SCROLLBAR_DOWN_ACTIVE,
	CHECKBUTTON_ON,
	CHECKBUTTON_OFF,
	RADIOBUTTON_ON,
	RADIOBUTTON_OFF,
	STIPPLE12,
	STIPPLE25,
	STIPPLE50,
	STIPPLE75,
	TABIMAGE,
	TABIMAGE_ACTIVE,
     VBACK_SLIDER,
     VFORE_SLIDER,
     HBACK_SLIDER,
     HFORE_SLIDER,
	SCROLLBAR_SLIDER,
	SCROLLBAR_NOTCH,
	SCROLLBAR_TROUGH,
	SCROLLBAR_TROUGH2,
	VSCROLLBAR_UP,
	VSCROLLBAR_UP_ACTIVE,
	VSCROLLBAR_DOWN,
	VSCROLLBAR_DOWN_ACTIVE,
	VSCROLLBAR_SLIDER,
	VSCROLLBAR_NOTCH,
	VSCROLLBAR_TROUGH,
	HSCROLLBAR_LEFT,
	HSCROLLBAR_LEFT_ACTIVE,
	HSCROLLBAR_RIGHT,
	HSCROLLBAR_RIGHT_ACTIVE,
	HSCROLLBAR_SLIDER,
	HSCROLLBAR_NOTCH,
	HSCROLLBAR_TROUGH,
	COMBOBOX_DOWN,
	COMBOBOX_DOWN_ACTIVE,
	NR_THEME_PIXMAPS
} theme_pixmaps;

// Typedefs
typedef struct _px_clipboard px_clipboard;
typedef struct _px_theme px_theme;
typedef struct _px_textlayout px_textlayout;
typedef struct _px_textchunk px_textchunk;
typedef struct _px_xynode px_xynode;
typedef struct _px_display px_display;
typedef struct _px_event px_event;
typedef struct _px_timer px_timer;
typedef struct _px_eventloop px_eventloop;
typedef struct _px_color px_color;
typedef struct _px_font px_font;
typedef struct _px_widget px_widget;
typedef struct _px_button px_button;
typedef struct _px_label px_label;
typedef struct _px_pixmap px_pixmap;
typedef struct _px_toplevel px_toplevel;
typedef struct _px_frame px_frame;
typedef struct _px_listbox px_listbox;
typedef struct _px_scrollbar px_scrollbar;
typedef struct _px_vscrollbar px_vscrollbar;
typedef struct _px_hscrollbar px_hscrollbar;
typedef struct _px_entry px_entry;
typedef struct _px_text px_text;
typedef struct _px_vslider px_vslider;
typedef struct _px_hslider px_hslider;
typedef struct _px_radiobutton px_radiobutton;
typedef struct _px_radiobuttongroup px_radiobuttongroup;
typedef struct _px_checkbutton px_checkbutton;
typedef struct _px_progressbar px_progressbar;
typedef struct _px_sketch px_sketch;
typedef struct _px_tab px_tab;
typedef struct _px_txtlb px_txtlb;
typedef struct _px_combobox px_combobox;

struct _px_clipboard {
	Atom selection;
	Window win;

	char* text;
	px_widget* selection_owner;

	void (*set_selection)(px_clipboard*, const char*);
	void (*set_selection_owner)(px_clipboard*, px_widget*);

	const char* (*get_selection)(px_clipboard*);
	px_widget* (*get_selection_owner)(px_clipboard*);

	void (*selection_clear)(px_clipboard*, px_event*);
	void (*send_selection)(px_clipboard*, px_event*);
	void (*selection_notify)(px_clipboard*, px_event*);
	void (*handle_incr_mechanism)(px_clipboard*, px_event*);

};

struct _px_theme {
	char *theme_path;							// Path of the theme file
	px_pixmap* pixmaps[NR_THEME_PIXMAPS];

	void (*load_theme)(px_theme *self, const char *theme_path);	// Initialize the theme
	px_pixmap* (*get_pixmap) (px_theme*, theme_pixmaps);
	void (*destroy)(px_theme*);
};

struct _px_textchunk
{
	char* start;
	int length;
	int text_width;
	px_textchunk* next;
};

struct _px_textlayout
{
	justification just;
	wrap_types wrap;
	char* text;
	int x, y, width, height;
	px_font* font;
	XColor* font_color;

	px_textchunk* chunks;
	int no_of_chunks;
};

typedef enum {XDIRECTION, YDIRECTION} Direction;
typedef enum {XAXIS, YAXIS} MajorAxis;

struct _px_xynode {
	unsigned int from, to;
	px_xynode* xnodes;
	
	px_widget** widget_list;
	unsigned int no_of_widgets;
	unsigned int allocated;
	px_xynode* next;
};

