/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#ifndef _PICOWIDGET_H_
#define _PICOWIDGET_H_

// Theme framework
// Types of possible colors
typedef enum _color_type {
	TRANSPARENT,
	SOLID,
	GRADIENTHOR,
	GRADIENTVERT
} color_type;

// Types of possible widgets in PicoWidget
typedef enum _widget_type {
	TOPLEVEL,
	FRAME,
	BUTTON,
	RADIOBUTTON,
	CHECKBUTTON,
	LABEL,
	ENTRY,
	LISTBOX,
	PROGRESSBAR,
	SCROLLBAR,
   	TAB,
	TEXT,
	VSLIDER,
     HSLIDER,
	VSCROLLBAR,
	HSCROLLBAR,
	TXTLB,
	SKETCH,
	COMBOBOX
} widget_type;

// Types of text justifications possible for labels
typedef enum _justification {
	JUSTIFY_LEFT,
	JUSTIFY_CENTER,
	JUSTIFY_RIGHT
} justification;

// List of events for which widgets call the callback
typedef enum _widget_callbacks {
	BUTTON_PRESSED, 	// 0
	BUTTON_RELEASED,	// 1
	MOTION,			// 2
	MOUSE_ENTER,		// 3
	MOUSE_LEAVE,		// 4
	KEY_PRESSED,		// 5
	KEY_RELEASED,		// 6
	EXPOSE,			// 7
	MAP,			// 8
	UNMAP,			// 9
	CONFIGURE,		// 10
	CLIENT_MESSAGE,		// 11
	SELECTION_CLEAR,	// 12
	SELECTION_NOTIFY,	// 13
	SELECTION_REQUEST,	// 14
	FOCUSIN,		// 15
	FOCUSOUT,
	VISIBILITY,		// 16
	MISCELLANIOUS,		// 17
	NR_CALLBACKS
} widget_callbacks;

#define STYLUS_PRESSED	BUTTON_PRESSED
#define STYLUS_RELEASED	BUTTON_RELEASED
#define STYLUS_MOTION	MOTION
#define STYLUS_ENTER	MOUSE_ENTER
#define STYLUS_LEAVE	MOUSE_LEAVE

typedef enum _widget_extra_callbacks {
	RETRACT_BAR_REL,
	UP_REL,
	TICK_REL,
	NEXT_PAGE,
	NEXT_PAGE_REL,
	PREV_PAGE,
	PREV_PAGE_REL,
	EMAIL_REL,
	PRINTER_REL,
	ROCKER_UP,
	ROCKER_UP_REL,
	ROCKER_DOWN,
	ROCKER_DOWN_REL,
	ROCKER_SELECT,
	ROCKER_SELECT_REL,
	ACCEL_RIGHT,
	ACCEL_LEFT,
	ACCEL_TOP,
	ACCEL_BOTTOM,
	ACCEL_RIGHT_EXTREME,
	ACCEL_LEFT_EXTREME,
	NR_EXTRA_CALLBACKS
} widget_extra_callbacks;

// Ways of catching event
typedef enum _event_masks {
	EVENT_SOLID,
	EVENT_TRANSPARENT,
	EVENT_MASK,
	NR_EVENTMASKS
} event_masks;

// Widget states
typedef enum _widget_states {
	NORMAL,
	ACTIVE,
	DISABLED,
} widget_states;

typedef enum _border_types {
	NO_BORDER,
	RECTANGLE_BORDER,
	BEVELLED_BORDER,
	ELLIPTIC_BORDER,
	NR_BORDERTYPES
} border_types;

typedef enum _wrap_types {
	NO_WRAP,
	WRAP_CHAR,
	WRAP_WORD
} wrap_types;

#include "intern_picowidget.h"

// Display
struct _px_display {
	Display* display;

	Atom* wm_protocols;
	char** protocol_names;
	unsigned int no_of_protocols;
	unsigned int protocols_allocated;
	
	px_clipboard* xa_primary, *xa_secondary;
	
	Bool (*open_display)(px_display*, char*);
	void (*close_display)(px_display*);
	void (*set_display)(px_display*, Display*);
	
	void (*register_wm_protocol)(px_display*, const char*);
	Atom (*get_atom)(px_display*, const char*);

	px_clipboard* (*get_clipboard)(px_display*, const char*);

	void (*flush)(px_display*);
	void (*sync)(px_display*);
};

// Color
struct _px_color {
	color_type type;		// Solid or gradient
	XColor first_color;		// If gradient, starting color, else solid color
	XColor second_color;	// If gradient, ending color, else undefined
};

// Font
struct _px_font{
	XFontStruct *font;		// Just the XFont structure for now
};

// Pixmap
struct _px_pixmap {
	int w,h;
	Pixmap pxmp;			// Mask for the pixmap
	Pixmap mask;			// Clipmask for the pixmap
};

// Event
struct _px_event {
	XEvent *xev;
	void *private_data;
	widget_callbacks type;
	int x;
	int y;
	unsigned char ch;
};

// Timer
struct _px_timer {
	struct timeval delay, fire_at;
	Bool (*function)(void*, void*);
	void* parameter_1, *parameter_2;
	Bool repeat;
	px_timer* next;
};

// Eventloop
struct _px_eventloop {
	px_widget* grabbed_widget;		//! Variable to keep track of which widget has the grab

	px_widget* focussed_widget;		//! Keeps track of which widget has the focus
	px_widget **toplevels;			//! List of toplevels created
	unsigned int no_of_toplevels;		//! Number of children.
	unsigned int toplevels_allocated;	//! No of pointers allocated for the list of children.

	px_widget **condemned;
	unsigned int no_of_condemned, condemned_allocated;
	
	int run;
	
	px_timer* timers, *secondary_timers, *timer_to_delete;
	Bool preserve_timers;
	struct timespec delay;

	void (*ignore)(px_eventloop*, widget_callbacks, int);
	void (*grab)(px_eventloop*, px_widget*);
	void (*focus)(px_eventloop*, px_widget*);
	void (*register_toplevel)(px_eventloop*, px_widget*);
	void (*remove_toplevel)(px_eventloop*, px_widget*);
	void (*set_delay)(px_eventloop*, const struct timespec*);
	px_widget* (*get_toplevel)(px_eventloop*, Window);

	px_timer* (*add_timer)(px_eventloop*, long, long, Bool (*function)(void*, void*), void*, void*,Bool);
	px_timer* (*add_timer_struct)(px_eventloop*, px_timer*, px_timer*);
	void (*delete_timer)(px_eventloop*, px_timer*);
	void (*fire_timers)(px_eventloop*);

	void (*condemn)(px_eventloop*, px_widget*);

	void (*destroy)(px_eventloop*);
	void (*exit)(px_eventloop*);

	void (*loop)(px_eventloop*);
	void (*do_one_event)(px_eventloop*);
	void (*handle_event)(px_eventloop*, px_event*, px_widget*);

	void *redraw_queue;
	void (*redraw_queue_add)(px_eventloop *eloop, px_widget *widget, int show);
	void (*redraw_queue_process)(px_eventloop *eloop);
};

// Toplevel
struct _px_toplevel {
	MajorAxis axis;
	px_xynode* ynodes;
	
	px_widget **widget_list;		//! List of widgets in this toplevel
	px_xynode* event_within;		//! Variable to keep track of the widget to give leave & enter events
	int count_within;
	int last_x, last_y;
	unsigned int no_of_children;		//! Number of children.
	unsigned int allocated;			//! No of pointers allocated for the list of children.
	void (*add_child) (px_widget*, px_widget*);
	void (*remove_child) (px_widget*, px_widget*);
	
	char* title;
	void (*set_title)(px_widget*, const char*);
};

// Frame
struct _px_frame {
	MajorAxis axis;
	px_xynode* ynodes;

	px_widget **widget_list;		//! List of widgets in this frame.
	px_xynode* event_within;		//! Variable to keep track of the widget to give leave & enter events.
	int count_within;
	int last_x, last_y;
	unsigned int no_of_children;		//! Number of children.
	unsigned int allocated;			//! No of pointers allocated for the list of children.

	void (*add_child) (px_widget*, px_widget*);
	void (*remove_child) (px_widget*, px_widget*);
};

//! Sketch widget
typedef enum _sketch_type {
	SKETCH_NORMAL,
	SKETCH_TIME_SERIES
} sketch_type;

struct _px_sketch {
	void *internal;

	void (*clear_sketches)(px_widget *wdgt);
	void (*set_pen_color)(px_widget *wdgt, unsigned long pcol);
	void (*set_pen_width)(px_widget *wdgt, unsigned int pwid);
	void (*set_bg_color)(px_widget *wdgt, unsigned long bcol);
	int (*load_from_file)(px_widget *wdgt, const char *filename);
	int (*save_to_file)(px_widget *wdgt, const char *filename);
};

// Button
struct _px_button {
	px_textlayout layout;	
	char* fontname;
	
	Bool change_size;
	
	Bool auto_repeat;
	unsigned int delay, speed;
	px_timer* auto_repeat_timer;

	XColor *normal_font;
	XColor *active_font;
	XColor *disabled_font;

	px_color* active_color;
	px_color* disabled_color;

	px_pixmap *normal; // Pixmaps for various states of the button
	px_pixmap *active;
	px_pixmap *disabled;

	XColor* active_border_color;
	XColor* disabled_border_color;

	void (*set_normal_pixmap_from_file)(px_widget*, char*);
	void (*set_active_pixmap_from_file)(px_widget*, char*);
	void (*set_disabled_pixmap_from_file)(px_widget*, char*);

	void (*set_normal_pixmap_from_pixmap)(px_widget*, const px_pixmap*);
	void (*set_active_pixmap_from_pixmap)(px_widget*, const px_pixmap*);
	void (*set_disabled_pixmap_from_pixmap)(px_widget*, const px_pixmap*);
	
	void (*set_justification)(px_widget*, justification);
	void (*set_text)(px_widget*, const char*);
	void (*set_font)(px_widget*, const char*);
	
	justification (*get_justification)(const px_widget*);
	const char* (*get_text)(const px_widget*);

	void (*set_auto_repeat)(px_widget*, Bool);
	void (*set_auto_repeat_delay)(px_widget*, unsigned int);
	void (*set_auto_repeat_speed)(px_widget*, unsigned int);
	
	void (*set_normal_font_color)(px_widget*, const char*);
	void (*set_active_font_color)(px_widget*, const char*);
	void (*set_disabled_font_color)(px_widget*, const char*);
	
	void (*set_active_background_color)(px_widget*, const px_color*);
	void (*set_disabled_background_color)(px_widget*, const px_color*);
	
	void (*set_active_border_color)(px_widget*, const char*);
	void (*set_disabled_border_color)(px_widget*, const char*);

	void (*set_wrap_type)(px_widget*, wrap_types);
	void (*size_depends_on_contents)(px_widget*, Bool);
};

// CheckButton
struct _px_checkbutton {
	px_textlayout layout;
	char* fontname;
	Bool indicator_on;
	Bool value;

	Bool change_size;
	
	XColor* normal_font;
	XColor* disabled_font;

	px_pixmap *normal; // Pixmaps for various states of the button
	px_pixmap *selected;
	px_pixmap *disabled;

	void (*set_normal_pixmap_from_file)(px_widget*, char*);
	void (*set_selected_pixmap_from_file)(px_widget*, char*);
	void (*set_disabled_pixmap_from_file)(px_widget*, char*);

	void (*set_normal_pixmap_from_pixmap)(px_widget*, const px_pixmap*);
	void (*set_selected_pixmap_from_pixmap)(px_widget*, const px_pixmap*);
	void (*set_disabled_pixmap_from_pixmap)(px_widget*, const px_pixmap*);
	
	void (*set_justification)(px_widget*, justification);

	void (*set_text)(px_widget*, const char*);
	void (*set_font)(px_widget*, const char*);
	
	justification (*get_justification)(const px_widget*);
	const char* (*get_text)(const px_widget*);
	
	void (*set_normal_font_color)(px_widget*, const char*);
	void (*set_disabled_font_color)(px_widget*, const char*);

	void (*show_indicator)(px_widget*);
	void (*hide_indicator)(px_widget*);
	
	void (*select)(px_widget*);
	void (*deselect)(px_widget*);
	void (*set_wrap_type)(px_widget*, wrap_types);
	void (*size_depends_on_contents)(px_widget*, Bool);
};

// RadioButton
struct _px_radiobuttongroup {
	int no_of_children;
	px_widget* selected;
};

struct _px_radiobutton {
	px_textlayout layout;
	char* fontname;
	Bool indicator_on;
	Bool value;

	Bool change_size;
	
	px_radiobuttongroup* group;

	XColor* normal_font;
	XColor* disabled_font;

	px_pixmap *normal; // Pixmaps for various states of the button
	px_pixmap *selected;
	px_pixmap *disabled;

	void (*set_normal_pixmap_from_file)(px_widget*, char*);
	void (*set_selected_pixmap_from_file)(px_widget*, char*);
	void (*set_disabled_pixmap_from_file)(px_widget*, char*);

	void (*set_normal_pixmap_from_pixmap)(px_widget*, const px_pixmap*);
	void (*set_selected_pixmap_from_pixmap)(px_widget*, const px_pixmap*);
	void (*set_disabled_pixmap_from_pixmap)(px_widget*, const px_pixmap*);
	
	void (*set_justification)(px_widget*, justification);

	void (*set_text)(px_widget*, const char*);
	void (*set_font)(px_widget*, const char*);

	justification (*get_justification)(const px_widget*);
	const char* (*get_text)(const px_widget*);
	
	void (*set_normal_font_color)(px_widget*, const char*);
	void (*set_disabled_font_color)(px_widget*, const char*);

	void (*show_indicator)(px_widget*);
	void (*hide_indicator)(px_widget*);
	
	void (*select)(px_widget*);
	void (*deselect)(px_widget*);

	void (*on_select)(px_widget*);
	void (*on_deselect)(px_widget*);
	void (*group_together) (px_widget*, px_widget*);
	void (*set_wrap_type)(px_widget*, wrap_types);
	void (*size_depends_on_contents)(px_widget*, Bool);
	
};

// Progressbar
struct _px_progressbar {
	unsigned int value;

	char* fontname;
	px_font *font;			// Font used to display the label

	Bool display_value;
	XColor* font_color;
	px_color fill_color;
	void (*set_font)(px_widget*, const char*);
	void (*set_font_color)(px_widget*, const char*);

	unsigned int (*get_value) (px_widget*);
	void (*set_value) (px_widget*, unsigned int);
	void (*show_value) (px_widget*, Bool);
	
	void (*set_fill_color)(px_widget*, px_color*);

	void (*value_changed)(px_widget*);

};

// Label
struct _px_label {
	char* fontname;
	px_textlayout layout;

	XColor *normal;
	XColor *active;
	XColor *disabled;
	
	Bool change_size;
	
	void (*set_justification)(px_widget*, justification);
	void (*set_text)(px_widget*, const char*);
	void (*set_font)(px_widget*, const char*);

	const char* (*get_text)(const px_widget*);
	justification (*get_justification)(const px_widget*);

	void (*set_normal_color)(px_widget*, const char*);
	void (*set_disabled_color)(px_widget*, const char*);
	void (*set_active_color)(px_widget*, const char*);
	void (*set_wrap_type)(px_widget*, wrap_types);
	void (*size_depends_on_contents)(px_widget*, Bool);
	
};

// Widget Listbox
struct _px_listbox {
	px_widget* scrollbar;
	px_widget **widget_list;
	px_widget* event_within;
	int event_within_index;

	unsigned int no_of_children;
	unsigned int allocated;
	unsigned int total_length_needed;
	int inter_widget_space;
	int start_of_viewport;


	void (*set_inter_widget_space)(px_widget* , int);
	void (*set_child)(px_widget*, px_widget*, unsigned int);
	void (*remove_child)(px_widget*, unsigned int);
	px_widget* (*get_child)(px_widget*, unsigned int);

	void (*see)(px_widget*, int);
};


// Text Listbox
typedef enum _txtlb_colors {
	TXTLB_BACKGROUND,
	TXTLB_FOREGROUND,
	TXTLB_ROW_BACKGROUND,
	TXTLB_PEAK,
	TXTLB_TROUGH,
	TXTLB_SELECN_HIGHLIGHT,
	TXTLB_SELECN_FOREGROUND,
	TXTLB_SCROLL_BG
} txtlb_colors;

typedef enum _txtlb_types {
	TXTLB_SINGLE_SELECN,
	TXTLB_MULTI_SELECN
} txtlb_types;

struct _px_txtlb {
	void *internal;

	int (*add_lb_column)(px_widget * wdgt, int colnum);
	int (*add_lb_elem)(px_widget * wdgt, char * txtstr, int colnum);
	int (*insert_lb_elem)(px_widget * wdgt, char * txtstr, int colnum, int rowidx);
	int (*remove_lb_row)(px_widget *wdgt, int rowidx, Bool redraw_flg);
	int (*change_lb_elem)(px_widget *wdgt, char *txtstr, int colnum, int rowidx);
	void (*change_lb_color)(px_widget *wdgt, txtlb_colors col, char *colstr);
	int (*remove_lb_elem)(px_widget * wdgt, int colnum, int rowidx);
	void (*show_lb_elems)(px_widget * wdgt);
	void (*intimate_selecn_change)(px_widget * wdgt, int rowidx);
};

// Combo Box
typedef enum _combo_colors {
	COMBO_LABEL_BACKGROUND,
	COMBO_LABEL_FOREGROUND,
	COMBO_LIST_BACKGROUND,
	COMBO_LIST_FOREGROUND,
	COMBO_LIST_ROW_BACKGROUND,
	COMBO_LIST_PEAK,
	COMBO_LIST_TROUGH,
	COMBO_LIST_SELECN_HIGHLIGHT,
	COMBO_LIST_SELECN_FOREGROUND
} combo_colors;

struct _px_combobox {
	void *internal;

	int (*add_option)(px_widget *cbox, char *option, Bool redraw);
	int (*remove_option)(px_widget *cbox, int selecn);
	int (*clear_options)(px_widget *cbox);
	void (*get_selecn)(px_widget *cbox, char ** option, int * idx);
	void (*set_selecn)(px_widget *cbox, int selecn);
	void (*change_color)(px_widget *wdgt, combo_colors col, char *colstr);
	void (*intimate_selecn_change)(px_widget *cbox, int selecn);
};

// Tab
struct _px_tab {
        void *internal;
        px_widget* (*add_tab)(px_widget *tabwidget, const char *title, Bool(*tab_callback)(px_widget*, px_event*));
	void (*raise_tab)(px_widget *tabwidget, int tabnumber);
	px_widget* (*get_active_widget)(px_widget *widget);
	px_widget* (*get_widget_pointer)(px_widget *widget, int tabnumber);
};

// Scrollbar
struct _px_scrollbar {
	float slider_start, slider_length, increment;	// Value to change when up/down arrows are pressed
	int x, y;

	px_timer* auto_repeat_timer;
	unsigned un_used1;

	void (*set_slider_position)(px_widget*, float);
	void (*set_slider_length)(px_widget*, float);

	void (*set_increment)(px_widget*, float);
	void (*intimate_viewport_changed)(px_widget*, float);

	void (*redraw_slider)(px_widget*);
	void (*redraw_top_button)(px_widget*);
	void (*redraw_bottom_button)(px_widget*);
	float (*get_slider_position)(px_widget *widget);
	float (*get_increment)(px_widget *widget);
};

// Entry
struct _px_entry {
	char *text;
	unsigned allocated;						// Space allocated for the text;
	unsigned int selection_start, selection_length, cursor_position;
	int show_from, show_length, show_width;

	unsigned int cursor_blink_time;
	px_timer* blink;
	Bool cursor_on;

	char* fontname;
	px_font *font;			// Font used to display the Entry
	px_color* selection_background;
	XColor* foreground;
	XColor* disabled_foreground;
	XColor* selection_foreground;

	void (*cut) (px_widget*, const char*);
	void (*copy) (px_widget*, const char*);
	void (*paste) (px_widget*, const char*);

	void (*set_foreground_color)(px_widget*, const char*);
	void (*set_disabled_foreground_color)(px_widget*, const char*);
	void (*set_selection_foreground_color)(px_widget*, const char*);
	void (*set_selection_background_color)(px_widget*, const px_color*);
	
	void (*insert_text)(px_widget*, const char*, unsigned int);
	void (*set_text)(px_widget*, const char*);
	void (*delete_text)(px_widget*, unsigned int, unsigned int);
	
	void (*set_font)(px_widget*, const char*);
	void (*see)(px_widget*, unsigned int);

	void (*on_change)(px_widget*);

	char password_char;
};

// Vertical Scrollbar
struct _px_vscrollbar {
	void *internal;

	void (*set_slider_position)(px_widget *widget, float position);
	void (*set_slider_length)(px_widget *widget, float length);
	void (*set_increment)(px_widget *widget, float increment);
	void (*intimate_change)(px_widget *widget, float position);
	float (*get_slider_position)(px_widget *widget);
	float (*get_increment)(px_widget *widget);
};

// Horizontal Scrollbar
struct _px_hscrollbar {
	void *internal;

	void (*set_slider_position)(px_widget *widget, float position);
	void (*set_slider_length)(px_widget *widget, float length);
	void (*set_increment)(px_widget *widget, float increment);
	void (*intimate_change)(px_widget *widget, float position);
	float (*get_slider_position)(px_widget *widget);
	float (*get_increment)(px_widget *widget);
};

// Text
struct _px_text {
	void *internal;

	void (*set_font)(px_widget *widget, const char *fontname);
	void (*set_wrap_type)(px_widget *widget, unsigned wrap_type);
	void (*set_foreground_color)(px_widget *widget, const char *color);
	
	int (*load_file)(px_widget *widget, const char *filename);
	int (*save_file)(px_widget *widget, const char *filename);
	int (*get_numlines)(px_widget *widget);
	const char *(*get_line)(px_widget *widget, int line_number);
	char *(*get_text)(px_widget *widget);
	int (*set_text)(px_widget *widget, const char *text);
};

// Vertical Slider
typedef enum _vslider_calibration_side {
     VSLIDER_NONE,
     SLIDER_LEFT,
     SLIDER_RIGHT
}vslider_calibration_side;

struct _px_vslider {
     void *internal;
     
	void (*set_position)(px_widget *widget, int value);
     void (*show_calibration)(px_widget *widget, vslider_calibration_side side);
     void (*set_range)(px_widget *widget, int min, int max);
     void (*intimate_slider_release)(int position);
	void (*intimate_slider_motion)(int position);
     int (*get_position)(px_widget *widget);
     void (*set_font)(px_widget *widget, const char* fontname);
     void (*show_value)(px_widget *widget, Bool state);
	void (*show_min_calibration)(px_widget *widget, Bool show);	
};

// Horizontal Slider
typedef enum _hslider_calibration_side {
     HSLIDER_NONE,
     SLIDER_TOP,
     SLIDER_BOTTOM
}hslider_calibration_side;

struct _px_hslider {
     void *internal;

     void (*set_position)(px_widget *widget, int value);
     void (*show_calibration)(px_widget *widget, hslider_calibration_side side);
     void (*set_range)(px_widget *widget, int min, int max);
	void (*intimate_slider_release)(int position);
	void (*intimate_slider_motion)(int position);
	int (*get_position)(px_widget *widget);
     void (*set_font)(px_widget *widget, const char* fontname);
     void (*show_value)(px_widget *widget, Bool state);
	void (*show_min_calibration)(px_widget *widget, Bool show);
};

// Widget
struct _px_widget {
	int depth;
	unsigned long wid;		// Widget ID
	char *tag;				// Tag/name for the widget
	int x,y,width,height;	// Position and dimension
	Window win;				// Widget Window
	px_widget *child_of;		// Parent frame
	widget_type wtype;		// Type of the widget
	px_pixmap backing_store;
	event_masks emask;
	XImage* mask_for_event;
	Bool hidden;
	widget_states state;
	Bool have_focus;
	
	border_types border;
	XColor* border_color;

	Bool condemned;
	Bool preserve_internal_structures;
	Bool dirty;
	
	union _u {				// Widget specific data
		px_toplevel toplevel;
		px_frame frame;
		px_label label;
		px_button button;
		px_listbox listbox;
		px_scrollbar scrollbar;
		px_entry entry;
		px_radiobutton radiobutton;
		px_checkbutton checkbutton;
		px_progressbar progressbar;
		px_sketch sketch;
		px_tab tab;
		px_text text;
		px_vslider vslider;
		px_hslider hslider;
		px_vscrollbar vscrollbar;
		px_hscrollbar hscrollbar;
		px_txtlb txtlb;
		px_combobox combobox;
	} specific;

	Bool use_pixmap;
	union _backround {			// Either color or pixmap
		px_color* background_color;
		px_pixmap* background_pixmap;
	} background;

	Bool (*default_callbacks[NR_CALLBACKS])(px_widget *w, px_event *ev);	// Callbacks on all widget events
	Bool (*callbacks[NR_CALLBACKS])(px_widget *w, px_event *ev);	// Callbacks on all widget events
	void *callback_parameters[NR_CALLBACKS];			// Parameters for callbacks on all widget events

	Bool (*within) (const px_widget*, px_event*);
	Bool (*is_transparent) (const px_widget*, int, int);
	Bool (*handle_event) (px_widget*, px_event*);

	void (*redraw)(px_widget *self);
	void (*show)(px_widget *self);
	void (*hide)(px_widget *self);
	
	void (*destroy)(px_widget *self);
	void (*destruct)(px_widget*);
	
	void (*set_size)(px_widget *self, unsigned long width, unsigned long height);
	void (*set_position)(px_widget *self, int x, int y);
	void (*get_size)(const px_widget *self, unsigned long *width, unsigned long *height);
	void (*get_position)(const px_widget *self, int *x, int *y);

	int (*get_absolute_x)(const px_widget*, Window);
	int (*get_absolute_y)(const px_widget*, Window);

	void (*set_event_mask) (px_widget*, event_masks);

	void (*set_background_pixmap_from_file)(px_widget*, char*);
	void (*set_background_pixmap_from_pixmap)(px_widget*, const px_pixmap*);
	void (*set_background_color)(px_widget*, const px_color*);
	
	void (*set_state)(px_widget*, widget_states);
	widget_states (*get_state)(const px_widget*);

	void (*set_focus)(px_widget*, Bool);

	void (*set_border_color)(px_widget*, const char*);
	void (*set_border_type)(px_widget*, border_types);

	void (*raise)(px_widget*);
	void (*lower)(px_widget*);

	void (*cleanup_internal_structures)(px_widget*);
	
	Bool (*window_delete_callback)(px_widget *w, px_event *ev);
	void *window_delete_callback_parameter;

	Bool (*extra_callbacks[NR_EXTRA_CALLBACKS])(px_widget *w, px_event *ev);
	void *extra_callbacks_parameters[NR_EXTRA_CALLBACKS];
};

px_display* get_display();

px_widget *new_widget(widget_type wt, px_widget *parent, int x, int y,
		int width, int height, char *tag);
px_widget* new_toplevel_frame(Window win, int x, int y, int width, int height, const char* tag);
int widget_xcolor_from_str(XColor *, const char *);

void picowidget_terminate();

#endif
