/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include <X11/extensions/shape.h>
#include <stdlib.h>
#include <string.h>
#include "progressbar.h"
#include "image.h"
#include "widget_factory.h"
#include "display.h"
#include "font.h"

px_widget* progressbar_create(px_widget* parent, int x, int y, int width, int height, const char* tag)
{
	px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
	Display* display = display_get_Xdisplay();
	int count;

	// Initialize common widget variables
	
	widget->win = XCreateWindow(display, parent->win, x, y, width, height, 0,
			CopyFromParent, CopyFromParent, CopyFromParent, 0, NULL);


	widget->x = x;
	widget->y = y;
	widget->height = height;
	widget->width = width;
	widget->wtype = PROGRESSBAR;
	widget->child_of = NULL;
	widget->use_pixmap = False;
	if (tag != NULL)
	{
		widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	}
	else
		widget->tag = NULL;

	widget->background.background_color = (px_color*)malloc(sizeof(px_color));
	widget->background.background_color->type = SOLID;
	widget_xcolor_from_str(&(widget->background.background_color->first_color), "white");

	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);

#ifdef __debug__
	printf("%s(): %p - Created Window with parent window = %lu, x = %d, y = %d, width = %d, height = %d, new window = %lu\n\n", 
			__func__, widget, parent->win, x, y, width, height, widget->win);
#endif
	return widget;
}

void progressbar_set_parameters(px_widget* widget)
{
	widget->default_callbacks[BUTTON_PRESSED] = progressbar_default_button_events;
	widget->default_callbacks[MOTION] = progressbar_default_button_events;
	widget->default_callbacks[BUTTON_RELEASED] = progressbar_default_button_events;
	widget->default_callbacks[MOUSE_ENTER] = progressbar_default_button_events;

	widget->redraw = progressbar_redraw;
	widget->destroy = progressbar_destroy;

	// Initialize widget specific variables 

	widget->specific.progressbar.value = 0;
	widget->specific.progressbar.display_value = False;
	widget->specific.progressbar.fill_color.type = SOLID;
	widget_xcolor_from_str(&(widget->specific.progressbar.fill_color.first_color), "blue");

	widget->specific.progressbar.font = NULL;
	widget->specific.progressbar.fontname = NULL;
	
	widget->specific.progressbar.font_color = NULL;

	widget->specific.progressbar.set_font = progressbar_set_font;
	widget->specific.progressbar.set_font_color = progressbar_set_font_color;
	
	widget->specific.progressbar.get_value = progressbar_get_value;
	widget->specific.progressbar.set_value = progressbar_set_value;
	widget->specific.progressbar.show_value = progressbar_show_value;
	
	widget->specific.progressbar.set_fill_color = progressbar_set_fill_color;

	widget->specific.progressbar.value_changed = NULL;

}

Bool progressbar_default_button_events(px_widget* widget, px_event* event)
{
	px_progressbar* progressbar = &( widget->specific.progressbar);
	unsigned int percent = (unsigned int)((float)(event->x)/(float)(widget->width) * 100);

	progressbar->value = percent;

	widget->redraw(widget);
	widget->show(widget);
	
	return True;
}

void progressbar_redraw(px_widget* widget)
{
	Display* display = display_get_Xdisplay();
	int fill_x;
	px_progressbar* progressbar = &( widget->specific.progressbar);
	
	
#ifdef __debug__
	printf("%s(): %p\n", __func__, widget);
#endif
	
	if (widget->use_pixmap == True)
	{
		image_copy_pixmap(widget->win, &(widget->backing_store), widget->background.background_pixmap);
	}
	else
	{
		widget_draw_px_color(&(widget->backing_store), widget->background.background_color);
	}

	fill_x = (progressbar->value * widget->width)/100;
	widget->backing_store.w = fill_x;

	widget_draw_px_color(&(widget->backing_store), &(progressbar->fill_color));
	widget->backing_store.w = widget->width;

	if (progressbar->display_value)
	{
		char ctemp[5];
		px_textlayout layout;
		sprintf(ctemp, "%d%%", progressbar->value);
	
		layout.text = ctemp;
		layout.font = progressbar->font;
		layout.just = JUSTIFY_CENTER;
		layout.font_color = progressbar->font_color;
		layout.x = 0;
		layout.y = 0;
		layout.width = widget->width;
		layout.height = widget->height;

		widget_draw_text(&layout, &(widget->backing_store));
	}

	widget_draw_border(widget);
	XClearWindow(display, widget->win);
	widget_set_event_mask(widget, widget->emask);

#ifdef __debug__
	printf("%s(): %p - done\n", __func__, widget);
#endif
}

void progressbar_destroy(px_widget* widget)
{
	px_progressbar* progressbar = &( widget->specific.progressbar);
	if (progressbar->font_color != NULL)
		free(progressbar->font_color);

	if (progressbar->fontname != NULL)
	{
		font_free_font(progressbar->fontname);
		free(progressbar->fontname);
	}

	widget_destroy(widget);
}

void progressbar_set_font_color(px_widget* widget, const char* colour)
{
	px_progressbar* progressbar = &( widget->specific.progressbar);
	if (progressbar->font_color == NULL)
	{
		progressbar->font_color = (XColor*)malloc(sizeof(XColor));
	}
	widget_xcolor_from_str(progressbar->font_color, colour);
}

void progressbar_set_font(px_widget* widget, const char* fontname)
{
	px_progressbar* progressbar = &( widget->specific.progressbar);
	if (progressbar->fontname != NULL)
	{
		font_free_font(progressbar->fontname);
		free(progressbar->fontname);
		progressbar->fontname = NULL;
	}

	progressbar->fontname = strdup(fontname);
	progressbar->font = font_load_font(progressbar->fontname);
}

void progressbar_set_value(px_widget* widget, unsigned int value)
{
	widget->specific.progressbar.value = value;
}

unsigned int progressbar_get_value(px_widget* widget)
{
	return widget->specific.progressbar.value;
}


void progressbar_show_value(px_widget* widget, Bool show)
{
	widget->specific.progressbar.display_value = show;
}

void progressbar_set_fill_color(px_widget* widget, px_color* col)
{
	px_progressbar* progressbar = &( widget->specific.progressbar);
	
	progressbar->fill_color  = (*col);
}

