/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include <X11/extensions/shape.h>
#include <X11/xpm.h>
#include <stdlib.h>
#include <string.h>
#include "radiobutton.h"
#include "image.h"
#include "font.h"
#include "theme.h"
#include "widget_factory.h"
#include "display.h"

px_widget* radiobutton_create(px_widget* parent, int x, int y, int width, int height, const char* tag)
{
	px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
	Display* display = display_get_Xdisplay();
	widget->win = XCreateWindow(display, parent->win, x, y, width, height, 0, CopyFromParent, CopyFromParent, CopyFromParent, 0, NULL);

#ifdef __debug__
	printf("%s(): %p - Created Window with parent window = %lu, x = %d, y = %d, width = %d, height = %d, new window = %lu\n\n", 
			__func__, widget, parent->win, x, y, width, height, widget->win);

	XSync(display, False);
#endif
	// Initialize common widget variables
	widget->x = x;
	widget->y = y;
	widget->height = height;
	widget->width = width;
	widget->child_of = parent;

	if (tag != NULL)
	{
		widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	}
	else
		widget->tag = NULL;
	widget->wtype = RADIOBUTTON;
	widget->child_of = NULL;

	widget->use_pixmap = False;
	widget->background.background_color = (px_color*)malloc(sizeof(px_color));
	widget->background.background_color->type = SOLID;
	widget_xcolor_from_str(&(widget->background.background_color->first_color), "white");

	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);

	return widget;
}

void radiobutton_set_parameters(px_widget* widget)
{
	px_radiobutton* radiobutton = &(widget->specific.radiobutton);
	
	widget->default_callbacks[BUTTON_PRESSED] = radiobutton_default_radiobutton_pressed;

	widget->set_size = radiobutton_set_size;
	widget->redraw = radiobutton_redraw;
	widget->destroy = radiobutton_destroy;
	
	
	// Initialize widget specific variables
	radiobutton->layout.just = JUSTIFY_CENTER;
	radiobutton->layout.wrap = WRAP_CHAR;
	radiobutton->layout.text = NULL;
	radiobutton->layout.x = 0;
	radiobutton->layout.y = 0;
	radiobutton->layout.width = widget->width;
	radiobutton->layout.height = widget->height;
	radiobutton->layout.font = NULL;
	radiobutton->layout.font_color = NULL;
	radiobutton->layout.chunks = NULL;
	radiobutton->layout.no_of_chunks = 0;

	radiobutton->fontname = NULL;
	radiobutton->indicator_on = True;
	radiobutton->value = 0;
	
	radiobutton->normal_font = NULL;
	radiobutton->disabled_font = NULL;
	
	radiobutton->normal = NULL;
	radiobutton->selected = NULL;
	radiobutton->disabled = NULL;

	radiobutton->group = NULL;

	radiobutton->change_size = False;
	
	radiobutton->set_normal_pixmap_from_file = radiobutton_set_normal_pixmap_from_file;
	radiobutton->set_selected_pixmap_from_file = radiobutton_set_selected_pixmap_from_file;
	radiobutton->set_disabled_pixmap_from_file = radiobutton_set_disabled_pixmap_from_file;
	
	radiobutton->set_normal_pixmap_from_pixmap = radiobutton_set_normal_pixmap_from_pixmap;
	radiobutton->set_selected_pixmap_from_pixmap = radiobutton_set_selected_pixmap_from_pixmap;
	radiobutton->set_disabled_pixmap_from_pixmap = radiobutton_set_disabled_pixmap_from_pixmap;

	radiobutton->set_text = radiobutton_set_text;
	radiobutton->set_justification = radiobutton_set_justification;
	radiobutton->set_font = radiobutton_set_font;

	radiobutton->get_justification = radiobutton_get_justification;
	radiobutton->get_text = radiobutton_get_text;

	radiobutton->set_normal_font_color = radiobutton_set_normal_font_color;
	radiobutton->set_disabled_font_color = radiobutton_set_disabled_font_color;
	
	radiobutton->show_indicator = radiobutton_show_indicator;
	radiobutton->hide_indicator = radiobutton_hide_indicator;

	radiobutton->select = radiobutton_select;
	radiobutton->deselect = radiobutton_deselect;

	radiobutton->on_select = NULL;
	radiobutton->on_deselect = NULL;
	
	radiobutton->group_together = radiobutton_group_together;

	radiobutton->set_wrap_type = widget_set_wrap_type;
	radiobutton->size_depends_on_contents = widget_size_depends_on_contents;
}

void radiobutton_set_size(px_widget* widget, unsigned long width, unsigned long height)
{
	widget_set_size(widget, width, height);
	widget->specific.radiobutton.layout.width = widget->width;
	widget->specific.radiobutton.layout.height = widget->height;

	widget_compute_textlayout(&(widget->specific.radiobutton.layout), widget);
}

Bool radiobutton_default_radiobutton_pressed(px_widget* widget, px_event* event)
{
	if (widget->specific.radiobutton.value)
	{
		if (widget->specific.radiobutton.group == NULL)
		{
			widget->specific.radiobutton.deselect(widget);
			widget->redraw(widget);
			widget->show(widget);
		}
	}
	else
	{
		widget->specific.radiobutton.select(widget);
		widget->redraw(widget);
		widget->show(widget);
	}

	return True;
}

void radiobutton_redraw(px_widget* widget)
{
	GC gc;
	px_pixmap* image_to_show = NULL;
	px_pixmap* indicator = NULL;
	unsigned int drawareax = 0 ;
	px_radiobutton* radiobutton = &(widget->specific.radiobutton);
	Display* display = display_get_Xdisplay();
	
#ifdef __debug__
	printf("%s() : %p : Begin\n", __func__, widget);
#endif
	gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
	XSetGraphicsExposures(display, gc, False);

	if (widget->state == DISABLED)
	{
		image_to_show = radiobutton->disabled;
		radiobutton->layout.font_color = radiobutton->disabled_font;
	}
	else
	{
		if (radiobutton->value)
			image_to_show = radiobutton->selected;

		if (image_to_show == NULL)
			image_to_show = radiobutton->normal;

		radiobutton->layout.font_color = radiobutton->normal_font;
	}

	if (widget->use_pixmap == True)
	{
		image_copy_pixmap(widget->win, &(widget->backing_store), widget->background.background_pixmap);
	}
	else
	{
		widget_draw_px_color(&(widget->backing_store), widget->background.background_color);
	}

	if (radiobutton->indicator_on)
	{
		px_theme* theme = theme_get_theme();
		
		if (radiobutton->value)
			indicator = theme->get_pixmap(theme, RADIOBUTTON_ON);
		else
			indicator = theme->get_pixmap(theme, RADIOBUTTON_OFF);

		if (indicator != NULL && indicator->pxmp != None)
		{
			int desty;

			desty = (widget->height - indicator->h) / 2;

			drawareax = indicator->w + 2;
			
			if (indicator->mask != None)
			{
				GC gcxor, gcand;
				gcxor = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
				XSetGraphicsExposures(display, gcxor, False);

			    	XSetForeground(display, gcxor,BlackPixel(display, DefaultScreen(display)));
				XSetBackground(display, gcxor,WhitePixel(display, DefaultScreen(display)));
				XSetFunction(display, gcxor, GXxor);
				XSetFillStyle(display, gcxor, FillSolid);
				XSetGraphicsExposures(display, gcxor, False);
	
				gcand = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
				XSetGraphicsExposures(display, gcand, False);

				XSetForeground(display, gcand, 0);
				XSetBackground(display, gcand, ~0);
				XSetFunction(display, gcand, GXand);
				XSetFillStyle(display, gcand, FillSolid);
				XSetGraphicsExposures(display, gcand, False);
	
				XCopyArea(display, indicator->pxmp, widget->backing_store.pxmp, gcxor, 
					0, 0, 
					indicator->w, indicator->h,
					1, desty); 

				XCopyPlane(display, indicator->mask, widget->backing_store.pxmp, gcand, 
					0, 0, 
					indicator->w, indicator->h,
					1, desty, 1L); 

				XCopyArea(display, indicator->pxmp, widget->backing_store.pxmp, gcxor, 
					0, 0,
					indicator->w, indicator->h,
					1, desty); 

				XFreeGC(display, gcxor);
				XFreeGC(display, gcand);
			}
			else
			{
				XCopyArea(display, indicator->pxmp, widget->backing_store.pxmp, gc,
					0, 0, 
					indicator->w, indicator->h,
					1, desty);
			}
		}

	}

	
	if (image_to_show != NULL)
	{
		int destx, desty;

		destx = ((widget->width - drawareax) - image_to_show->w) / 2 + drawareax;
		desty = (widget->height - image_to_show->h) / 2;

		if (image_to_show->pxmp != None)
			XCopyArea(display, image_to_show->pxmp, widget->backing_store.pxmp, gc,
				0, 0, 
				image_to_show->w, image_to_show->h,
				destx, desty);
		if (image_to_show->mask != None)
		{
			GC mask_gc;
			int created = 0;
			
			if (widget->backing_store.mask == None)
			{
				widget->backing_store.mask = XCreatePixmap(display, widget->win, 
						widget->backing_store.w, widget->backing_store.h, 1);
				created = 1;
			}
				
			mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
			XSetGraphicsExposures(display, mask_gc, False);
			
			XSetForeground(display, mask_gc, 1);
			
			if (created)
			{
				XFillRectangle(display, widget->backing_store.mask, mask_gc, 0, 0, 
						widget->backing_store.w, widget->backing_store.h);
			}
			
			XCopyArea(display, image_to_show->mask, widget->backing_store.mask, mask_gc,
					0, 0, 
					image_to_show->w, image_to_show->h,
					destx, desty);
			XFreeGC(display, mask_gc);
		}
		else if (widget->backing_store.mask != None)
		{
			XFreePixmap(display, widget->backing_store.mask);
			widget->backing_store.mask = None;
		}
	}

	if (radiobutton->indicator_on && widget->backing_store.mask != None && 
			indicator != NULL && indicator->mask != None)
	{
		int desty;
		GC mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
		XSetGraphicsExposures(display, mask_gc, False);

		desty = (widget->height - indicator->h) / 2;
		drawareax = indicator->w + 2;
		XCopyArea(display, indicator->mask, widget->backing_store.mask, mask_gc,
			0, 0, 
			indicator->w, indicator->h,
			1, desty);
		XFreeGC(display, mask_gc);
	}
	radiobutton->layout.x = drawareax;
	widget_draw_text(&(radiobutton->layout), &(widget->backing_store));

	XFreeGC(display, gc);
	widget_draw_border(widget);
	XClearWindow(display, widget->win);
	widget_set_event_mask(widget, widget->emask);

#ifdef __debug__
	XSync(display, False);
	printf("%s() : %p : End\n", __func__, widget);
#endif
}

void radiobutton_destroy(px_widget* widget)
{
	px_radiobutton* radiobutton = &(widget->specific.radiobutton);
	px_textchunk* chunk, *next;

	if (radiobutton->layout.text != NULL)
		free(radiobutton->layout.text);

	if (widget->specific.radiobutton.normal_font != NULL)
	{
		free(widget->specific.radiobutton.normal);
	}
	
	if (widget->specific.radiobutton.disabled_font != NULL)
	{
		free(widget->specific.radiobutton.disabled);
	}

	if (widget->specific.radiobutton.fontname != NULL)
	{
		font_free_font(widget->specific.radiobutton.fontname);
		free(widget->specific.radiobutton.fontname);
	}

	if (widget->specific.radiobutton.normal != NULL)
	{
		image_destroy(widget->specific.radiobutton.normal);
		free(widget->specific.radiobutton.normal);
	}
	
	if (widget->specific.radiobutton.selected != NULL)
	{
		image_destroy(widget->specific.radiobutton.selected);
		free(widget->specific.radiobutton.selected);
	}
	
	if (widget->specific.radiobutton.disabled != NULL)
	{
		image_destroy(widget->specific.radiobutton.disabled);
		free(widget->specific.radiobutton.disabled);
	}

	if (radiobutton->group != NULL)
	{
		radiobutton->group->no_of_children--;

		if (radiobutton->group->no_of_children == 0)
		{
			free(radiobutton->group);
		}
	}

	chunk = radiobutton->layout.chunks;
	while (chunk != NULL)
	{
		next = chunk->next;
		free(chunk);
		chunk = next;
	}

	widget_destroy(widget);
}

void radiobutton_set_normal_pixmap_from_file(px_widget* widget, char*path)
{
	px_radiobutton* specific = &(widget->specific.radiobutton);
	
	if (specific->normal == NULL)
	{
		specific->normal = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->normal);
	}

	if (!image_load_file(widget->win, specific->normal, path))
	{
		image_destroy(specific->normal);
		free(specific->normal);
		specific->normal = NULL;
	}
}

void radiobutton_set_selected_pixmap_from_file(px_widget* widget, char*path)
{
	px_radiobutton* specific = &(widget->specific.radiobutton);
	
	if (specific->selected == NULL)
	{
		specific->selected = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->selected);
	}

	if (!image_load_file(widget->win, specific->selected, path))
	{
		image_destroy(specific->selected);
		free(specific->selected);
		specific->selected = NULL;
	}
}

void radiobutton_set_disabled_pixmap_from_file(px_widget* widget, char*path)
{
	px_radiobutton* specific = &(widget->specific.radiobutton);
	
	if (specific->disabled == NULL)
	{
		specific->disabled = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->disabled);
	}

	image_load_file(widget->win, specific->disabled, path);
	if (!image_load_file(widget->win, specific->disabled, path))
	{
		image_destroy(specific->disabled);
		free(specific->disabled);
		specific->disabled = NULL;
	}
}

void radiobutton_set_normal_pixmap_from_pixmap(px_widget* widget, const px_pixmap* pixmap)
{
	px_radiobutton* specific = &(widget->specific.radiobutton);
	
	if (specific->normal == NULL)
	{
		specific->normal = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->normal);
	}

	if (!image_copy_pixmap(widget->win, specific->normal, pixmap))
	{
		image_destroy(specific->normal);
		free(specific->normal);
		specific->normal = NULL;
	}
}

void radiobutton_set_selected_pixmap_from_pixmap(px_widget* widget, const px_pixmap* pixmap)
{
	px_radiobutton* specific = &(widget->specific.radiobutton);
	
	if (specific->selected == NULL)
	{
		specific->selected = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->selected);
	}

	if (!image_copy_pixmap(widget->win, specific->selected, pixmap))
	{
		image_destroy(specific->selected);
		free(specific->selected);
		specific->selected = NULL;
	}
}

void radiobutton_set_disabled_pixmap_from_pixmap(px_widget* widget, const px_pixmap* pixmap)
{
	px_radiobutton* specific = &(widget->specific.radiobutton);
	
	if (specific->disabled == NULL)
	{
		specific->disabled = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(specific->disabled);
	}

	if (!image_copy_pixmap(widget->win, specific->disabled, pixmap))
	{
		image_destroy(specific->disabled);
		free(specific->disabled);
		specific->disabled = NULL;
	}
}

void radiobutton_set_normal_font_color(px_widget* widget, const char* colour)
{
	if (widget->specific.radiobutton.normal_font == NULL)
	{
		widget->specific.radiobutton.normal_font  = (XColor*)malloc(sizeof(XColor));
	}
	widget_xcolor_from_str(widget->specific.radiobutton.normal_font, colour);
}


void radiobutton_set_disabled_font_color(px_widget* widget, const char* colour)
{
	if (widget->specific.radiobutton.disabled_font == NULL)
	{
		widget->specific.radiobutton.disabled_font = (XColor*)malloc(sizeof(XColor));
	}
	widget_xcolor_from_str(widget->specific.radiobutton.disabled_font, colour);
}

void radiobutton_set_text(px_widget* widget, const char* t)
{
	if (widget->specific.radiobutton.layout.text != NULL)
	{
		free(widget->specific.radiobutton.layout.text);
		widget->specific.radiobutton.layout.text = NULL;
	};

	if (t != NULL)
	{
		widget->specific.radiobutton.layout.text = strdup(t);
		widget_compute_textlayout(&(widget->specific.radiobutton.layout), widget);
	}
}

void radiobutton_set_font(px_widget* widget, const char* fontname)
{
	if (widget->specific.radiobutton.fontname != NULL)
	{
		font_free_font(widget->specific.radiobutton.fontname);
		free(widget->specific.radiobutton.fontname);
		widget->specific.radiobutton.fontname = NULL;
	}

	widget->specific.radiobutton.fontname = strdup(fontname);
	widget->specific.radiobutton.layout.font = font_load_font(widget->specific.radiobutton.fontname);
	widget_compute_textlayout(&(widget->specific.radiobutton.layout), widget);
}

void radiobutton_set_justification(px_widget* widget, justification j)
{
	widget->specific.radiobutton.layout.just = j;
}

justification radiobutton_get_justification(const px_widget* widget)
{
	return widget->specific.radiobutton.layout.just;
}

const char* radiobutton_get_text(const px_widget* widget)
{
	return widget->specific.radiobutton.layout.text;
}

void radiobutton_show_indicator(px_widget* widget)
{
	widget->specific.radiobutton.indicator_on = True;
}

void radiobutton_hide_indicator(px_widget* widget)
{
	widget->specific.radiobutton.indicator_on = False;
}

void radiobutton_select(px_widget* widget)
{
	if (!widget->specific.radiobutton.value)
	{
		if (widget->specific.radiobutton.group != NULL)
		{
			px_widget* sibling = widget->specific.radiobutton.group->selected;
			
			if (sibling != NULL)
			{
				sibling->specific.radiobutton.value = False;
				sibling->redraw(sibling);
				sibling->show(sibling);
				if (sibling->specific.radiobutton.on_deselect != NULL)
					sibling->specific.radiobutton.on_deselect(sibling);
			}
			widget->specific.radiobutton.group->selected = widget;
		}

		widget->specific.radiobutton.value = True;
		if (widget->specific.radiobutton.on_select != NULL)
			widget->specific.radiobutton.on_select(widget);
	}
}

void radiobutton_deselect(px_widget* widget)
{
	if (widget->specific.radiobutton.group == NULL)
	{
		widget->specific.radiobutton.value = False;
		if (widget->specific.radiobutton.on_deselect != NULL)
			widget->specific.radiobutton.on_deselect(widget);
	}
}

void radiobutton_group_together(px_widget* orphan, px_widget* sibling)
{
	if (sibling->specific.radiobutton.group == NULL)
	{
		sibling->specific.radiobutton.group = (px_radiobuttongroup*)malloc(sizeof(px_radiobuttongroup));
		sibling->specific.radiobutton.group->no_of_children = 1;

		if (sibling->specific.radiobutton.value)
		{
			sibling->specific.radiobutton.group->selected = sibling;
		}
		else
		{
			sibling->specific.radiobutton.group->selected = NULL;
		}
	}
	if (orphan->specific.radiobutton.group != NULL)
	{
		orphan->specific.radiobutton.group->no_of_children--;

		if (orphan->specific.radiobutton.group->no_of_children == 0)
		{
			free(orphan->specific.radiobutton.group);
		}
	}
	orphan->specific.radiobutton.group = sibling->specific.radiobutton.group;
	sibling->specific.radiobutton.group->no_of_children++;

	if (orphan->specific.radiobutton.value)
	{
		if (orphan->specific.radiobutton.group->selected != NULL)
		{
			orphan->specific.radiobutton.group->selected->specific.radiobutton.value = False;
			orphan->specific.radiobutton.group->selected->redraw(orphan->specific.radiobutton.group->selected);
		}
		
		orphan->specific.radiobutton.group->selected = orphan;
	}
}
