/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include <X11/extensions/shape.h>
#include <X11/xpm.h>
#include <stdlib.h>
#include <string.h>
#include <image.h>
#include <theme.h>
#include "tabs.h"
#include "font.h"
#include "eventloop.h"
#include "widget_factory.h"
#include "display.h"

#define NORMAL_FONT "-picopeta-simputer-medium-r-normal--11-*-*-*-p-*-fontspecific-0"
#define BOLD_FONT "-picopeta-simputer-bold-r-normal--11-*-*-*-p-*-fontspecific-0"


px_widget *tab_create(px_widget *parent, int x, int y, int width, int height, const char* tag)
{
        px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
        Display* display = display_get_Xdisplay();

        widget->win = XCreateWindow(display, parent->win, x, y, width, height, 0, CopyFromParent, CopyFromParent, CopyFromParent, 0, NULL);
#ifdef __debug__
        printf("%s(): %p - Created Window with parent window = %lu, x = %d, y = %d, width = %d, height = %d, new window = %lu\n\n", 
                __func__, widget, parent->win, x, y, width, height, widget->win);

        XSync(display, False);
#endif

        widget->x = x;
        widget->y = y;
        widget->height = height;
        widget->width = width;
        widget->use_pixmap = False;

        if (tag != NULL)
        {
                widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	 }
	 else
		widget->tag = NULL;
	
        widget->wtype = TAB;
	widget->child_of = NULL;

	widget->background.background_color = (px_color*)malloc(sizeof(px_color));
	widget->background.background_color->type = SOLID;
	widget_xcolor_from_str(&(widget->background.background_color->first_color), "white");

	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);

#ifdef __debug__
	printf("%s(): XCreateWindow : %p - Created Window with parent window = %lu, x = %d, y = %d, width = %d, height = %d, new window = %lu\n", 
			__func__, widget, widget->win, x, y, width, height, widget->win);
	XSync(display, False);
#endif

	return widget;
}

Bool tab_handle_event(px_widget* widget, px_event* event)
{
        px_tab* tab = &(widget->specific.tab);
        pw_tab *internal = tab->internal;
	Bool return_value = True;
#ifdef __debug__
	
        printf("%s() : %p Event type = %d at %d %d\n", __func__, widget, event->type, event->xev->xbutton.x, event->xev->xbutton.y);
#endif

	if (event->type != NR_CALLBACKS && (event->type == EXPOSE || widget->get_state(widget) != DISABLED))
	{
                if (event->y > 25)
                {
                         event->y -= internal->widget_list[internal->widget_selected]->y + internal->widget_list[internal->widget_selected]->child_of->y;
                         event->x -= internal->widget_list[internal->widget_selected]->x + internal->widget_list[internal->widget_selected]->child_of->x;
                         return_value = internal->widget_list[internal->widget_selected]->handle_event(internal->widget_list[internal->widget_selected], event);
                         event->y += internal->widget_list[internal->widget_selected]->y + internal->widget_list[internal->widget_selected]->child_of->y;
                         event->x += internal->widget_list[internal->widget_selected]->x + internal->widget_list[internal->widget_selected]->child_of->x;
                } else 
                {
                         return_value = widget_handle_event(widget, event);
                }
	}
	return return_value;
}

void tab_set_parameters(px_widget *widget)
{
        px_tab* tab = &(widget->specific.tab);
        int i;

        tab->internal = (pw_tab *)malloc(sizeof(pw_tab));
        pw_tab *internal = tab->internal;

        px_theme* theme = theme_get_theme();
        internal->normal_pixmap= NULL;
        internal->active_pixmap = NULL;


        internal->content_frame = new_widget (FRAME, widget, 5, 25, widget->width-10, widget->height-35, "Content Frame");
        internal->content_frame->set_border_type(internal->content_frame, RECTANGLE_BORDER);
        internal->content_frame->redraw(internal->content_frame);
        internal->content_frame->show(internal->content_frame);



//        widget->default_callbacks[BUTTON_PRESSED] = tab_default_button_pressed;
        widget->default_callbacks[BUTTON_RELEASED] = tab_default_button_released;
    
//        widget->set_size = tab_set_size;
        widget->redraw = tab_redraw;
        widget->destroy = tab_destroy;
        internal->event_within = NULL;
        internal->normal_pixmap = theme_get_pixmap(theme, TABIMAGE);
        internal->active_pixmap = theme_get_pixmap(theme, TABIMAGE_ACTIVE);

        widget->handle_event = tab_handle_event;
        internal->ends_at = 5;

        internal->widget_list = (px_widget**)malloc(sizeof(px_widget*) * 5);
        internal->tablist = (px_tab_holder**)malloc(sizeof(px_tab_holder*) * 5);
	
        for (i = 0; i < 5; i++)
        {
                internal->widget_list[i] = NULL;
                internal->tablist[i] = NULL;
        }

	
        internal->allocated = 5;
	internal->no_of_children= 0;
        internal->widget_selected = 0;
        tab->add_tab= tab_add_tab;
        tab->raise_tab = tab_raise_tab;
	tab->get_active_widget = tab_get_active_widget;
	tab->get_widget_pointer = tab_get_widget_pointer;
}

px_widget *tab_get_active_widget(px_widget *widget)
{
        px_tab* tab = &(widget->specific.tab);
        pw_tab *internal = tab->internal;
        return internal->widget_list[internal->widget_selected];
}

px_widget *tab_get_widget_pointer(px_widget *widget, int tabnumber)
{
        px_tab* tab = &(widget->specific.tab);
        pw_tab *internal = tab->internal;
        return internal->widget_list[tabnumber];

}


px_widget* tab_add_tab (px_widget *widget, const char *title, Bool (*tab_callback)(px_widget*, px_event*))
{
        px_tab* tab = &(widget->specific.tab);
        Display *display = display_get_Xdisplay();
        GC gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
        // static int x = 5;
        int font_width;
        pw_tab *internal = tab->internal;

        XFontStruct *normal_font, *selected_font;

        normal_font = XLoadQueryFont(display, NORMAL_FONT);
        selected_font = XLoadQueryFont(display, BOLD_FONT);

        if (internal->allocated == internal->no_of_children)
        {
                unsigned int count;
                
                px_widget** new_list = (px_widget**)malloc(sizeof(px_widget*)*(internal->allocated + 10));
                px_tab_holder** new_tablist = (px_tab_holder**)malloc(sizeof(px_tab_holder*) * (internal->allocated + 10));
             
                for (count = 0; count < internal->no_of_children; count++)
                {
                        new_list[count] = internal->widget_list[count];
                        new_tablist[count] = internal->tablist[count];
                }

                free(internal->widget_list);
                free(internal->tablist);
                internal->widget_list = new_list;
                internal->tablist = new_tablist;
                internal->allocated += 10;
        }

    
        font_width = font_get_width(&normal_font, title, strlen(title)); 

        if (internal->no_of_children == 0)
        {
        
                XCopyArea(display, internal->active_pixmap->pxmp, widget->backing_store.pxmp, gc, 0, 0, 20, 20, internal->ends_at, 5);
                XCopyArea(display, internal->active_pixmap->pxmp, widget->backing_store.pxmp, gc, 20, 0, font_width, 20, internal->ends_at+20, 5);
                XCopyArea(display, internal->active_pixmap->pxmp, widget->backing_store.pxmp, gc, 310, 0, 20, 20, internal->ends_at+font_width+20, 5);
                XSetFont(display, gc, selected_font->fid);
        } else
        {
                XCopyArea(display, internal->normal_pixmap->pxmp, widget->backing_store.pxmp, gc, 0, 0, 20, 20, internal->ends_at, 5);
                XCopyArea(display, internal->normal_pixmap->pxmp, widget->backing_store.pxmp, gc, 20, 0, font_width, 20, internal->ends_at+20, 5);
                XCopyArea(display, internal->normal_pixmap->pxmp, widget->backing_store.pxmp, gc, 310, 0, 20, 20, internal->ends_at+font_width+20, 5);
                XSetFont(display, gc, normal_font->fid);
        }
   
        XDrawString(display, widget->backing_store.pxmp, gc, internal->ends_at+15, 20, title, strlen(title));
   
        internal->ends_at = internal->ends_at + font_width+ 30;

        XFreeGC(display, gc);

        internal->tablist[internal->no_of_children] = (px_tab_holder*)malloc(sizeof(px_tab_holder));
        internal->tablist[internal->no_of_children]->tab_position = internal->no_of_children;
        internal->tablist[internal->no_of_children]->xend_position = internal->ends_at;
        internal->tablist[internal->no_of_children]->callbacks = tab_callback;
        internal->tablist[internal->no_of_children]->tab_title = title;

        internal->widget_list[internal->no_of_children] = new_widget(FRAME, internal->content_frame, 2, 2, widget->width-15, widget->height-40, NULL);
        internal->no_of_children++;
	
	if (internal->no_of_children)
		tab_raise_tab(widget, 0);

        XClearWindow(display, widget->win);
        return internal->widget_list[internal->no_of_children-1];

}


void tab_raise_tab(px_widget *widget, int tabnumber)
{
        px_tab* tab = &(widget->specific.tab);
        pw_tab *internal = tab->internal;

	   redraw_tabs(widget, tab, internal->widget_selected, tabnumber);

	   internal->widget_list[internal->widget_selected]->hide(internal->widget_list[internal->widget_selected]);
        internal->widget_selected = tabnumber;
        internal->widget_list[internal->widget_selected]->redraw(internal->widget_list[internal->widget_selected]);
        internal->widget_list[internal->widget_selected]->show(internal->widget_list[internal->widget_selected]);
}


void tab_show (px_widget* widget)
{
        Display* display = display_get_Xdisplay();

    #ifdef __debug__
    	printf("%s(): %p Start\n", __func__, widget);
    #endif
	 
        XSetWindowBackgroundPixmap(display, widget->win, widget->backing_store.pxmp);
	
        if (widget->backing_store.mask != None)
                XShapeCombineMask (display, widget->win, ShapeBounding, 0, 0, 
			widget->backing_store.mask, ShapeSet);

	        XMapWindow(display, widget->win);
	        widget->hidden = False;


#ifdef __debug__
	        XSync(display, False);
	printf("%s() : %p : End\n", __func__, widget);
#endif

}

void tab_redraw(px_widget* widget)
{
        unsigned int count;
	Display* display = display_get_Xdisplay();
        px_tab* tab = &(widget->specific.tab);
        pw_tab *internal = tab->internal;
	int i;

#ifdef __debug__
	printf("%s(): %p\n", __func__, widget);
#endif
	
	if (widget->use_pixmap == True)
	{
		image_copy_pixmap(widget->win, &(widget->backing_store), widget->background.background_pixmap);
	}
	else
	{
		widget_draw_px_color(&(widget->backing_store), widget->background.background_color);
	}

	if (widget->use_pixmap == False && widget->background.background_color->type == TRANSPARENT)
	{
                GC  mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
		XSetGraphicsExposures(display, mask_gc, False);
		XSetForeground(display, mask_gc, 1);
		for (count = 0; count < internal->no_of_children; count++)
		{
			px_widget* child = internal->widget_list[count];
		
			XCopyArea(display, child->backing_store.mask,
					widget->backing_store.mask, mask_gc, 0, 0, 
					child->backing_store.w,
					child->backing_store.h,
					child->x,
					child->y);
		}
		XFreeGC(display, mask_gc);
	}
	widget_draw_border(widget);
	
	if (internal->no_of_children)
		tab_raise_tab(widget, 0);

	for (i=internal->no_of_children; i>0; i--)
	{
		redraw_tabs(widget, tab, i, i-1);
	}
	
	XClearWindow(display, widget->win);
	widget_set_event_mask(widget, widget->emask);

#ifdef __debug__
	printf("%s(): %p - done\n", __func__, widget);
#endif

}

void tab_destroy(px_widget* widget)
{
        px_tab *tab = &(widget->specific.tab);
        pw_tab *internal = tab->internal;
        int count;

        for (count = 0; count < internal->no_of_children; count++)
        {
                free(internal->tablist[count]);
                internal->widget_list[count]->destroy(internal->widget_list[count]);
                free(internal->widget_list[count]);
        }

        free(internal->tablist);
        free(internal->widget_list);
        widget_destroy(widget); 
}

Bool tab_default_button_released(px_widget* widget, px_event *event)
{
        px_tab* tab = &(widget->specific.tab);
        pw_tab *internal = tab->internal;

        if (event->y <= 25)
        {
                if (event->x > 5 && event->x < internal->ends_at)
                {
                        int count;
                        for (count=0; count < internal->no_of_children; count++)
                        {
                                if (event->x <= internal->tablist[count]->xend_position)
                                {
                                        if (internal->widget_selected == internal->tablist[count]->tab_position)
                                                break;
                                        redraw_tabs(widget, tab, internal->widget_selected, internal->tablist[count]->tab_position);
                                        internal->widget_list[internal->widget_selected]->hide(internal->widget_list[internal->widget_selected]);
                                        internal->widget_selected = internal->tablist[count]->tab_position;
                                        internal->widget_list[internal->widget_selected]->redraw(internal->widget_list[internal->widget_selected]);
                                        internal->widget_list[internal->widget_selected]->show(internal->widget_list[internal->widget_selected]);
								if(internal->tablist[internal->widget_selected]->callbacks)
	                                        internal->tablist[internal->widget_selected]->callbacks(internal->widget_list[internal->widget_selected], event);
                                        break;
                                }
                        }
                }

        }

        return True;
}

void redraw_tabs( px_widget *widget, px_tab *tab, int prev_tab, int act_tab)
{
        Display *display = display_get_Xdisplay();
        GC gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
        int i, x = 5;
        int font_width;
        XFontStruct *normal_font, *selected_font;
        pw_tab *internal = tab->internal;

        normal_font = XLoadQueryFont(display, NORMAL_FONT);
        selected_font = XLoadQueryFont(display, BOLD_FONT);

        if (prev_tab == act_tab)
                return;

        XClearWindow(display, widget->win);

        for (i=0; i < internal->no_of_children; i++)
        {
                font_width = font_get_width(&normal_font, internal->tablist[i]->tab_title, strlen(internal->tablist[i]->tab_title)); 
                if (prev_tab == i)
                {
                        XCopyArea(display, internal->normal_pixmap->pxmp, widget->backing_store.pxmp, gc, 0, 0, 20, 20, x, 5);
                        XCopyArea(display, internal->normal_pixmap->pxmp, widget->backing_store.pxmp, gc, 20, 0, font_width, 20, x+20, 5);
                        XCopyArea(display, internal->normal_pixmap->pxmp, widget->backing_store.pxmp, gc, 310, 0, 20, 20, x+font_width+20, 5);
                        XSetFont(display, gc, normal_font->fid);
                        XDrawString(display, widget->backing_store.pxmp, gc, x+15, 21, internal->tablist[i]->tab_title, strlen(internal->tablist[i]->tab_title));
                } else if (act_tab == i)
                {
                        XCopyArea(display, internal->active_pixmap->pxmp, widget->backing_store.pxmp, gc, 0, 0, 20, 20, x, 5);
                        XCopyArea(display, internal->active_pixmap->pxmp, widget->backing_store.pxmp, gc, 20, 0, font_width, 20, x+20, 5);
                        XCopyArea(display, internal->active_pixmap->pxmp, widget->backing_store.pxmp, gc, 310, 0, 20, 20, x+font_width+20, 5);
   
                        XSetFont(display, gc, selected_font->fid);
                        XDrawString(display, widget->backing_store.pxmp, gc, x+15, 21, internal->tablist[i]->tab_title, strlen(internal->tablist[i]->tab_title));
                }
        
                x = x + font_width+ 30;
        }
        XClearWindow(display, widget->win);
        XFreeGC(display, gc);
}
