/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

// PicoWidget : Text
// Copyrights (c) PicoPeta Simputers Pvt. Ltd.
// Programmer: Santhosh N

#include "text.h"

#define BLINK_TIME	300000

struct _paragraph {
	char *text;						// Text data.
	unsigned allocated;					// Number of bytes allocated, to store text.

	unsigned *wrap_positions;			// Has positions where the text will be wrapped.
	unsigned num_lines;					// Will have a value of atleast 1.

	struct _paragraph *next;
	struct _paragraph *prev;
};

struct _text {
	struct _paragraph *data;						// Pointer to the paragraph linked list.
	struct _paragraph *row_ptr;					// Pointer to the current edit paragraph.
	unsigned cursor_pos;						// Cursor postion in the current edit paragraph.
	int top_x, top_y;							// Total text area.
	unsigned total_width, total_height;
	px_widget *hscrollbar, *vscrollbar;			// Scrollbars.
	unsigned width, height;						// Dimensions of the widget. Used to resize widget.
	GC gc;

     unsigned wraptype;
	unsigned interline_space;
     char *fontname;
     px_font *font;
     XColor *foreground;
	XColor *disabled_foreground;
};
	
static px_timer *cursor_blink=NULL;
static int cursor_x, cursor_y, blink_fg=0;

// Function declarations
static void cursor_blink_timer(void*, void*);
static Bool text_handle_event(px_widget* widget, px_event* event);
static void text_hviewport_changed(px_widget* widget, float slider_pos);
static void text_vviewport_changed(px_widget* widget, float slider_pos);
static unsigned text_display_width(px_widget *widget);
static unsigned text_display_height(px_widget *widget);
static void text_reset_view(px_widget *widget);

static struct _paragraph *paragraph_create_newnode();
static void paragraph_node_free(struct _paragraph **node);
static void paragraph_list_free(struct _paragraph **list);
static unsigned paragraph_list_numlines(struct _paragraph **list);
static struct _paragraph *load_file(const char *filename);
static struct _paragraph *load_text(const char *text);
static void paragraph_list_wrapnode(struct _paragraph **list, unsigned wraptype, px_font *font, unsigned width);
static void paragraph_list_wrap(struct _paragraph **list, unsigned wraptype, px_font *font, unsigned width);
static void paragraph_insert_char(struct _paragraph **paragraph, int pos, char ch);
static void paragraph_delete_char(struct _paragraph **list, struct _paragraph **node, int pos);

#if 0
void print_llist(struct _paragraph **paragraph)
{
	struct _paragraph *p=*paragraph;

	printf("=====================================\n");
	while(p) {
		printf("me = %p, prev = %p, next = %p, numlines = %d text = %s\n", p, p->prev, p->next, p->num_lines, p->text);
		p = p->next;
	}
	printf("=====================================\n");
}
#endif

px_widget* text_create(px_widget* parent, int x, int y, int width, int height, const char* tag)
{
	px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
	Display* display = display_get_Xdisplay();

	// Initialize common widget variables
	widget->win = XCreateWindow(display, parent->win, x, y, width, height, 
							0, CopyFromParent, CopyFromParent, CopyFromParent, 0, NULL);

	XSelectInput(display, widget->win, PropertyChangeMask);
	
	widget->x = x;
	widget->y = y;
	widget->height = height;
	widget->width = width;

	if(tag != NULL) {
		widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	} else
		widget->tag = NULL;

	widget->wtype = TEXT;

	widget->use_pixmap = False;
	widget->background.background_color = (px_color*)malloc(sizeof(px_color));
	widget->background.background_color->type = SOLID;
	widget_xcolor_from_str(&(widget->background.background_color->first_color), "#ffffff");

	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);

	return widget;
}

void text_set_parameters(px_widget* widget)
{
	px_text *text = &(widget->specific.text);
	struct _text *internal;
	Display* display = display_get_Xdisplay();

	// Initialize common widget default callbacks
	widget->default_callbacks[BUTTON_PRESSED] = text_default_button_pressed;
	widget->default_callbacks[BUTTON_RELEASED] = text_default_button_released;
	widget->default_callbacks[MOTION] = text_default_motion;
	widget->default_callbacks[KEY_PRESSED] = text_default_key_pressed;
	widget->default_callbacks[KEY_RELEASED] = text_default_key_released;

	widget->handle_event = text_handle_event;

	// Create internal data structure
	text->internal = internal = (struct _text *)malloc(sizeof(struct _text));
	
	// Create Scrollbars
	internal->hscrollbar = new_widget(HSCROLLBAR, widget, 0, widget->height-16, widget->width-16, 16, "");
	internal->hscrollbar->specific.hscrollbar.intimate_change = text_hviewport_changed;
	internal->vscrollbar = new_widget(VSCROLLBAR, widget, widget->width-16, 0, 16, widget->height-16, "");
	internal->vscrollbar->specific.vscrollbar.intimate_change = text_vviewport_changed;
	
	// Initialize internal widget structure variables
	internal->data = (struct _paragraph *)paragraph_create_newnode();
	internal->row_ptr = internal->data;
	internal->wraptype = NO_WRAP;
	internal->top_x = internal->top_y = 0;
	internal->cursor_pos = 0;
	internal->total_width = widget->width-16;
	internal->total_height = widget->height-16;
	internal->interline_space = 2;
	internal->width = widget->width;
	internal->height = widget->height;
	internal->gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
	
	internal->fontname = strdup("-picopeta-simputer-medium-r-normal--11-120-101-108-p-120-fontspecific-0");
	internal->font = font_load_font(internal->fontname);
	
	internal->foreground = (XColor*)malloc(sizeof(XColor));
	widget_xcolor_from_str(internal->foreground, "black");
	internal->disabled_foreground = (XColor*)malloc(sizeof(XColor));
	widget_xcolor_from_str(internal->disabled_foreground, "white");

	// Initialize all function pointers
	widget->redraw = text_redraw;
	widget->destroy = text_destroy;
	text->set_font = text_set_font;
	text->set_wrap_type = text_set_wrap_type;
	text->set_foreground_color = text_set_foreground_color;
	text->load_file = text_load_file;
	text->save_file = text_save_file;
	text->get_numlines = text_get_numlines;
	text->get_line = text_get_line;
	text->get_text = text_get_text;
	text->set_text = text_set_text;

	// Set some attributes of the widget
	XSetFont(display, internal->gc, internal->font->font->fid);
	
	text_reset_view(widget);
}

static Bool text_handle_event(px_widget* widget, px_event* event)
{
	struct _text *internal = (struct _text *)widget->specific.text.internal;
	Bool return_value = True;

	if(event->type!=NR_CALLBACKS && (event->type==EXPOSE || widget->get_state(widget)!=DISABLED)) {
		// Check if within the scrollbars
		if(event->x > widget->width-16 || event->y > widget->height-16) {
			px_widget *scrollbar;
		
			if(event->x>widget->width-16)
				scrollbar = internal->vscrollbar;
			else
				scrollbar = internal->hscrollbar;
			
			event->x -= scrollbar->x; event->y -= scrollbar->y; 
			return_value = scrollbar->handle_event(scrollbar, event);
			event->x += scrollbar->x; event->y += scrollbar->y;
		} else {
			return_value = widget_handle_event(widget, event);
		}
	}

	return return_value;
}

void text_redraw(px_widget *widget)
{
	struct _text *internal = widget->specific.text.internal;
	Display* display = display_get_Xdisplay();
	widget_states state;
	int font_height, line_incr, y, cursor_pos=0;
	struct _paragraph *p = (struct _paragraph *)internal->data;

	state = (*widget->get_state)(widget);
	font_height = font_get_height(internal->font, "");
	line_incr = font_height + internal->interline_space;

	// Draw the background color
	widget_draw_px_color(&(widget->backing_store), widget->background.background_color);

	// If the size of the widget has changed, update the contents.
	if(internal->width!=widget->width || internal->height!=widget->height) {
		internal->width = widget->width;
		internal->height = widget->height;
		
		paragraph_list_wrap((struct _paragraph **)&(internal->data), internal->wraptype, 
											internal->font, internal->width-16-4);
	
		widget_set_size(internal->vscrollbar, 16, internal->height-16);
		widget_set_size(internal->hscrollbar, internal->width-16, 16);
		widget_set_position(internal->hscrollbar, 0, internal->height-16);
		widget_set_position(internal->vscrollbar, internal->width-16, 0);
		
		internal->vscrollbar->redraw(internal->vscrollbar);
		internal->hscrollbar->redraw(internal->hscrollbar);
	}
	
	// Set foreground color
	switch(state) {
		case DISABLED:
			if(internal->disabled_foreground)
				XSetForeground(display, internal->gc, internal->disabled_foreground->pixel);
			break;
		case NORMAL:
			if(internal->foreground)
				XSetForeground(display, internal->gc, internal->foreground->pixel);
			break;
		default:
			XSetForeground(display, internal->gc, 0);
	}

	// Draw the text
	y = internal->top_y;
	while(p) {
		// Go through each paragraph
		if(p->text) {
			int i, off=0, slen;

			slen = strlen(p->text);

			// Draw the text line-by-line	
			if(p->num_lines!=1) {
				for(i=0; i<p->num_lines-1; i++) {
					// Get the cursor x and y position
					if(internal->row_ptr==p) {
						cursor_pos = internal->cursor_pos;
						if(cursor_pos>=off && cursor_pos<p->wrap_positions[i]) {
							cursor_y = y;
							cursor_x = font_get_width(internal->font, p->text+off, cursor_pos-off);
						}
					}
					// Draw the string
					if(y+font_height>=0 && y<widget->height-16)
						XDrawString(display, widget->backing_store.pxmp, internal->gc,
									internal->top_x+2, y+font_height, 
									p->text+off, p->wrap_positions[i]-off);
					y += line_incr;
					off = p->wrap_positions[i];
				}

				// Get the cursor x and y position
				if(internal->row_ptr==p) {
					cursor_pos = internal->cursor_pos;
					if(cursor_pos>=off && cursor_pos<=slen) {
						cursor_y = y;
						cursor_x = font_get_width(internal->font, p->text+off, cursor_pos-off);
					}
				}
				// Draw the string
				if(y+font_height>=0 && y<widget->height-16)
					XDrawString(display, widget->backing_store.pxmp, internal->gc,
							internal->top_x+2, y+font_height, p->text+off, slen-off);
				y += line_incr;
			} else {
				// Get the cursor x and y position
				if(internal->row_ptr==p) {
					cursor_pos = internal->cursor_pos;
					cursor_y = y;
					cursor_x = internal->top_x + font_get_width(internal->font, p->text, cursor_pos);
				}
				// Draw the string
				if(y+font_height>=0 && y<widget->height-16)
					XDrawString(display, widget->backing_store.pxmp, internal->gc,
							internal->top_x+2, y+font_height, p->text, slen);
				y += line_incr;
			}
		} else {
			// Get the cursor x and y position
			if(internal->row_ptr==p) {
				cursor_x = internal->top_x;
				cursor_y = y;
			}
			y += line_incr;
		}
		
		p = p->next;
	}

	// Draw the cursor
	if(!widget->have_focus)
		XSetForeground(display, internal->gc, widget->background.background_color->first_color.pixel);
	XDrawLine(display, widget->backing_store.pxmp, internal->gc, 
					cursor_x+1, cursor_y, cursor_x+1, cursor_y+font_height);
		
	// Fill the bottom left portion with a rectangle
	XSetForeground(display, internal->gc, widget->background.background_color->first_color.pixel);
	XFillRectangle(display, widget->backing_store.pxmp, internal->gc, 
								widget->width-16, widget->height-16, 16, 16);

	XClearWindow(display, widget->win);
}

Bool text_default_button_pressed(px_widget *widget, px_event *event)
{
	struct _text *internal = widget->specific.text.internal;
	px_eventloop* eventloop = eventloop_get_eventloop();
	int font_height, line_incr, i, j, y, slen;
	struct _paragraph *p = (struct _paragraph *)internal->data;
	px_widget *prev_widget;

	eventloop->grab(eventloop, widget);
	
	// Set the focus, if not set already
	if(eventloop->focussed_widget != widget) {
		prev_widget = eventloop->focussed_widget;
     	eventloop->focus(eventloop, widget);
		if(prev_widget)
			prev_widget->redraw(prev_widget);
	}
	
	font_height = font_get_height(internal->font, "");
	line_incr = font_height + internal->interline_space;

	// Find out the x,y position and set the new cursor position.
	y = internal->top_y;
	while(p) {
		for(i=0; i<p->num_lines; i++) {
			if(event->y>=y && event->y<y+line_incr) {
				internal->row_ptr = p;

				if(p->text) {
					slen = strlen(p->text);
						
					if(p->num_lines==1) {
						for(j=0; j<slen; j++)
							if((-internal->top_x)+event->x < font_get_width(internal->font, p->text, j))
								break;
									
						if(j==slen && (-internal->top_x)+event->x > font_get_width(internal->font, p->text, j))
							internal->cursor_pos = slen;
						else
							internal->cursor_pos = j-1;
					} else {
						if(i==0) {
							// First line
							for(j=0; j<p->wrap_positions[0]; j++)
								if(event->x < font_get_width(internal->font, p->text, j))
									break;

							if(j==p->wrap_positions[0] && event->x > font_get_width(internal->font, p->text, j))
								internal->cursor_pos = p->wrap_positions[0]-1;
							else
								internal->cursor_pos = j-1;
						} else if(i==p->num_lines-1) {
							int sp = p->wrap_positions[i-1];
							// Last line
							for(j=sp; j<slen; j++)
								if(event->x < font_get_width(internal->font, p->text+sp, j-sp))
									break;

							if(j==slen && event->x > font_get_width(internal->font, p->text+sp, j-sp))
								internal->cursor_pos = slen;
							else
								internal->cursor_pos = j-1;
						} else {
							int sp = p->wrap_positions[i-1], ep = p->wrap_positions[i];
							// In-between lines
							for(j=sp; j<ep; j++)
								if(event->x < font_get_width(internal->font, p->text+sp, j-sp))
									break;

							if(j==ep && event->x > font_get_width(internal->font, p->text+sp, j-sp))
								internal->cursor_pos = ep-1;
							else
								internal->cursor_pos = j-1;
						}
					}
				} else {
					internal->cursor_pos = 0;
				}
				goto end;
			}
			y += line_incr;
		}
		p = p->next;
	}

end:
	widget->redraw(widget);

	return True;
}

Bool text_default_button_released(px_widget* widget, px_event* event)
{
	px_eventloop* eventloop = eventloop_get_eventloop();

	eventloop->grabbed_widget = NULL;

	return True;
}

Bool text_default_motion(px_widget* widget, px_event* event)
{
	return True;
}

Bool text_default_key_pressed(px_widget* widget, px_event* event)
{
	KeySym keysym;
	char buffer[10];
	int bufflen=0, slen=0;
	struct _text *internal = widget->specific.text.internal;
	XEvent xev;
	int font_height = font_get_height(internal->font, "");
	int line_incr = font_height + internal->interline_space;
	float hlen, vlen;
	Display* display = display_get_Xdisplay();

	bufflen = XLookupString(&event->xev->xkey, buffer, 10, &keysym, NULL);

	event->ch = buffer[0];

	switch(keysym)	{
		case XK_Shift_L:
		case XK_Shift_R:
			return True;

		case XK_Left:
			if(internal->cursor_pos>0) {
				internal->cursor_pos--;
			} else {
				// Cursor reached the beginning of the line.
				// Move to the previous line, if it exists.
				if(internal->row_ptr->prev) {
					internal->row_ptr = internal->row_ptr->prev;
			
					if(internal->row_ptr->text)
						slen = strlen(internal->row_ptr->text);

					if(internal->row_ptr->text)
						internal->cursor_pos = slen;
					else
						internal->cursor_pos = 0;
				}
			}
			break;

		case XK_Right:
			if(internal->row_ptr->text)
				slen = strlen(internal->row_ptr->text);

			if(internal->cursor_pos < slen) {
				internal->cursor_pos++;
			} else {
				// Cursor reached the end of the line.
				// Move to the next line, if it exists.
				if(internal->row_ptr->next) {
					internal->row_ptr = internal->row_ptr->next;
					internal->cursor_pos = 0;
				}
			}
			break;

		case XK_Up:
				if(internal->row_ptr->prev) {
					internal->row_ptr = internal->row_ptr->prev;

					if(internal->row_ptr->text)
						slen = strlen(internal->row_ptr->text);

					// If the next line is wrapped and the cursor position is
					// more than the line length, set the cursor position to 
					// the first wrap point.
					if(internal->row_ptr->wrap_positions) {
						if(internal->cursor_pos >= internal->row_ptr->wrap_positions[0])
							internal->cursor_pos = internal->row_ptr->wrap_positions[0]-1;
					} else {
						if(internal->cursor_pos >= slen)
							internal->cursor_pos = slen;
					}
				}
			break;

		case XK_Down:
				if(internal->row_ptr->next) {
					internal->row_ptr = internal->row_ptr->next;

					if(internal->row_ptr->text)
						slen = strlen(internal->row_ptr->text);

					// If the next line is wrapped and the cursor position is
					// more than the line length, set the cursor position to 
					// the first wrap point.
					if(internal->row_ptr->wrap_positions) {
						if(internal->cursor_pos >= internal->row_ptr->wrap_positions[0])
							internal->cursor_pos = internal->row_ptr->wrap_positions[0]-1;
					} else {
						if(internal->cursor_pos >= slen)
							internal->cursor_pos = slen;
					}
				}
			break;

		case XK_BackSpace:
			if(internal->cursor_pos>0) {
				internal->cursor_pos--;
			
				paragraph_delete_char(&(internal->data), &(internal->row_ptr), 
												internal->cursor_pos);
			} else {
				// Cursor reached the beginning of the line.
				// Move to the previous line, if it exists.
				if(internal->row_ptr->prev) {
					internal->row_ptr = internal->row_ptr->prev;
			
					if(internal->row_ptr->text)
						slen = strlen(internal->row_ptr->text);

					if(internal->row_ptr->text)
						internal->cursor_pos = slen;
					else
						internal->cursor_pos = 0;
				
					paragraph_delete_char(&(internal->data), &(internal->row_ptr), 
												internal->cursor_pos);
				}
			}

			paragraph_list_wrapnode(&(internal->row_ptr), internal->wraptype, 
											internal->font, widget->width-16);
			break;

		case XK_Delete:
			paragraph_delete_char(&(internal->data), &(internal->row_ptr), 
													internal->cursor_pos);

			paragraph_list_wrapnode(&(internal->row_ptr), internal->wraptype, 
											internal->font, widget->width-16);
			break;
		
		case XK_Return:
			paragraph_insert_char(&(internal->row_ptr), internal->cursor_pos, '\n');

			paragraph_list_wrapnode(&(internal->row_ptr), internal->wraptype, 
											internal->font, widget->width-16);

			internal->cursor_pos = 0;
			internal->row_ptr = internal->row_ptr->next;

			if(internal->row_ptr) {
				paragraph_list_wrapnode(&(internal->row_ptr), internal->wraptype, 
											internal->font, widget->width-16);
			}

			break;

		default :
			if(keysym>=XK_space && keysym<=XK_asciitilde) {
				paragraph_insert_char(&(internal->row_ptr), internal->cursor_pos, buffer[0]);
				internal->cursor_pos++;
	
				paragraph_list_wrapnode(&(internal->row_ptr), internal->wraptype, 
												internal->font, widget->width-16);
			}
	}

	widget->redraw(widget);
	// Update scrollbar length
	internal->total_width = text_display_width(widget);
	internal->total_height = text_display_height(widget);
	if(!internal->total_height)
		internal->total_height = widget->height-16;

	hlen = (float)(widget->width-16)/internal->total_width;
	vlen = (float)(widget->height-16)/internal->total_height;
	internal->hscrollbar->specific.hscrollbar.set_slider_length(internal->hscrollbar, hlen); 
	internal->vscrollbar->specific.vscrollbar.set_slider_length(internal->vscrollbar, vlen);
	
	// If the cursor has gone out of view, bring it into view
	if(cursor_y<0) {
		float pos;
		
		internal->top_y += -cursor_y;
		widget->redraw(widget);

		internal->vscrollbar->specific.vscrollbar.intimate_change = NULL;
		pos = (float)abs(internal->top_y)/(float)abs(internal->total_height - widget->height + 16);
		internal->vscrollbar->specific.vscrollbar.set_slider_position(internal->vscrollbar, pos);
		internal->vscrollbar->specific.vscrollbar.intimate_change = text_vviewport_changed;
	}
	if(cursor_y+line_incr > widget->height-16) {
		float pos;
		
		internal->top_y += (widget->height-16-cursor_y-line_incr-internal->interline_space);
		widget->redraw(widget);

		internal->vscrollbar->specific.vscrollbar.intimate_change = NULL;
		pos = (float)abs(internal->top_y)/(float)abs(internal->total_height - widget->height + 16);
		internal->vscrollbar->specific.vscrollbar.set_slider_position(internal->vscrollbar, pos);
		internal->vscrollbar->specific.vscrollbar.intimate_change = text_vviewport_changed;
	}
	if(cursor_x > widget->width-16) {
		float pos;

		internal->top_x += (widget->width-16-cursor_x)-2;
		widget->redraw(widget);

		internal->hscrollbar->specific.hscrollbar.intimate_change = NULL;
		pos = (float)abs(internal->top_x)/(float)abs(internal->total_width - widget->width + 16);
		internal->hscrollbar->specific.hscrollbar.set_slider_position(internal->hscrollbar, pos);
		internal->hscrollbar->specific.hscrollbar.intimate_change = text_hviewport_changed;
	}
	if(cursor_x <= 0) {
		float pos;

		internal->top_x += -cursor_x;
		widget->redraw(widget);

		internal->hscrollbar->specific.hscrollbar.intimate_change = NULL;
		pos = (float)abs(internal->top_x)/(float)abs(internal->total_width - widget->width + 16);
		internal->hscrollbar->specific.hscrollbar.set_slider_position(internal->hscrollbar, pos);
		internal->hscrollbar->specific.hscrollbar.intimate_change = text_hviewport_changed;
	}
	
	internal->hscrollbar->redraw(internal->hscrollbar);
	internal->hscrollbar->show(internal->hscrollbar);
	internal->vscrollbar->redraw(internal->vscrollbar);
	internal->vscrollbar->show(internal->vscrollbar);

	// Bypass extra key events
	while(XPending(display))
		XNextEvent(display, &xev);

	return True;
}

Bool text_default_key_released(px_widget* widget, px_event* event)
{
	KeySym keysym;
	char buffer[10];

	XLookupString(&event->xev->xkey, buffer, 10, &keysym, NULL);

	event->ch = buffer[0];
	
	return True;
}

void text_destroy(px_widget* widget)
{
	struct _text *internal = widget->specific.text.internal;
	Display* display = display_get_Xdisplay();
	px_eventloop* eloop = eventloop_get_eventloop();

	if(eloop->focussed_widget == widget)
     	eloop->focus(eloop, NULL);
		
	if(cursor_blink) {
		eloop->delete_timer(eloop, cursor_blink);
		cursor_blink = NULL;
	}
	
	XFreeGC(display, internal->gc);
}

void text_set_font(px_widget* widget, const char *fontname)
{
	struct _text *internal = widget->specific.text.internal;
	Display* display = display_get_Xdisplay();
	
	font_free_font(internal->fontname);	
	free(internal->fontname);
	internal->font = font_load_font(fontname);
	internal->fontname = strdup(fontname);

	XSetFont(display, internal->gc, internal->font->font->fid);

	text_reset_view(widget);
}
	
void text_set_wrap_type(px_widget *widget, unsigned wrap_type)
{
	struct _text *internal = widget->specific.text.internal;
	
	internal->wraptype = wrap_type;
	
	paragraph_list_wrap((struct _paragraph **)&(internal->data), internal->wraptype, 
											internal->font, widget->width-16-4);

	text_reset_view(widget);
}
	
int text_load_file(px_widget* widget, const char *filename)
{
	struct _text *internal = widget->specific.text.internal;

	if(internal->data)
		paragraph_list_free((struct _paragraph **)&(internal->data));

	internal->data = (struct _paragraph *)load_file(filename);

	if(!internal->data)
		internal->data = (struct _paragraph *)paragraph_create_newnode();

	internal->row_ptr = internal->data;

	text_reset_view(widget);

	return 0;
}
	
int text_save_file(px_widget *widget, const char *filename)
{
	FILE *fp;
	struct _text *internal = widget->specific.text.internal;
	struct _paragraph *p= (struct _paragraph *)internal->data;

	if(p) {
		fp = fopen(filename, "w");
		if(fp) {
			while(p) {
				if(p->text)
					fprintf(fp, "%s\n", p->text);
				else
					fprintf(fp, "\n");
				p = p->next;
			}

			fclose(fp);
		}

		return 0;
	} else {
		return -1;
	}
}

int text_get_numlines(px_widget *widget)
{
	struct _text *internal = widget->specific.text.internal;
	struct _paragraph *p=(struct _paragraph *)(internal->data);
	unsigned len=0;
	
	while(p) {
		len ++;
		p = p->next;
	}

	return len;
}

const char *text_get_line(px_widget *widget, unsigned linenum)
{
	struct _text *internal = widget->specific.text.internal;
	struct _paragraph *p=(struct _paragraph *)(internal->data);
	unsigned len=0;
	
	while(p) {
		if(len==linenum)
			return (p->text) ? (const char*)p->text : "\n";
		len ++;
		p = p->next;
	}

	return NULL;
}

char *text_get_text(px_widget *widget)
{
	char *buff;
	struct _text *internal = widget->specific.text.internal;
	struct _paragraph *p;
	unsigned len=0;
	
	p=(struct _paragraph *)(internal->data);
	while(p) {
		len += (p->text) ? strlen(p->text)+2 : 3;
		p = p->next;
	}

	buff = (char*)malloc(len);
	if(buff) {
		int pos=0;

		p=(struct _paragraph *)(internal->data);
		while(p) {
			strcpy(buff+pos, (p->text) ? p->text : "\n");
			if(p->text)
				strcat(buff, "\n");
			pos += (p->text) ? strlen(p->text)+1 : strlen("\n");
			p = p->next;
		}
		buff[pos-1] = '\0';

		return buff;
	} else {
		return NULL;
	}
}

int text_set_text(px_widget *widget, const char *text)
{
	struct _text *internal = widget->specific.text.internal;

	if(internal->data)
		paragraph_list_free((struct _paragraph **)&(internal->data));

	internal->data = (struct _paragraph *)load_text(text);

	if(!internal->data)
		internal->data = (struct _paragraph *)paragraph_create_newnode();

	internal->row_ptr = internal->data;

	text_reset_view(widget);

	return 0;
}

void text_set_foreground_color(px_widget *widget, const char *color)
{
	struct _text *internal = widget->specific.text.internal;

	widget_xcolor_from_str(internal->foreground, color);
}
	
void text_set_focus(px_widget* widget, Bool f)
{
	px_eventloop* eloop = eventloop_get_eventloop();
 
	if(cursor_blink) {
		eloop->delete_timer(eloop, cursor_blink);
		cursor_blink = NULL;
	}
 
	widget->have_focus = f;
	
     if(widget->have_focus)
		cursor_blink = eloop->add_timer(eloop, 0, BLINK_TIME, cursor_blink_timer, widget, NULL, True);

	widget->redraw(widget);
}

static void cursor_blink_timer(void *var1, void *var2)
{
	px_widget *widget = (px_widget*)var1; 
	struct _text *internal = widget->specific.text.internal;
	Display* display = display_get_Xdisplay();
	widget_states state;
	int font_height;

	state = (*widget->get_state)(widget);

	if(state==DISABLED)
		return;

	font_height = font_get_height(internal->font, "");

	if(blink_fg) {
		XSetForeground(display, internal->gc, widget->background.background_color->first_color.pixel);
		blink_fg = 0;
	} else {
		XSetForeground(display, internal->gc, internal->foreground->pixel);
		blink_fg = 1;
	}

	XDrawLine(display, widget->backing_store.pxmp, internal->gc, 
					cursor_x+1, cursor_y, cursor_x+1, cursor_y+font_height);
	XClearWindow(display, widget->win);
}

static void text_hviewport_changed(px_widget* widget, float slider_pos)
{
	struct _text *internal = widget->specific.text.internal;

	internal->top_x = slider_pos * (internal->total_width - widget->width + 16) * -1;
	widget->redraw(widget);
}

static void text_vviewport_changed(px_widget* widget, float slider_pos)
{
	struct _text *internal = widget->specific.text.internal;

	internal->top_y = slider_pos * (internal->total_height - widget->height + 16) * -1;
	widget->redraw(widget);
}

static unsigned text_display_width(px_widget *widget)
{
	struct _text *internal = widget->specific.text.internal;
	unsigned max_width=1;
	struct _paragraph *p=(struct _paragraph *)(internal->data);
	
	while(p) {
		if(p->text) {
			int i, off=0, width;

			if(p->num_lines!=1) {
				for(i=0; i<p->num_lines-1; i++) {
					width = font_get_width(internal->font, p->text+off, p->wrap_positions[i]-off);
					off = p->wrap_positions[i];
					max_width = (max_width>width) ? max_width : width;
				}
			} else {
				width = font_get_width(internal->font, p->text, strlen(p->text));
				max_width = (max_width>width) ? max_width : width;
			}
		}
		p = p->next;
	}

	return max_width;
}

static unsigned text_display_height(px_widget *widget)
{
	struct _text *internal = widget->specific.text.internal;
	unsigned numlines;

	numlines = paragraph_list_numlines((struct _paragraph **)&(internal->data));

	return numlines*font_get_height(internal->font, "") + numlines*internal->interline_space;
}

static void text_reset_view(px_widget *widget)
{
	struct _text *internal = widget->specific.text.internal;
	float hlen, vlen;
	
	internal->top_x = internal->top_y = 0;
	internal->total_width = text_display_width(widget);
	internal->total_height = text_display_height(widget);
	if(!internal->total_height)
		internal->total_height = widget->height-16;

	hlen = (float)(widget->width-16)/internal->total_width;
	vlen = (float)(widget->height-16)/internal->total_height;
	internal->hscrollbar->specific.hscrollbar.set_slider_position(internal->hscrollbar, 0.0);
	internal->vscrollbar->specific.vscrollbar.set_slider_position(internal->vscrollbar, 0.0);
	internal->hscrollbar->specific.hscrollbar.set_slider_length(internal->hscrollbar, hlen);
	internal->vscrollbar->specific.vscrollbar.set_slider_length(internal->vscrollbar, vlen);

	internal->hscrollbar->redraw(internal->hscrollbar);
	internal->hscrollbar->show(internal->hscrollbar);
	internal->vscrollbar->redraw(internal->vscrollbar);
	internal->vscrollbar->show(internal->vscrollbar);
}

/////////////// Text Widget data structure related
static struct _paragraph *paragraph_create_newnode()
{
	struct _paragraph *node;
			
	node = (struct _paragraph*)malloc(sizeof(struct _paragraph));
	node->text = NULL;
	node->allocated = 0;
	node->wrap_positions = NULL;
	node->num_lines = 1;
	node->next = node->prev = NULL;

	return node;
}

static void paragraph_list_insertend(struct _paragraph **list, struct _paragraph *node)
{
	struct _paragraph *p;
	
	p = *list;

	if(p==NULL) {
		*list = node;
	} else {
		while(p->next)
			p = p->next;
		p->next = node;
		node->prev = p;
	}
}

static void paragraph_node_free(struct _paragraph **node)
{
	struct _paragraph *p = *node;

	if(p->text)
		free(p->text);
	if(p->wrap_positions)
		free(p->wrap_positions);
	free(p);
}

static void paragraph_list_free(struct _paragraph **list)
{
	struct _paragraph *p;

	while(*list) {
		p = *list;
		(*list)= (*list)->next;
		if(p->text)
			free(p->text);
		if(p->wrap_positions)
			free(p->wrap_positions);
		free(p);
	}
}

static unsigned paragraph_list_numlines(struct _paragraph **list)
{
	unsigned len=0;
	struct _paragraph *p;
	
	p = *list;

	while(p) {
		len += p->num_lines;
		p = p->next;
	}

	return len;
}

static void paragraph_list_wrapnode(struct _paragraph **list, unsigned wraptype, 
												px_font *font, unsigned width)
{
	struct _paragraph *p=*list;

	if(!p->text) {
		if(p->wrap_positions) {
			free(p->wrap_positions);
			p->wrap_positions = NULL;
		}
		p->num_lines = 1;
		return;
	}

	if(wraptype==WRAP_CHAR || wraptype==WRAP_WORD) {
		unsigned i, t=0, slen, off=0, alloc=5;

		if(p->wrap_positions) {
			free(p->wrap_positions);
			p->num_lines = 1;
		}
		p->wrap_positions = (unsigned*)malloc(alloc*sizeof(unsigned));
		slen = strlen(p->text);
		
		for(i=0; i<slen; i++) {
			// Using the offset find the new starting point
			// to wrap the line.
			char *text = p->text + off;
			unsigned len = i - off + 1;

			if(font_get_width(font, text, len)>=width) {
				if(wraptype==WRAP_CHAR) {
					// We are at a point, where the text width is 
					// more than the window width. Wrap at this
					// position and set the offset to this.
					off = p->wrap_positions[t++] = i--;
				} else {
					int tmp = i;
			
					// We are at a point, where the text width is 
					// more than the window width. Start looking
					// backwards for a space to wrap. Set the
					// offset to the position, where wrap is done.
					// For words longer than window width, do a 
					// character wrap.
					while(i-off>0 && text[i-off]!=' ')
						i--;
				
					if(i-off==0) {
						off = p->wrap_positions[t++] = i = tmp;
					} else {
						off = p->wrap_positions[t++] = i+1;
					}
				}

				// If the number of wrap positions allocated is
				// less, reallocate more.
				if(t==alloc) {
					unsigned *tmp;
					
					alloc += 5;
					tmp = (unsigned*)realloc(p->wrap_positions, alloc*sizeof(unsigned));
					if(!tmp) {
						t--;
						break;
					}
					p->wrap_positions = tmp;
				}
			}
		}

		p->num_lines = t+1;
		if(p->num_lines==1) {
			free(p->wrap_positions);
			p->wrap_positions = NULL;
		}
	}
	
	if(wraptype==NO_WRAP) {
		if(p->wrap_positions)
			free(p->wrap_positions);
		p->wrap_positions = NULL;
		p->num_lines = 1;
	}
}

static void paragraph_list_wrap(struct _paragraph **list, unsigned wraptype, 
											px_font *font, unsigned width)
{
	struct _paragraph *p=*list;

	// Wrap the paragraphs one-by-one.
	while(p) {
		paragraph_list_wrapnode(&p, wraptype, font, width);
		p = p->next;
	}
}

static void paragraph_insert_char(struct _paragraph **paragraph, int pos, char ch)
{
	struct _paragraph *p=*paragraph;
	int slen=0, i;

	if(ch=='\n') {	// For \n create a new paragraph node
		struct _paragraph *node;

		node = paragraph_create_newnode();
		node->prev = p;
		node->next = p->next;
		if(p->next)
			p->next->prev = node;
		p->next = node;

		if(p->text) {
			slen = strlen(p->text);

		 	if(pos!=slen) {	// If \n is inserted in between the paragraph,
							// copy the contents after the middle-half in the 
							// current paragraph to the newly added paragraph.
				int nlen = (slen-pos)*2.5;

				node->text = (unsigned char*)malloc(nlen);
				if(node->text) {
					unsigned char *tmp;

					node->allocated = nlen;
					strcpy(node->text, p->text+pos);

					if(pos==0) {
						free(p->text);
						p->text = NULL;
						p->allocated = 0;
					} else {
						p->text[pos] = 0;
						slen = strlen(p->text)*2.5;
						tmp = (unsigned char*)realloc(p->text, slen);
						if(tmp) {
							p->text = tmp;
							p->allocated = slen;
						}
					}
				}
			}
		}
	} else {	// Add alphets to the paragraph
		if(p->text) {
			slen = strlen(p->text);

			if(slen+1 == p->allocated-2) {
				unsigned char *tmp;

				tmp = (unsigned char*)realloc(p->text, p->allocated+128); 
				if(tmp) {
					p->text = tmp;
					p->allocated += 128;
				} else
					return;
			}

			if(pos>slen)
				pos = slen;
			for(i=slen+1; i>pos; i--)
				p->text[i] = p->text[i-1];

			p->text[i] = ch; 
		} else {
			p->text = (unsigned char*)malloc(128);
			if(p->text) {
				p->allocated = 128;
				p->text[0] = ch; 
				p->text[1] = '\0';
			}
		}
	}
}

static void paragraph_delete_char(struct _paragraph **list, struct _paragraph **node, int pos)
{
	struct _paragraph *p=*node;
	int slen=0, i;

	if(p->text) {	// Handle paragraphs having text.
		slen = strlen(p->text);

		if(pos>=slen) {	// Handle deletion of characters in between the paragraph
			if(p->next) {
				struct _paragraph *tp;

				if(p->next->text) {
					int nlen;
					char *tmp;

					nlen = (slen + strlen(p->next->text))*2.5;
					tmp = (char*)realloc(p->text, nlen);
					if(tmp) {
						p->text = tmp;
						strcpy(p->text+slen, p->next->text);
					}
				}
			
				if(p->next->next)
					p->next->next->prev = p;
				tp = p->next;
				p->next = p->next->next;	
				paragraph_node_free(&tp);
			}
		} else {		// Handle deletion of the last character in the paragraph
			for(i=pos; i<=slen; i++)
				p->text[i] = p->text[i+1];

			if(strlen(p->text) == 0) {
				free(p->text);
				p->text = NULL;
				p->allocated = 0;
				if(p->wrap_positions) {
					free(p->wrap_positions);
					p->wrap_positions = NULL;
				}
				p->num_lines = 1;
			}
		}
	} else {	// Handle empty paragraphs
		if(p->prev && p->next) {
			p->prev->next = p->next;
			p->next->prev = p->prev;
			*node = p->prev->next;
			paragraph_node_free(&p);
		} else if(!p->prev && p->next) {
			*list = p->next;
			*node = p->next;
			p->next->prev = NULL;
			paragraph_node_free(&p);
		}
	}
}

static struct _paragraph *load_file(const char *filename)
{
	struct _paragraph *list=NULL, *node;
	FILE *fp;

	fp = fopen(filename, "r");
	if(fp) {
		while(!feof(fp)) {
			unsigned slen, i=0, alloc=64;
			char *line, *tmp;

			// Read a line from the file.
			// Treat this as a pragraph.
			line = (char*)malloc(alloc);
			while(!feof(fp)) {
				line[i] = getc(fp);
				if(feof(fp) || line[i]=='\n')
					break;
				i++;
				
				if(i==alloc-2) {
					alloc += 64;
					tmp = (char*)realloc(line, alloc); 
					if(tmp)
						line = tmp;
					else
						break;
				}
			}
			line[i] = '\0';

			// Create a paragraph node.
			node = paragraph_create_newnode();
			slen = strlen(line);
			node->allocated = slen*2.5;
			if(node->allocated) {
				node->text = (char*)malloc(node->allocated);
				memcpy(node->text, line, slen+1);
			}

			// Add to the list of paragraphs.
			paragraph_list_insertend(&list, node);
			
			free(line);
		}
		
		fclose(fp);
	}

	return list;
}

static struct _paragraph *load_text(const char *text)
{
	int off=0, slen;
	struct _paragraph *list=NULL, *node;
	char *ptr=text;

	if(text==NULL)
		return NULL;

	slen = strlen(text);

	while(off < slen) {
		int len;
		char *line;

		len = strcspn(ptr+off, "\n");
		line = (char*)malloc(len+2);
		strncpy(line, ptr+off, len);
		line[len] = '\0';
		off += len+1;

		// Create a paragraph node.
		node = paragraph_create_newnode();
		node->allocated = len*2.5;
		if(node->allocated) {
			node->text = (char*)malloc(node->allocated);
			memcpy(node->text, line, len+1);
		}

		// Add to the list of paragraphs.
		paragraph_list_insertend(&list, node);

		free(line);
	}

	return list;
}
