/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include <X11/extensions/shape.h>
#include <stdlib.h>
#include <string.h>
#include "toplevel.h"
#include "image.h"
#include "eventloop.h"
#include "widget_factory.h"
#include "display.h"

px_widget* toplevel_create(int x, int y, int width, int height, const char* tag)
{
	px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
	px_eventloop* eventloop = NULL;
	px_display* pxdisplay = display_get_display();
	
	int screen_num = DefaultScreen(pxdisplay->display);
	// Initialize common widget variables
	widget->win = XCreateSimpleWindow(pxdisplay->display,
				RootWindow(pxdisplay->display, screen_num),
				x, y,
				width, height,
				0, BlackPixel(pxdisplay->display, screen_num),
				WhitePixel(pxdisplay->display, screen_num));

	XSelectInput(pxdisplay->display, widget->win, 
		   ExposureMask |
		   ButtonPressMask |
		   ButtonReleaseMask |
		   Button1MotionMask |
		   KeyPressMask |
		   KeyReleaseMask|
		   FocusChangeMask |
		   StructureNotifyMask |
		   VisibilityChangeMask |
   		   SubstructureRedirectMask);

	widget->x = x;
	widget->y = y;
	widget->height = height;
	widget->width = width;
	widget->wtype = TOPLEVEL;
	widget->child_of = NULL;
	widget->use_pixmap = False;
	if (tag != NULL)
	{
		widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	}
	else
		widget->tag = NULL;

	widget->background.background_color = (px_color*)malloc(sizeof(px_color));
	widget->background.background_color->type = SOLID;
	widget_xcolor_from_str(&(widget->background.background_color->first_color), "white");

	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);
	
#ifdef __debug__
	printf("%s(): %p - Created Window %lu with x = %d, y = %d, width = %d, height = %d, new window = %lu\n\n", 
			__func__, widget, widget->win, x, y, width, height, widget->win);
#endif
	eventloop = eventloop_get_eventloop();
	eventloop->register_toplevel(eventloop, widget);
	XSetWMProtocols(pxdisplay->display, widget->win, &(pxdisplay->wm_protocols[0]), 1);
	return widget;
}

void toplevel_set_parameters(px_widget* widget)
{
	int i;
	widget->default_callbacks[BUTTON_PRESSED] = toplevel_default_button_press;
	widget->default_callbacks[MOTION] = toplevel_default_motion;
	widget->default_callbacks[BUTTON_RELEASED] = toplevel_default_button_release;
	widget->default_callbacks[MOUSE_ENTER] = toplevel_default_button_press;
	widget->default_callbacks[MOUSE_LEAVE] = toplevel_default_mouse_leave;
	widget->default_callbacks[EXPOSE] = toplevel_default_expose;
	widget->default_callbacks[CLIENT_MESSAGE] = toplevel_default_client_message;
	widget->default_callbacks[CONFIGURE] = toplevel_default_configure_event;
	
	widget->handle_event = toplevel_handle_event;
	widget->redraw = toplevel_redraw;
	widget->show = toplevel_show;
	widget->destroy = toplevel_destroy;
	widget->cleanup_internal_structures = toplevel_cleanup_data_structures;

	// Initialize widget specific variables 

	widget->specific.toplevel.widget_list = (px_widget**)malloc(sizeof(px_widget*) * 5);

	//added by Jyotirmoy
	for (i = 0; i < 5; i++)
		widget->specific.toplevel.widget_list[i] = NULL;
	
	
	widget->specific.toplevel.allocated = 5;
	widget->specific.toplevel.no_of_children = 0;
	widget->specific.toplevel.event_within = NULL;
	widget->specific.toplevel.title = NULL;
	widget->specific.toplevel.ynodes = NULL;
	widget->specific.toplevel.axis = YAXIS;
	widget->specific.toplevel.count_within = -1;
	
	widget->specific.toplevel.add_child = toplevel_add_child;
	widget->specific.toplevel.remove_child = toplevel_remove_child;
	widget->specific.toplevel.set_title = toplevel_set_title;
}

void toplevel_add_child(px_widget* widget, px_widget* child)
{
	px_toplevel* toplevel = &(widget->specific.toplevel);

	if (toplevel->allocated == toplevel->no_of_children)
	{
		unsigned int count;
				
		px_widget** new_list = (px_widget**)malloc(sizeof(px_widget*) * 
				(toplevel->allocated + 10));
			
		for (count = 0; count < toplevel->no_of_children; count++)
			new_list[count] = toplevel->widget_list[count];

		free(toplevel->widget_list);
		toplevel->widget_list = new_list;
		toplevel->allocated += 10;
	}
	toplevel->widget_list[toplevel->no_of_children] = child;
	toplevel->no_of_children++;

	if (!widget->preserve_internal_structures)
	{
		toplevel->ynodes = widget_add_child(toplevel->ynodes, child, 
			toplevel->axis == YAXIS ? YDIRECTION : XDIRECTION, toplevel->axis);
#ifdef __debug__
		printf("-------- For toplevel %p --------------\n", widget);
		widget_print_frame_buffer(toplevel->ynodes, toplevel->axis == YAXIS ? YDIRECTION : XDIRECTION, toplevel->axis);
		printf("-------- ENd For toplevel %p --------------\n", widget);
#endif
	}
	else
	{
		widget->dirty = True;
	}

}

void toplevel_remove_child(px_widget* widget, px_widget* child)
{
	px_toplevel* toplevel = &(widget->specific.toplevel);
	unsigned int count;

	for (count = 0; count < toplevel->no_of_children && toplevel->widget_list[count] != child ; count++);

	if (count < toplevel->no_of_children)
	{
		if (!widget->preserve_internal_structures)
		{
			toplevel->ynodes = widget_remove_child(toplevel->ynodes, child, 
				toplevel->axis == YAXIS ? YDIRECTION : XDIRECTION, toplevel->axis);
#ifdef __debug__
		printf("-------- For toplevel %p --------------\n", widget);
			widget_print_frame_buffer(toplevel->ynodes, toplevel->axis == YAXIS ? YDIRECTION : XDIRECTION, toplevel->axis);
		printf("-------- End For toplevel %p --------------\n", widget);
#endif
		}
		else
		{
			widget->dirty = True;
		}
				
		for (count++;count < toplevel->no_of_children;count++)
			toplevel->widget_list[count-1] = toplevel->widget_list[count];
		toplevel->no_of_children--;
	}
}

Bool toplevel_handle_event(px_widget* widget, px_event* event)
{
	unsigned int count = 0;
	Bool return_value = True;
	
#ifdef __debug__
	printf("%s() : %p Event type = %d at %d %d, toplevel->win %lu, Event window = %lu\n", __func__, widget, event->type, event->xev->xbutton.x, event->xev->xbutton.y, widget->win, event->xev->xany.window);
#endif
	if (event->xev->xany.window != widget->win)
		return True;

	if (event->type != NR_CALLBACKS && (event->type == EXPOSE || widget->get_state(widget) != DISABLED))
	{

		return_value = widget_handle_event(widget, event);

		if (event->type > 6 && event->type != CLIENT_MESSAGE && return_value)
		{
			for (count = 0; count < widget->specific.toplevel.no_of_children; count++)
			{
				px_widget* child = widget->specific.toplevel.widget_list[count];
#ifdef __debug__
				printf("%s() : %p Found widget %p at %d : hidden %d\n", __func__,widget, child, 
						count, child->hidden);
#endif
				if (!child->hidden)
					return_value = child->handle_event(child, event);
			}
		}
	}
	return return_value;
}

Bool toplevel_default_button_press(px_widget* widget, px_event* event)
{
	px_toplevel* toplevel = &( widget->specific.toplevel);
	px_xynode* node = widget_get_xynode(toplevel->ynodes, event->x, event->y, 
			toplevel->axis == YAXIS ? YDIRECTION : XDIRECTION,
			toplevel->axis);
	
	Bool return_value = True;
	widget->preserve_internal_structures = True;
	
	if (node != NULL)
	{
		int count;
		int event_given_to = -1;
		
		for (count = 0; count < node->no_of_widgets;count++)
		{
			if (!node->widget_list[count]->hidden && 
					node->widget_list[count]->get_state(node->widget_list[count]) != DISABLED)
			{
				int wx, wy;
				Bool transparency;
				
				wx = event->x - node->widget_list[count]->x;
				wy = event->y - node->widget_list[count]->y;
	
				transparency = node->widget_list[count]->is_transparent(node->widget_list[count], wx, wy);

				if (node->widget_list[count]->emask != EVENT_TRANSPARENT && 
						!(node->widget_list[count]->emask == EVENT_MASK && transparency))
				{
					event->x = wx;
					event->y = wy;

					if (return_value)
					{
						return_value = node->widget_list[count]->handle_event
							(node->widget_list[count], event);
					}
					event_given_to = count;
					event->x += node->widget_list[count]->x;
					event->y += node->widget_list[count]->y;
				}

				if (!transparency && !node->widget_list[count]->hidden)
				{
					break;
				}
			}
		}

		toplevel->count_within = event_given_to;
		toplevel->event_within = node;
		toplevel->last_x = event->x;
		toplevel->last_y = event->y;
	}

	widget->preserve_internal_structures = False;
	return return_value;
}

Bool toplevel_default_motion(px_widget* widget, px_event* event)
{
	px_toplevel* toplevel = &( widget->specific.toplevel);

	px_xynode* node = widget_get_xynode(toplevel->ynodes, event->x, event->y, 
			toplevel->axis == YAXIS ? YDIRECTION : XDIRECTION,
			toplevel->axis);
	px_xynode* event_within = toplevel->event_within;
	int count_within = toplevel->count_within;
	Bool return_value = True;
	
	widget->preserve_internal_structures = True;
	if (!(node == NULL && event_within == NULL))
	{
		if (node == event_within)
		{
			// Give the motion event to all children
			int count;
			int event_given_to = -1;

			for (count = 0; count < node->no_of_widgets;count++)
			{
				if (!node->widget_list[count]->hidden && 
					node->widget_list[count]->get_state(node->widget_list[count]) != DISABLED)
				{
					int wx, wy;
					Bool transparency;
					wx = event->x - node->widget_list[count]->x;
					wy = event->y - node->widget_list[count]->y;
	
					transparency = node->widget_list[count]->is_transparent(node->widget_list[count], wx, wy);

					if (node->widget_list[count]->emask != EVENT_TRANSPARENT && 
							!(node->widget_list[count]->emask == EVENT_MASK && transparency))
					{
						event->x = wx;
						event->y = wy;

						if (return_value && count <= count_within)
						{
							return_value = node->widget_list[count]->handle_event
								(node->widget_list[count], event);
						}
						
						event_given_to = count;
						event->x += node->widget_list[count]->x;
						event->y += node->widget_list[count]->y;
					}

					if (!transparency && !node->widget_list[count]->hidden)
					{
						break;
					}
				}
			}

			if (event_given_to < count_within)
			{
				int ctr;
				Bool leave_return_value = True;	
				event->type = MOUSE_LEAVE;
				for (ctr = event_given_to + 1; leave_return_value && ctr <= count_within;ctr++)
				{
					if (!node->widget_list[ctr]->hidden && 
							node->widget_list[ctr]->get_state(node->widget_list[ctr]) != DISABLED)
					{
						int wx, wy;
						Bool transparency;
						wx = toplevel->last_x - node->widget_list[ctr]->x;
						wy = toplevel->last_y - node->widget_list[ctr]->y;
		
						transparency = node->widget_list[ctr]->is_transparent
							(node->widget_list[ctr], wx, wy);
						
						if (node->widget_list[ctr]->emask != EVENT_TRANSPARENT && 
							!(node->widget_list[ctr]->emask == EVENT_MASK && transparency))
						{
#ifdef __debug__				
							printf("%s() 1. : %p Generating leave event on %p\n", __func__, widget, 
								node->widget_list[ctr]);
#endif
							leave_return_value = node->widget_list[ctr]->handle_event
								(node->widget_list[ctr], event);
						}
					}
				}
				event->type = MOTION;
			}
			else if (event_given_to > count_within)
			{
				int ctr;
				Bool enter_return_value = True;
				
				event->type = MOUSE_ENTER;
				for (ctr = count_within + 1; enter_return_value && ctr <= event_given_to;ctr++)
				{
					if (!node->widget_list[ctr]->hidden && 
							node->widget_list[ctr]->get_state(node->widget_list[ctr]) != DISABLED)
					{
						int wx, wy;
						Bool transparency;
						wx = event->x - node->widget_list[ctr]->x;
						wy = event->y - node->widget_list[ctr]->y;
	
						transparency = node->widget_list[ctr]->is_transparent
							(node->widget_list[ctr], wx, wy);
#ifdef __debug__				
						printf("%s() 1. : %p Generating Enter event on %p\n", __func__, widget, 
								node->widget_list[ctr]);
#endif
						if (node->widget_list[ctr]->emask != EVENT_TRANSPARENT && 
							!(node->widget_list[ctr]->emask == EVENT_MASK && transparency))
						{
							event->x = wx;
							event->y = wy;
							enter_return_value = node->widget_list[ctr]->handle_event
								(node->widget_list[ctr], event);
							event->x += node->widget_list[ctr]->x;
							event->y += node->widget_list[ctr]->y;
						}
					}
				}
				event->type = MOTION;
			}

			toplevel->count_within = event_given_to;
			toplevel->last_x = event->x;
			toplevel->last_y = event->y;
		}
		else if (node == NULL)
		{
			// Give leave events for all the widgets in event_within
			int count;
			Bool leave_return_value = True;
			event->type = MOUSE_LEAVE;

			for (count = 0; leave_return_value && count <= count_within;count++)
			{
				if (!event_within->widget_list[count]->hidden &&
						event_within->widget_list[count]->
						get_state(event_within->widget_list[count]) != DISABLED)
				{
					int wx, wy;
					Bool transparency;
					wx = toplevel->last_x - event_within->widget_list[count]->x;
					wy = toplevel->last_y - event_within->widget_list[count]->y;
		
					transparency = event_within->widget_list[count]->is_transparent
							(event_within->widget_list[count], wx, wy);
#ifdef __debug__				
					printf("%s() 2. : %p Generating leave event on %p\n", __func__, widget, 
							event_within->widget_list[count]);
#endif
					
					leave_return_value = event_within->widget_list[count]->handle_event
							(event_within->widget_list[count], event);
		
				}
			}
			event->type = MOTION;
			toplevel->event_within = NULL;
			toplevel->count_within = -1;
		}
		else if (event_within == NULL)
		{
			// Give Enter events for all the widgets in node
			int count;
			int event_given_to = -1;
			Bool enter_return_value = True;
			
			event->type = MOUSE_ENTER;
			for (count = 0; enter_return_value && count < node->no_of_widgets;count++)
			{
				if (!node->widget_list[count]->hidden && 
					node->widget_list[count]->get_state(node->widget_list[count]) != DISABLED)
				{
					int wx, wy;
					Bool transparency;
				
					wx = event->x - node->widget_list[count]->x;
					wy = event->y - node->widget_list[count]->y;
	
					transparency = node->widget_list[count]->is_transparent(node->widget_list[count], wx, wy);

					if (node->widget_list[count]->emask != EVENT_TRANSPARENT && 
							!(node->widget_list[count]->emask == EVENT_MASK && transparency))
					{
#ifdef __debug__				
						printf("%s() 2. : %p Generating Enter event on %p\n", __func__, widget, 
							node->widget_list[count]);
#endif
						event->x = wx;
						event->y = wy;
						enter_return_value = node->widget_list[count]->handle_event
							(node->widget_list[count], event);
						event->x += node->widget_list[count]->x;
						event->y += node->widget_list[count]->y;
						event_given_to = count;
					}

					if (!transparency && !node->widget_list[count]->hidden)
					{
						break;
					}
				}
			}
			event->type = MOTION;
			toplevel->count_within = event_given_to;
			toplevel->event_within = node;
			toplevel->last_x = event->x;
			toplevel->last_y = event->y;
		}
		else
		{
			// Find out the widgets that are there in node and not in event_within
			// and give them enter events
			
			int count;
			int new_count_within;
			int event_given_to = -1;
			Bool local_return_value = True;

			for (count = 0; local_return_value && count < node->no_of_widgets;count++)
			{
				Bool transparency;
				int ctr;
				int wx, wy;

				wx = event->x - node->widget_list[count]->x;
				wy = event->y - node->widget_list[count]->y;
	
				transparency = node->widget_list[count]->is_transparent(node->widget_list[count], wx, wy);
			
				if (!node->widget_list[count]->hidden && 
					node->widget_list[count]->get_state(node->widget_list[count]) != DISABLED &&
					node->widget_list[count]->emask != EVENT_TRANSPARENT && 
					!(node->widget_list[count]->emask == EVENT_MASK && transparency))
				{
					for (ctr = 0; ctr <= count_within && local_return_value && 
						event_within->widget_list[ctr]->depth  >=  node->widget_list[count]->depth &&
						node->widget_list[count] != event_within->widget_list[ctr];ctr++);

					if (ctr > count_within || 
							(node->widget_list[count] != event_within->widget_list[ctr]))
					{
#ifdef __debug__
						printf("%s() 3. : %p Generating ENter event on %p\n", __func__, widget, 
								node->widget_list[count]);
#endif
						event->x = wx;
						event->y = wy;
						event->type = MOUSE_ENTER;
						local_return_value = node->widget_list[count]->handle_event
							(node->widget_list[count], event);
						event->x += node->widget_list[count]->x;
						event->y += node->widget_list[count]->y;
						event_given_to = count;
					}
					else
					{
						event->x = wx;
						event->y = wy;
						event->type = MOTION;
						local_return_value = node->widget_list[count]->handle_event
							(node->widget_list[count], event);
						event->x += node->widget_list[count]->x;
						event->y += node->widget_list[count]->y;
						event_given_to = count;
					}
				}
				if (!transparency && !node->widget_list[count]->hidden)
					break;
			}
			event->type = MOTION;
			
			new_count_within = event_given_to;

			// Find out the widgets that are there in event_within and not in node
			// and give them leave events

			local_return_value = True;
			event->type = MOUSE_LEAVE;
			for (count = 0; local_return_value && count <= count_within;count++)
			{
				Bool transparency;
				int ctr;
				int wx, wy;

				wx = toplevel->last_x - event_within->widget_list[count]->x;
				wy = toplevel->last_y - event_within->widget_list[count]->y;
	
				transparency = event_within->widget_list[count]->is_transparent
					(event_within->widget_list[count], wx, wy);

				if (!event_within->widget_list[count]->hidden &&
					event_within->widget_list[count]->
					get_state(event_within->widget_list[count]) != DISABLED
					&& event_within->widget_list[count]->emask != EVENT_TRANSPARENT && 
					!(event_within->widget_list[count]->emask == EVENT_MASK && transparency))
				{
					for (ctr = 0; ctr <= new_count_within && local_return_value &&  
						node->widget_list[ctr]->depth >= event_within->widget_list[count]->depth &&
						node->widget_list[ctr] != event_within->widget_list[count];ctr++);

					if (ctr > new_count_within || 
							node->widget_list[ctr] != event_within->widget_list[count])
					{
						
#ifdef __debug__				
						printf("%s() 3. : %p Generating leave event on %p\n", __func__, widget, 
								event_within->widget_list[count]);
#endif
					
						local_return_value = event_within->widget_list[count]->handle_event
								(event_within->widget_list[count], event);
	
					}
				}
			}
			event->type = MOTION;

			toplevel->count_within = new_count_within;
			toplevel->event_within = node;
			toplevel->last_x = event->x;
			toplevel->last_y = event->y;
		}
	}
	widget->preserve_internal_structures = False;
	return return_value;
}

Bool toplevel_default_button_release(px_widget* widget, px_event* event)
{
	px_toplevel* toplevel = &( widget->specific.toplevel);

	int event_given_to = -1;
	
	px_xynode* node = widget_get_xynode(toplevel->ynodes, event->x, event->y, 
			toplevel->axis == YAXIS ? YDIRECTION : XDIRECTION,
			toplevel->axis);

	Bool return_value = True;
	widget->preserve_internal_structures = True;
	
	if (node != NULL)
	{
		int count;

		for (count = 0; return_value && count < node->no_of_widgets;count++)
		{
			if (!node->widget_list[count]->hidden && 
				node->widget_list[count]->get_state(node->widget_list[count]) != DISABLED)
			{
				int wx, wy;
				Bool transparency;
				
				wx = event->x - node->widget_list[count]->x;
				wy = event->y - node->widget_list[count]->y;
	
				transparency = node->widget_list[count]->is_transparent(node->widget_list[count], wx, wy);

				if (node->widget_list[count]->emask != EVENT_TRANSPARENT && 
						!(node->widget_list[count]->emask == EVENT_MASK && transparency))
				{
					event->x = wx;
					event->y = wy;
					return_value = node->widget_list[count]->handle_event(node->widget_list[count], event);
					event_given_to = -1;
					event->x += node->widget_list[count]->x;
					event->y += node->widget_list[count]->y;
				}

				if (!transparency && !node->widget_list[count]->hidden)
				{
					count = node->no_of_widgets;
				}
			}
		}
	}

	if (toplevel->event_within != NULL && toplevel->event_within != node)
	{
		
		event->type = MOUSE_LEAVE;
		if (node != NULL)
		{
			// Find out the widgets that are there in event_within and not in node
			// and give them leave events

			int count;
			px_xynode* event_within = toplevel->event_within;
			Bool leave_return_value = True;
			
			for (count = 0; leave_return_value && count <= toplevel->count_within;count++)
			{
				int wx, wy;
				Bool transparency;
				int ctr;
				
				wx = toplevel->last_x - event_within->widget_list[count]->x;
				wy = toplevel->last_y - event_within->widget_list[count]->y;
	
				transparency = event_within->widget_list[count]->is_transparent
					(event_within->widget_list[count], wx, wy);


				if (!event_within->widget_list[count]->hidden &&
						event_within->widget_list[count]->
						get_state(event_within->widget_list[count]) != DISABLED &&
						event_within->widget_list[count]->emask != EVENT_TRANSPARENT && 
						!(event_within->widget_list[count]->emask == EVENT_MASK && transparency))
				{
					for (ctr = 0; ctr < event_given_to && leave_return_value && 
						node->widget_list[ctr]->depth >= event_within->widget_list[count]->depth &&
						node->widget_list[ctr] != event_within->widget_list[count];ctr++);

					if (ctr >= event_given_to || 
							node->widget_list[ctr] != event_within->widget_list[count])
					{
#ifdef __debug__				
						printf("%s() : %p Generating leave event on %p\n", __func__, widget, 
								event_within->widget_list[count]);
#endif
					
						leave_return_value = event_within->widget_list[count]->handle_event
								(toplevel->event_within->widget_list[count], event);
					}
				}
			}
		}
		else
		{
			// Have to give leave events for everything;
			int count;
			px_xynode* event_within = toplevel->event_within;
			Bool leave_return_value = True;
			
			for (count = 0; leave_return_value && count <= toplevel->count_within;count++)
			{
				if (!event_within->widget_list[count]->hidden &&
						event_within->widget_list[count]->
						get_state(event_within->widget_list[count]) != DISABLED)
				{
					int wx, wy;
					Bool transparency;
				
					wx = toplevel->last_x - event_within->widget_list[count]->x;
					wy = toplevel->last_y - event_within->widget_list[count]->y;
	
					transparency = event_within->widget_list[count]->is_transparent
						(event_within->widget_list[count], wx, wy);

					if (event_within->widget_list[count]->emask != EVENT_TRANSPARENT && 
							!(event_within->widget_list[count]->emask == EVENT_MASK && transparency))
					{
#ifdef __debug__				
						printf("%s() : %p Generating leave event on %p\n", __func__, widget, 
							event_within->widget_list[count]);
#endif
					
						leave_return_value = event_within->widget_list[count]->handle_event
							(event_within->widget_list[count], event);
					}
				}
			}
		}
		event->type = BUTTON_RELEASED;
	}
		
		
	toplevel->event_within = NULL;
	toplevel->count_within = -1;
	widget->preserve_internal_structures = False;

	return return_value;
}
	
	
Bool toplevel_default_mouse_leave(px_widget* widget, px_event* event)
{
	px_toplevel* toplevel = &( widget->specific.toplevel);
	Bool return_value = True;
	widget->preserve_internal_structures = True;
	if (toplevel->event_within != NULL)
	{
		// Have to give leave events for everything;
		int count;
		px_xynode* event_within = toplevel->event_within;

		for (count = 0; return_value && count <= toplevel->count_within;count++)
		{
			if (!event_within->widget_list[count]->hidden &&
					event_within->widget_list[count]->
					get_state(event_within->widget_list[count]) != DISABLED)
			{
				int wx, wy;
				Bool transparency;
				
				wx = toplevel->last_x - event_within->widget_list[count]->x;
				wy = toplevel->last_y - event_within->widget_list[count]->y;
	
				transparency = event_within->widget_list[count]->is_transparent
					(event_within->widget_list[count], wx, wy);

				if (event_within->widget_list[count]->emask != EVENT_TRANSPARENT && 
						!(event_within->widget_list[count]->emask == EVENT_MASK && transparency))
				{
#ifdef __debug__				
					printf("%s() : %p Generating leave event on %p\n", __func__, widget, 
						event_within->widget_list[count]);
#endif
					
					return_value = event_within->widget_list[count]->handle_event
						(event_within->widget_list[count], event);
				}
			}
		}
		toplevel->event_within = NULL;
		toplevel->count_within = -1;
	}
	widget->preserve_internal_structures = False;

	return return_value;
}

Bool toplevel_default_configure_event(px_widget* widget, px_event* event)
{
	XEvent xev;
	while(XPending(display_get_Xdisplay()))
		XNextEvent(display_get_Xdisplay(), &xev);

	widget->preserve_internal_structures = True;
	if (widget->width != event->xev->xconfigure.width || widget->height != event->xev->xconfigure.height)
	{
		widget->set_size(widget, event->xev->xconfigure.width, event->xev->xconfigure.height);
		widget->redraw(widget);

		if (!widget->hidden)
			widget->show(widget);
	}
	widget->preserve_internal_structures = False;

	return True;
}

Bool toplevel_default_client_message(px_widget* widget, px_event* event)
{
	px_display* pxdisplay = display_get_display();
	Atom del_window = pxdisplay->get_atom(pxdisplay, "WM_DELETE_WINDOW");
	Atom protocol_atom = pxdisplay->get_atom(pxdisplay, "WM_PROTOCOLS");
	int i;
	struct _extra_event {
		int event_type;
		char msg[40];
	} extra_event[NR_EXTRA_CALLBACKS] = {
									{RETRACT_BAR_REL, "_AC_BUTTON_RETRACTBAR_REL"},
									{UP_REL, "_AC_BUTTON_UP_REL"},
									{TICK_REL, "_AC_BUTTON_TICK_REL"},
									{NEXT_PAGE, "_AC_BUTTON_NEXTPAGE"},
									{NEXT_PAGE_REL, "_AC_BUTTON_NEXTPAGE_REL"},
									{PREV_PAGE, "_AC_BUTTON_PREVPAGE"},
									{PREV_PAGE_REL, "_AC_BUTTON_PREVPAGE_REL"},
									{EMAIL_REL, "_AC_BUTTON_EMAIL_REL"},
									{PRINTER_REL, "_AC_BUTTON_PRINTER_REL"},
									{ROCKER_UP, "_AC_BUTTON_ROCKER_UP"},
									{ROCKER_UP_REL, "_AC_BUTTON_ROCKER_UP_REL"},
									{ROCKER_DOWN, "_AC_BUTTON_ROCKER_DOWN"},
									{ROCKER_DOWN_REL, "_AC_BUTTON_ROCKER_DOWN_REL"},
									{ROCKER_SELECT, "_AC_BUTTON_ROCKER_SELECT"},
									{ROCKER_SELECT_REL, "_AC_BUTTON_ROCKER_SELECT_REL"},
									{ACCEL_RIGHT, "_AC_ACCEL_RIGHT_FLICK"},
									{ACCEL_LEFT, "_AC_ACCEL_LEFT_FLICK"},
									{ACCEL_TOP, "_AC_ACCEL_TOP_FLICK"},
									{ACCEL_BOTTOM, "_AC_ACCEL_BOTTOM_FLICK"},
									{ACCEL_RIGHT_EXTREME, "_AC_ACCEL_RIGHT_EXTREME"},
									{ACCEL_LEFT_EXTREME, "_AC_ACCEL_LEFT_EXTREME"}
								};

	if((event->xev->xclient.message_type == protocol_atom)	
			&& (event->xev->xclient.data.l[0] == del_window)) 
	{
		if(widget->window_delete_callback)
			widget->window_delete_callback(widget, event);
			
		widget->destruct(widget);

		exit(0);
//		return False;
	}

	for(i=0; i<NR_EXTRA_CALLBACKS-1; i++)
		if(event->xev->xclient.data.l[0] == pxdisplay->get_atom(pxdisplay, extra_event[i].msg)) {
			if(widget->extra_callbacks[i])
				widget->extra_callbacks[i](widget, event);
		}

	return True;
}

Bool toplevel_default_expose(px_widget* widget, px_event* event)
{
	widget->show(widget);
	return True;
}

void toplevel_redraw(px_widget* widget)
{
	unsigned int count;
	Display* display = display_get_Xdisplay();

	if (widget->use_pixmap == True)
	{
		image_copy_pixmap(widget->win, &(widget->backing_store), widget->background.background_pixmap);
	}
	else
	{
		widget_draw_px_color(&(widget->backing_store), widget->background.background_color);
	}

	if (widget->use_pixmap == False && widget->background.background_color->type == TRANSPARENT)
	{
		GC  mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
		
		XSetForeground(display, mask_gc, 1);
		for (count = 0; count < widget->specific.toplevel.no_of_children; count++)
		{
			px_widget* child = widget->specific.toplevel.widget_list[count];
		
			XCopyArea(display, child->backing_store.mask,
					widget->backing_store.mask, mask_gc, 0, 0, 
					child->backing_store.w,
					child->backing_store.h,
					child->x,
					child->y);
		}
		XFreeGC(display, mask_gc);
	}
	widget_draw_border(widget);
	XClearWindow(display, widget->win);
	widget_set_event_mask(widget, widget->emask);
}


void toplevel_show(px_widget* widget)
{
	unsigned int count;
	Display* display = display_get_Xdisplay();
	XSetWindowBackgroundPixmap(display, widget->win, widget->backing_store.pxmp);

	if (widget->backing_store.mask != None)
		XShapeCombineMask (display, widget->win, ShapeBounding, 0, 0, 
			widget->backing_store.mask, ShapeSet);

	XMapWindow(display, widget->win);
	for (count = 0; count < widget->specific.toplevel.no_of_children; count++)
	{
		px_widget* child = widget->specific.toplevel.widget_list[count];

		if (!child->hidden)
			child->show(child);
	}
	widget->hidden = False;
}

void toplevel_destroy(px_widget* widget)
{
	int count;
	px_toplevel *toplevel = &(widget->specific.toplevel);
	px_eventloop* eventloop = eventloop_get_eventloop();
	
	for (count = 0; count < toplevel->no_of_children; count++)
	{
		toplevel->ynodes = widget_remove_child(toplevel->ynodes, toplevel->widget_list[count], 
			toplevel->axis == YAXIS ? YDIRECTION : XDIRECTION, toplevel->axis);
		toplevel->widget_list[count]->destroy(toplevel->widget_list[count]);
		free(toplevel->widget_list[count]);
	}
	eventloop->remove_toplevel(eventloop, widget);

	widget_destroy(widget);
}

void toplevel_set_title(px_widget* widget, const char* title)
{
	px_toplevel *toplevel = &(widget->specific.toplevel);
	Display* display = display_get_Xdisplay();
	XTextProperty window_name_property;
	
	if (toplevel->title != NULL)
	{
		free(toplevel->title);
	}

	toplevel->title = strdup(title);
	XStringListToTextProperty(&(toplevel->title), 1, &window_name_property);
	XSetWMProperties(display, widget->win, &window_name_property, NULL, NULL, 0, NULL, NULL, NULL); 
}

void toplevel_cleanup_data_structures(px_widget* widget)
{
	px_toplevel* toplevel = &(widget->specific.toplevel);
	int count;
	
	px_xynode* ynode = toplevel->ynodes, *next = NULL;

	while (ynode != NULL)
	{
		px_xynode* xnode = ynode->xnodes, *xnext = NULL;
		next = ynode->next;

		free(ynode->widget_list);
		
		while (xnode != NULL)
		{
			free(xnode->widget_list);
			xnext = xnode->next;
			free(xnode);
			xnode = xnext;
		}
		free(ynode);
		ynode = next;
	}
	
	toplevel->ynodes= NULL;
	
	for (count = 0; count < toplevel->no_of_children; count++)
	{
		if (!toplevel->widget_list[count]->condemned)
			toplevel->ynodes = widget_add_child(toplevel->ynodes, toplevel->widget_list[count], 
				toplevel->axis == YAXIS ? YDIRECTION : XDIRECTION, toplevel->axis);
	}
}

