/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

// PicoWidget : Text Listbox
// Copyrights (c) PicoPeta Simputers Pvt. Ltd.
// Programmer: P R Subramanya

#include <stdlib.h>
#include <string.h>
#include "txtlb.h"
#include "vscrollbar.h"
#include "widget_factory.h"
#include "display.h"
#include "image.h"
#include "eventloop.h"

#define LB_START 0
#define LB_END -1

#define TXTLB_REPEAT_DELAY 120000

#define TXTLB_NO_DIRECN 0
#define TXTLB_UP_DIRECN 1
#define TXTLB_DOWN_DIRECN 2

#define DFLT_TXTLB_FONT "-picopeta-simputer-medium-r-normal--9-*-*-*-p-*-fontspecific-0"

typedef struct _txtlb_elem txtlb_elem;
typedef struct _txtlb_data txtlb_data;

struct _txtlb_elem {
	char *data;
	short int len;
	struct _txtlb_elem *next;
};

struct _txtlb_data {
	px_widget *sbar;

	GC mgc;
	XColor col_tab[TXTLB_SCROLL_BG+1];
	XFontStruct *fnt;
	int ex, ey;
	txtlb_types type;
	int max_elems, sidx, selecn, ncols;
	int max_selecn, selecn_count;
	txtlb_elem **elems;
	int *n_elems;
	int *col_wids;
	int *selecn_arr;
	void (*intimate_view_change)(px_widget * wdgt, int rowidx);
};

static px_timer *rpt_timer=NULL;
static int rpt_direcn=TXTLB_NO_DIRECN;

px_widget * txtlb_create(px_widget *parent, int x, int y, int wid, int ht,
			const char *tag)
{
	px_widget* wdgt = (px_widget*)malloc(sizeof(px_widget));
	Display* disp = display_get_Xdisplay();
	int new_ht = ((ht-8)/20)*20 + 8;

	wdgt->win = XCreateWindow(disp, parent->win, x, y, wid, new_ht, 0,
			CopyFromParent,CopyFromParent,CopyFromParent,0,NULL);
	
	wdgt->x = x; wdgt->y = y;
	wdgt->width = wid; wdgt->height = new_ht;
	wdgt->child_of = NULL;
	wdgt->wtype = TXTLB;
	wdgt->tag = NULL;

	wdgt->use_pixmap = False;
	wdgt->background.background_color = (px_color*)malloc(sizeof(px_color));
	wdgt->background.background_color->type = SOLID;
	widget_xcolor_from_str(&(wdgt->background.background_color->first_color), "#d3d3dd");

	image_initialize_pixmap(&(wdgt->backing_store));
	image_create_pixmap(wdgt->win, &(wdgt->backing_store), wid, ht);
	widget_draw_px_color(&(wdgt->backing_store), wdgt->background.background_color);

	return(wdgt);
}

static void txtlb_draw_outline(txtlb_data *intern, Pixmap pix, int wid, int ht)
{
	int sx, sy, ex, ey;
	Display *disp=display_get_Xdisplay();

	sx = 2; sy = 2; ex = intern->ex; ey = intern->ey;
	XSetForeground(disp, intern->mgc,
			intern->col_tab[TXTLB_BACKGROUND].pixel);
	XFillRectangle(disp,pix,intern->mgc,0,0,wid-17,ht);
	XFillRectangle(disp,pix,intern->mgc,0,0,wid-17,ht);
	XSetForeground(disp, intern->mgc,
			intern->col_tab[TXTLB_SCROLL_BG].pixel);
//		WhitePixel(disp,DefaultScreen(disp)));
	XFillRectangle(disp,pix,intern->mgc,wid-17,0,17,ht);
	XSetForeground(disp, intern->mgc, intern->col_tab[TXTLB_TROUGH].pixel);
	XDrawLine(disp, pix, intern->mgc, sx, sy, ex, sy);
	XDrawLine(disp, pix, intern->mgc, sx, sy, sx, ey);
	XSetForeground(disp, intern->mgc, intern->col_tab[TXTLB_PEAK].pixel);
	XDrawLine(disp, pix, intern->mgc, sx, ey, ex, ey);
	XDrawLine(disp, pix, intern->mgc, ex, sy, ex, ey);
	sx += 2; ex -= 1; sy += 2;
	for (; (sy+20) <= ey; sy += 20) {
		XSetForeground(disp, intern->mgc,
				intern->col_tab[TXTLB_BACKGROUND].pixel);
		XDrawLine(disp, pix, intern->mgc, sx, sy, ex, sy);
		XDrawLine(disp, pix, intern->mgc, sx, sy+1, ex, sy+1);
		XSetForeground(disp, intern->mgc,
				intern->col_tab[TXTLB_ROW_BACKGROUND].pixel);
		XFillRectangle(disp,pix,intern->mgc,sx,sy,ex-sx,18);
	}
}

static void txtlb_row_select(txtlb_data *intern, int idx, Bool flg)
{
	if (flg == False) {	// Deselect row with idx
		int i;

		for (i=0; i < intern->selecn_count; i++) {
			if (intern->selecn_arr[i] == idx) {
				i++;
				for (; i < intern->selecn_count; i++) {
					intern->selecn_arr[i-1] = intern->selecn_arr[i];
				}
				intern->selecn_count--;
			}
		}
	} else {		// Select the row with idx
		if (intern->selecn_count == intern->max_selecn) {
			intern->selecn_arr = (int *)realloc(intern->selecn_arr,
					sizeof(int)*(intern->max_selecn+5));
			intern->max_selecn += 5;
		}
		intern->selecn_arr[intern->selecn_count] = idx;
		intern->selecn_count++;
	}
}

static Bool txtlb_is_row_selected(txtlb_data *intern, int idx)
{
	if (intern->selecn_count == 0)
		return(False);
	if (intern->type == TXTLB_SINGLE_SELECN) {
		if (idx == intern->selecn_arr[0])
			return(True);
		else
			return(False);
	} else {
		int i;

		for (i=0; i < intern->selecn_count; i++) {
			if (idx == intern->selecn_arr[i])
				return(True);
		}
		return(False);
	}
}

static void show_txtlb_elem(px_widget *wdgt,
           short int x, short int y, short int wid,
           char *str, short int str_sz, Bool sel_flg)
{
	txtlb_data *intern;
	Pixmap pix=wdgt->backing_store.pxmp;
	Display *disp=display_get_Xdisplay();

	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	XSetForeground(disp, intern->mgc,
			intern->col_tab[TXTLB_BACKGROUND].pixel);
	XDrawLine(disp, pix, intern->mgc, x+1, y+2, x+(wid-3), y+2);
	XDrawLine(disp, pix, intern->mgc, x+1, y+3, x+(wid-3), y+3);
	if (sel_flg)
		XSetForeground(disp, intern->mgc,
				intern->col_tab[TXTLB_SELECN_HIGHLIGHT].pixel);
	else
		XSetForeground(disp, intern->mgc,
				intern->col_tab[TXTLB_ROW_BACKGROUND].pixel);
	XFillRectangle(disp,pix,intern->mgc,x+2,y-(20-4),
                           (wid-5),(20-2));
	if (sel_flg)
		XSetForeground(disp,intern->mgc,
				intern->col_tab[TXTLB_SELECN_FOREGROUND].pixel);
	else
		XSetForeground(disp,intern->mgc,
				intern->col_tab[TXTLB_FOREGROUND].pixel);
	if (intern->fnt != NULL) {
		int str_wid=0, str_len=str_sz;

		XSetFont(disp, intern->mgc, intern->fnt->fid);
		str_wid = XTextWidth(intern->fnt, str, str_sz);
		if (str_wid > (wid-7)) {
			str_len = ((wid-18)*str_sz)/str_wid;
			str_wid = XTextWidth(intern->fnt,str,str_len);
			if (str_wid > (wid-18)) {
				while (str_wid > (wid-18)) {
					str_len--;
					str_wid = XTextWidth(intern->fnt,str,str_len);
				}
			} else if (str_wid < (wid-18)) {
				while (str_wid < (wid-18)) {
					str_len++;
					str_wid = XTextWidth(intern->fnt,str,str_len);
				}
				str_len--;
				str_wid = XTextWidth(intern->fnt,str,str_len);
			}
			XDrawString(disp,pix,intern->mgc,x+4,y-2,str,str_len);
			XDrawString(disp,pix,intern->mgc,x+4+str_wid,y-2,"...",3);
		} else 
			XDrawString(disp,pix,intern->mgc,x+4,y-2,str,str_len);
	} else if (str_sz > 0)
		XDrawString(disp,pix,intern->mgc,x+4,y-2,str,str_sz);
}

void txtlb_show_elems(px_widget *wdgt)
{
	int idx, colidx;
	int y, x;
	int maxy;
	txtlb_elem **curr_elems=NULL;
	txtlb_data *intern;
	Pixmap pix=wdgt->backing_store.pxmp;
	Display *disp=display_get_Xdisplay();
	Bool sel_flg;

	y = 20; x = 3;
	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	maxy = intern->ey-4;
	curr_elems = malloc(intern->ncols * sizeof(txtlb_elem *));
	for (colidx=0; colidx < intern->ncols; colidx++) {
		curr_elems[colidx] = intern->elems[colidx*2];
	}
	if (((maxy/20)+intern->sidx) > intern->max_elems) {
		intern->sidx = intern->max_elems - maxy/20;
		if (intern->sidx < 0)
			intern->sidx = 0;
	}
	for (idx=0; (idx < intern->max_elems) && (y <= maxy); idx++) {
		if (idx >= intern->sidx) {
			sel_flg = txtlb_is_row_selected(intern, idx);
//			if (idx == intern->selecn)
//				sel_flg = True;
//			else
//				sel_flg = False;
			x = 3;
			for (colidx=0; colidx < intern->ncols; colidx++) {
				if (curr_elems[colidx] != NULL) {
					show_txtlb_elem(wdgt,x,y,intern->col_wids[colidx],curr_elems[colidx]->data,curr_elems[colidx]->len,sel_flg);
					curr_elems[colidx] = curr_elems[colidx]->next;
				} else {
					show_txtlb_elem(wdgt,x,y,intern->col_wids[colidx],NULL,0,sel_flg);
				}
				x += intern->col_wids[colidx];
			}
			y += 20;
		} else {
			for (colidx=0; colidx < intern->ncols; colidx++) {
				if (curr_elems[colidx] != NULL) {
					curr_elems[colidx] = curr_elems[colidx]->next;
				}
			}
		}
	}
	if (y < maxy) {
		for (; y < maxy; y+=20) {
			x = 3;
			for (colidx=0; colidx < intern->ncols; colidx++) {
				show_txtlb_elem(wdgt,x,y,intern->col_wids[colidx],NULL,0,False);
				x += intern->col_wids[colidx];
			}
		}
	}
	if (intern->ncols > 1) {
		y = 2; maxy = intern->ey - 2;
		for (colidx=0, x=2; colidx < (intern->ncols-1); colidx++) {
			x += intern->col_wids[colidx];
			XSetForeground(disp, intern->mgc,
					intern->col_tab[TXTLB_TROUGH].pixel);
			XDrawLine(disp, pix, intern->mgc, x, y, x, maxy);
			XSetForeground(disp, intern->mgc,
					intern->col_tab[TXTLB_PEAK].pixel);
			XDrawLine(disp, pix, intern->mgc, x+1, y, x+1, maxy);
		}
	}
//	XClearWindow(disp, wdgt->win);
	free(curr_elems);
}

void txtlb_sbar_change(px_widget *wdgt, float val)
{
	txtlb_data *intern;
	int len, new_sidx;

	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	len = (intern->ey-4)/20;
	new_sidx = (val * (intern->max_elems - len));
	if (intern->sidx != new_sidx) {
		intern->sidx = new_sidx;
		txtlb_show_elems(wdgt);
		XClearWindow(display_get_Xdisplay(), wdgt->win);
		if (intern->intimate_view_change != NULL)
			intern->intimate_view_change(wdgt, intern->sidx);
	}
}

void txtlb_redraw(px_widget *wdgt)
{
	txtlb_data *intern=wdgt->specific.txtlb.internal;

	txtlb_draw_outline(intern, wdgt->backing_store.pxmp,
		wdgt->width, wdgt->height);
	txtlb_show_elems(wdgt);
	if (wdgt->hidden == False) {
		XClearWindow(display_get_Xdisplay(), wdgt->win);
	}
	if ((intern->ey-4)/20 <= intern->max_elems) {
		if (intern->sbar == NULL) {
			px_widget *sbar;
			float len=(intern->ey-4)/20;

			sbar = new_widget(VSCROLLBAR, wdgt, wdgt->width-16, 2, 16, wdgt->height-4, NULL);
			sbar->specific.vscrollbar.set_slider_position(sbar,0.0);
			sbar->specific.vscrollbar.set_slider_length(sbar,
					len/intern->max_elems);
			if (len == intern->max_elems) {
				sbar->specific.vscrollbar.set_increment(sbar,1.0);
			} else {
				sbar->specific.vscrollbar.set_increment(sbar,
						1.0/(intern->max_elems-len));
			}
			sbar->specific.vscrollbar.intimate_change = txtlb_sbar_change;
			sbar->redraw(sbar);
			sbar->show(sbar);
			intern->sbar = sbar;
		} else {
			float len=(intern->ey-4)/20;

			intern->sbar->specific.vscrollbar.set_slider_length(intern->sbar, len/intern->max_elems);
			intern->sbar->redraw(intern->sbar);
		}
        } else if ((intern->sbar != NULL) && (intern->sbar->hidden == False)) {
		intern->sbar->hide(intern->sbar);
	}
}

void txtlb_show(px_widget *wdgt)
{
	Display *disp=display_get_Xdisplay();
	txtlb_data *intern=(txtlb_data *)(wdgt->specific.txtlb.internal);

	XSetWindowBackgroundPixmap(disp, wdgt->win, wdgt->backing_store.pxmp);
	XMapWindow(disp, wdgt->win);
	if ((intern->sbar != NULL) && 
		(intern->max_elems >= (intern->ey-2)/20)) {
		intern->sbar->show(intern->sbar);
	}
	wdgt->hidden = False;
}

void txtlb_destroy(px_widget *wdgt)
{
	int colidx;
	txtlb_data *intern=NULL;
	txtlb_elem *curr=NULL;

	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	if (intern->max_elems > 0) {
		for (colidx=0; colidx < intern->ncols; colidx++) {
			for (curr=intern->elems[2*colidx]; curr != NULL;) {
				if ((curr->len > 0) && (curr->data != NULL))
					free(curr->data);
				intern->elems[2*colidx] = curr->next;
				free(curr);
				curr = intern->elems[2*colidx];
			}
		}
	}
	if (intern->sbar != NULL) {
		intern->sbar->hide(intern->sbar);
		intern->sbar->destroy(intern->sbar);
		intern->sbar->destruct(intern->sbar);
	}
	if (intern->elems != NULL)
		free(intern->elems);
	if (intern->n_elems != NULL)
		free(intern->n_elems);
	if (intern->col_wids != NULL)
		free(intern->col_wids);

	widget_destroy(wdgt);
}

static void set_txtlb_selecn(px_widget *wdgt, int selecn, Bool flg)
{
	short int x, y;
	int i, colidx, maxidx;
	txtlb_data *intern;
	txtlb_elem **curr_elems=NULL;

	intern = (txtlb_data *)(wdgt->specific.txtlb.internal);
	maxidx = intern->sidx + (intern->ey-2)/20;
	curr_elems = (txtlb_elem **)malloc(intern->ncols*sizeof(txtlb_elem *));
	for (colidx=0; colidx < intern->ncols; colidx++) {
		curr_elems[colidx] = intern->elems[colidx*2];
	}

	y = 20;
	for (i=0; i <= maxidx; i++) {
		if ((i == selecn) && (i >= intern->sidx)) {
			x = 3;
			for (colidx=0; colidx < intern->ncols; colidx++) {
				if (curr_elems[colidx] != NULL) {
					show_txtlb_elem(wdgt,x,y,intern->col_wids[colidx],curr_elems[colidx]->data,curr_elems[colidx]->len,flg);
				} else {
					show_txtlb_elem(wdgt,x,y,intern->col_wids[colidx],NULL,0,flg);
				}
				x += intern->col_wids[colidx];
			}
			break;
		} else if (i == selecn) {
			break;
		} else {
			for (colidx=0; colidx < intern->ncols; colidx++) {
				if (curr_elems[colidx] != NULL) {
					curr_elems[colidx] = curr_elems[colidx]->next;
				}
			}
			if (i >= intern->sidx) {
				y += 20;
			}
		}
	}
	free(curr_elems);
}

// flg == True =====> called from a button press.
// flg == False ======> called from button motion.
static void change_txtlb_selecn(px_widget *wdgt, short int selecn, Bool flg)
{
	txtlb_data *intern=NULL;
	Display *disp=display_get_Xdisplay();

	intern = (txtlb_data *)(wdgt->specific.txtlb.internal);
	if ((intern->sidx+selecn) >= intern->max_elems)
		return;
	if (intern->type == TXTLB_SINGLE_SELECN) {
		if (intern->selecn_count != 0) {
			if (intern->selecn_arr[0] != (intern->sidx+selecn)) {
				set_txtlb_selecn(wdgt,intern->selecn_arr[0],False);
				set_txtlb_selecn(wdgt,intern->sidx+selecn,True);
//				intern->selecn = intern->sidx + selecn;
				intern->selecn_arr[0] = intern->sidx + selecn;
			} else if (flg == False)
				return;
		} else {
			set_txtlb_selecn(wdgt, intern->sidx+selecn, True);
//			intern->selecn = intern->sidx + selecn;
			intern->selecn_arr[0] = intern->sidx + selecn;
			intern->selecn_count = 1;
		}
	} else if (intern->type == TXTLB_MULTI_SELECN) {
		if (txtlb_is_row_selected(intern,intern->sidx+selecn) == True) {
			if (flg == True) {
				set_txtlb_selecn(wdgt,intern->sidx+selecn,False);
				txtlb_row_select(intern, intern->sidx+selecn, False);
			}
		} else {
			set_txtlb_selecn(wdgt,intern->sidx+selecn,True);
			txtlb_row_select(intern, intern->sidx+selecn, True);
		}
	}
	XClearWindow(disp, wdgt->win);
//	if (lbox->intimate_selecn_change != NULL) {
//		lbox->intimate_selecn_change(wdgt, intern->selecn);
//	}
}

Bool txtlb_ev_in_sbar(px_widget *wdgt, px_event *ev)
{
	txtlb_data *intern;

	intern = (txtlb_data *)wdgt->specific.txtlb.internal;

	if (intern->sbar == NULL) {
		return(False);
	}
	if ((ev->type == BUTTON_PRESSED) || (ev->type == BUTTON_RELEASED) ||
		(ev->type == MOTION)) {
		if ((ev->x >= (wdgt->width-16)) && (ev->x <= (wdgt->width-1))) {
			return(True);
		}
	}
	return(False);
}

Bool txtlb_handle_ev(px_widget *wdgt, px_event *ev)
{
	if ((ev->type != NR_CALLBACKS) &&
	((ev->type == EXPOSE) || (wdgt->get_state(wdgt) != DISABLED))) {
		return(widget_handle_event(wdgt,ev));
	}
	return(True);
}

Bool txtlb_dflt_bpress(px_widget *wdgt, px_event *ev)
{
	txtlb_data *intern;
	px_eventloop *eloop= eventloop_get_eventloop();

	eloop->grab(eloop, wdgt);

	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	if (txtlb_ev_in_sbar(wdgt, ev)) {
		px_widget *sbar = intern->sbar;
		Bool ret_val;

		ev->x -= sbar->x; ev->y -= sbar->y;
		ret_val = sbar->handle_event(sbar, ev);
		ev->x += sbar->x; ev->y += sbar->y;
		return(ret_val);
	}
	if ((ev->x >= 2) && (ev->x <= intern->ex) &&
		(ev->y > 2) && (ev->y < (intern->ey-2))) {
		change_txtlb_selecn(wdgt, (ev->y-2)/20, True);
	}
	return(True);
}

Bool handle_repeat_timer(px_widget *wdgt, px_event *ev)
{
	int len;
	txtlb_data *intern=wdgt->specific.txtlb.internal;

	if (rpt_direcn == TXTLB_DOWN_DIRECN) {
		len = (intern->ey-4)/20;
		if ((len < intern->max_elems) &&
			((intern->sidx+len) < intern->max_elems)) {
			txtlb_set_view(wdgt, intern->sidx+1);
			change_txtlb_selecn(wdgt, len-1, False);
			if (intern->intimate_view_change != NULL)
				intern->intimate_view_change(wdgt,intern->sidx);
		} else {
			rpt_direcn = TXTLB_NO_DIRECN;
		}
	} else if (rpt_direcn == TXTLB_UP_DIRECN) {
		if (intern->sidx > 0) {
			txtlb_set_view(wdgt, intern->sidx-1);
			change_txtlb_selecn(wdgt, 0, False);
			if (intern->intimate_view_change != NULL)
				intern->intimate_view_change(wdgt,intern->sidx);
		} else {
			rpt_direcn = TXTLB_NO_DIRECN;
		}
	}
	return(True);
}

Bool txtlb_dflt_motion(px_widget *wdgt, px_event *ev)
{
	txtlb_data *intern;

	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	if (txtlb_ev_in_sbar(wdgt, ev)) {
		px_widget *sbar = intern->sbar;
		Bool ret_val;

		ev->x -= sbar->x; ev->y -= sbar->y;
		ret_val = sbar->handle_event(sbar, ev);
		ev->x += sbar->x; ev->y += sbar->y;
		return(ret_val);
	}
	if ((ev->x >= 2) && (ev->x <= intern->ex) &&
		(ev->y > 2) && (ev->y < (intern->ey-2))) {
		if (rpt_timer != NULL) {
			px_eventloop* el = eventloop_get_eventloop();

			el->delete_timer(el, rpt_timer);
			rpt_timer = NULL; rpt_direcn = TXTLB_NO_DIRECN;

		}
		change_txtlb_selecn(wdgt, (ev->y-2)/20, False);
	} else {
		if (rpt_timer == NULL) {
			px_eventloop* el = eventloop_get_eventloop();

			rpt_timer = el->add_timer(el,0,TXTLB_REPEAT_DELAY,handle_repeat_timer, wdgt, ev, True);
			if (ev->y <= 2)
				rpt_direcn = TXTLB_UP_DIRECN;
			else if (ev->y >= (intern->ey-2))
				rpt_direcn = TXTLB_DOWN_DIRECN;
		}
	}
	{
		int fg;
		XEvent xev, xev_copy;

		fg = 0;
		while(XPending(display_get_Xdisplay())) {
			XNextEvent(display_get_Xdisplay(), &xev);
			if ((!fg) && (xev.type==MotionNotify)) {
				fg = 1; xev_copy = xev;
			}
			if (xev.type==ButtonRelease) {
				fg = 1; xev_copy = xev; break;
			}
		}
		if (fg) {
			XPutBackEvent(display_get_Xdisplay(), &xev_copy);
		}
	}
	return(True);
}

Bool txtlb_dflt_brelease(px_widget *wdgt, px_event *ev)
{
	txtlb_data *intern;
	px_eventloop *eloop= eventloop_get_eventloop();
	px_txtlb *lbox=&(wdgt->specific.txtlb);

	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	if (txtlb_ev_in_sbar(wdgt, ev)) {
		px_widget *sbar = intern->sbar;
		Bool ret_val;

		ev->x -= sbar->x; ev->y -= sbar->y;
		ret_val = sbar->handle_event(sbar, ev);
		ev->x += sbar->x; ev->y += sbar->y;
		return(ret_val);
	}
	eloop->grab(eloop, NULL);
	if (rpt_timer != NULL) {
		px_eventloop* el = eventloop_get_eventloop();

		el->delete_timer(el, rpt_timer);
		rpt_timer = NULL; rpt_direcn = TXTLB_NO_DIRECN;
	}
	if (lbox->intimate_selecn_change != NULL) {
		if (intern->type == TXTLB_SINGLE_SELECN) {
			lbox->intimate_selecn_change(wdgt,intern->selecn_arr[0]);
		} else if ((intern->type == TXTLB_MULTI_SELECN) &&
				(intern->selecn_count > 0)) {
			lbox->intimate_selecn_change(wdgt,
				intern->selecn_arr[intern->selecn_count-1]);
		}
	}
	return(True);
}

int add_lb_elem(px_widget *wdgt, char *str, int colnum)
{
	txtlb_data *intern;
	int sz, i;
	int colidx=colnum*2;
	txtlb_elem *new=NULL;

	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	if ((colnum >= intern->ncols) || (colnum < 0))
		return(-1);
	new = (txtlb_elem *)malloc(sizeof(txtlb_elem));
	if (new == NULL) {
		printf("Unable to alloc mem for txtlb elem\n");
		exit(0);
	}
	sz = strlen(str);
	new->data = (char *)malloc(sizeof(char)*(sz+1));
	if (new->data == NULL) {
		printf("Unable to alloc mem for txtlb elem\n");
		exit(0);
	}
	new->len = sz; new->next = NULL;

	for (i=0; i < sz; i++) {
		new->data[i] = str[i];
	}
	new->data[i] = 0;

	if (intern->elems[colidx] == NULL) {
		intern->elems[colidx] = new;
		intern->elems[colidx+1] = new;
	} else {
		intern->elems[colidx+1]->next = new;
		intern->elems[colidx+1] = new;
	}
	intern->n_elems[colnum]++;
	if (intern->n_elems[colnum] > intern->max_elems)
		intern->max_elems = intern->n_elems[colnum];
	return(intern->n_elems[colnum]);
}

int insert_lb_elem(px_widget *wdgt, char *str, int colnum, int idx)
{
	int sz, i, col;
	txtlb_data *intern;
	txtlb_elem *new=NULL;

	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	if ((colnum >= intern->ncols) || (colnum < 0)) {
		printf("Incorrect column number=%d, max=%d\n", colnum, intern->ncols);
		return(-1);
	}
	new = (txtlb_elem *)malloc(sizeof(txtlb_elem));
	if (new == NULL) {
		printf("Unable to alloc mem for txtlb elem\n");
		exit(0);
	}
	sz = strlen(str);
	new->data = (char *)malloc(sizeof(char)*(sz+1));
	if (new->data == NULL) {
		printf("Unable to alloc mem for txtlb elem\n");
		exit(0);
	}
	new->len = sz; new->next = NULL;

	for (i=0; i < sz; i++) {
		new->data[i] = str[i];
	}
	new->data[i] = 0;
	col = 2*colnum;
	if ((idx == LB_END) || (idx < LB_START) || (idx >= intern->max_elems)) {
		if (intern->elems[col] == NULL) {
			intern->elems[col] = new;
			intern->elems[col+1] = new;
		} else {
			intern->elems[col+1]->next = new;
			intern->elems[col+1] = new;
		}
	} else if (idx == LB_START) {
		new->next = intern->elems[col];
		intern->elems[col] = new;
		if (intern->elems[col+1] == NULL) {
			intern->elems[col+1] = new;
		}
	} else {
		txtlb_elem *curr=NULL, *prev=NULL;

		if (idx >= intern->n_elems[colnum]) {
			if (intern->elems[col] == NULL) {
				intern->elems[col] = new;
				intern->elems[col+1] = new;
			} else {
				intern->elems[col+1]->next = new;
				intern->elems[col+1] = new;
			}
		} else {
			for (i=LB_START,curr=intern->elems[col],prev=curr;
				curr != NULL; prev=curr,curr=curr->next,i++) {
				if (i == idx) {
					prev->next = new;
					new->next = curr; break;
				}
			}
		}
	}
	intern->n_elems[colnum]++;
	if (intern->n_elems[colnum] > intern->max_elems) {
		intern->max_elems = intern->n_elems[colnum];
	}
	return(intern->n_elems[colnum]);
}

static void txtlb_set_max_no_of_elems( txtlb_data * intern )
{
	int colidx=0;

	intern->max_elems = 0;
	for (colidx=0; colidx < intern->ncols; colidx++) {
		if (intern->n_elems[colidx] >= intern->max_elems) {
			intern->max_elems = intern->n_elems[colidx];
		}
	}
}

void txtlb_set_selection(px_widget *wdgt, int sel_idx)
{
	txtlb_data *intern;
	int len;
	Bool chng_view=False;

	if (wdgt->wtype != TXTLB)
		return;
	intern = wdgt->specific.txtlb.internal;
	len = (intern->ey-4)/20;

	if ((sel_idx < 0) || (sel_idx >= intern->max_elems)) {
		if (intern->type == TXTLB_SINGLE_SELECN) {
			if (intern->selecn_count != 0) {
				set_txtlb_selecn(wdgt,intern->selecn_arr[0],False);
				intern->selecn_arr[0] = -1;
			}
			intern->selecn_count = 0;
		}
		chng_view = True;
//		if (intern->selecn != -1)
//			set_txtlb_selecn(wdgt, intern->selecn, False);
//		intern->selecn = -1;
	} else {
		if (intern->type == TXTLB_SINGLE_SELECN) {
			if (intern->selecn_count != 0)
				set_txtlb_selecn(wdgt, intern->selecn, False);
			else
				intern->selecn_count = 1;
			intern->selecn_arr[0] = sel_idx;
			if ((intern->selecn_arr[0] >= intern->sidx) && 
				(intern->selecn_arr[0] < (intern->sidx+len))) {
				set_txtlb_selecn(wdgt,intern->selecn_arr[0],True);
				chng_view = False;
			} else {
				chng_view = True;
			}
		} else {
			txtlb_row_select(intern, sel_idx, True);
			if ((sel_idx >= intern->sidx) && 
				(sel_idx < (intern->sidx+len))) {
				set_txtlb_selecn(wdgt,sel_idx,True);
				chng_view = False;
			} else {
				chng_view = True;
			}
		}
	}
	if (chng_view == True) {
		if (sel_idx < intern->sidx) {
			txtlb_set_view(wdgt, sel_idx);
			if (intern->intimate_view_change != NULL)
				intern->intimate_view_change(wdgt,intern->sidx);
		} else if (sel_idx >= (intern->sidx+len)) {
			if ((sel_idx+len) <= intern->max_elems) {
				txtlb_set_view(wdgt, sel_idx-len+1);
			} else {
				txtlb_set_view(wdgt, intern->max_elems-len);
			}
			if (intern->intimate_view_change != NULL)
				intern->intimate_view_change(wdgt,intern->sidx);
		} else {
			XClearWindow(display_get_Xdisplay(), wdgt->win);
		}
	} else {
		XClearWindow(display_get_Xdisplay(), wdgt->win);
	}
}

int remove_lb_elem(px_widget *wdgt, int colnum, int idx)
{
	txtlb_data *intern;
	txtlb_elem *curr, *prev, *next;
	int colidx=0;

	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	if ((colnum >= intern->ncols) || (colnum < 0))
		return(-1);
	if ((idx >= intern->max_elems) || (idx < 0))
		return(-1);
	colidx = colnum*2;
	if (idx == 0) {
		curr = intern->elems[colidx];
		intern->elems[colidx] = curr->next;
		if (intern->elems[colidx] == NULL) {
			intern->elems[colidx+1] = NULL;
		}
		if (curr->data != NULL) {
			free(curr->data);
		}
		free(curr); intern->n_elems[colnum]--;
		txtlb_set_max_no_of_elems(intern);
		return(intern->n_elems[colnum]);
	} else {
		int i=1;

		prev = intern->elems[colidx];
		next = intern->elems[colidx]->next;
		for (i=1;(i < intern->n_elems[colnum]) && (curr != NULL);
			i++, prev=curr, curr=curr->next) {
          if (i == idx) {
             prev->next = curr->next;
             if (curr->next == NULL) {
                intern->elems[colidx+1] = prev;
             }
             if (curr->data != NULL) {
                free(curr->data);
             }
             free(curr); intern->n_elems[colnum]--;
             txtlb_set_max_no_of_elems(intern);
             return(intern->max_elems);
          }
      }
      return(-1);
   }
}

int add_lb_column(px_widget *wdgt, int wid)
{
	txtlb_data *intern=NULL;
	int idx;

	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	if (intern->col_wids[0] > wid) {
		intern->elems = realloc(intern->elems,
				(intern->ncols+1)*2*sizeof(txtlb_elem *));
		intern->n_elems = (int *)realloc(intern->n_elems,
                              sizeof(int)*(intern->ncols+1));
		intern->col_wids = (int *)realloc(intern->col_wids,
				sizeof(int)*(intern->ncols+1));
		idx = intern->ncols*2;
		intern->elems[idx] = NULL; intern->elems[idx+1] = NULL;
		intern->n_elems[intern->ncols] = 0;
		intern->col_wids[intern->ncols] = wid;
		intern->col_wids[0] -= wid;
		intern->ncols++; return(intern->ncols);
	} else
		return(-1);
}

void free_txtlb_elem(txtlb_data *intern, txtlb_elem *curr, txtlb_elem *prev, int colidx, int rowidx)
{
	if (curr == NULL)
		return;
	if (rowidx != 0) {
		prev->next = curr->next;
		if (curr->next == NULL) {
			intern->elems[2*colidx+1] = prev;
		}
		intern->n_elems[colidx]--;
	} else {
		intern->elems[colidx*2] = curr->next;
		intern->n_elems[colidx]--;
		if (intern->n_elems[colidx] <= 0) {
			intern->elems[colidx*2+1] = NULL;
		}
	}
	if (curr->len > 0) {
		free(curr->data);
	}
	free(curr);
}

int remove_lb_row(px_widget *wdgt, int rowidx, Bool redraw_flg)
{
	txtlb_data *intern;
	txtlb_elem **curr_elems=NULL, **prev_elems=NULL;
	int colidx, rowcount;

	if (wdgt->wtype != TXTLB)
		return(-1);
	intern = wdgt->specific.txtlb.internal;
	if (rowidx >= intern->max_elems)
		return(-1);
	curr_elems = malloc(intern->ncols * sizeof(txtlb_elem *));
	prev_elems = malloc(intern->ncols * sizeof(txtlb_elem *));
	for (colidx=0; colidx < intern->ncols; colidx++) {
		curr_elems[colidx] = intern->elems[colidx*2];
	}
	for (rowcount=0; rowcount < intern->max_elems; rowcount++) {
		if (rowcount == rowidx) {
			for (colidx=0; colidx < intern->ncols; colidx++) {
				free_txtlb_elem(intern, curr_elems[colidx], prev_elems[colidx], colidx, rowidx);
			}
			break;
		} else {
			for (colidx=0; colidx < intern->ncols; colidx++) {
				prev_elems[colidx] = curr_elems[colidx];
				if (curr_elems[colidx] != NULL)
					curr_elems[colidx] = curr_elems[colidx]->next;
			}
		}
	}
	txtlb_set_max_no_of_elems(intern);
	if (intern->type == TXTLB_SINGLE_SELECN) {
		intern->selecn_arr[0] = -1;
		intern->selecn_count = 0;
	} else if (intern->type == TXTLB_MULTI_SELECN) {
		int i;

		for (i=0; i < intern->selecn_count; i++)
			intern->selecn_arr[i] = -1;
		intern->selecn_count = 0;
	}
	if (redraw_flg == True)
		wdgt->redraw(wdgt);
	return(intern->max_elems);
}

int txtlb_clear_all_rows(px_widget *wdgt)
{
	txtlb_data *intern;
	txtlb_elem **curr_elems=NULL, *curr=NULL;
	int colidx, rowcount;

	if (wdgt->wtype != TXTLB)
		return(-1);
	intern = wdgt->specific.txtlb.internal;
	curr_elems = malloc(intern->ncols * sizeof(txtlb_elem *));
	for (colidx=0; colidx < intern->ncols; colidx++) {
		curr_elems[colidx] = intern->elems[colidx*2];
	}
	for (rowcount=0; rowcount < intern->max_elems; rowcount++) {
		for (colidx=0; colidx < intern->ncols; colidx++) {
			if (curr_elems[colidx] != NULL) {
				curr = curr_elems[colidx];
				curr_elems[colidx] = curr->next;
				if (curr->len > 0)
					free(curr->data);
				free(curr);
			}
		}
	}
	for (colidx=0;colidx < intern->ncols; colidx++) {
		intern->elems[colidx*2] = NULL;
		intern->elems[colidx*2+1] = NULL;
		intern->n_elems[colidx] = 0;
	}
	intern->max_elems = 0;
	if (intern->type == TXTLB_SINGLE_SELECN) {
		intern->selecn_arr[0] = -1;
		intern->selecn_count = 0;
	} else if (intern->type == TXTLB_MULTI_SELECN) {
		int i;

		for (i=0; i < intern->selecn_count; i++)
			intern->selecn_arr[i] = -1;
		intern->selecn_count = 0;
	}
	wdgt->redraw(wdgt);
	return(0);
}

int change_lb_elem(px_widget *wdgt, char *txtstr, int colnum, int rowidx)
{
	txtlb_data *intern;
	txtlb_elem *curr;
	int rowcount, len, i;

	if (wdgt->wtype != TXTLB)
		return(-1);
	intern = wdgt->specific.txtlb.internal;
	if (intern->ncols <= colnum)
		return(-1);
	if (intern->n_elems[colnum] <= rowidx)
		return(-1);
	for (curr=intern->elems[colnum*2], rowcount=0; curr != NULL;
		curr=curr->next, rowcount++) {
		if (rowcount == rowidx) {
			len = strlen(txtstr);
			if (len > curr->len) {
				curr->data = (char *)realloc(curr->data,
						sizeof(char)*(len+1));
			}
			for (i=0; i < len; i++)
				curr->data[i] = txtstr[i];
			curr->data[i] = 0;
			curr->len = len;
			return(intern->n_elems[colnum]);
		}
	}
	return(-1);
}

void change_lb_color(px_widget *wdgt, txtlb_colors col, char *colstr)
{
	txtlb_data *intern=(txtlb_data *)wdgt->specific.txtlb.internal;
	Display *disp=display_get_Xdisplay();
	Colormap a_cmap=DefaultColormap(disp,DefaultScreen(disp));

	XParseColor(disp, a_cmap, colstr, &(intern->col_tab[col]));
	XAllocColor(disp, a_cmap, &(intern->col_tab[col]));
}

void txtlb_set_size(px_widget* wdgt, unsigned long width,
		unsigned long height)
{
	int new_ht=((height-8)/20)*20 + 8;
	int width_dx=wdgt->width-width;
	txtlb_data *intern=wdgt->specific.txtlb.internal;

	if (width_dx >= intern->col_wids[0])
		return;

	intern->ex = width - 20;
	intern->ey = new_ht-4;

	widget_set_size(wdgt, width, new_ht);
	intern->col_wids[0] -= width_dx;
	if (intern->sbar != NULL) {
		widget_set_size(intern->sbar, 16, new_ht-4);
		if (intern->sbar->x != width-16)
			intern->sbar->set_position(intern->sbar, width-16, 2);
	}
//	wdgt->child_of->redraw(wdgt->child_of);
	txtlb_redraw(wdgt);
}

void txtlb_set_parameters(px_widget *wdgt)
{
	px_txtlb *lbox;
	txtlb_data *intern=NULL;

	wdgt->default_callbacks[BUTTON_PRESSED] = txtlb_dflt_bpress;
	wdgt->default_callbacks[BUTTON_RELEASED] = txtlb_dflt_brelease;
	wdgt->default_callbacks[MOTION] = txtlb_dflt_motion;

	wdgt->handle_event = txtlb_handle_ev;
	wdgt->redraw = txtlb_redraw;
	wdgt->show = txtlb_show;
	wdgt->destroy = txtlb_destroy;

//	wdgt->set_size = widget_set_size;
	wdgt->set_size = txtlb_set_size;

	lbox = &(wdgt->specific.txtlb);

	intern = (txtlb_data *)malloc(sizeof(txtlb_data));
	lbox->internal = (void *)intern;

	intern->sbar = NULL;
	
	{
		Display *disp=display_get_Xdisplay();
		Colormap a_cmap=DefaultColormap(disp,DefaultScreen(disp));

		intern->mgc = XCreateGC(disp, wdgt->win, 0, NULL);
		XParseColor(disp, a_cmap, "#d3d3dd",
				&(intern->col_tab[TXTLB_BACKGROUND]));
		XAllocColor(disp, a_cmap, &(intern->col_tab[TXTLB_BACKGROUND]));
		XParseColor(disp, a_cmap, "#000000",
				&(intern->col_tab[TXTLB_FOREGROUND]));
		XAllocColor(disp, a_cmap, &(intern->col_tab[TXTLB_FOREGROUND]));
		XParseColor(disp, a_cmap, "#ffffff",
				&(intern->col_tab[TXTLB_PEAK]));
		XAllocColor(disp, a_cmap, &(intern->col_tab[TXTLB_PEAK]));
		XParseColor(disp, a_cmap, "#8e8ea1",
				&(intern->col_tab[TXTLB_TROUGH]));
		XAllocColor(disp, a_cmap, &(intern->col_tab[TXTLB_TROUGH]));
		XParseColor(disp, a_cmap, "#336699",
				&(intern->col_tab[TXTLB_SELECN_HIGHLIGHT]));
		XAllocColor(disp, a_cmap,
				&(intern->col_tab[TXTLB_SELECN_HIGHLIGHT]));
		XParseColor(disp, a_cmap, "#ffffff",
				&(intern->col_tab[TXTLB_SELECN_FOREGROUND]));
		XAllocColor(disp, a_cmap,
				&(intern->col_tab[TXTLB_SELECN_FOREGROUND]));
		XParseColor(disp, a_cmap, "#e5e5f7",
				&(intern->col_tab[TXTLB_ROW_BACKGROUND]));
		XAllocColor(disp, a_cmap,
				&(intern->col_tab[TXTLB_ROW_BACKGROUND]));
		XParseColor(disp, a_cmap, "#d3d3dd",
				&(intern->col_tab[TXTLB_SCROLL_BG]));
		XAllocColor(disp, a_cmap, &(intern->col_tab[TXTLB_SCROLL_BG]));
		intern->fnt = XLoadQueryFont(disp, DFLT_TXTLB_FONT);

		if (intern->fnt == NULL) {
			printf("Unable to load default txtlb font.\n");
		}
	}
	intern->ex = wdgt->width - 20;
//	intern->ey = ((wdgt->height-8)/20)*20 + 4;
	intern->ey = wdgt->height-4;
	intern->max_elems = 0; intern->sidx = 0;
	intern->selecn = -1; intern->ncols = 1;
	intern->elems = (txtlb_elem **)malloc(sizeof(txtlb_elem *)*2);
	intern->elems[0] = NULL; intern->elems[1] = NULL;
	intern->n_elems = (int *)malloc(sizeof(int));
	intern->n_elems[0] = 0;
	intern->col_wids = (int *)malloc(sizeof(int));
	intern->col_wids[0] = wdgt->width - 21;

	{
		short int i;

		intern->type = TXTLB_SINGLE_SELECN;
		intern->selecn_arr = (int *)malloc(sizeof(int) * 5);
		for (i=0; i < 5; i++) {
			intern->selecn_arr[i] = -1;
		}
		intern->max_selecn = 5;
		intern->selecn_count = 0;
	}
	intern->intimate_view_change = NULL;

	lbox->add_lb_column = add_lb_column;
	lbox->add_lb_elem = add_lb_elem;
	lbox->insert_lb_elem = insert_lb_elem;
	lbox->remove_lb_row = remove_lb_row;
	lbox->change_lb_elem = change_lb_elem;
	lbox->change_lb_color = change_lb_color;
	lbox->remove_lb_elem = remove_lb_elem;
	lbox->show_lb_elems = txtlb_show_elems;
	lbox->intimate_selecn_change = NULL;
}

void txtlb_set_view(px_widget *wdgt, int sidx)
{
	txtlb_data *intern=NULL;
	int len;
	Display *disp;

	if (wdgt->wtype != TXTLB)
		return;
	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	len = (intern->ey-4)/20;
	disp = display_get_Xdisplay();
	if (len >= intern->max_elems)
		return;
	if (sidx < 0) {
		intern->sidx = 0;
		txtlb_show_elems(wdgt);
		intern->sbar->specific.vscrollbar.intimate_change = NULL;
		intern->sbar->specific.vscrollbar.set_slider_position(intern->sbar,0.0);
		intern->sbar->redraw(intern->sbar);
		XClearWindow(disp, wdgt->win);
		intern->sbar->specific.vscrollbar.intimate_change = txtlb_sbar_change;
	} else if ((sidx+len) <= intern->max_elems) {
		intern->sidx = sidx;
		txtlb_show_elems(wdgt);
		intern->sbar->specific.vscrollbar.intimate_change = NULL;
		intern->sbar->specific.vscrollbar.set_slider_position(intern->sbar,(sidx * 1.0)/(intern->max_elems-len));
		intern->sbar->redraw(intern->sbar);
		XClearWindow(disp, wdgt->win);
		intern->sbar->specific.vscrollbar.intimate_change = txtlb_sbar_change;
	} else if (sidx != intern->sidx) {
		if ((intern->max_elems-len) != intern->sidx) {
			intern->sidx = intern->max_elems - len;
			txtlb_show_elems(wdgt);
			intern->sbar->specific.vscrollbar.intimate_change = NULL;
			intern->sbar->specific.vscrollbar.set_slider_position(intern->sbar,1.0);
			intern->sbar->redraw(intern->sbar);
			XClearWindow(disp, wdgt->win);
			intern->sbar->specific.vscrollbar.intimate_change = txtlb_sbar_change;
		}
	}
}

void txtlb_get_element(px_widget *wdgt, char **str, int idx, int colnum)
{
	txtlb_data *intern=NULL;
	txtlb_elem **elems=NULL, *curr=NULL;
	int i, j;

	
	if (wdgt->wtype != TXTLB)
		return;
	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	if (colnum >= intern->ncols) {
		*str = NULL;
		return;
	}
	if (idx >= intern->n_elems[colnum]) {
		*str = NULL;
		return;
	}
	elems = intern->elems;
	for (i=0,curr=elems[colnum*2]; curr != NULL; i++, curr=curr->next) {
		if (i == idx) {
			(*str) = (char *)malloc(sizeof(char)*curr->len+1);
			for (j=0; j < curr->len; j++) {
				(*str)[j] = curr->data[j];
			}
			(*str)[j] = 0;
			break;
		}
	}
	if (curr == NULL) {
		*str = NULL;
	}
}

int txtlb_get_selecn(px_widget *wdgt)
{
	txtlb_data *intern;

	if (wdgt->wtype != TXTLB)
		return(-1);
	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	if (intern->type == TXTLB_SINGLE_SELECN)
		return(intern->selecn_arr[0]);
	else if ((intern->type == TXTLB_MULTI_SELECN) &&
			(intern->selecn_count > 0)) {
		return(intern->selecn_arr[intern->selecn_count-1]);
	} else
		return(-1);
}

int txtlb_get_all_selecn(px_widget *wdgt, int ** selecn_arr)
{
	txtlb_data *intern;

	if (wdgt->wtype != TXTLB)
		return(-1);
	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	if (intern->type == TXTLB_SINGLE_SELECN) {
		(*selecn_arr) = (int *)malloc(sizeof(int));
		(*selecn_arr)[0] = intern->selecn_arr[0];
		return(intern->selecn_count);
	} else if (intern->type == TXTLB_MULTI_SELECN) {
		int i;

		if (intern->selecn_count == 0) {
			(*selecn_arr) = NULL;
			return(0);
		}
		(*selecn_arr) = (int *)malloc(sizeof(int)*intern->selecn_count);
		for (i=0; i < intern->selecn_count; i++)
			(*selecn_arr)[i] = intern->selecn_arr[i];
		return(intern->selecn_count);
	} else {
		return(-1);
	}
}

void txtlb_set_intimate_change(px_widget *wdgt, 
		void (*change_fn)(px_widget *, int))
{
	if (wdgt->wtype != TXTLB)
		return;
	wdgt->specific.txtlb.intimate_selecn_change = change_fn;;
}

void txtlb_set_intimate_view_change(px_widget *wdgt, 
		void (*change_fn)(px_widget *, int))
{
	txtlb_data *intern;

	if (wdgt->wtype != TXTLB)
		return;
	intern = wdgt->specific.txtlb.internal;
	intern->intimate_view_change = change_fn;
}

void txtlb_set_font(px_widget *wdgt, const char *fnt_name)
{
	XFontStruct *new;
	txtlb_data *intern;

	if (wdgt->wtype != TXTLB)
		return;
	new = XLoadQueryFont(display_get_Xdisplay(), fnt_name);
	if (new == NULL) {
		printf("Could not load font:%s\n", fnt_name);
		return;
	}
	intern = wdgt->specific.txtlb.internal;
	if (intern->fnt != NULL) {
		XFreeFont(display_get_Xdisplay(), intern->fnt);
	}
	intern->fnt = new;
}

int txtlb_get_no_of_rows(px_widget *wdgt)
{
	txtlb_data *intern;

	if (wdgt->wtype != TXTLB)
		return(-1);
	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	return(intern->max_elems);
}

void txtlb_set_type(px_widget *wdgt, txtlb_types type)
{
	txtlb_data *intern;

	if (wdgt->wtype != TXTLB)
		return;
	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	if (type == TXTLB_SINGLE_SELECN) {
		if (intern->type == TXTLB_MULTI_SELECN) {
			int i;

			for (i=0; i < intern->selecn_count; i++)
				intern->selecn_arr[i] = -1;
			intern->selecn_count = 0;
			intern->type = type;
		}
	} else if (type == TXTLB_MULTI_SELECN) {
		intern->type = type;
	}
}

unsigned long txtlb_get_fg_color(px_widget *wdgt)
{
	txtlb_data *intern;
	unsigned long col=0L;

	if ((wdgt == NULL) || (wdgt->wtype != TXTLB))
		return(0L);
	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	col = (((intern->col_tab[TXTLB_FOREGROUND].red * 255)/0xffff) << 16)
	+ (((intern->col_tab[TXTLB_FOREGROUND].green * 255)/0xffff) << 8)
	+ ((intern->col_tab[TXTLB_FOREGROUND].blue * 255)/0xffff);

	return(col);
}

unsigned long txtlb_get_bg_color(px_widget *wdgt)
{
	txtlb_data *intern;
	unsigned long col=0L;

	if ((wdgt == NULL) || (wdgt->wtype != TXTLB))
		return(0L);
	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	col = (((intern->col_tab[TXTLB_BACKGROUND].red * 255)/0xffff) << 16)
	+ (((intern->col_tab[TXTLB_BACKGROUND].green * 255)/0xffff) << 8)
	+ ((intern->col_tab[TXTLB_BACKGROUND].blue * 255)/0xffff);

	return(col);
}

void txtlb_get_view_port(px_widget *wdgt, int *sidx, int *eidx)
{
	txtlb_data *intern;

	if ((wdgt == NULL) || (wdgt->wtype != TXTLB)) {
		(*sidx) = -1; (*eidx) = -1;
		return;
	}
	intern = (txtlb_data *)wdgt->specific.txtlb.internal;
	(*sidx) = intern->sidx;
	if (((intern->ey-4)/20) < intern->max_elems)
		(*eidx) = (intern->ey-4)/20 + intern->sidx;
	else
		(*eidx) = intern->max_elems;
}

void txtlb_set_visible_no_of_rows(px_widget *wdgt, int num)
{
	int new_ht;

	if ((wdgt == NULL) || (wdgt->wtype != TXTLB) || (num <= 0))
		return;

	new_ht = num*20 + 8;
	txtlb_set_size(wdgt, wdgt->width, new_ht);
}
