/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

// PicoWidget : Vertical Scrollbar
// Copyrights (c) PicoPeta Simputers Pvt. Ltd.
// Programmer: Santhosh N

#include "vscrollbar.h"
#include "image.h"
#include "widget_factory.h"
#include <X11/extensions/shape.h>
#include <stdlib.h>
#include <string.h>
#include "theme.h"
#include "eventloop.h"
#include "display.h"

#define VSCROLL_DELAY 130000

static struct _vscrollbar {
	float slider_start, slider_length, increment;
};

typedef enum _vscrollbar_states 
{
	VSCROLL_NONE, VSCROLL_UP, VSCROLL_DOWN, VSCROLL_SLIDER, VSCROLL_TROUGH
} vscrollbar_states;

static vscrollbar_states sbar_state;
static int prev_y;
static px_timer *timer=NULL;

px_widget* vscrollbar_create(px_widget* parent, int x, int y, int width, int height, const char *tag)
{
	px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
	Display* display = display_get_Xdisplay();
	px_theme* theme = theme_get_theme();

	widget->win = XCreateWindow(display, parent->win, x, y, width, height, 0,
							CopyFromParent, CopyFromParent, CopyFromParent, 0, NULL);

	// Initialize common widget variables
	widget->x = x;
	widget->y = y;
	widget->height = height;
	widget->width = 16;
	widget->wtype = VSCROLLBAR;
	widget->child_of = NULL;
	widget->use_pixmap = True;
	if(tag != NULL) {
		widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	}else
		widget->tag = NULL;

	widget->background.background_pixmap = theme_get_pixmap(theme, VSCROLLBAR_TROUGH);

	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);
	
	return widget;
}

void vscrollbar_set_parameters(px_widget* widget)
{
	px_vscrollbar *vscrollbar = &(widget->specific.vscrollbar);
	struct _vscrollbar *internal;
	
	// Initialize common widget default callbacks
	widget->default_callbacks[BUTTON_PRESSED] = vscrollbar_default_button_press;
	widget->default_callbacks[BUTTON_RELEASED] = vscrollbar_default_button_release;
	widget->default_callbacks[MOUSE_LEAVE] = vscrollbar_default_button_release;
	widget->default_callbacks[MOTION] = vscrollbar_default_button_motion;

	// Create internal data structure
	vscrollbar->internal = internal = (struct _vscrollbar *)malloc(sizeof(struct _vscrollbar));
	
	// Initialize internal widget structure variables
	internal->slider_start = 0.0;
	internal->slider_length = 1.0;
	internal->increment = 0.1;

	// Initialize all function pointers
	widget->redraw = vscrollbar_redraw;
	widget->destroy = vscrollbar_destroy;
	vscrollbar->set_slider_position = vscrollbar_set_slider_position;
	vscrollbar->set_slider_length = vscrollbar_set_slider_length;
	vscrollbar->set_increment = vscrollbar_set_increment;
	vscrollbar->intimate_change = NULL;
	vscrollbar->get_slider_position = vscrollbar_get_slider_position;
	vscrollbar->get_increment = vscrollbar_get_increment;
}

Bool vscrollbar_handle_up_arrow(px_widget* widget, px_event* event)
{
	px_vscrollbar *vscrollbar = &(widget->specific.vscrollbar);
	struct _vscrollbar *internal = (struct _vscrollbar *)vscrollbar->internal;
	float nstart;

	nstart = internal->slider_start - internal->increment;

	if(nstart < 0.0) {
		nstart = 0.0;
		if(timer != NULL) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			eventloop->delete_timer(eventloop, timer);
			timer = NULL;
		}
	}

	internal->slider_start = nstart;
	vscrollbar_redraw_slider(widget);
		
	if(vscrollbar->intimate_change != NULL)
		vscrollbar->intimate_change(widget->child_of, internal->slider_start);

	return True;
}

Bool vscrollbar_handle_down_arrow(px_widget* widget, px_event* event)
{
	px_vscrollbar *vscrollbar = &(widget->specific.vscrollbar);
	struct _vscrollbar *internal = (struct _vscrollbar *)vscrollbar->internal;
	float nstart;

	nstart = internal->slider_start + internal->increment;

	if(nstart > 1.0) {
		nstart = 1.0 ;
		if(timer != NULL) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			eventloop->delete_timer(eventloop, timer);
			timer = NULL;
		}
	}

	internal->slider_start = nstart;
	vscrollbar_redraw_slider(widget);
		
	if(vscrollbar->intimate_change != NULL)
		vscrollbar->intimate_change(widget->child_of, internal->slider_start);

	return True;
}
	
Bool vscrollbar_default_button_press(px_widget* widget, px_event* event)
{
	struct _vscrollbar *internal = (struct _vscrollbar *)widget->specific.vscrollbar.internal;
	int slider_start, slider_length;
	px_eventloop* eventloop = eventloop_get_eventloop();

	slider_length = (int)(internal->slider_length*(widget->height-32));
	if(slider_length<24)
		slider_length = 24;
	slider_start = (int)(internal->slider_start*(widget->height-32-slider_length));

	if(event->y >= 0 && event->y <= 16) {	// Up arrow
		sbar_state = VSCROLL_UP;
		if(internal->slider_start > 0.0) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			
			vscrollbar_handle_up_arrow(widget, event);
			timer = eventloop->add_timer(eventloop, 0, VSCROLL_DELAY, vscrollbar_handle_up_arrow, widget, event, True);
		}
		vscrollbar_redraw_top_button(widget);
	} else if(event->y >= (widget->height-16) && event->y <= (widget->height-1)) {	// Down arrow
		sbar_state = VSCROLL_DOWN;
		if(widget->height-16 > slider_start+slider_length+16) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			
			vscrollbar_handle_down_arrow(widget, event);
			timer = eventloop->add_timer(eventloop, 0, VSCROLL_DELAY, vscrollbar_handle_down_arrow, widget, event, True);
		}
		vscrollbar_redraw_bottom_button(widget);
	} else if(event->y > 11 && event->y < slider_start+16) {	// Trough above slider
		sbar_state = VSCROLL_TROUGH;
		if(internal->slider_start > 0) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			
			vscrollbar_handle_up_arrow(widget, event);
			timer = eventloop->add_timer(eventloop, 0, VSCROLL_DELAY, vscrollbar_handle_up_arrow, widget, event, True);
		}
	} else if(event->y > slider_start+slider_length+16 && event->y < widget->height-16) { // Trough below slider
		sbar_state = VSCROLL_TROUGH;
		if(slider_start+slider_length+16 < widget->height-16) {
			px_eventloop* eventloop = eventloop_get_eventloop();
			
			vscrollbar_handle_down_arrow(widget, event);
			timer = eventloop->add_timer(eventloop, 0, VSCROLL_DELAY, vscrollbar_handle_down_arrow, widget, event, True);
		}
	} else {	// Slider
		sbar_state = VSCROLL_SLIDER;
		prev_y = event->y;
	}

	eventloop->grab(eventloop, widget);
	
	return True;
}
		
Bool vscrollbar_default_button_release(px_widget* widget, px_event* event)
{
	px_eventloop* eventloop = eventloop_get_eventloop();

	eventloop->grab(eventloop, NULL);
	
	if(timer != NULL) {
		eventloop->delete_timer(eventloop, timer);
		timer = NULL;
	}

	sbar_state = VSCROLL_NONE;
	prev_y = 0;

	widget->redraw(widget);
	widget->show(widget);

	return True;
}

Bool vscrollbar_default_button_motion(px_widget* widget, px_event* event)
{
	px_vscrollbar* vscrollbar = &(widget->specific.vscrollbar);
	struct _vscrollbar *internal = (struct _vscrollbar *)vscrollbar->internal;
	int fg;
	XEvent xev, xev_copy;

	// Check if within Slider
	if(sbar_state == VSCROLL_SLIDER && event->y >= 16 && event->y <= (widget->height-16)) {
		float nstart;
		int slider_length;

		slider_length = (int)(internal->slider_length*(widget->height-32));
		if(slider_length<24)
			slider_length = 24;
		nstart = (internal->slider_start*(widget->height-32-slider_length)+(event->y-prev_y))/(widget->height-32-slider_length);

		// Make sure that slider is within the designated area
		if(nstart<0.0)
			nstart = 0.0;
		else if(nstart>1.0)
			nstart = 1.0;

		internal->slider_start = nstart;
		vscrollbar_redraw_slider(widget);
		
		if(vscrollbar->intimate_change != NULL)
			vscrollbar->intimate_change(widget->child_of, internal->slider_start);

		prev_y = event->y;
	}

	// Bypass extra motion events
	fg = 0;
	while(XPending(display_get_Xdisplay())) {
		XNextEvent(display_get_Xdisplay(), &xev);
		if(!fg && xev.type==MotionNotify) {
			fg = 1;
			xev_copy = xev;
		}
		if(xev.type==ButtonRelease) {
			fg = 1;
			xev_copy = xev;
			break;
		}
	}
	if(fg)
		XPutBackEvent(display_get_Xdisplay(), &xev_copy);

	return True;
}

void vscrollbar_redraw_top_button(px_widget* widget)
{
	Display* display = display_get_Xdisplay();
	GC gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
	px_pixmap* show_pixmap = NULL;
	px_theme* theme = theme_get_theme();

	if(sbar_state == VSCROLL_UP)
		show_pixmap = theme_get_pixmap(theme, VSCROLLBAR_UP_ACTIVE);
	else
		show_pixmap = theme_get_pixmap(theme, VSCROLLBAR_UP);
	
	XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc, 
				0, 0, 16, 16, 0, 0); 
	
	XClearArea(display, widget->win, 0, 0, 16, 16, False);

	XFreeGC(display, gc);
}

void vscrollbar_redraw_bottom_button(px_widget* widget)
{
	Display* display = display_get_Xdisplay();
	GC gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
	px_pixmap* show_pixmap = NULL;
	px_theme* theme = theme_get_theme();
	
	if(sbar_state == VSCROLL_DOWN)
		show_pixmap = theme_get_pixmap(theme, VSCROLLBAR_DOWN_ACTIVE);
	else
		show_pixmap = theme_get_pixmap(theme, VSCROLLBAR_DOWN);
	
	XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc, 
				0, 0, 16, 16, 0, widget->height - 16); 

	XClearArea(display, widget->win,  0, widget->height-16, 16, 16, False);
	
	XFreeGC(display, gc);
}
	
void vscrollbar_redraw_slider(px_widget* widget)
{
	struct _vscrollbar *internal = (struct _vscrollbar *)widget->specific.vscrollbar.internal;
	Display* display = display_get_Xdisplay();
	px_theme* theme = theme_get_theme();
	px_pixmap *show_pixmap = NULL;
	GC gc, gcxor, gcand;
	int slider_start, slider_length;

	slider_length = (int)(internal->slider_length*(widget->height-32));
	if(slider_length<24) 
		slider_length = 24;
	slider_start = (int)(internal->slider_start*(widget->height-32-slider_length));

	if(widget->use_pixmap && (slider_start+slider_length)<widget->height) {
		// Create required GCs
		gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);

		gcxor = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
		XSetForeground(display, gcxor,WhitePixel(display, DefaultScreen(display)));
		XSetBackground(display, gcxor,WhitePixel(display, DefaultScreen(display)));
		XSetFunction(display, gcxor, GXxor);
		XSetFillStyle(display, gcxor, FillSolid);
		XSetGraphicsExposures(display, gcxor, False);

		gcand = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
		XSetForeground(display, gcand, 0);
		XSetBackground(display, gcand, ~0);
		XSetFunction(display, gcand, GXand);
		XSetFillStyle(display, gcand, FillSolid);
		XSetGraphicsExposures(display, gcand, False);

		// Draw the vscrollbar pixmaps
		show_pixmap = theme_get_pixmap(theme, VSCROLLBAR_TROUGH);
		XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc,
							0, 0, 16, widget->height-32, 0, 16);
		
		show_pixmap = theme_get_pixmap(theme, VSCROLLBAR_SLIDER);
		if(show_pixmap->mask!=None) {		// Transperant Slider
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 0, 16, slider_length/2+1, 0, slider_start+16);
			XCopyPlane(display, show_pixmap->mask, widget->backing_store.pxmp, gcand,
								0, 0, 16, slider_length/2+1, 0, slider_start+16, 1L);
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 0, 16, slider_length/2+1, 0, slider_start+16);

			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 320-slider_length/2, 16, slider_length/2, 0, slider_start+16+slider_length/2);
			XCopyPlane(display, show_pixmap->mask, widget->backing_store.pxmp, gcand,
								0, 320-slider_length/2, 16, slider_length/2, 0, slider_start+16+slider_length/2, 1L);
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 320-slider_length/2, 16, slider_length/2, 0, slider_start+16+slider_length/2);
		} else {
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc,
								0, 0, 16, slider_length/2+1, 0, slider_start+16);
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc,
								0, 320-slider_length/2, 16, slider_length/2, 0, slider_start+16+slider_length/2);
		}

		show_pixmap = theme_get_pixmap(theme, VSCROLLBAR_NOTCH);
		if(show_pixmap->mask!=None) {		// Transperant Slider Notch
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 0, 10, 10, 3, (2*slider_start+slider_length)/2-5+16);
			XCopyPlane(display, show_pixmap->mask, widget->backing_store.pxmp, gcand,
								0, 0, 10, 10, 3, (2*slider_start+slider_length)/2-5+16, 1L);
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor,
								0, 0, 10, 10, 3, (2*slider_start+slider_length)/2-5+16);
		} else {
			XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gc,
								0, 0, 10, 10, 3, (2*slider_start+slider_length)/2-5+16);
		}
		
		XClearArea(display, widget->win, 0, 16, widget->width, widget->height-32, False);

		XFreeGC(display, gcand);
		XFreeGC(display, gcxor);
		XFreeGC(display, gc);
	}
}

void vscrollbar_redraw(px_widget* widget)
{
	// Draw the arrows
	vscrollbar_redraw_top_button(widget);
	vscrollbar_redraw_bottom_button(widget);

	// Draw the slider
	vscrollbar_redraw_slider(widget);

	widget_set_event_mask(widget, widget->emask);
}

void vscrollbar_destroy(px_widget* widget)
{
	widget_destroy(widget);
}

void vscrollbar_set_slider_position(px_widget* widget, float s)
{
	px_vscrollbar* vscrollbar = &(widget->specific.vscrollbar);
	struct _vscrollbar *internal = (struct _vscrollbar *)widget->specific.vscrollbar.internal;

	if(s<0.0)
		internal->slider_start = 0.0;
	else if(s>1.0)
		internal->slider_start = 1.0;
	else
		internal->slider_start = s;
	
	if(vscrollbar->intimate_change != NULL)
		vscrollbar->intimate_change(widget->child_of, internal->slider_start);
}

void vscrollbar_set_slider_length(px_widget* widget, float s)
{
	struct _vscrollbar *internal = (struct _vscrollbar *)widget->specific.vscrollbar.internal;
	
	if(s<0.0)
		internal->slider_length = 0.0;
	else if(s>1.0)
		internal->slider_length = 1.0;
	else
		internal->slider_length = s;
}

void vscrollbar_set_increment(px_widget* widget, float i)
{
	struct _vscrollbar *internal = (struct _vscrollbar *)widget->specific.vscrollbar.internal;

	if(i>=0.0 && i<=1.0)
		internal->increment = i;
}

float vscrollbar_get_slider_position(px_widget* widget)
{
	struct _vscrollbar *internal = (struct _vscrollbar *)widget->specific.vscrollbar.internal;
	
	return internal->slider_start;
}

float vscrollbar_get_increment(px_widget* widget)
{
	struct _vscrollbar *internal = (struct _vscrollbar *)widget->specific.vscrollbar.internal;

	return internal->increment;
}
