/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

// PicoWidget : Vertical Slider
// Copyrights (c) PicoPeta Simputers Pvt. Ltd.
// Programmer: Vishweshwarayya. G. Mathad

#include "vslider.h"
#include "image.h"
#include "widget_factory.h"
#include <X11/extensions/shape.h>
#include <stdlib.h>
#include <string.h>
#include<math.h>
#include "theme.h"
#include "eventloop.h"
#include "display.h"
#include "font.h"

static struct _vslider {
	int vslider_start, vslider_length,min,max;
     Bool vslider_view,min_calib;
     vslider_calibration_side calib;
     char* fontname;
     px_font* font;
};

typedef struct _vslider pw_vslider;
static int vslider_y;
typedef enum _vslider_states 
{
	VSLIDER_SCROLL_NONE, 
	VSLIDER_SCROLL_SLIDER, 
	VSLIDER_SCROLL_TROUGH
} vslider_states;

vslider_states state;
int trough_width,vslider_height,vslider_width;

px_widget* vslider_create(px_widget* parent, int x, int y, int width, int height, const char* tag)
{
	px_widget* widget = (px_widget*)malloc(sizeof(px_widget));
	px_theme* theme = theme_get_theme();
	px_pixmap* show_pixmap = NULL;
	Display* display = display_get_Xdisplay();
	widget->win = XCreateWindow(display, parent->win, x, y, width, height, 0, CopyFromParent, CopyFromParent, CopyFromParent, 0, NULL);
		
	show_pixmap = theme_get_pixmap(theme, VBACK_SLIDER);
	trough_width=show_pixmap->w;
	show_pixmap = theme_get_pixmap(theme, VFORE_SLIDER);
	vslider_height=show_pixmap->h;
	vslider_width=show_pixmap->w;
#ifdef __debug__
	printf("%s(): %p - Created Window with parent window = %lu, x = %d, y = %d, width = %d, height = %d, new window = %lu\n\n", 
			__func__, widget, parent->win, x, y, width, height, widget->win);

	XSync(display, False);
#endif
	// Initialize common widget variables
	widget->x = x;
	widget->y = y;
	widget->height = height;
	widget->width = width;
	widget->child_of = parent;

	if (tag != NULL)
	{
		widget->tag = (char*)malloc(sizeof(char) * (strlen(tag) + 1));
		strcpy(widget->tag, tag);
	}
	else
		widget->tag = NULL;
	widget->wtype = VSLIDER;
	widget->child_of = NULL;

	
	widget->use_pixmap = False;
	widget->background.background_color = (px_color*)malloc(sizeof(px_color));
	widget->background.background_color->type = SOLID;
	widget_xcolor_from_str(&(widget->background.background_color->first_color), "white");


	
	// Initializing backing store
	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);
		
	return widget;
}

void vslider_set_parameters(px_widget* widget)
{
	px_vslider *vslider=&(widget->specific.vslider);
	pw_vslider *internal;
	
	widget->default_callbacks[BUTTON_PRESSED] = vslider_default_button_press;
	widget->default_callbacks[BUTTON_RELEASED] = vslider_default_button_release;
	widget->default_callbacks[MOUSE_LEAVE] = vslider_default_button_release;
	widget->default_callbacks[MOTION] = vslider_default_button_motion;

	widget->redraw = vslider_redraw;
	widget->destroy = vslider_destroy;

	vslider->internal=(pw_vslider *)malloc(sizeof(pw_vslider));
	internal=vslider->internal;

	internal->vslider_start = 15;
	internal->vslider_view=0;
	internal->min_calib=False;
	internal->min = 0;
	internal->max = 100;
	internal->calib=VSLIDER_NONE;
	internal->vslider_length =vslider_height;
	vslider_y = 15;
	internal->fontname  = strdup("-picopeta-simputer-medium-r-normal--11-120-101-108-p-120-fontspecific-0");
	internal->font = font_load_font(internal->fontname);

	// Initalizing widget specific callbacks
	widget->specific.vslider.get_position = vslider_get_position;
	widget->specific.vslider.show_calibration=vslider_show_calibration;
	widget->specific.vslider.show_min_calibration=vslider_show_minor_calibration;
	widget->specific.vslider.set_position = vslider_set_value;
	widget->specific.vslider.show_value=vslider_show_value;
	widget->specific.vslider.set_range=vslider_set_range;
	widget->specific.vslider.set_font = vslider_set_font;
	widget->specific.vslider.intimate_slider_release = NULL;
	widget->specific.vslider.intimate_slider_motion = NULL;
	
}

Bool vslider_default_button_press(px_widget* widget, px_event* event)
{
	pw_vslider* vslider = widget->specific.vslider.internal;	
	if (!(event->x >= ((widget->width/2)-(vslider_width/2)) && event->x <= ((widget->width/2) +(vslider_width/2)) &&
			event->y >= 15 && event->y <= (widget->height - 15)))
	{
		state = VSLIDER_SCROLL_NONE;
		widget->show(widget);

	}
	else if ( event->x >= (widget->width/2-(vslider_width/2)) && event->x <= (widget->width/2 +(vslider_width/2)) && event->y >= 15 && event->y <= (widget->height-15))
	{
		
		int new_start;	
		if(event->y>15 && event->y< (widget->height-15))
		{
			state =VSLIDER_SCROLL_TROUGH;

			if(event->y<= (15+vslider->vslider_length))
				new_start=15;
			else if (event->y>= (widget->height - 15 -vslider->vslider_length))
				new_start=(widget->height -15-vslider->vslider_length);
			else
				new_start=event->y-vslider->vslider_length/2;
			widget->redraw(widget);
			widget->show(widget);
			vslider->vslider_start = new_start;
				
						
			state =VSLIDER_SCROLL_SLIDER;	
			px_eventloop* eventloop = eventloop_get_eventloop();
			eventloop->grab(eventloop, widget);
			vslider_y = event->y;

			widget->redraw(widget);
			widget->show(widget);
		}			
			
	}

	return True;
}

Bool vslider_default_button_release(px_widget* widget, px_event* event)
{
	px_eventloop* eventloop = eventloop_get_eventloop();
	
	eventloop->grab(eventloop, NULL);
	
	if (widget->specific.vslider.intimate_slider_release != NULL)
				widget->specific.vslider.intimate_slider_release(vslider_get_position(widget));
	state = VSLIDER_SCROLL_NONE;
	vslider_y = 15;
	widget->redraw(widget);
	widget->show(widget);

	return True;
}

Bool vslider_default_button_motion(px_widget* widget, px_event* event)
{
	pw_vslider* vslider = widget->specific.vslider.internal;
	int fg=0;
	XEvent xev, xev_copy;
	
	if (state == VSLIDER_SCROLL_SLIDER && event->y >= 15 && event->y <= (widget->height-15))
	{
		{
			int diffy = (event->y - vslider_y);
			int endy = (vslider->vslider_start + vslider->vslider_length);
			
			if ( diffy< 0 && vslider->vslider_start >=15)
			{
				int new_start;
				if (vslider->vslider_start >= (-diffy) && (vslider->vslider_start + diffy) >= 15)
					new_start = vslider->vslider_start + diffy;
				else
					new_start = 15;
				
				widget->redraw(widget);
				widget->show(widget);
				vslider->vslider_start = new_start;
				
				if (widget->specific.vslider.intimate_slider_motion != NULL)
					widget->specific.vslider.intimate_slider_motion(vslider_get_position(widget));
			}
			else if (diffy >= 0 && endy<= (widget->height-15))
			{
				int diff = widget->height-15-endy, new_start;
		
				if (diff >=diffy && (vslider->vslider_start+diffy)<=(widget->height-15))
				{
					new_start = vslider->vslider_start + diffy;
				}
				else
				{
					new_start = (widget->height-15) - vslider->vslider_length;
				}
				
				widget->redraw(widget);
				widget->show(widget);
				vslider->vslider_start = new_start;
				
				if (widget->specific.vslider.intimate_slider_motion != NULL)
					widget->specific.vslider.intimate_slider_motion(vslider_get_position(widget));

			}
			vslider_y = event->y;
		}
	}

	// Bypass extra motion events
	while(XPending(display_get_Xdisplay())) {
		XNextEvent(display_get_Xdisplay(), &xev);
		if(xev.type==MotionNotify) {
			fg = 1;
			xev_copy = xev;
		}
		if(xev.type==ButtonRelease) {
			fg = 1;
			xev_copy = xev;
			break;
		}
	}
	if(fg) {
		XPutBackEvent(display_get_Xdisplay(), &xev_copy);
	}
	
	
	return True;
}

void vslider_redraw(px_widget* widget)
{
	unsigned int starty;
	char buffer[20];
	Display* display = display_get_Xdisplay();
	GC gcxor, gcand;
	px_pixmap* show_pixmap = NULL;
	GC gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
	px_theme* theme = theme_get_theme();
	pw_vslider *internal=widget->specific.vslider.internal;


	if (widget->use_pixmap)
	{
		image_copy_pixmap(widget->win, &(widget->backing_store), widget->background.background_pixmap);
	}
	else
	{
		widget_draw_px_color(&(widget->backing_store), widget->background.background_color);
	}

	gcxor = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
    	XSetForeground(display, gcxor,WhitePixel(display, DefaultScreen(display)));
	XSetBackground(display, gcxor,WhitePixel(display, DefaultScreen(display)));
	XSetFunction(display, gcxor, GXxor);
	XSetFillStyle(display, gcxor, FillSolid);
	XSetGraphicsExposures(display, gcxor, False);

	gcand = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
	XSetForeground(display, gcand, 0);
	XSetBackground(display, gcand, ~0);
	XSetFunction(display, gcand, GXand);
	XSetFillStyle(display, gcand, FillSolid);
	XSetGraphicsExposures(display, gcand, False);
	
	XSetFont(display, gc, internal->font->font->fid);	
	XSetForeground(display, gc, BlackPixel(display, DefaultScreen(display)));
	show_pixmap = theme_get_pixmap(theme, VBACK_SLIDER);
	if(show_pixmap->mask!=None)
	{
		XCopyArea(display,show_pixmap->pxmp, widget->backing_store.pxmp, gcxor, 0, 0,trough_width, widget->height/2-15, widget->width/2-trough_width/2,15);
		XCopyPlane(display,show_pixmap->mask, widget->backing_store.pxmp, gcand, 0, 0, trough_width, widget->height/2-15, widget->width/2-trough_width/2,15,1L);
		XCopyArea(display,show_pixmap->pxmp, widget->backing_store.pxmp, gcxor, 0, 0, trough_width, widget->height/2-15, widget->width/2-trough_width/2,15);
		XCopyArea(display,show_pixmap->pxmp, widget->backing_store.pxmp, gcxor, 0, 320-widget->height/2+15, trough_width, widget->height/2-15, 
						widget->width/2-trough_width/2,widget->height/2);
		XCopyPlane(display,show_pixmap->mask, widget->backing_store.pxmp, gcand, 0, 320-widget->height/2+15, trough_width, widget->height/2-15,
						widget->width/2-trough_width/2,widget->height/2,1L);
		XCopyArea(display,show_pixmap->pxmp, widget->backing_store.pxmp, gcxor, 0, 320-widget->height/2+15, trough_width, widget->height/2-15, 
						widget->width/2-trough_width/2,widget->height/2);
	}
	else
	{
		XCopyArea(display,show_pixmap->pxmp, widget->backing_store.pxmp, gc, 0, 0, trough_width, widget->height/2-15, widget->width/2-trough_width/2,15);
		XCopyArea(display,show_pixmap->pxmp, widget->backing_store.pxmp, gc, 0, 320-widget->height/2+15, trough_width, widget->height/2-15, 
						widget->width/2-trough_width/2,widget->height/2);
	}
		
	

	starty = (internal->vslider_start );
	show_pixmap = theme_get_pixmap(theme, VFORE_SLIDER);
	if(show_pixmap->mask!=None)
	{
		XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor, 0, 0,vslider_width , vslider_height, widget->width/2-vslider_width/2, starty);
		XCopyPlane(display, show_pixmap->mask, widget->backing_store.pxmp, gcand, 0, 0, vslider_width, vslider_height, widget->width/2-vslider_width/2, starty,1L);
		XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor, 0, 0, vslider_width, vslider_height, widget->width/2-vslider_width/2, starty);
	}
	else
		XCopyArea(display, show_pixmap->pxmp, widget->backing_store.pxmp, gcxor, 0, 0, vslider_width, vslider_height, widget->width/2-vslider_width/2, starty);

	if(internal->calib==SLIDER_RIGHT)
	{
		XDrawLine(display, widget->backing_store.pxmp, gc,widget->width/2+trough_width/2+vslider_width/2 , 15 , widget->width/2+trough_width/2+vslider_width/2 , widget->height-16);
		XDrawLine(display, widget->backing_store.pxmp, gc,widget->width/2+trough_width/2+vslider_width/2 , 15 , widget->width/2+trough_width/2+vslider_width/2+10 ,15 );
		if(internal->min_calib==True)
		{
			XDrawLine(display, widget->backing_store.pxmp, gc,widget->width/2+trough_width/2+vslider_width/2 , widget->height/4+ 7 , widget->width/2+trough_width/2+vslider_width/2+10 ,widget->height/4+7 );
			XDrawLine(display, widget->backing_store.pxmp, gc,widget->width/2+trough_width/2+vslider_width/2 ,widget->height-(widget->height)/4-7 , widget->width/2+trough_width/2+vslider_width/2+10 ,widget->height-(widget->height)/4-7  );
			XDrawLine(display, widget->backing_store.pxmp, gc,widget->width/2+trough_width/2+vslider_width/2 , widget->height/2 , widget->width/2+trough_width/2+vslider_width/2+10 , widget->height/2);
		}
		XDrawLine(display, widget->backing_store.pxmp, gc,widget->width/2+trough_width/2+vslider_width/2 , widget->height-16 , widget->width/2+trough_width/2+vslider_width/2+10 ,widget->height-16 );
		sprintf(buffer,"%d",(int)internal->min);
		XDrawString(display,widget->backing_store.pxmp,gc,widget->width/2+trough_width/2+14+vslider_width/2,20,buffer,strlen(buffer));
		sprintf(buffer,"%d",(int)internal->max);
		XDrawString(display,widget->backing_store.pxmp,gc,widget->width/2+trough_width/2+14+vslider_width/2,widget->height-12,buffer,strlen(buffer));
	}
	if(internal->calib==SLIDER_LEFT)
	{
		XDrawLine(display, widget->backing_store.pxmp, gc,widget->width/2-trough_width/2-vslider_width/2 , 15 , widget->width/2-trough_width/2-vslider_width/2 , widget->height-16);
		XDrawLine(display, widget->backing_store.pxmp, gc,widget->width/2-trough_width/2-vslider_width/2-10 , 15 , widget->width/2-trough_width/2-vslider_width/2 ,15 );
		if(internal->min_calib==True)
		{
			XDrawLine(display, widget->backing_store.pxmp, gc,widget->width/2-trough_width/2-vslider_width/2-10 , widget->height/4+ 7 , widget->width/2-trough_width/2-vslider_width/2 ,widget->height/4+7 );
			XDrawLine(display, widget->backing_store.pxmp, gc,widget->width/2-trough_width/2-vslider_width/2-10 ,widget->height-(widget->height)/4-7 , widget->width/2-trough_width/2-vslider_width/2 ,widget->height-(widget->height)/4-7  );
			XDrawLine(display, widget->backing_store.pxmp, gc,widget->width/2-trough_width/2-vslider_width/2-10 , widget->height/2 , widget->width/2-trough_width/2-vslider_width/2 ,widget->height/2 );
		}
		XDrawLine(display, widget->backing_store.pxmp, gc,widget->width/2-trough_width/2-vslider_width/2-10 , widget->height-16 , widget->width/2-trough_width/2-vslider_width/2 ,widget->height-16 );
		sprintf(buffer,"%d",(int)internal->min);
		XDrawString(display,widget->backing_store.pxmp,gc,widget->width/2-trough_width/2-vslider_width/2-font_get_width(internal->font, buffer, strlen(buffer))-14,20,buffer,strlen(buffer));
		sprintf(buffer,"%d",(int)internal->max);
		XDrawString(display,widget->backing_store.pxmp,gc,widget->width/2-trough_width/2-vslider_width/2-font_get_width(internal->font, buffer, strlen(buffer))-14,widget->height-12,buffer,strlen(buffer));
	}
		
	sprintf(buffer,"%d",vslider_get_position(widget));
	if(internal->vslider_view && internal->calib==SLIDER_LEFT)
		XDrawString(display,widget->backing_store.pxmp,gc,widget->width/2+trough_width/2+10,internal->vslider_start+vslider_height/2+5,buffer,strlen(buffer));
	else if(internal->vslider_view && internal->calib==SLIDER_RIGHT)
		XDrawString(display,widget->backing_store.pxmp,gc,widget->width/2-trough_width/2-font_get_width(internal->font, buffer, strlen(buffer))-10,internal->vslider_start+vslider_height/2+5,buffer,strlen(buffer));
	
	XClearWindow(display, widget->win);
		
	widget_set_event_mask(widget, widget->emask);
	
	XFreeGC(display, gc);
	XSync(display, False);
#ifdef __debug__
	XSync(display, False);
	printf("%s() : %p\n", __func__, widget);
#endif
}

void vslider_destroy(px_widget* widget)
{
	widget_destroy(widget);
}

void vslider_set_range(px_widget* widget,int min, int max)
{
	pw_vslider *internal=widget->specific.vslider.internal;
	if(min<max)
	{
		internal->min=min;
		internal->max=max;
	}
}
void vslider_show_calibration(px_widget* widget,vslider_calibration_side side)
{
	pw_vslider *internal=widget->specific.vslider.internal;
	internal->calib=side;
}
void vslider_show_minor_calibration(px_widget* widget,Bool show)
{
	pw_vslider *internal=widget->specific.vslider.internal;
	internal->min_calib=show;
}
void vslider_show_value(px_widget* widget, Bool state)
{
	pw_vslider *internal=widget->specific.vslider.internal;
	internal->vslider_view=state;
}
void vslider_set_value(px_widget* widget, int value)
{
	int vslider_length, vslider_max, vslider_min;
	pw_vslider *internal=widget->specific.vslider.internal;

	vslider_length=internal->vslider_length;
	vslider_max=internal->max;
	vslider_min=internal->min;
	
	if(value>=vslider_min && value<=vslider_max)
	{
		internal->vslider_start=(value-vslider_min)*((float)((widget->height-30)-vslider_length)/(float)(vslider_max-vslider_min))+15;
	}
	else if (value<vslider_min)
		internal->vslider_start=15;
	else if (value>vslider_max)
		internal->vslider_start=(widget->height-15)-vslider_length;

}

int vslider_get_position(px_widget* widget)
{
	int vslider_length, vslider_max, vslider_min;
	float pos;
	pw_vslider *internal=widget->specific.vslider.internal;

	vslider_length=internal->vslider_length;
	vslider_max=internal->max;
	vslider_min=internal->min;
	
	pos=(internal->vslider_start-15)/((float)((widget->height-30)-vslider_length)/(float)(vslider_max-vslider_min));
	pos=ceilf(pos);
	
	pos=pos+vslider_min;
	return(pos);
}
void vslider_set_font(px_widget* widget, const char* fontname)
{
	pw_vslider *internal=widget->specific.vslider.internal;
	if(internal->fontname == NULL) {
		internal->fontname = (char*)malloc(sizeof(char) * (strlen(fontname) + 1));
	} else {
		font_free_font(internal->fontname);	
	}
	internal->font = font_load_font(fontname);
}
