/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include "widget_factory.h"
#include "frame.h"
#include "toplevel.h"
#include "button.h"
#include "checkbutton.h"
#include "radiobutton.h"
#include "label.h"
#include "listbox.h"
#include "scrollbar.h"
#include "hscrollbar.h"
#include "vscrollbar.h"
#include "picowidget.h"
#include "image.h"
#include "entry.h"
#include "tabs.h"
#include "text.h"
#include "vslider.h"
#include "hslider.h"
#include "font.h"
#include "eventloop.h"
#include "display.h"
#include "theme.h"
#include "progressbar.h"
#include "txtlb.h"
#include "sketch.h"

#include <stdio.h>
#include <stdlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>
#include <string.h>
#include <ctype.h>

px_display* get_display()
{
	return display_get_display();
}

px_widget *new_widget(widget_type wtype, px_widget *parent, int x, int y, 
									int width, int height, char *tag)
{
	px_widget* widget = NULL;

	switch(wtype) {
		case TOPLEVEL:
			widget = toplevel_create(x, y, width, height, tag);
			break;
		case BUTTON:
			widget = button_create(parent, x, y, width, height, tag);
			break;
		case CHECKBUTTON:
			widget = checkbutton_create(parent, x, y, width, height, tag);
			break;
		case RADIOBUTTON:
			widget = radiobutton_create(parent, x, y, width, height, tag);
			break;
		case LABEL:
			widget = label_create(parent, x, y, width, height, tag);
			break;
		case FRAME:
			widget = frame_create(parent, x, y, width, height, tag);
			break;
		case LISTBOX:
			widget = listbox_create(parent, x, y, width, height, tag);
			break;
		case SCROLLBAR:
			widget = scrollbar_create(parent, x, y, width, height, tag);
			break;
		case ENTRY:
			widget = entry_create(parent, x, y, width, height, tag);
			break;
		case TEXT:
			widget = text_create(parent, x, y, width, height, tag);
			break;
		case VSLIDER:
			widget = vslider_create(parent, x, y, width, height, tag);
			break;
		case HSLIDER:
			widget = hslider_create(parent, x, y, width, height, tag);
			break;
		case HSCROLLBAR:
			widget = hscrollbar_create(parent, x, y, width, height, tag);
			break;
		case VSCROLLBAR:
			widget = vscrollbar_create(parent, x, y, width, height, tag);
			break;
		case PROGRESSBAR:
			widget = progressbar_create(parent, x, y, width, height, tag);
			break;
		case TAB:
			widget = tab_create(parent, x, y, width, height, tag);
			break;
		case TXTLB:
			widget = txtlb_create(parent,x,y,width,height,tag);
			break;
		case SKETCH:
			widget = sketch_create(parent,x,y,width,height,tag);
			break;
		case COMBOBOX:
			widget = combobox_create(parent,x,y,width,height,tag);
			break;
		default :
			fprintf(stderr, "Type not supported\n");
			break;

	}

	if(widget != NULL) {
		widget->child_of = NULL;
		widget_set_common_parameters(widget);
		widget_set_specific_parameters(widget);

		if(parent!=NULL) {
			switch(parent->wtype) {
				case FRAME:
					parent->specific.frame.add_child(parent, widget);
					break;

				case TOPLEVEL:
					parent->specific.toplevel.add_child(parent, widget);
					break;
				case LISTBOX:
					break;
				default :
#ifdef __debug__
					printf("Are you sure the parent %p can take children ? \n", parent);
#endif
					break;
			}
			widget->child_of = parent;
		}
	}

	return widget;
}

void toplevel_frame_destroy(px_widget* widget)
{
	px_eventloop* eventloop = eventloop_get_eventloop();
	frame_destroy(widget);

	eventloop->remove_toplevel(eventloop, widget);
}

	
px_widget* new_toplevel_frame(Window win, int x, int y, int width, int height, const char* tag)
{
	px_widget* widget = _frame_create(win, x, y, width, height, tag);
	Display* display = display_get_Xdisplay();

	if (widget != NULL)
	{
		px_eventloop* eventloop = eventloop_get_eventloop();
		widget->child_of = NULL;
		widget_set_common_parameters(widget);
		widget_set_specific_parameters(widget);

		widget->destroy = toplevel_frame_destroy;

		XSelectInput(display, widget->win, 
			   ExposureMask |
			   ButtonPressMask |
			   ButtonReleaseMask |
			   Button1MotionMask |
			   StructureNotifyMask |
			   VisibilityChangeMask |
			   KeyPressMask | 
			   KeyReleaseMask |
			   SubstructureRedirectMask |
			   SubstructureNotifyMask
		);
		eventloop->register_toplevel(eventloop, widget);
	}

	return widget;
}

void widget_set_common_parameters(px_widget* widget)
{
	unsigned int count;
	
	widget->depth = 0;
//	Not setting widget->wid
//	widget->tag set when the widget is created.
//	widget->x, widget->y, widget->width, widget->height set when widget is created.
//	widget->win set when widget is created
//	widget->child_of set in new_widget()
//	widget->wtype set when widget is created.
//	widget->backing_store set when widget is created.

	widget->emask = EVENT_SOLID;
	widget->mask_for_event = NULL;
	widget->hidden = False;
	widget->state = NORMAL;
	widget->have_focus = False;
	widget->border = NO_BORDER;
	widget->border_color = NULL;
	
	widget->condemned = False;
	widget->preserve_internal_structures = False;
	widget->dirty = False;

//	widget->specific set in widget_set_specific_parameters
//	widget->use_pixmap set when widget is created
//	widget->background set when widget is created

	for (count = BUTTON_PRESSED; count < NR_CALLBACKS;count++)
	{
		widget->callbacks[count] = NULL;
		widget->default_callbacks[count] = NULL;
		widget->callback_parameters[count] = NULL;
	}
	widget->window_delete_callback = NULL;
	widget->window_delete_callback_parameter = NULL;
	for (count = BUTTON_PRESSED; count < NR_EXTRA_CALLBACKS;count++)
	{
		widget->extra_callbacks[count] = NULL;
		widget->extra_callbacks_parameters[count] = NULL;
	}

	widget->within = widget_within;
	widget->is_transparent = widget_is_transparent;
	widget->handle_event = widget_handle_event;
	
	widget->redraw = NULL;
	widget->show = widget_show;
	widget->hide = widget_hide;

	widget->destroy = widget_destroy;
	widget->destruct = widget_destruct;
	
	widget->set_size = widget_set_size;
	widget->set_position = widget_set_position;
	widget->get_size = widget_get_size;
	widget->get_position = widget_get_postion;

	widget->get_absolute_x = widget_get_absolute_x;
	widget->get_absolute_y = widget_get_absolute_y;

	widget->set_event_mask = widget_set_event_mask;

	widget->set_background_pixmap_from_file = widget_set_background_pixmap_from_file;
	widget->set_background_pixmap_from_pixmap = widget_set_background_pixmap_from_pixmap;
	widget->set_background_color = widget_set_background_color;

	widget->set_state = widget_set_state;
	widget->get_state = widget_get_state;

	widget->set_focus = widget_set_focus;

	widget->set_border_color = widget_set_border_color;
	widget->set_border_type = widget_set_border_type;

	widget->raise = widget_raise;
	widget->lower = widget_lower;

	widget->cleanup_internal_structures = NULL;
}

void widget_set_specific_parameters(px_widget* widget)
{
	switch (widget->wtype)
	{
		case TOPLEVEL :
			toplevel_set_parameters(widget);
			break;

		case FRAME :
			frame_set_parameters(widget);
			break;

		case BUTTON :
			button_set_parameters(widget);
			break;

		case RADIOBUTTON :
			radiobutton_set_parameters(widget);
			break;

		case CHECKBUTTON :
			checkbutton_set_parameters(widget);
			break;

		case LABEL :
			label_set_parameters(widget);
			break;

		case ENTRY :
			entry_set_parameters(widget);
			break;

		case TEXT:
			text_set_parameters(widget);
			break;

      case VSLIDER :
         vslider_set_parameters(widget);
         break;

       case HSLIDER :
         hslider_set_parameters(widget);
         break;

		case LISTBOX :
			listbox_set_parameters(widget);
			break;

		case HSCROLLBAR:
			hscrollbar_set_parameters(widget);
			break;
			
		case VSCROLLBAR:
			vscrollbar_set_parameters(widget);
			break;

		case PROGRESSBAR :
			progressbar_set_parameters(widget);
			break;

		case SCROLLBAR :
			scrollbar_set_parameters(widget);
			break;
		case TAB :
			tab_set_parameters(widget);
			break; 
		case TXTLB:
			txtlb_set_parameters(widget);
			break;
		case SKETCH:
			sketch_set_parameters(widget);
			break;
		case COMBOBOX:
			widget = combobox_set_parameters(widget);
			break;
		default:
			break;
	}
}

void widget_set_wrap_type(px_widget* widget, wrap_types wrap)
{
	switch (widget->wtype)
	{
		case BUTTON:
			widget->specific.button.layout.wrap = wrap;
			widget_compute_textlayout(&(widget->specific.button.layout), widget);
			break;

		case CHECKBUTTON:
			widget->specific.checkbutton.layout.wrap = wrap;
			widget_compute_textlayout(&(widget->specific.checkbutton.layout), widget);
			break;

		case RADIOBUTTON:
			widget->specific.radiobutton.layout.wrap = wrap;
			widget_compute_textlayout(&(widget->specific.radiobutton.layout), widget);
			break;

		case LABEL:
			widget->specific.label.layout.wrap = wrap;
			widget_compute_textlayout(&(widget->specific.label.layout), widget);
			break;

		default:
			break;
	}
}

void widget_size_depends_on_contents(px_widget* widget, Bool value)
{
	switch (widget->wtype)
	{
		case BUTTON:
			widget->specific.button.change_size = value;
			break;

		case CHECKBUTTON:
			widget->specific.checkbutton.change_size = value;
			break;

		case RADIOBUTTON:
			widget->specific.radiobutton.change_size = value;
			break;

		case LABEL:
			widget->specific.label.change_size = value;
			break;

		default:
			break;
	}
}

void generate_map_events(px_xynode* node, px_widget* widget, Bool is_major)
{
	if (!is_major)
	{
		int count;

		for (count = 0; count < node->no_of_widgets && node->widget_list[count] != widget;count++);

		if (count < node->no_of_widgets - 1)
		{
			int ctr;

			for (ctr = count+1; ctr < node->no_of_widgets;ctr++)
			{
				if (!node->widget_list[ctr]->hidden)
				{
					px_event event;
					XEvent xev;
					Display* display = display_get_Xdisplay();

					xev.xexpose.type = Expose;
					xev.xexpose.window = node->widget_list[ctr]->win ;
					xev.xexpose.x = node->widget_list[ctr]->x;
					xev.xexpose.y = node->widget_list[ctr]->y;
					xev.xexpose.width = node->widget_list[ctr]->width;
					xev.xexpose.height = node->widget_list[ctr]->height;
					xev.xexpose.count = 0;
					xev.xexpose.display = display;
					event.xev = &xev;

					event.type = EXPOSE;
					event.private_data = NULL;
					event.x = 0;
					event.y = 0;

					node->widget_list[ctr]->handle_event(node->widget_list[ctr], &event);
				}
			}
		}
	}
}

void widget_destruct(px_widget* widget)
{
	px_eventloop* eventloop = eventloop_get_eventloop();
	px_widget* parent = widget->child_of;
	eventloop->condemn(eventloop, widget);
	widget->condemned = True;
	if(parent) { // A check to find whether, it is main toplevel or not
		switch (parent->wtype)
		{
			case FRAME:
				parent->specific.frame.remove_child(parent, widget);
				break;

			case TOPLEVEL:
				parent->specific.toplevel.remove_child(parent, widget);
				break;
			case LISTBOX:
				break;
			default :
#ifdef __debug__
				printf("Are you sure the parent %p can take children ? \n", parent);
#endif
				break;
		}
	} else {
		if(widget->wtype==TOPLEVEL) {
			if(widget->specific.toplevel.title) {
				free(widget->specific.toplevel.title);
			}
		}
	}
}

void widget_show(px_widget* widget)
{
	Display* display = display_get_Xdisplay();
#ifdef __debug__
	printf("%s() : %p : type = %d : Start\n", __func__, widget, widget->wtype);
#endif

	XSetWindowBackgroundPixmap(display, widget->win, widget->backing_store.pxmp);

	if (widget->backing_store.mask != None)
		XShapeCombineMask (display, widget->win, ShapeBounding, 0, 0, 
			widget->backing_store.mask, ShapeSet);

	XMapWindow(display, widget->win);
	widget->hidden = False;
#ifdef __debug__
	XSync(display, False);
	printf("%s() : %p : End\n", __func__, widget);
#endif
	
}

void widget_hide(px_widget* widget)
{
	Display* display = display_get_Xdisplay();
	XUnmapWindow(display, widget->win);
	widget->hidden = True;

	if (widget->child_of != NULL && widget->child_of->wtype != LISTBOX)
	{
		px_widget* parent = widget->child_of;
		px_xynode* ynodes = NULL;
		MajorAxis axis;
		
		switch (parent->wtype)
		{
			case FRAME:
				ynodes = parent->specific.frame.ynodes;
				axis = parent->specific.frame.axis;
				break;

			case TOPLEVEL:
				ynodes = parent->specific.toplevel.ynodes;
				axis = parent->specific.toplevel.axis;
				break;

			default:
				break;
		}

	//	XSync(display, False);
	//	widget_get_xynode_with_child(ynodes, widget, &generate_map_events, 
	//			axis == YAXIS ? YDIRECTION : XDIRECTION, axis);
	}
}

void widget_destroy(px_widget* widget)
{

	Display* display = display_get_Xdisplay();
	if (widget->tag != NULL)
		free(widget->tag);

	image_destroy(&(widget->backing_store));

	if (widget->use_pixmap == True && widget->background.background_pixmap != NULL)
	{
		image_destroy(widget->background.background_pixmap);
		free(widget->background.background_pixmap);
	}
	else if (widget->use_pixmap == False && widget->background.background_color != NULL)
	{
		free(widget->background.background_color);
	}

	if (widget->mask_for_event != NULL)
	{
		XDestroyImage(widget->mask_for_event);
	}

	XDestroyWindow(display, widget->win);
}

void widget_set_size(px_widget* widget, unsigned long width, unsigned long height)
{
	Display* display = display_get_Xdisplay();
	Bool position_changed = False;
	px_xynode* ynodes = NULL;
	MajorAxis axis;
	XWindowAttributes attr;

	if (widget->width == width && widget->height == height)
		return;

	if (widget->child_of != NULL && widget->child_of->wtype != LISTBOX)
	{
		px_widget* parent = widget->child_of;
		switch (parent->wtype)
		{
			case FRAME:
				ynodes = parent->specific.frame.ynodes;
				axis = parent->specific.frame.axis;
				break;

			case TOPLEVEL:
				ynodes = parent->specific.toplevel.ynodes;
				axis = parent->specific.toplevel.axis;
				break;

			default:
				break;
		}
		
		ynodes = widget_remove_child(ynodes, widget, axis == YAXIS ? YDIRECTION : XDIRECTION, axis);
		position_changed = True;
	}

	widget->width = width;
	widget->height = height;
	image_destroy(&(widget->backing_store));

	image_initialize_pixmap(&(widget->backing_store));
	image_create_pixmap(widget->win, &(widget->backing_store), width, height);
	
	XGetWindowAttributes(display, widget->win, &attr);

	if (attr.width != width || attr.height != height)
		XResizeWindow(display, widget->win, width, height);
	
	XSetWindowBackgroundPixmap(display, widget->win, widget->backing_store.pxmp);
	
	if (position_changed)
	{
		px_widget* parent = widget->child_of;
		
		ynodes = widget_add_child(ynodes, widget, axis == YAXIS ? YDIRECTION : XDIRECTION, axis);
		switch (parent->wtype)
		{
			case FRAME:
				parent->specific.frame.ynodes = ynodes;
				break;

			case TOPLEVEL:
				parent->specific.toplevel.ynodes = ynodes;
				break;

			default:
				break;
		}

	}
}

void widget_set_position(px_widget* wid, int x, int y)
{
	Display* display = display_get_Xdisplay();
	Bool position_changed = False;
	px_xynode* ynodes;
	MajorAxis axis;

	if (wid->x == x && wid->y == y)
		return;
	
	if (wid->child_of != NULL && wid->child_of->wtype != LISTBOX)
	{
		px_widget* parent = wid->child_of;
		switch (parent->wtype)
		{
			case FRAME:
				ynodes = parent->specific.frame.ynodes;
				axis = parent->specific.frame.axis;
				break;

			case TOPLEVEL:
				ynodes = parent->specific.toplevel.ynodes;
				axis = parent->specific.toplevel.axis;
				break;

			default:
				break;
		}
		
		ynodes = widget_remove_child(ynodes, wid, axis == YAXIS ? YDIRECTION : XDIRECTION, axis);
#ifdef __debug__
		printf("-------- For toplevel/frame %p after removing widget %p who has x = %d, y = %d, width = %d, height = %d--------------\n", parent, wid, wid->x, wid->y, wid->width, wid->height);
	widget_print_frame_buffer(ynodes, axis == YAXIS ? YDIRECTION : XDIRECTION, axis);
	printf("-------- ENd For toplevel/frame %p --------------\n", parent);
#endif
		position_changed = True;
	}
			
	wid->x = x;
	wid->y = y;

	XMoveWindow(display, wid->win, x, y);

	if (position_changed)
	{
		px_widget* parent = wid->child_of;
		
		ynodes = widget_add_child(ynodes, wid, axis == YAXIS ? YDIRECTION : XDIRECTION, axis);

		switch (parent->wtype)
		{
			case FRAME:
				parent->specific.frame.ynodes = ynodes;
				break;

			case TOPLEVEL:
				parent->specific.toplevel.ynodes = ynodes;
				break;

			default:
				break;
		}
#ifdef __debug__
		printf("-------- For toplevel/frame %p after adding widget %p who has x = %d, y = %d, width = %d, height = %d--------------\n", parent, wid, wid->x, wid->y, wid->width, wid->height);
	widget_print_frame_buffer(ynodes, axis == YAXIS ? YDIRECTION : XDIRECTION, axis);
	printf("-------- ENd For toplevel/frame %p --------------\n", parent);
#endif
	}
	
}

void widget_get_size(const px_widget* wid, unsigned long* width, unsigned long* height)
{
	(*width) = wid->width;
	(*height) = wid->height;
}

void widget_get_postion(const px_widget*wid, int* x, int* y)
{
	(*x) = wid->x;
	(*y) = wid->y;
}

Bool widget_within(const px_widget* widget, px_event* event)
{
	if (event->xev->type == ButtonPress || event->xev->type == ButtonRelease || event->xev->type == MotionNotify)
	{

	//	printf ("%s : Abs x = %d, Abs y = %d, width = %d, height = %d, x = %d, y = %d, Pixel = %d\n", widget->tag, x, y, widget->width, widget->height, event->xev->xmotion.x, event->xev->xmotion.y, 
	//			widget->mask_for_event == NULL ? -1 : XGetPixel(widget->mask_for_event, event->xev->xmotion.x - x, event->xev->xmotion.y - y));
		
		if (widget->emask != EVENT_TRANSPARENT && 
				event->x >= widget->x && event->x <= (widget->x + widget->width-1) &&
				event->y >= widget->y && event->y <= (widget->y + widget->height-1))
		{
			if (widget->emask == EVENT_SOLID || (widget->emask == EVENT_MASK &&   // Checking if pixel is transparent
						widget->mask_for_event != NULL &&
						XGetPixel(widget->mask_for_event, 
							event->x - widget->x, event->y - widget->y)))
			{
				return True;
			}
		}
		return False;
	}
	return True;
}

Bool widget_is_transparent(const px_widget* widget, int x, int y)
{
	if (x < 0 || x >= widget->width || y < 0 || y >= widget->height)
		return True;

			
	if (widget->mask_for_event != NULL)
	{
		register char byte = widget->mask_for_event->data[y * widget->mask_for_event->bytes_per_line + x / 8];
		register char mask = 0x80 >> x%8;
	
//		printf("Checking byte no. %d for x = %d, y = %d, bytes_pre_line = %d, byte = %x, mask = %x\n", y * widget->mask_for_event->bytes_per_line + x / 8, x, y, widget->mask_for_event->bytes_per_line, byte, mask);
		if (!(byte & mask))
			return True;
	}
	
	return False;
}

Bool toplevel_frame_is_transparent(const px_widget* widget, int x, int y)
{
	if (widget->mask_for_event != NULL && XGetPixel(widget->mask_for_event, x, y))
	{
		px_xynode* ynodes = NULL, *node;
		MajorAxis axis;

		if (widget->wtype == FRAME)
		{
			ynodes = widget->specific.frame.ynodes;
			axis = widget->specific.frame.axis;
		}
		else if (widget->wtype == TOPLEVEL)
		{
			ynodes = widget->specific.toplevel.ynodes;
			axis = widget->specific.toplevel.axis;
		}
		else
		{
			return True;
		}
			
			
		node = widget_get_xynode(ynodes, x, y, axis == YAXIS ? YDIRECTION : XDIRECTION, axis);

		if (node != NULL)
		{
			unsigned int count;
			
			for (count = 0;count < node->no_of_widgets;count++)
			{
				if (!node->widget_list[count]->hidden && 
						!node->widget_list[count]->is_transparent(node->widget_list[count], x, y))
					return False;
			}
		}
		return True;
	}
	
	return False;
}

int widget_get_absolute_x(const px_widget* widget, Window window)
{
	px_widget* parent = widget;
	int x = 0;
	
	while (parent != NULL && parent->win != window)
	{
		x += parent->x;
		parent = parent->child_of;
	}

	return x;
}

int widget_get_absolute_y(const px_widget* widget, Window window)
{
	px_widget* parent = widget;
	int y = 0;
	
	while (parent != NULL && parent->win != window)
	{
		y += parent->y;
		parent = parent->child_of;
	}

	return y;
}

void widget_set_event_mask(px_widget* widget, event_masks e)
{
	Display* display = display_get_Xdisplay();
	if (widget->mask_for_event != NULL)
		XDestroyImage(widget->mask_for_event);
		
	if (widget->backing_store.mask != None)
	{
		widget->mask_for_event = XGetImage(display, widget->backing_store.mask, 0, 0, 
			widget->width, widget->height, 1, XYPixmap);
	}
	else
	{
		widget->mask_for_event = NULL;
	}

	widget->emask = e;
}

Bool widget_handle_event(px_widget* widget, px_event* event)
{
	Bool return_value = True;
	if (!widget->condemned && (event->type == EXPOSE || (*widget->get_state)(widget) != DISABLED))
	{
		Bool changed_private_data = False;
		
		if (event->private_data == NULL)
		{
			event->private_data = widget->callback_parameters[event->type];
			changed_private_data = True;
		}

		if (widget->default_callbacks[event->type] != NULL)
			return_value = widget->default_callbacks[event->type] (widget, event);

		if (return_value && widget->callbacks[event->type] != NULL)
			return_value = widget->callbacks[event->type](widget, event);

		if (changed_private_data)
			event->private_data = NULL;

		if (widget->dirty && widget->cleanup_internal_structures != NULL)
			widget->cleanup_internal_structures(widget);
		widget->dirty = False;
	}
	return return_value;
}

void widget_draw_px_color(px_pixmap* dst, const px_color* col)
{
	Display* display = display_get_Xdisplay();
	GC gc, mask_gc;
	gc = XCreateGC(display, dst->pxmp, 0 , NULL);
	XSetGraphicsExposures(display, gc, False);
	
	if (col->type == SOLID)
	{
		XSetForeground(display, gc, col->first_color.pixel);
		XFillRectangle(display, dst->pxmp, gc, 0, 0, dst->w, dst->h);
	}
	else if (col->type == GRADIENTHOR)
	{
		int no_of_regions = 20, width, height;
		int count;
		unsigned short startr, startg, startb, endr, endg, endb;
		float dr, dg, db; 
		
		if (dst->w < no_of_regions)
			no_of_regions = dst->w;

		startr = col->first_color.red;
		startg = col->first_color.green;
		startb = col->first_color.blue;
			
		endr = col->second_color.red;
		endg = col->second_color.green;
		endb = col->second_color.blue;

		dr = (float)(endr - startr)/(float)no_of_regions;
		dg = (float)(endg - startg)/(float)no_of_regions;
		db = (float)(endb - startb)/(float)no_of_regions;

		width = dst->w / no_of_regions;
		height = dst->h;
			
		for (count = 0; count < no_of_regions; count++)
		{
			unsigned long colour;

			colour = ((startr + (unsigned short)(dr * count)) << 16) +
				((startg + (unsigned short)(dg * count)) << 8) +
				(startb + (unsigned short)(db * count));
			
			XSetForeground(display, gc, colour);
			XFillRectangle(display, dst->pxmp, gc, 
						width * count, 0, 
						width, height);
		}
	}
	else if (col->type == GRADIENTVERT)
	{
		int no_of_regions = 20, width, height;
		int count;
		unsigned short startr, startg, startb, endr, endg, endb;
		float dr, dg, db; 
		
		if (dst->h < no_of_regions)
			no_of_regions = dst->h;

		startr = col->first_color.red;
		startg = col->first_color.green;
		startb = col->first_color.blue;
			
		endr = col->second_color.red;
		endg = col->second_color.green;
		endb = col->second_color.blue;

		dr = (float)(endr - startr)/(float)no_of_regions;
		dg = (float)(endg - startg)/(float)no_of_regions;
		db = (float)(endb - startb)/(float)no_of_regions;

		width = dst->w;
		height = dst->h / no_of_regions;
			
		for (count = 0; count < no_of_regions; count++)
		{
			unsigned long colour;

			colour = ((startr + (unsigned short)(dr * count)) << 16) +
				((startg + (unsigned short)(dg * count)) << 8) +
				(startb + (unsigned short)(db * count));
				
			XSetForeground(display, gc, colour);
			XFillRectangle(display, dst->pxmp, gc, 
					0, height * count, 
					width, height);
		}
	}
	
	if (col->type == TRANSPARENT)
	{
		if (dst->mask == None)
			dst->mask = XCreatePixmap(display, DefaultRootWindow(display), dst->w, dst->h, 1);
		mask_gc = XCreateGC(display, dst->mask, 0 , NULL);
		XSetGraphicsExposures(display, mask_gc, False);
		XSetForeground(display, mask_gc, 0);

		XFillRectangle(display, dst->mask, mask_gc, 0, 0, dst->w, dst->h);
		XFreeGC(display, mask_gc);
	} 
	else if (dst->mask != None)
	{
		//XFreePixmap(display, dst->mask);
		//dst->mask = None;
		mask_gc = XCreateGC(display, dst->mask, 0 , NULL);
		XSetGraphicsExposures(display, mask_gc, False);
		XSetForeground(display, mask_gc, 1);

		XFillRectangle(display, dst->mask, mask_gc, 0, 0, dst->w, dst->h);
		XFreeGC(display, mask_gc);
		
	}
	XFreeGC(display, gc);

}

int widget_xcolor_from_str(XColor *col, const char *defstr)
{
  	char *str;
  	const char delim[] = ",:";
  	char *token;
  	XColor exact;
	Display* display = display_get_Xdisplay();

  	if ((strchr(defstr, delim[0]) != NULL) || (strchr(defstr, delim[1]) != NULL) )
  	{
  		str = strdup(defstr);
     		token = strsep (&str, delim); 
     		col->red = ( atoi(token) * 65535 ) / 255; 
     		token = strsep (&str, delim); 
     		col->green = ( atoi(token) * 65535 ) / 255;
     		token = strsep (&str, delim); 
     		col->blue = ( atoi(token) * 65535 ) / 255;
		free(str);

     		return XAllocColor(display,DefaultColormap(display, DefaultScreen(display)),col);
  	} 
	else 
	{
          	return XAllocNamedColor(display,DefaultColormap(display, DefaultScreen(display)), defstr, col, &exact);
  	}
}

void widget_set_background_pixmap_from_file(px_widget* widget, char* path)
{
	if (widget->use_pixmap == False)
	{
		free(widget->background.background_color);
		widget->background.background_pixmap = NULL;
	}
	
	if (widget->background.background_pixmap == NULL)
	{
		widget->background.background_pixmap = (px_pixmap*)malloc(sizeof(px_pixmap));
	
		image_initialize_pixmap(widget->background.background_pixmap);
	}
	
	image_load_file(widget->win, widget->background.background_pixmap, path);

	widget->use_pixmap = True;
}

void widget_set_background_pixmap_from_pixmap(px_widget* widget, const px_pixmap* pix)
{
	if (widget->use_pixmap == False)
	{
		free(widget->background.background_color);
		widget->background.background_pixmap = NULL;
	}
	
	if (widget->background.background_pixmap == NULL)
	{
		widget->background.background_pixmap = (px_pixmap*)malloc(sizeof(px_pixmap));
		image_initialize_pixmap(widget->background.background_pixmap);
	} 
	else if (pix->w != widget->background.background_pixmap->w || 
		pix->h != widget->background.background_pixmap->h)
	{
		image_destroy(widget->background.background_pixmap);
	}
	
	image_copy_pixmap(widget->win, widget->background.background_pixmap, pix);

	widget->use_pixmap = True;
}

void widget_set_background_color(px_widget* widget, const px_color* col)
{
	if (widget->use_pixmap == True)
	{
		image_destroy(widget->background.background_pixmap);
		
		free(widget->background.background_pixmap);
		widget->background.background_color = NULL;
	}
	
	widget->use_pixmap = False;

	if (widget->background.background_color == NULL)
	{
		widget->background.background_color = (px_color*)malloc(sizeof(px_color));
	}
	
	(*widget->background.background_color) = (*col);
}

void widget_set_border_color(px_widget* widget, const char* colour)
{
	if (widget->border_color == NULL)
	{
		widget->border_color = (XColor*)malloc(sizeof(XColor));
	}
	widget_xcolor_from_str(widget->border_color, colour);
}

void widget_set_border_type(px_widget* widget, border_types b)
{
	widget->border = b;
}

void raise_the_widget(px_xynode* node, px_widget* widget, Bool is_major)
{

	int count, ctr;

	for (count = 0;count < node->no_of_widgets && node->widget_list[count] != widget;count++);
	
	if (count > 0 && count < node->no_of_widgets)
	{
		for (ctr = count; ctr > 0; ctr--)
			node->widget_list[ctr] = node->widget_list[ctr-1];

		node->widget_list[0] = widget;
	}
}

void widget_raise(px_widget* widget)
{
	px_xynode* ynodes = NULL;
	MajorAxis axis;
	Display* display = display_get_Xdisplay();

	if (widget->child_of != NULL && widget->child_of->wtype != LISTBOX)
	{
		px_widget* parent = widget->child_of;
		switch (parent->wtype)
		{
			case FRAME:
				ynodes = parent->specific.frame.ynodes;
				axis = parent->specific.frame.axis;
				break;

			case TOPLEVEL:
				ynodes = parent->specific.toplevel.ynodes;
				axis = parent->specific.toplevel.axis;
				break;

			default:
				break;
		}
	
		widget_get_xynode_with_child(ynodes, widget, &raise_the_widget, 
				axis == YAXIS ? YDIRECTION : XDIRECTION, axis);
	}
			
	XRaiseWindow(display, widget->win);
}

void lower_the_widget(px_xynode* node, px_widget* widget, Bool is_major)
{
	int count, ctr;
	int less_by_one;
			
	less_by_one = node->no_of_widgets - 1;

	for (count = 0;count < node->no_of_widgets && node->widget_list[count] != widget;count++);

	if (count < less_by_one)
	{
		for (ctr = count; ctr < less_by_one; ctr++)
			node->widget_list[ctr] = node->widget_list[ctr+1];

		node->widget_list[less_by_one] = widget;
	}
}
	
void widget_lower(px_widget* widget)
{
	px_xynode* ynodes = NULL;
	MajorAxis axis;
	Display* display = display_get_Xdisplay();

	if (widget->child_of != NULL && widget->child_of->wtype != LISTBOX)
	{
		px_widget* parent = widget->child_of;
		switch (parent->wtype)
		{
			case FRAME:
				ynodes = parent->specific.frame.ynodes;
				axis = parent->specific.frame.axis;
				break;

			case TOPLEVEL:
				ynodes = parent->specific.toplevel.ynodes;
				axis = parent->specific.toplevel.axis;
				break;

			default:
				break;
		}
		
		widget_get_xynode_with_child(ynodes, widget, &lower_the_widget, 
				axis == YAXIS ? YDIRECTION : XDIRECTION, axis);
	}
			
	XLowerWindow(display, widget->win);
}

void widget_draw_border(px_widget* widget)
{
	Display* display = display_get_Xdisplay();
	if (widget->border == RECTANGLE_BORDER)
	{
		GC gc;
		unsigned long fg = 0;

		gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
		XSetGraphicsExposures(display, gc, False);
		if (widget->border_color != NULL)
			fg = widget->border_color->pixel;
		
		XSetForeground(display, gc, fg);
		XDrawRectangle(display, widget->backing_store.pxmp, gc, 0, 0, widget->width-1, widget->height-1);
		XFreeGC(display, gc);

		if (widget->backing_store.mask != None)
		{
			GC mask_gc;
			
			mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
			XSetGraphicsExposures(display, mask_gc, False);
			XSetForeground(display, mask_gc, 1);
			XDrawRectangle(display, widget->backing_store.mask, mask_gc, 0, 0, widget->width-1, widget->height-1);
			XFreeGC(display, mask_gc);
		}
	}
	else if (widget->border == ELLIPTIC_BORDER)
	{
		GC gc;
		unsigned long fg = 0;

		gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
		XSetGraphicsExposures(display, gc, False);
		if (widget->border_color != NULL)
			fg = widget->border_color->pixel;
		
		XSetForeground(display, gc, fg);

		XDrawArc(display, widget->backing_store.pxmp, gc, 
				0, 0, widget->width-1, widget->height-1,0, 64*360);
		XFreeGC(display, gc);

		if (widget->backing_store.mask == None)
		{
			GC mask_gc;
			widget->backing_store.mask = XCreatePixmap(display, widget->win, widget->width, widget->height, 1);

			mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
			XSetGraphicsExposures(display, mask_gc, False);
			XSetForeground(display, mask_gc, 0);
			XFillRectangle(display, widget->backing_store.mask, mask_gc, 0, 0, widget->width, widget->height);

			XSetForeground(display, mask_gc, 1);
			XDrawArc(display, widget->backing_store.mask, mask_gc, 
				0, 0, widget->width-1, widget->height-1,0, 64*360);
			XFillArc(display, widget->backing_store.mask, mask_gc, 
				0, 0, widget->width-1, widget->height-1, 0, 64*360);
			XFreeGC(display, mask_gc);
		}
		else
		{
			GC mask_gc;
		
			Pixmap m = XCreatePixmap(display, widget->win, widget->width, widget->height, 1);
			
			mask_gc = XCreateGC(display,widget->backing_store.mask, 0, NULL);
			XSetGraphicsExposures(display, mask_gc, False);
			XSetForeground(display, mask_gc, 1);
			XDrawArc(display, widget->backing_store.mask, mask_gc, 
				0, 0, widget->width-1, widget->height-1,0, 64*360);
			XFreeGC(display, mask_gc);

			mask_gc = XCreateGC(display, m, 0, NULL);
			XSetGraphicsExposures(display, mask_gc, False);
			XSetForeground(display, mask_gc, 0);
			XFillRectangle(display, m, mask_gc, 0, 0, widget->width, widget->height);

			XSetForeground(display, mask_gc, 1);
			XDrawArc(display, m, mask_gc, 
				0, 0, widget->width-1, widget->height-1,0, 64*360);
			XFillArc(display, m, mask_gc, 
				0, 0, widget->width-1, widget->height-1, 0, 64*360);

			XSetForeground(display, mask_gc, 1);
			XSetBackground(display, mask_gc, 0);

			XSetFunction(display, mask_gc, GXand);
			XCopyArea(display, m, widget->backing_store.mask, mask_gc, 
				0, 0, widget->width, widget->height, 0, 0);

			XFreeGC(display, mask_gc);
			XFreePixmap(display, m);
		}
	}
	else if (widget->border == BEVELLED_BORDER)
	{
		GC gc;
		unsigned long fg = 0;
		int radius = 16;

		gc = XCreateGC(display, widget->backing_store.pxmp, 0, NULL);
		XSetGraphicsExposures(display, gc, False);
		if (widget->border_color != NULL)
			fg = widget->border_color->pixel;
		
		XSetForeground(display, gc, fg);

		// These functions draw the outline of the bevelled border on the pixmap

		//Draws the arc at the top left corner

		XDrawArc(display, widget->backing_store.pxmp, gc, 0, 0, radius, radius, 90 * 64, 90 * 64);
		
		//Draws the left vertical edge.
		XDrawLine(display, widget->backing_store.pxmp, gc, 0, (radius >> 1) + 1, 
				0, widget->height - (radius>>1) - 1);

		// Draws the arc at the bottom left corner.

		XDrawArc(display, widget->backing_store.pxmp, gc, 0, widget->height - radius -1, 
				radius, radius, 180 * 64, 90 * 64);

		// Draws the bottom horizontal edge.

		XDrawLine(display, widget->backing_store.pxmp, gc, (radius>>1), widget->height - 1,
				widget->width - (radius>>1) - 1,  widget->height - 1);

		// Draws the arc at the bottom right corner.
		
		XDrawArc(display, widget->backing_store.pxmp, gc, widget->width - radius - 1, widget->height - radius - 1,
				radius, radius, 270 * 64, 90 * 64);

		//Draws the right vertical edge.
		XDrawLine(display, widget->backing_store.pxmp, gc, widget->width - 1, widget->height - (radius>>1) - 1, 
				widget->width - 1, radius - (radius >>1));

		// Draws the arc at top right corner.

		XDrawArc(display, widget->backing_store.pxmp, gc, widget->width - radius -1, 0, 
				radius, radius, 0 * 64, 90 * 64);
		
		//Draws the top edge.

		XDrawLine(display, widget->backing_store.pxmp, gc, widget->width - (radius >>1) - 1, 0,  
				radius-(radius>>1) , 0);


		XFreeGC(display, gc);

		if (widget->backing_store.mask == None)
		{
			GC mask_gc;
			widget->backing_store.mask = XCreatePixmap(display, widget->win, widget->width, widget->height, 1);

			mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
			XSetGraphicsExposures(display, mask_gc, False);
			XSetForeground(display, mask_gc, 0);
			XFillRectangle(display, widget->backing_store.mask, mask_gc, 0, 0, widget->width-1, widget->height-1);

			XSetForeground(display, mask_gc, 1);

			// Fills arc in top left corner.
			
			XFillArc(display, widget->backing_store.mask, mask_gc, 0, 0, radius, radius, 90 * 64, 90 * 64);

			// Fills arc in bottom left corner.

			XFillArc(display, widget->backing_store.mask, mask_gc, 0, widget->height - radius -1, 
					radius, radius, 180 * 64, 90 * 64);

			// Fills arc in bottom right corner.
			
			XFillArc(display, widget->backing_store.mask, mask_gc, widget->width-radius-1, widget->height-radius-1,
					radius, radius, 270 * 64, 90 * 64);

			// Fills arc in top right corner.
			
			XFillArc(display, widget->backing_store.mask, mask_gc, widget->width - radius -1, 0, 
					radius, radius, 0 * 64, 90 * 64);

			// Fills the middle section of the widget.
			
			XFillRectangle(display, widget->backing_store.mask, mask_gc, radius-(radius>>1) - 2 , 0,
					widget->width - radius + 4,  widget->height);

			// Fills the rectangle b/w the left arcs.
			
			XFillRectangle(display, widget->backing_store.mask, mask_gc, 0, (radius >> 1) - 2, 
					radius-(radius>>1), widget->height - (radius) + 4);

			// Fills the rectangle b/w the right arcs.
			
			XFillRectangle(display, widget->backing_store.mask, mask_gc, widget->width - (radius>>1), 
					(radius >> 1) -2,
					widget->width - 1, widget->height - (radius) + 4);

			//Draws the arc at the top left corner
			XDrawArc(display, widget->backing_store.mask, mask_gc, 0, 0, radius, radius, 90 * 64, 90 * 64);


			//Draws the left vertical edge.
			XDrawLine(display, widget->backing_store.mask, mask_gc, 0, (radius >> 1) + 1, 
					0, widget->height - (radius>>1) - 1);


			// Draws the arc at the bottom left corner.
			XDrawArc(display, widget->backing_store.mask, mask_gc, 0, widget->height - radius -1,
					radius, radius, 180 * 64, 90 * 64);

			// Draws the bottom horizontal edge.
			XDrawLine(display, widget->backing_store.mask, mask_gc, (radius>>1), widget->height - 1, 
					widget->width - (radius>>1) - 1,  widget->height - 1);


			// Draws the arc at the bottom right corner.
			XDrawArc(display, widget->backing_store.mask, mask_gc, widget->width-radius-1, 
					widget->height-radius-1,
					radius, radius, 270 * 64, 90 * 64);


			//Draws the right vertical edge.
			XDrawLine(display, widget->backing_store.mask, mask_gc, widget->width - 1, 
					widget->height - (radius>>1) - 1, 
					widget->width - 1, radius - (radius >>1));


			// Draws the arc at top right corner.
			XDrawArc(display, widget->backing_store.mask, mask_gc, widget->width - radius -1, 0,
					radius, radius, 0 * 64, 90 * 64);

			
			//Draws the top edge.
			XDrawLine(display, widget->backing_store.mask, mask_gc, widget->width - (radius >>1) - 1, 0,  
					radius-(radius>>1) , 0);  //  top

			XFreeGC(display, mask_gc);
		}
		else
		{
			GC mask_gc;
			
			Pixmap m = XCreatePixmap(display, widget->win, widget->width, widget->height, 1);

			mask_gc = XCreateGC(display, widget->backing_store.mask, 0, NULL);
			XSetGraphicsExposures(display, mask_gc, False);
			XSetForeground(display, mask_gc, 1);

			//Draws the arc at the top left corner
			XDrawArc(display, widget->backing_store.mask, mask_gc, 0, 0, radius, radius, 90 * 64, 90 * 64);


			//Draws the left vertical edge.
			XDrawLine(display, widget->backing_store.mask, mask_gc, 0, (radius >> 1) + 1, 
					0, widget->height - (radius>>1) - 1);


			// Draws the arc at the bottom left corner.
			XDrawArc(display, widget->backing_store.mask, mask_gc, 0, widget->height - radius -1,
					radius, radius, 180 * 64, 90 * 64);

			// Draws the bottom horizontal edge.
			XDrawLine(display, widget->backing_store.mask, mask_gc, (radius>>1), widget->height - 1, 
					widget->width - (radius>>1) - 1,  widget->height - 1);


			// Draws the arc at the bottom right corner.
			XDrawArc(display, widget->backing_store.mask, mask_gc, widget->width-radius-1, 
					widget->height-radius-1,
					radius, radius, 270 * 64, 90 * 64);


			//Draws the right vertical edge.
			XDrawLine(display, widget->backing_store.mask, mask_gc, widget->width - 1, 
					widget->height - (radius>>1) - 1, 
					widget->width - 1, radius - (radius >>1));


			// Draws the arc at top right corner.
			XDrawArc(display, widget->backing_store.mask, mask_gc, widget->width - radius -1, 0,
					radius, radius, 0 * 64, 90 * 64);

			
			//Draws the top edge.
			XDrawLine(display, widget->backing_store.mask, mask_gc, widget->width - (radius >>1) - 1, 0,  
					radius-(radius>>1) , 0);  //  top

			XFreeGC(display, mask_gc);

			mask_gc = XCreateGC(display, m, 0, NULL);
			XSetGraphicsExposures(display, mask_gc, False);
			XSetForeground(display, mask_gc, 0);
			XFillRectangle(display, m, mask_gc, 0, 0, widget->width, widget->height);

			XSetForeground(display, mask_gc, 1);

			// Fills arc in top left corner.
			
			XFillArc(display, m, mask_gc, 0, 0, radius, radius, 90 * 64, 90 * 64);

			// Fills arc in bottom left corner.

			XFillArc(display, m, mask_gc, 0, widget->height - radius -1, 
					radius, radius, 180 * 64, 90 * 64);

			// Fills arc in bottom right corner.
			
			XFillArc(display, m, mask_gc, widget->width-radius-1, widget->height-radius-1,
					radius, radius, 270 * 64, 90 * 64);

			// Fills arc in top right corner.
			
			XFillArc(display, m, mask_gc, widget->width - radius -1, 0, 
					radius, radius, 0 * 64, 90 * 64);

			// Fills the middle section of the widget.
			
			XFillRectangle(display, m, mask_gc, radius-(radius>>1) - 2 , 0,
					widget->width - radius + 4,  widget->height);

			// Fills the rectangle b/w the left arcs.
			
			XFillRectangle(display, m, mask_gc, 0, (radius >> 1) - 2, 
					radius-(radius>>1), widget->height - (radius) + 4);

			// Fills the rectangle b/w the right arcs.
			
			XFillRectangle(display, m, mask_gc, widget->width - (radius>>1), 
					(radius >> 1) -2,
					widget->width - 1, widget->height - (radius) + 4);

			//Draws the arc at the top left corner
			XDrawArc(display, m, mask_gc, 0, 0, radius, radius, 90 * 64, 90 * 64);


			//Draws the left vertical edge.
			XDrawLine(display, m, mask_gc, 0, (radius >> 1) + 1, 
					0, widget->height - (radius>>1) - 1);


			// Draws the arc at the bottom left corner.
			XDrawArc(display, m, mask_gc, 0, widget->height - radius -1,
					radius, radius, 180 * 64, 90 * 64);

			// Draws the bottom horizontal edge.
			XDrawLine(display, m, mask_gc, (radius>>1), widget->height - 1, 
					widget->width - (radius>>1) - 1,  widget->height - 1);


			// Draws the arc at the bottom right corner.
			XDrawArc(display, m, mask_gc, widget->width-radius-1, 
					widget->height-radius-1,
					radius, radius, 270 * 64, 90 * 64);


			//Draws the right vertical edge.
			XDrawLine(display, m, mask_gc, widget->width - 1, 
					widget->height - (radius>>1) - 1, 
					widget->width - 1, radius - (radius >>1));


			// Draws the arc at top right corner.
			XDrawArc(display, m, mask_gc, widget->width - radius -1, 0,
					radius, radius, 0 * 64, 90 * 64);

			
			//Draws the top edge.
			XDrawLine(display, m, mask_gc, widget->width - (radius >>1) - 1, 0,  
					radius-(radius>>1) , 0);  //  top

			XSetForeground(display, mask_gc, 1);
			XSetBackground(display, mask_gc, 0);

			XSetFunction(display, mask_gc, GXand);
			XCopyArea(display, m, widget->backing_store.mask, mask_gc, 
				0, 0, widget->width, widget->height, 0, 0);

			XFreeGC(display, mask_gc);
			XFreePixmap(display, m);
		}
	}
}
		
		
void widget_set_state(px_widget* widget, widget_states state)
{
	widget->state = state;
}

widget_states widget_get_state(const px_widget* widget)
{
	return widget->state;
}

void widget_set_focus(px_widget* widget, Bool focus)
{
	if (widget->wtype == ENTRY)
		entry_set_focus(widget, focus);
	else if(widget->wtype == TEXT)
		text_set_focus(widget, focus);
	else
		widget->have_focus = focus;
}

void widget_compute_textlayout(px_textlayout* layout, px_widget* widget)
{
	int length;
	int count;
	px_textchunk *chunk = NULL, *last_chunk = NULL;

	chunk = layout->chunks;
	
	while(chunk != NULL)
	{
		last_chunk = chunk->next;
		free(chunk);
		chunk = last_chunk;
	}

	chunk = NULL;
	last_chunk = NULL;

	if (layout->text == NULL || !strlen(layout->text) || layout->font == NULL || layout->font->font == NULL)
		return;

	length = strlen(layout->text);
	if (layout->wrap == NO_WRAP)
	{
		chunk = (px_textchunk*)malloc(sizeof(px_textchunk));
		chunk->start = layout->text;
		chunk->length = length;
		chunk->text_width = font_get_width(layout->font, chunk->start, length);
		layout->no_of_chunks = 1;
		layout->height = font_get_height(layout->font, chunk->start) + 4;
		chunk->next = NULL;
		layout->chunks = chunk;
		return;
	}
		
	layout->no_of_chunks = 0;
	
	for (count = 0; count < length; count++)
	{
		if (chunk == NULL)
		{
			chunk = (px_textchunk*)malloc(sizeof(px_textchunk));
			chunk->start = NULL;
			chunk->length = 0;
			chunk->text_width = 0;
			chunk->next = NULL;
			count--;
		}
		else
		{
			int char_width = font_get_width_of_char(layout->font, layout->text[count]);

			if (chunk->text_width + char_width < layout->width && layout->text[count] != '\n')
			{
				if (chunk->start == NULL)
					chunk->start = layout->text + count;
				chunk->length++;
				chunk->text_width += char_width;
			}
			else
			{
				if (layout->wrap == WRAP_WORD && chunk->start != NULL && 
						layout->text[count] != '\n')
				{
					int l = chunk->length - 1;
					
					while (l > 1 && (isalnum(chunk->start[l]) || chunk->start[l] == '_'))
					{
						chunk->text_width -= 
							font_get_width_of_char(layout->font,chunk->start[l]);
						l--;
					}
						

					if (l > 0)
					{
						count -= (chunk->length - l - 1);
						chunk->length = l + 1;
					}
				}
				if (last_chunk == NULL)
					layout->chunks = chunk;
				else
					last_chunk->next = chunk;
				last_chunk = chunk;
				chunk = NULL;
				if (layout->text[count] != '\n' && char_width < layout->width)
					count--;
				layout->no_of_chunks++;
			}
		}
	}

	if (chunk != NULL)
	{
		if (last_chunk == NULL)
			layout->chunks = chunk;
		else
			last_chunk->next = chunk;
		last_chunk = chunk;
		chunk = NULL;
		layout->no_of_chunks++;
	}

	layout->height = layout->no_of_chunks * font_get_height(layout->font, layout->text) + 
		(layout->no_of_chunks + 1) * 2;

}
		
void widget_draw_text(px_textlayout* layout, px_pixmap* drawable)
{
	if (layout->text != NULL && layout->font != NULL && layout->font->font != NULL && layout->chunks != NULL)
	{
		int text_x, text_y, text_height, line_height;
		GC mask_gc, gc;
		Display* display = display_get_Xdisplay();
		px_textchunk* chunk = layout->chunks;
		
		line_height = font_get_height(layout->font, layout->text);
		text_height = layout->no_of_chunks * line_height;

		gc = XCreateGC(display, drawable->pxmp, 0, NULL);
		XSetGraphicsExposures(display, gc, False);
		if (layout->font_color != NULL)
			XSetForeground(display, gc, layout->font_color->pixel);
		else
			XSetForeground(display, gc, 0);

		XSetFont(display, gc, layout->font->font->fid);
		
		if (drawable->mask != None)
		{
			mask_gc	=  XCreateGC(display, drawable->mask, 0, NULL);
			XSetGraphicsExposures(display, mask_gc, False);
			XSetFont(display, mask_gc, layout->font->font->fid);
			XSetForeground(display, mask_gc, 1);
		}

		text_y = (drawable->h - layout->y - (text_height + (layout->no_of_chunks - 1) * 2))/2 + layout->y;
		text_y -= layout->font->font->descent;
		text_y += line_height;
	
		
		while (chunk != NULL)
		{
			switch (layout->just)
			{
				case JUSTIFY_LEFT :
					text_x = layout->x + 3;
					break;

				case JUSTIFY_RIGHT :
					text_x = drawable->w - chunk->text_width;
					if (text_x <= layout->x)
						text_x = layout->x + 3;
					break;

				default :
					text_x = (drawable->w - layout->x - chunk->text_width)/2 + layout->x;
					break;
			}
			
			if (text_x < layout->x)
				text_x = layout->x;

			if (chunk->start != NULL)
			{
				XDrawString(display, drawable->pxmp, gc, text_x, text_y, chunk->start, chunk->length);

				if (drawable->mask != None)
				{
					XDrawString(display, drawable->mask, mask_gc, text_x, text_y, 
						chunk->start, chunk->length);
				}
			}
			
			text_y += 2 + line_height;
			chunk = chunk->next;
		}
		XFreeGC(display, gc);

		if (drawable->mask != None)
			XFreeGC(display, mask_gc);
	}
}

/*
void widget_draw_text(px_textlayout* layout, px_pixmap* drawable)
{
	if (layout->text != NULL && layout->font != NULL && layout->font->font != NULL)
	{
		int text_x, text_y, no_of_lines = 0;
		char* text = strdup(layout->text);
		char* line_start = text;
		GC mask_gc, gc;
		Display* display = display_get_Xdisplay();
		
		gc = XCreateGC(display, drawable->pxmp, 0, NULL);
		XSetGraphicsExposures(display, gc, False);
		if (layout->font_color != NULL)
			XSetForeground(display, gc, layout->font_color->pixel);
		else
			XSetForeground(display, gc, 0);

		XSetFont(display, gc, layout->font->font->fid);
		
		if (drawable->mask != None)
		{
			mask_gc	=  XCreateGC(display, drawable->mask, 0, NULL);
			XSetGraphicsExposures(display, mask_gc, False);
			XSetFont(display, mask_gc, layout->font->font->fid);
			XSetForeground(display, mask_gc, 1);
		}

		
		while (line_start != NULL)
		{
			char* line_break = strchr(line_start, '\n');
			int text_width;
			if (line_break != NULL)
				(*line_break) = '\0';
			 
			text_width = font_get_width(layout->font, line_start, strlen(line_start));

			switch (layout->just)
			{
				case JUSTIFY_LEFT :
					text_x = layout->x + 3;
					break;

				case JUSTIFY_RIGHT :
					text_x = layout->width - text_width;
					if (text_x <= layout->x)
						text_x = layout->x + 3;
					break;

				default :
					text_x = (layout->width - layout->x - text_width)/2 + layout->x;
					break;
			}
			
			if (no_of_lines == 0 && line_break == NULL)
			{
				text_y = layout->height - (layout->height - font_get_height(layout->font, line_start))/2;
				text_y -= layout->font->font->descent;
			}
			else
				text_y = 2 + (no_of_lines + 1) * font_get_height(layout->font, line_start);

			XDrawString(display, drawable->pxmp, gc, text_x, text_y, line_start, strlen(line_start));

			if (drawable->mask != None)
			{
				XDrawString(display, drawable->mask, mask_gc, text_x, text_y, 
					line_start, strlen(line_start));
			}

			if (line_break != NULL)
				line_start = line_break + 1;
			else
				line_start = NULL;
			no_of_lines++;
			
		}
		XFreeGC(display, gc);

		if (drawable->mask != None)
			XFreeGC(display, mask_gc);

		free(text);
	}
}
*/

px_xynode* widget_add_child(px_xynode* ynode_start, px_widget* child, Direction dir, MajorAxis majoraxis)
{
	px_xynode* ynode, *prev_ynode;
	unsigned child_from; // Stores the unprocessed start value of the child;
	unsigned child_to; // Stores the end value of the child;
	
	/****************************************************************
	 * 1. We check if the child falls into any of the existing	*
	 *    ynodes.							*
	 *								*
	 * 2. If it doesnt fall into any ynode ... then it is quite	*
	 *    straight forward ... create a new ynode, xnode, fill	*
	 *    them up and return.					*
	 *								*
	 * 3. If it falls within a ynode or overlaps a number of 	*
	 *    ynodes, we have to start splitting ynodes ... which 	*
	 *    can get tricky. See later comments for more explanations	*
	 ****************************************************************/
	ynode = ynode_start;
	prev_ynode = NULL;
	if (dir == XDIRECTION)
	{
		child_from = child->x;
		child_to = child->x + child->width - 1;
	}
	else
	{
		child_from = child->y;
		child_to = child->y + child->height - 1;
	}

	while (ynode != NULL && ynode->to < child_from)
	{
		// Uninteresting nodes ... they are much above the new child
		prev_ynode = ynode;
		ynode = ynode->next;
	}


	while (ynode != NULL && child_to >= ynode->from)
	{
		if (child_from < ynode->from)
		{
			/********************************************************
			 * There is a gap in between the nodes that we need to	*
			 * fill up.		 				*
			 * ******************************************************/
	
			px_xynode* new_ynode;
			
			new_ynode = (px_xynode*)malloc(sizeof(px_xynode));
			new_ynode->from = child_from;
			if (child_to < ynode->from)
			{
				new_ynode->to = child_to;
			}
			else
			{
				new_ynode->to = ynode->from - 1;
			}
			child_from = new_ynode->to + 1;
			
			new_ynode->next = NULL;
			
			if ((dir == YDIRECTION && majoraxis == YAXIS) || 
					(dir == XDIRECTION && majoraxis == XAXIS))
			{
				new_ynode->xnodes = widget_add_child(NULL, child, !dir, majoraxis);
			}
			else
			{
				new_ynode->xnodes = NULL;
			}
			
			// add child to this node
			new_ynode->allocated = 3;
	
			new_ynode->widget_list = (px_widget**)malloc(sizeof(px_widget*) * new_ynode->allocated);
			new_ynode->widget_list[0] = child;
			new_ynode->no_of_widgets = 1;

			if (prev_ynode != NULL)
			{
				new_ynode->next = prev_ynode->next;
				prev_ynode->next = new_ynode;
			}
			else
			{
				new_ynode->next = ynode_start;
				ynode_start = new_ynode;
			}
			ynode = new_ynode;
	
		}
		else if (child_from > ynode->from)
		{
			/****************************************************	*
			 * We need to split the node into top & bottom where 	*
			 * the top part is the same as the original node except	*
			 * for y-range. We need to add 'child' to the bottom 	*
			 * part some time later.				*
			 * ******************************************************/

			px_xynode* new_node;
			unsigned int count;
			
			new_node = (px_xynode*)malloc(sizeof(px_xynode));
			new_node->from = child_from;
			new_node->to = ynode->to;
			new_node->next = ynode->next;
			
			if ((dir == YDIRECTION && majoraxis == YAXIS) || 
					(dir == XDIRECTION && majoraxis == XAXIS))
			{
				// Replicate the xnodes in the new ynode
				px_xynode* new_xnode = NULL, *xnode, *prev_new_xnode = NULL;
				xnode = ynode->xnodes;

				while(xnode != NULL)
				{
				
					new_xnode = (px_xynode*)malloc(sizeof(px_xynode));
					new_xnode->from = xnode->from;
					new_xnode->to = xnode->to;
					new_xnode->xnodes = NULL;
					new_xnode->widget_list = (px_widget**)malloc(sizeof(px_widget*) * xnode->allocated);

					for (count = 0; count < xnode->no_of_widgets; count++)
						new_xnode->widget_list[count] = xnode->widget_list[count];

					new_xnode->no_of_widgets = xnode->no_of_widgets;
					new_xnode->allocated = xnode->allocated;
					new_xnode->next = NULL;

					if (prev_new_xnode == NULL)
					{
						new_node->xnodes = new_xnode;
					}
					else
					{
						prev_new_xnode->next = new_xnode;
					}

					prev_new_xnode = new_xnode;
					xnode = xnode->next;
				}
			}
			else
			{
				new_node->xnodes = NULL;
			}
			
			new_node->widget_list = (px_widget**)malloc(sizeof(px_widget*) * (ynode->allocated));

			for (count = 0; count < ynode->no_of_widgets;count++)
				new_node->widget_list[count] = ynode->widget_list[count];

			new_node->allocated = ynode->allocated;
			new_node->no_of_widgets = ynode->no_of_widgets;
					
			// Add the new node to the list.
			ynode->to = child_from - 1;
			new_node->next = ynode->next;
			ynode->next = new_node;
		}
		else if (child_to < ynode->to)
		{
			/****************************************************	*
			 * We need to split the node into top & bottom where 	*
			 * the bottom part is the same as the original node 	*
			 * except for y-range. We need to add 'child' to the 	*
			 * top part some time later.				*
			 * ******************************************************/

			px_xynode* new_node;
			int count, index;
			
			new_node = (px_xynode*)malloc(sizeof(px_xynode));
			new_node->from = child_to + 1;
			new_node->to = ynode->to;

			if ((dir == YDIRECTION && majoraxis == YAXIS) || 
					(dir == XDIRECTION && majoraxis == XAXIS))
			{
				// Replicate the xnodes in the new ynode
				px_xynode* new_xnode = NULL, *xnode, *prev_new_xnode = NULL;
				xnode = ynode->xnodes;

				while(xnode != NULL)
				{
					unsigned int count = 0;
				
					new_xnode = (px_xynode*)malloc(sizeof(px_xynode));
					new_xnode->from = xnode->from;
					new_xnode->to = xnode->to;
					new_xnode->xnodes = NULL;
					new_xnode->widget_list = (px_widget**)malloc(sizeof(px_widget*) * xnode->allocated);

					for (count = 0; count < xnode->no_of_widgets; count++)
						new_xnode->widget_list[count] = xnode->widget_list[count];

					new_xnode->no_of_widgets = xnode->no_of_widgets;
					new_xnode->allocated = xnode->allocated;
					new_xnode->next = NULL;

					if (prev_new_xnode == NULL)
					{
						new_node->xnodes = new_xnode;
					}
					else
					{
						prev_new_xnode->next = new_xnode;
					}

					prev_new_xnode = new_xnode;
					xnode = xnode->next;
				}
			}
			else
			{
				new_node->xnodes = NULL;
			}
			
			new_node->widget_list = (px_widget**)malloc(sizeof(px_widget*) * (ynode->allocated));

			for (count = 0; count < ynode->no_of_widgets;count++)
				new_node->widget_list[count] = ynode->widget_list[count];

			new_node->allocated = ynode->allocated;
			new_node->no_of_widgets = ynode->no_of_widgets;

			// Add the new node to the list.
			ynode->to = child_to;
			new_node->next = ynode->next;
			ynode->next = new_node;

			if ((dir == YDIRECTION && majoraxis == YAXIS) || 
					(dir == XDIRECTION && majoraxis == XAXIS))
			{
				ynode->xnodes = widget_add_child(ynode->xnodes, child, !dir, majoraxis);
			}
			
			// add child to this node
			if (ynode->no_of_widgets == ynode->allocated)
			{
				px_widget** list = (px_widget**)malloc(sizeof(px_widget*) * (ynode->allocated + 3));

				for (count = 0; count < ynode->no_of_widgets;count++)
					list[count] = ynode->widget_list[count];

				free(ynode->widget_list);
				ynode->widget_list = list;
				ynode->allocated += 3;
			}
			
			count = 0;
				
			while (count < ynode->no_of_widgets && ynode->widget_list[count]->depth > child->depth)
				count++;

			index = count;

			for (count = ynode->no_of_widgets - 1; count >= index;count--)
				ynode->widget_list[count+1] = ynode->widget_list[count];

			ynode->widget_list[index] = child;
			ynode->no_of_widgets++;
			
			child_from = ynode->to + 1;

		}
		else if (child_to >= ynode->to)
		{
			/************************************************
			 * Either the 'child' fits exactly into this 	*
			 * node or is bigger and overlaps this node. 	*
			 * So we have to child to this node.		*
			 * **********************************************/
			int count, index;
			if ((dir == YDIRECTION && majoraxis == YAXIS) || 
					(dir == XDIRECTION && majoraxis == XAXIS))
			{
				ynode->xnodes = widget_add_child(ynode->xnodes, child, !dir, majoraxis);
			}
			
			// add child to this node
			if (ynode->no_of_widgets == ynode->allocated)
			{
				px_widget** list = (px_widget**)malloc(sizeof(px_widget*) * (ynode->allocated + 3));

				for (count = 0; count < ynode->no_of_widgets;count++)
					list[count] = ynode->widget_list[count];

				free(ynode->widget_list);
				ynode->widget_list = list;
				ynode->allocated += 3;
			}
				
			count = 0;
				
			while (count < ynode->no_of_widgets && ynode->widget_list[count]->depth > child->depth)
				count++;

			index = count;

			for (count = ynode->no_of_widgets - 1; count >= index;count--)
				ynode->widget_list[count+1] = ynode->widget_list[count];

			ynode->widget_list[index] = child;
			ynode->no_of_widgets++;
					
			child_from = ynode->to + 1;
		}

		prev_ynode = ynode;
		ynode = ynode->next;
	}

	if (child_from <= child_to && (ynode == NULL || ynode->from > child_to))
	{
		/********************************************************
		 * Ah ... this is too cheap.		 		*
		 * It means that the parent never had any children	*
		 * to start of with or we need to create a new node and	*
		 * add it somewhere in the list given by prev_ynode. 	*
		 * So create a ynode and a xnode, fill it up and return.*
		 * ******************************************************/

		ynode = (px_xynode*)malloc(sizeof(px_xynode));
		ynode->from = child_from;
		ynode->to = child_to;
		ynode->next = NULL;
		if ((dir == YDIRECTION && majoraxis == YAXIS) || 
					(dir == XDIRECTION && majoraxis == XAXIS))
		{
			ynode->xnodes = widget_add_child(NULL, child, !dir, majoraxis);
		}
		else
		{
			ynode->xnodes = NULL;
		}
		
		// add child to this node
		ynode->allocated = 3;
			
		ynode->widget_list = (px_widget**)malloc(sizeof(px_widget*) * ynode->allocated);
		ynode->widget_list[0] = child;
		ynode->no_of_widgets = 1;
		
		if (prev_ynode != NULL)
		{
			ynode->next = prev_ynode->next;
			prev_ynode->next = ynode;
		}
		else
		{
			ynode->next = ynode_start;
			ynode_start = ynode;
		}

	}
	return ynode_start;
}

px_xynode* widget_remove_child(px_xynode* ynode_start, px_widget* child, Direction dir, MajorAxis majoraxis)
{
	px_xynode* ynode, *prev_ynode;
	unsigned child_from; // Stores the unprocessed start value of the child;
	unsigned child_to; // Stores the end value of the child;
	
	ynode = ynode_start;
	prev_ynode = NULL;
	if (dir == XDIRECTION)
	{
		child_from = child->x;
		child_to = child->x + child->width - 1;
	}
	else
	{
		child_from = child->y;
		child_to = child->y + child->height - 1;
	}

	while (ynode != NULL && ynode->to < child_from)
	{
		// Uninteresting nodes ... they are much above the new child
		prev_ynode = ynode;
		ynode = ynode->next;
	}


	while (ynode != NULL && child_to >= ynode->from)
	{
		int count;
		Bool dont_change = False;
		
		for (count = 0; count < ynode->no_of_widgets;count++)
			if (child == ynode->widget_list[count])
				break;

		/********************************************************
		 * This node contains the child. We have to process it.	*
		 *							*
		 * 1. If this is the only child in the node, we can 	*
		 *    delete the node itself from the list.		*
		 *							*
		 * 2. If this is not the only node we might have to 	*
		 *    merge the node with the one above or one below	*
		 ********************************************************/
		
		if ((dir == YDIRECTION && majoraxis == YAXIS) || 
				(dir == XDIRECTION && majoraxis == XAXIS))
		{
			ynode->xnodes = widget_remove_child(ynode->xnodes, child, !dir, majoraxis);
		}

		if (ynode->no_of_widgets == 1)
		{
			if (prev_ynode != NULL)
			{
				prev_ynode->next = ynode->next;
			}
			else
			{
				ynode_start = ynode->next;
			}

			free(ynode->widget_list);
			free(ynode);
			// We arent freeing ynodes->xnodes here because if it existed it would have been deleted
			// and freed a few statments above where widget_remove_child is called.
			
			
			if (prev_ynode != NULL)
			{
				ynode = prev_ynode;
			}
			else
			{
				ynode = ynode_start;
				dont_change = True;
			}
		}
		else
		{
			int ctr;

			for (ctr = count + 1;ctr < ynode->no_of_widgets;ctr++)
				ynode->widget_list[ctr-1] = ynode->widget_list[ctr];

			ynode->no_of_widgets--;

			// Check if we need to merge with the node above

			if (prev_ynode != NULL && prev_ynode->no_of_widgets == ynode->no_of_widgets)
			{
				Bool same = True;

				for (ctr = 0; ctr < ynode->no_of_widgets;ctr++)
				{
					if (prev_ynode->widget_list[ctr] != ynode->widget_list[ctr])
					{
						same = False;
						break;
					}
				}

				if (same)
				{
					px_xynode* xnode;
					prev_ynode->next = ynode->next;
					prev_ynode->to = ynode->to;

					while (ynode->xnodes != NULL)
					{
						xnode = ynode->xnodes->next;

						free(ynode->xnodes->widget_list);
						free(ynode->xnodes);

						ynode->xnodes = xnode;
					}

					free(ynode->widget_list);
					free(ynode);

					ynode = prev_ynode;
				}
			}

			// Check if we need to merge with the node below

			if (ynode->next != NULL && ynode->next->no_of_widgets == ynode->no_of_widgets)
			{
				Bool same = True;
				px_xynode* next = ynode->next;

				for (ctr = 0; ctr < ynode->no_of_widgets;ctr++)
				{
					if (next->widget_list[ctr] != ynode->widget_list[ctr])
					{
						same = False;
						break;
					}
				}

				if (same)
				{
					px_xynode* xnode;
					ynode->next = next->next;
					ynode->to = next->to;

					while (next->xnodes != NULL)
					{
						xnode = next->xnodes->next;

						free(next->xnodes->widget_list);
						free(next->xnodes);

						next->xnodes = xnode;
					}

					free(next->widget_list);
					free(next);

				}
			}
		}
			

		if (!dont_change)
		{
			prev_ynode = ynode;
			ynode = ynode->next;
		}
	}

	return ynode_start;
}

px_xynode* widget_get_xynode(px_xynode* ynode_start, int x, int y, Direction dir, MajorAxis majoraxis)
{
	int value;
	px_xynode* node = NULL;

	if (dir == YDIRECTION)
		value = y;
	else
		value = x;

	node = ynode_start;

	while (node != NULL)
	{
		if (value >= node->from && value <= node->to)
		{
			if ((dir == YDIRECTION && majoraxis == YAXIS) || 
					(dir == XDIRECTION && majoraxis == XAXIS))
				return widget_get_xynode(node->xnodes, x, y, !dir, majoraxis);
			else
			{
				return node;
			}
		}

		node = node->next;
	}

	return NULL;
}

void widget_get_xynode_with_child(px_xynode* ynode_start, px_widget* child, void (*func)(px_xynode*, px_widget*, Bool),
		Direction dir, MajorAxis majoraxis)
{

	int child_from, child_to;
	px_xynode* ynodes = ynode_start;

	if (dir == YDIRECTION)
	{
		child_from = child->y;
		child_to = child->y + child->height - 1;
	}
	else
	{
		child_from = child->x;
		child_to = child->x + child->width - 1;
	}
			
	while (ynodes != NULL && ynodes->to < child_from)
		ynodes = ynodes->next;

	while (ynodes != NULL && ynodes->from <= child_to)
	{
		Bool is_major = (dir == YDIRECTION && majoraxis == YAXIS) || 
				(dir == XDIRECTION && majoraxis == XAXIS);
			
		if (func != NULL)
			func(ynodes, child, is_major);

		if (is_major)
			widget_get_xynode_with_child(ynodes->xnodes, child, func, !dir, majoraxis);
		ynodes = ynodes->next;
	}
}
	
#ifdef __debug__
void widget_print_frame_buffer(px_xynode* ynode, Direction dir, MajorAxis majoraxis)
{
	while (ynode != NULL)
	{

		unsigned int count;
		if (!((dir == YDIRECTION && majoraxis == YAXIS) || 
					(dir == XDIRECTION && majoraxis == XAXIS)))
		{
			printf("\t");
		}

		printf("%s : %p : %d, %d : ", dir == YDIRECTION ? "Y-Direction" : "X-Direction", ynode, 
				ynode->from, ynode->to);
		for (count = 0; count < ynode->no_of_widgets;count++)
			printf("%p (%s) : ", ynode->widget_list[count], ynode->widget_list[count]->tag);
		printf("\n");
		
		widget_print_frame_buffer(ynode->xnodes, !dir, majoraxis);
		ynode = ynode->next;
	}
}
#endif
		
void picowidget_terminate()
{
	px_eventloop* eventloop = eventloop_get_eventloop();
	px_theme* theme = theme_get_theme();

	eventloop->destroy(eventloop);
	font_destroy_all();
	theme->destroy(theme);
	display_destroy();
}

