/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#ifndef __WIDGETFACTORY_H
#define __WIDGETFACTORY_H

#include "picowidget.h"

void widget_show(px_widget*);
void widget_hide(px_widget*);
void widget_destroy(px_widget*);

void widget_set_common_parameters(px_widget*);
void widget_set_specific_parameters(px_widget*);

Bool widget_within(const px_widget*, px_event*);
Bool widget_is_transparent(const px_widget*, int x, int y);
Bool toplevel_frame_is_transparent(const px_widget* widget, int x, int y);
void widget_set_size(px_widget* wid, unsigned long width, unsigned long height);
void widget_set_position(px_widget* wid, int x, int y);

void widget_get_size(const px_widget* wid, unsigned long* width, unsigned long* height);
void widget_get_postion(const px_widget*wid, int* x, int* y);

int widget_get_absolute_x(const px_widget*, Window);
int widget_get_absolute_y(const px_widget*, Window);
void widget_set_event_mask(px_widget*, event_masks);

Bool widget_handle_event(px_widget*, px_event*);

void widget_set_background_pixmap_from_file(px_widget* , char* );
void widget_set_background_pixmap_from_pixmap(px_widget* , const px_pixmap* );
void widget_set_background_color(px_widget* widget, const px_color*);
	
void widget_set_border_color(px_widget*, const char*);
void widget_set_border_type(px_widget*, border_types);
void widget_draw_border(px_widget*);

void widget_raise(px_widget*);
void widget_lower(px_widget*);

void widget_draw_px_color(px_pixmap*, const px_color*);

void widget_set_state(px_widget*, widget_states);
widget_states widget_get_state(const px_widget*);
void widget_set_focus(px_widget*, Bool);

void widget_compute_textlayout(px_textlayout* layout, px_widget*);
void widget_draw_text(px_textlayout* layout, px_pixmap* drawable);

px_xynode* widget_add_child(px_xynode*, px_widget*, Direction, MajorAxis);
px_xynode* widget_remove_child(px_xynode*, px_widget*, Direction, MajorAxis);
px_xynode* widget_get_xynode(px_xynode*, int x, int y, Direction, MajorAxis);
void widget_get_xynode_with_child(px_xynode*, px_widget*, void (*func)(px_xynode*, px_widget*, Bool), 
		Direction, MajorAxis);

void widget_destruct(px_widget*);
void widget_set_wrap_type(px_widget* widget, wrap_types wrap);
void widget_size_depends_on_contents(px_widget* widget, Bool value);

#ifdef __debug__
void widget_print_frame_buffer(px_xynode* ynode, Direction dir, MajorAxis majoraxis);
#endif
#endif
